#ifndef CGUL_BITWISE_TRIE_H
#define CGUL_BITWISE_TRIE_H

#include "cgul_common.h"
#include "cgul_exception.h"

/**
 * Bitwise trie implementation. <p>
 *
 * @file
 * @author Paul Serice
 * @brief bitwise trie implementation
 */

#include "cgul_common.h"
#include "cgul_cache.h"
#include "cgul_exception.h"
#include "cgul_bitwise_trie_node.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_bitwise_trie</code> instance. */
typedef struct cgul_bitwise_trie* cgul_bitwise_trie_t;

/**
 * Create a new <code>cgul_bitwise_trie</code> object.  If memory
 * cannot be allocated, <code>NULL</code> is returned, and an
 * exception is thrown.  The client is responsible for freeing the
 * object by calling <code>cgul_bitwise_trie__delete()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_bitwise_trie</code> instance
 */
extern CGUL_EXPORT cgul_bitwise_trie_t
cgul_bitwise_trie__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  This does not
 * include the values stored in the trie.  The client is responsible
 * for freeing the values when the client thinks it is convenient and
 * safe to do so. <p>
 *
 * As a convenience, you may want to call
 * <code>cgul_bitwise_trie__clear()</code> before calling this method
 * in order to properly put the values back on its
 * <code>cgul_cache</code> object (if a <code>cgul_cache</code> object
 * is being used) before you delete the trie.  If a
 * <code>cgul_cache</code> object is not being used, the client needs
 * to arrange some other mechanism to free the values. <p>
 *
 * Do not try to access <code>t</code> after calling this method. <p>
 *
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @see cgul_bitwise_trie__free_values()
 */
extern CGUL_EXPORT void
cgul_bitwise_trie__delete(cgul_bitwise_trie_t t);

/**
 * This method calls <code>free()</code> on all the values in the trie
 * <code>t</code>.  Because this is such a common operation, it is an
 * exception to the rule that cgul containers never free values.  This
 * method should only ever be called immediately before calling
 * <code>cgul_bitwise_trie__delete()</code> because it otherwise
 * invalidates the trie.  Because this function is basically an
 * extension of <code>cgul_bitwise_trie__delete()</code>, it does not
 * accept a <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 */
extern CGUL_EXPORT void
cgul_bitwise_trie__free_values(cgul_bitwise_trie_t t);

/**
 * Get the size of the cache of nodes. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @return size of the cache of nodes
 */
extern CGUL_EXPORT unsigned long int
cgul_bitwise_trie__get_cache_size(cgul_exception_t* cex,
                                  cgul_bitwise_trie_t t);

/**
 * Set the size of the cache of nodes.
 *
 * For efficiency, the <code>cgul_bitwise_trie</code> can keep a cache
 * of nodes that can be reused.  In many situations, this can greatly
 * reduce the number of calls to <code>malloc()</code> which can
 * greatly improve performance and reduce memory fragmentation. <p>
 *
 * You can release all the cached nodes and disable caching of nodes
 * by setting the cache reserve to 0 and the cache size to 0 (in that
 * order).  This is the default. <p>
 *
 * If an error occurs (while enlarging the cache), an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @param[in] size of the cache of nodes
 */
extern CGUL_EXPORT void
cgul_bitwise_trie__set_cache_size(cgul_exception_t* cex,
                                  cgul_bitwise_trie_t t,
                                  unsigned long int size);

/**
 * Get the reserve limit of the cache of nodes. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @return reserve limit of the cache of nodes
 */
extern CGUL_EXPORT unsigned long int
cgul_bitwise_trie__get_cache_reserve(cgul_exception_t* cex,
                                     cgul_bitwise_trie_t t);

/**
 * Set the reserve limit of the cache of nodes and guarantee that at
 * least that many nodes are allocated and waiting in reserve. <p>
 *
 * When you use this method to increase the reserve, you are
 * guaranteed that at least <code>reserve</code> count of nodes will
 * be allocated and reserved for future use. <p>
 *
 * This is useful because <code>cgul_bitwise_trie__insert()</code> can
 * throw an exception if it cannot allocate a node to hold the
 * index/value pair.  Thus, by allocating and reserving a node now, we
 * know that <code>cgul_bitwise_trie__insert()</code> will be able to
 * run without error later when the node is "unreserved".  Normally,
 * this is not an issue because you can almost always just insert at
 * the moment when it is requested, but if the
 * <code>cgul_bitwise_trie</code> is busy at that moment (perhaps it
 * is iterating over its nodes), you'll have to queue insert (and
 * remove) requests for later execution. <p>
 *
 * To "unreserve" a node so that it can be used, you simply call this
 * method again with a value for <code>reserve</code> that is less
 * than the current reserve limit. <p>
 *
 * If an error occurs while allocating the reserved nodes, an
 * exception is thrown. <p>
 *
 * WARNING: If you reserve nodes without "unreserving" them, you will
 * introduce a memory leak that is difficult to track down because
 * this class is careful to free all of the cached nodes (including
 * the reserved nodes) when it is deleted, but the reserved nodes will
 * just sit around in memory until that time.  Thus, after reserving
 * nodes by calling this method, you must remember to timely set the
 * reserve level back to zero in order to make those nodes available.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @param[in] reserve of the cache of nodes
 * @see cgul_cache__set_reserve
 */
extern CGUL_EXPORT void
cgul_bitwise_trie__set_cache_reserve(cgul_exception_t* cex,
                                     cgul_bitwise_trie_t t,
                                     unsigned long int reserve);

/**
 * Return 1 if the trie is empty; otherwise, return 0.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @return whether the trie is empty
 */
extern CGUL_EXPORT int
cgul_bitwise_trie__is_empty(cgul_exception_t* cex,
                            cgul_bitwise_trie_t t);

/**
 * Assign <code>ptr</code> to <code>t[index]</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @param[in] index index into <code>t</code>
 * @param[in] ptr pointer to assign to <code>t[index]</code>
 */
extern CGUL_EXPORT void
cgul_bitwise_trie__assign(cgul_exception_t* cex,
                          cgul_bitwise_trie_t t,
                          unsigned long int index,
                          void* ptr);

/**
 * This method returns the element at <code>t[index]</code>.  If
 * <code>index</code> cannot be found in the trie, <code>NULL</code>
 * is returned. <p>
 *
 * If and only if <code>exists</code> is not <code>NULL</code>, this
 * method will set it to <code>1</code> if a node exists for
 * <code>index</code> and to <code>0</code> otherwise.  This lets the
 * client distinguished between a <code>NULL</code> value that is
 * returned because <code>index</code> could not be found and a
 * <code>NULL</code> value that is returned because a it is being
 * stored in the trie. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @param[in] index index into <code>t</code>
 * @param[out] exists whether a node exists for <code>index</code>
 * @return element at <code>t[index]</code>
 */
extern CGUL_EXPORT void*
cgul_bitwise_trie__at(cgul_exception_t* cex,
                      cgul_bitwise_trie_t t,
                      unsigned long int index,
                      int* exists);

/**
 * Remove the node in <code>t</code> associated with <code>index</code>.
 * The value pointer stored in the node that is to be removed will be
 * returned in <code>value_out</code> if you pass in a pointer that is
 * not <code>NULL</code>.  This method returns 1 if the node was
 * removed; it returns 0 otherwise.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @param[in] index index into <code>t</code>
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_bitwise_trie__remove(cgul_exception_t* cex,
                          cgul_bitwise_trie_t t,
                          unsigned long int index,
                          void** value_out);

/**
 * Return the oldest node according to chronological order (i.e., the
 * order in which the nodes are inserted).  This operation is O(1).
 * If the trie is empty, <code>NULL</code> is returned.  This class
 * does not have to search the trie in order to find the oldest node
 * because it efficiently keeps a direct pointer to the oldest node up
 * to date. <p>
 *
 * The following example shows how to iterate over the entire trie in
 * chronological order:
 *
 * <pre>
 *     cgul_bitwise_trie_node_t n = cgul_bitwise_trie__get_oldest(cex, t);
 *     for ( ; n ; n = cgul_bitwise_trie_node__get_younger(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @return oldest node
 * @see cgul_bitwise_trie_node__get_younger()
 */
extern CGUL_EXPORT cgul_bitwise_trie_node_t
cgul_bitwise_trie__get_oldest(cgul_exception_t* cex,
                              cgul_bitwise_trie_t t);

/**
 * Set the oldest node in the trie to be <code>n</code>.  This method
 * could be used, for example, if your code expires the oldest node in
 * the trie, and you want to force <code>n</code> to be the next node
 * to expire. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @param[in] n node
 * @see cgul_bitwise_trie__set_youngest()
 */
extern CGUL_EXPORT void
cgul_bitwise_trie__set_oldest(cgul_exception_t* cex,
                              cgul_bitwise_trie_t t,
                              cgul_bitwise_trie_node_t n);

/**
 * Return the youngest node according to chronological order (i.e.,
 * the order in which the nodes are inserted).  This operation is
 * O(1).  If the trie is empty, <code>NULL</code> is returned.  This
 * class does not have to search the trie in order to find the
 * youngest node because it efficiently keeps a direct pointer to the
 * youngest node up to date. <p>
 *
 * The following example shows how to iterate over the entire trie in
 * reverse chronological order:
 *
 * <pre>
 *     cgul_bitwise_trie_node_t n = cgul_bitwise_trie__get_youngest(cex, t);
 *     for ( ; n ; n = cgul_bitwise_trie_node__get_older(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @return youngest node
 * @see cgul_bitwise_trie_node__get_older()
 */
extern CGUL_EXPORT cgul_bitwise_trie_node_t
cgul_bitwise_trie__get_youngest(cgul_exception_t* cex,
                                cgul_bitwise_trie_t t);

/**
 * Set the youngest node in the trie to be <code>n</code>.  This
 * method could be used, for example, if your code expires the
 * least-recently used (LRU) node.  By calling
 * <code>cgul_bitwise_trie__set_youngest()</code> each time a node is
 * used, the least-recently used node will be the oldest node in the
 * trie. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @param[in] n node
 * @see cgul_bitwise_trie__set_oldest()
 */
extern CGUL_EXPORT void
cgul_bitwise_trie__set_youngest(cgul_exception_t* cex,
                                cgul_bitwise_trie_t t,
                                cgul_bitwise_trie_node_t n);

/**
 * Return the total number of elements stored in <code>t</code>.  This
 * value can be less than than the maximum index stored in
 * <code>t</code> because in general <code>cgul_bitwise_trie</code>
 * instances are sparse.
 *
 * @param[in] cex c-style exception
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 * @return size of the trie
 */
extern CGUL_EXPORT unsigned long int
cgul_bitwise_trie__get_size(cgul_exception_t* cex,
                            cgul_bitwise_trie_t t);

/**
 * Swap the underlying data for <code>t1</code> and <code>t2</code>.
 * For large bitwise tries, this should be much faster than trying to
 * do the same thing using removes and inserts.
 *
 * @param[in] cex c-style exception
 * @param[in] t1 first <code>cgul_bitwise_trie</code> instance
 * @param[in] t2 second <code>cgul_bitwise_trie</code> instance
 */
extern CGUL_EXPORT void
cgul_bitwise_trie__swap(cgul_exception_t* cex,
                        cgul_bitwise_trie_t t1,
                        cgul_bitwise_trie_t t2);

CGUL_END_C

#endif  /*  CGUL_BITWISE_TRIE_H  */
