#ifndef CGUL_BLOCK_READER_FILE_H
#define CGUL_BLOCK_READER_FILE_H

/**
 * Implementation of the <code>cgul_block_reader</code> interface for
 * standard C <code>FILE</code> objects.
 *
 * @file
 * @author Paul Serice
 * @brief file block reader
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_block_reader.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_block_reader_file</code> instance. */
typedef struct cgul_block_reader_file* cgul_block_reader_file_t;

/**
 * Safely upcast <code>brfile</code> to the
 * <code>cgul_block_reader</code> interface.
 *
 * @param[in] cex c-style exception
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 * @see cgul_block_reader
 */
extern CGUL_EXPORT cgul_block_reader_t
cgul_block_reader_file__as_block_reader(cgul_exception_t* cex,
                                        cgul_block_reader_file_t brfile);

/**
 * Create a new <code>cgul_block_reader_file</code> instance.  After
 * this method returns,
 * <code>cgul_block_reader_file__open_fname()</code> or
 * <code>cgul_block_reader_file__open_file()</code> should be called
 * before calling
 * <code>cgul_block_reader_file__read_next_block()</code>.  The client
 * is responsible for calling
 * <code>cgul_block_reader_file__delete()</code> on the object
 * returned.  If an error occurs, <code>NULL</code> is returned, and
 * an exception is thrown. <p>
 *
 * @note It is safe to cast the object returned directly to a
 * <code>cgul_block_reader</code> (i.e., to the base type). <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_block_reader_file</code> instance
 */
extern CGUL_EXPORT cgul_block_reader_file_t
cgul_block_reader_file__new(cgul_exception_t* cex);

/**
 * This method creates a new <code>cgul_block_reader_file</code>
 * instance and calls
 * <code>cgul_block_reader_file__open_fname()</code> passing it
 * <code>fname</code>.  The file will be closed when this instance is
 * deleted.  The client is responsible for calling
 * <code>cgul_block_reader_file__delete()</code> on the object
 * returned.  If an error occurs, <code>NULL</code> is returned, and
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname file name
 * @return new <code>cgul_block_reader_file</code> instance
 */
extern CGUL_EXPORT cgul_block_reader_file_t
cgul_block_reader_file__new_from_fname(cgul_exception_t* cex,
                                       const char* fname);

/**
 * This method creates a new <code>cgul_block_reader_file</code>
 * instance and calls <code>cgul_block_reader_file__open_file()</code>
 * passing it <code>f</code>.  The class does not take ownership of
 * <code>f</code>.  Thus, the client is still responsible for calling
 * <code>fclose()</code> on it.  The client is also responsible for
 * calling <code>cgul_block_reader_file__delete()</code> on the object
 * returned.  If an error occurs, <code>NULL</code> is returned, and
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] f file
 * @return new <code>cgul_block_reader_file</code> instance
 */
extern CGUL_EXPORT cgul_block_reader_file_t
cgul_block_reader_file__new_from_file(cgul_exception_t* cex,
                                      FILE* f);

/**
 * This virtual method deletes the <code>brfile</code> instance
 * freeing all internally allocated resources.  This does not include
 * closing the underlying file if it was obtained from the client by a
 * call to <code>cgul_block_reader_file__new_from_file()</code> or
 * <code>cgul_block_reader_file__open_file()</code>.  The client must
 * not use <code>brfile</code> after calling this method.
 *
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 */
extern CGUL_EXPORT void
cgul_block_reader_file__delete(cgul_block_reader_file_t brfile);

/**
 * Open the file with name <code>fname</code> and use it when reading
 * blocks of data.  If a data source is already open, it will be
 * closed before this method attempts to open the new data source.
 * The new file will be closed when this instance is deleted.  If an
 * error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 * @param[in] fname file name
 */
extern CGUL_EXPORT void
cgul_block_reader_file__open_fname(cgul_exception_t* cex,
                                   cgul_block_reader_file_t brfile,
                                   const char* fname);

/**
 * Use the file <code>f</code> when reading blocks of data.  If a data
 * source is already open, it is closed before attempting to open the
 * new data source.  This class does not take ownership of
 * <code>f</code>.  Thus, the client is still responsible for calling
 * <code>fclose()</code> on <code>f</code>.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 * @param[in] f file
 */
extern CGUL_EXPORT void
cgul_block_reader_file__open_file(cgul_exception_t* cex,
                                  cgul_block_reader_file_t brfile,
                                  FILE* f);

/**
 * Close the file.  After this method returns,
 * <code>cgul_block_reader_file__open_fname()</code> or
 * <code>cgul_block_reader_file__open_file()</code> should be called
 * before calling any other method.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 */
extern CGUL_EXPORT void
cgul_block_reader_file__close(cgul_exception_t* cex,
                              cgul_block_reader_file_t brfile);

/**
 * Virtual method used to read one block of data <code>block</code> of
 * size <code>block_size</code> bytes from the underlying file.  The
 * user is responsible for making sure <code>block</code> points to at
 * least <code>block_size</code> contiguous bytes of memory.  This
 * method returns the number of bytes read or <code>0</code> if EOF is
 * reached without error.  If an error occurs, <code>0</code> is
 * returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 * @param[in] block block
 * @param[in] block_size block size in bytes
 * @return number of bytes read
 */
extern CGUL_EXPORT size_t
cgul_block_reader_file__read_next_block(cgul_exception_t* cex,
                                        cgul_block_reader_file_t brfile,
                                        void* block,
                                        size_t block_size);

/**
 * Return the name of underlying file.  If
 * <code>cgul_block_reader_file__new_from_file()</code> or
 * <code>cgul_block_reader_file__open_file()</code> was used to
 * initialize the object, <code>NULL</code> is returned instead.  This
 * method throws an exception if no file is currently open.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 * @return name of underlying file or <code>NULL</code>
 */
extern CGUL_EXPORT const char*
cgul_block_reader_file__get_fname(cgul_exception_t* cex,
                                  cgul_block_reader_file_t brfile);

/**
 * Virtual method that returns the presentation name.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 * @return presentation name
 */
extern CGUL_EXPORT const char*
cgul_block_reader_file__get_presentation_name(cgul_exception_t* cex,
                                              cgul_block_reader_file_t brfile);

/**
 * Virtual method that sets the presentation name.  The presentation
 * name is primarily used for error reporting and can be set to
 * anything, but it does need to be set again if the input file
 * changes.  If the presentation name is not set, <code>"FILE"</code>
 * is returned.  An exception is thrown if an attempt is made to set
 * the presentation name without first setting the input file.
 *
 * @param[in,out] cex c-style exception
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 * @param[in] presentation_name presentation name
 */
extern CGUL_EXPORT void
cgul_block_reader_file__set_presentation_name(
    cgul_exception_t* cex,
    cgul_block_reader_file_t brfile,
    const char* presentation_name);

/**
 * Return the currently opened file.  The client must not do anything
 * that would invalidate the pointer or corrupt the associated data
 * stream.  This method throws an exception if no file is currently
 * open.
 *
 * @param[in] cex c-style exception
 * @param[in] brfile <code>cgul_block_reader_file</code> instance
 * @return currently opened file or memory
 */
extern CGUL_EXPORT FILE*
cgul_block_reader_file__get_file(cgul_exception_t* cex,
                                 cgul_block_reader_file_t brfile);

CGUL_END_C

#endif  /*  CGUL_BLOCK_READER_FILE_H  */
