#ifndef CGUL_CRC32_INCR_H
#define CGUL_CRC32_INCR_H

/**
 * This class is a wrapper around the <code>cgul_crc32</code> class.
 * It is designed so that CRC-32 values can be calculated
 * incrementally.  The problem is that <code>cgul_crc32__crc()</code>
 * must be called to calculate the final CRC-32 value, but this
 * operation destroys the previous state making it otherwise
 * impossible to continue a CRC-32 calculation based on data already
 * consumed.  Without anything else, you would have to re-calculate
 * the entire CRC-32 value from scratch.  This class lets you
 * incrementally calculate CRC-32 values by saving the intermediate
 * state between calls to <code>cgul_crc32__crc()</code>.
 *
 * @file
 * @brief incremental CRC-32
 * @see cgul_crc32
 */

#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_crc32_incr</code> instance. */
typedef struct cgul_crc32_incr* cgul_crc32_incr_t;

/**
 * Create a new <code>cgul_crc32_incr</code> object.  Objects created
 * with this constructor do not append the size to the data stream
 * that is CRCed.  The client is responsible for calling
 * <code>cgul_crc32_incr__delete()</code> on the object returned.  If
 * an error occurs, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] g generator polynomial
 * @param[in] init initial value loaded into the accumulator
 * @param[in] final_xor value to xor with the calculated crc
 * @param[in] reflected whether to use the reflected algorithm
 * @return new <code>cgul_crc32_incr</code> instance
 * @see cgul_crc32__new()
 */
extern CGUL_EXPORT cgul_crc32_incr_t
cgul_crc32_incr__new(cgul_exception_t* cex,
                     unsigned long g,
                     unsigned long init,
                     unsigned long final_xor,
                     int reflected);

/**
 * Create a new <code>cgul_crc32_incr</code> object.  Objects created with
 * this constructor append the size to the data stream that is CRCed.
 * The client is responsible for calling
 * <code>cgul_crc32_incr__delete()</code> on the object returned.  If an
 * error occurs, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] g polynomial
 * @param[in] init initial value loaded into the accumulator
 * @param[in] final_xor value to xor with the calculated crc
 * @param[in] reflected whether to use the reflected algorithm
 * @param[in] size_xor value to xor with the size before appending
 * @return new <code>cgul_crc32_incr</code> instance
 * @see cgul_crc32__new_appends_size()
 */
extern CGUL_EXPORT cgul_crc32_incr_t
cgul_crc32_incr__new_appends_size(cgul_exception_t* cex,
                                  unsigned long g,
                                  unsigned long init,
                                  unsigned long final_xor,
                                  int reflected,
                                  size_t size_xor);

/**
 * This function frees all internally allocated resources associated
 * with the <code>crc</code> instance.  The client must not attempt to
 * use <code>crc</code> after calling this method.
 *
 * @param[in] crc <code>cgul_crc32_incr</code> instance
 */
extern CGUL_EXPORT void
cgul_crc32_incr__delete(cgul_crc32_incr_t crc);

/**
 * Assign the current state of of <code>rhs</code> to
 * <code>lhs</code>.  This operation probably only makes sense if
 * <code>lhs</code> and <code>rhs</code> have been constructed
 * identically.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] lhs left-hand side
 * @param[in] rhs right-hand side
 */
extern CGUL_EXPORT void
cgul_crc32_incr__assign(cgul_exception_t* cex,
                        cgul_crc32_incr_t lhs,
                        cgul_crc32_incr_t rhs);

/**
 * This method completes the CRC calculation by performing the size
 * XOR and the final XOR.  Unlike <code>cgul_crc32__crc()</code>, this
 * method does not reset the CRC-32 calculation.  Instead, the current
 * CRC-32 calculation can be continued with subsequent calls to
 * <code>cgul_crc32_incr__update_block()</code>.  The return value is
 * the current value for the entire CRC-32 calculation.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc32_incr</code> instance
 * @return final crc value
 */
extern CGUL_EXPORT unsigned long
cgul_crc32_incr__crc(cgul_exception_t* cex,
                     cgul_crc32_incr_t crc);

/**
 * This method updates the CRC by adding all the bytes in
 * <code>block</code> to the calculation.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc32_incr</code> instance
 * @param[in] block block of data to add to the crc
 * @param[in] length length of <code>block</code>
 */
extern CGUL_EXPORT void
cgul_crc32_incr__update_block(cgul_exception_t* cex,
                              cgul_crc32_incr_t crc,
                              const void* block,
                              size_t length);

/**
 * This method updates the CRC by adding all the bytes (starting at
 * the current offset) in the file <code>f</code> to the calculation.
 * If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc32_incr</code> instance
 * @param[in] f input file
 */
extern CGUL_EXPORT void
cgul_crc32_incr__update_file(cgul_exception_t* cex,
                             cgul_crc32_incr_t crc,
                             FILE* f);

/**
 * This method updates the CRC by adding all the bytes in the file
 * with name <code>fname</code> to the calculation.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] crc <code>cgul_crc32_incr</code> instance
 * @param[in] fname input file name
 */
extern CGUL_EXPORT void
cgul_crc32_incr__update_fname(cgul_exception_t* cex,
                              cgul_crc32_incr_t crc,
                              const char* fname);

/**
 * Reset to the initial internal state.
 *
 * @param[in] cex c-style exception
 * @param[in] crc <code>cgul_crc32_incr</code> instance
 */
extern CGUL_EXPORT void
cgul_crc32_incr__reset(cgul_exception_t* cex,
                       cgul_crc32_incr_t crc);

CGUL_END_C

#endif  /*  CGUL_CRC32_INCR_H  */
