#ifndef CGUL_HASH_NODE_H
#define CGUL_HASH_NODE_H

/**
 * This class implements a hash node which holds one key/value pair.
 *
 * @file
 * @author Paul Serice
 * @brief hash node
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

typedef struct cgul_hash_node* cgul_hash_node_t;

/**
 * Get the key for this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_hash_node_t</code> instance
 * @return key
 */
extern CGUL_EXPORT const void*
cgul_hash_node__get_key(cgul_exception_t* cex,
                        cgul_hash_node_t n);

/**
 * Get the value for this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_hash_node_t</code> instance
 * @return value
 */
extern CGUL_EXPORT void*
cgul_hash_node__get_value(cgul_exception_t* cex,
                          cgul_hash_node_t n);

/**
 * Set the value for this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_hash_node_t</code> instance
 * @param[in] value new value
 */
extern CGUL_EXPORT void
cgul_hash_node__set_value(cgul_exception_t* cex,
                          cgul_hash_node_t n,
                          const void* value);

/**
 * Get the previous node in insertion order.  This method is a synonym
 * for <code>cgul_hash_node__get_older()</code>.  You can use this
 * method to iterate over the hash in reverse insertion order.  When
 * there are no more nodes, <code>NULL</code> is returned.  You can
 * use <code>cgul_hash__get_back()</code> to get the last node in a
 * cgul_hash in order to start iterating.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_hash_node_t</code> instance
 * @return previous node
 * @see cgul_hash_node__get_older()
 */
extern CGUL_EXPORT cgul_hash_node_t
cgul_hash_node__get_prev(cgul_exception_t* cex,
                         cgul_hash_node_t n);

/**
 * Get the next node in insertion order.  This method is a synonym for
 * <code>cgul_hash_node__get_younger()</code>.  You can use this
 * method to iterate over the hash in insertion order.  When there are
 * no more nodes, <code>NULL</code> is returned.  You can use
 * <code>cgul_hash__get_front()</code> to get the first node in a
 * cgul_hash in order to start iterating.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_hash_node_t</code> instance
 * @return previous node
 * @see cgul_hash_node__get_younger()
 */
extern CGUL_EXPORT cgul_hash_node_t
cgul_hash_node__get_next(cgul_exception_t* cex,
                         cgul_hash_node_t n);

/**
 * Return the next older node.  Together with
 * <code>cgul_hash_node__get_younger()</code>, this method lets you
 * traverse the hash in chronological order. If there is no older
 * node, this method returns <code>NULL</code>. <p>
 *
 * The following example shows how to iterate over the entire hash in
 * reverse chronological order.  Notice that you have to start with
 * the youngest node when calling
 * <code>cgul_hash_node__get_older()</code>:
 *
 * <pre>
 *     cgul_hash_node_t n = cgul_hash__get_youngest(cex, h);
 *     for ( ; n ; n = cgul_hash_node__get_older(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_hash_node</code> instance
 * @return older node
 */
extern CGUL_EXPORT cgul_hash_node_t
cgul_hash_node__get_older(cgul_exception_t* cex,
                          cgul_hash_node_t n);

/**
 * Return the next younger node.  Together with
 * <code>cgul_hash_node__get_older()</code>, this method lets you
 * traverse the hash in chronological order. If there is no younger
 * node, this method returns <code>NULL</code>. <p>
 *
 * The following example shows how to iterate over the entire hash in
 * chronological order.  Notice that you have to start with the oldest
 * node when calling <code>cgul_hash_node__get_younger()</code>:
 *
 * <pre>
 *     cgul_hash_node_t n = cgul_hash__get_oldest(cex, h);
 *     for ( ; n ; n = cgul_hash_node__get_younger(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_hash_node</code> instance
 * @return younger node
 */
extern CGUL_EXPORT cgul_hash_node_t
cgul_hash_node__get_younger(cgul_exception_t* cex,
                            cgul_hash_node_t n);

CGUL_END_C

#endif  /*  CGUL_HASH_NODE_H  */
