#ifndef CGUL_LIST_H
#define CGUL_LIST_H

/**
 * Linked list implementation.
 *
 * @file
 * @author Paul Serice
 * @brief linked list implementation
 */

#include "cgul_cache.h"
#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_list_node.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_list</code> instance. */
typedef struct cgul_list* cgul_list_t;

/**
 * This typedef is the interface for the comparison function used by
 * <code>cgul_list__sort()</code> and
 * <code>cgul_list__insert_sorted()</code>.  When you define your
 * comparison function, you should compare <code>value1</code> to
 * <code>value2</code> and return less than zero, zero, or greater
 * than zero if <code>value1</code> is less than, equal to, or greater
 * than <code>value2</code>.
 *
 * @param[in] value1 first value
 * @param[in] value2 second value
 * @return result of comparison
 */
typedef int (*cgul_list__compare_t)(const void* value1,
                                    const void* value2);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_list__foldl()</code>
 *     <code>cgul_list__foldr()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_list__fold_t)(cgul_exception_t* cex,
                                 void* value,
                                 void* data);

/**
 * This typedef is the interface for the callback function used by the
 * following methods: <p>
 *
 * <pre>
 *     <code>cgul_list__traverse()</code>
 *     <code>cgul_list__traverse_range()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n node
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_list__traverse_t)(cgul_exception_t* cex,
                                     cgul_list_t l,
                                     cgul_list_node_t n,
                                     void* data);

/**
 * Create a new <code>cgul_list</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_list__delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_list</code> instance
 */
extern CGUL_EXPORT cgul_list_t
cgul_list__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory associated with
 * <code>l</code>.  Do not try to access <code>l</code> after calling
 * this method.
 *
 * @param[in] l <code>cgul_list</code> instance
 * @see cgul_list__free_values()
 */
extern CGUL_EXPORT void
cgul_list__delete(cgul_list_t l);

/**
 * This method calls <code>free()</code> on all the values in the list
 * <code>l</code>.  Because this is such a common operation, it is an
 * exception to the rule that cgul containers never free values.  This
 * method should only ever be called immediately before calling
 * <code>cgul_list__delete()</code> because it otherwise invalidates
 * the list.  Because this function is basically an extension of
 * <code>cgul_list__delete()</code>, it does not accept a
 * <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] l <code>cgul_list</code> instance
 */
extern CGUL_EXPORT void
cgul_list__free_values(cgul_list_t l);

/**
 * Get the size of the cache of nodes. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @return size of the cache of nodes
 */
extern CGUL_EXPORT unsigned long int
cgul_list__get_cache_size(cgul_exception_t* cex,
                          cgul_list_t l);

/**
 * Set the size of the cache of nodes.
 *
 * For efficiency, the <code>cgul_list</code> can keep a cache of
 * nodes that can be reused.  In many situations, this can greatly
 * reduce the number of calls to <code>malloc()</code> which can
 * greatly improve performance and reduce memory fragmentation. <p>
 *
 * You can release all the cached nodes and disable caching of nodes
 * by setting the cache size to 0.  This is the default. <p>
 *
 * If an error occurs (while enlarging the cache), an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] size of the cache of nodes
 */
extern CGUL_EXPORT void
cgul_list__set_cache_size(cgul_exception_t* cex,
                          cgul_list_t l,
                          unsigned long int size);

/**
 * Return true if the list <code>l</code> is empty.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @return whether the list is empty
 */
extern CGUL_EXPORT int
cgul_list__is_empty(cgul_exception_t* cex,
                    cgul_list_t l);

/**
 * Return the first element on the list <code>l</code>.  You can use
 * <code>cgul_list_node_next()</code> to iterate over the list by
 * starting with the value returned by this method.  If the list is
 * empty, <code>NULL</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @return first element
 */
extern CGUL_EXPORT cgul_list_node_t
cgul_list__get_front(cgul_exception_t* cex,
                     cgul_list_t l);

/**
 * Return the last element on the list <code>l</code>.  You can use
 * <code>cgul_list_node_prev()</code> to iterate over the list by
 * starting with the value returned by this method.  If the list is
 * empty, <code>NULL</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @return last element
 */
extern CGUL_EXPORT cgul_list_node_t
cgul_list__get_back(cgul_exception_t* cex,
                    cgul_list_t l);

/**
 * Insert <code>value</code> at the front of the list <code>l</code>.
 * If successful, the inserted node is returned; otherwise,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] value value to add at the front of the list
 * @return node that holds <code>value</code>
 */
extern CGUL_EXPORT cgul_list_node_t
cgul_list__push_front(cgul_exception_t* cex,
                      cgul_list_t l,
                      const void* value);

/**
 * Remove the first node from the list.  Return the value that was
 * stored in node.  If the list is empty, <code>NULL</code> is
 * returned.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @return value that was stored in the front node
 */
extern CGUL_EXPORT void*
cgul_list__pop_front(cgul_exception_t* cex,
                     cgul_list_t l);

/**
 * Insert <code>value</code> at the back of the list <code>l</code>.
 * If successful, the return value is the inserted node; otherwise,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] value value to add at the back of the list
 * @return node that holds <code>value</code>
 */
extern CGUL_EXPORT cgul_list_node_t
cgul_list__push_back(cgul_exception_t* cex,
                     cgul_list_t l,
                     const void* value);

/**
 * Remove the last value from the list.  Return the value that was
 * stored in node.  If the list is empty, <code>NULL</code> is
 * returned.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @return value that was stored in the back node
 */
extern CGUL_EXPORT void*
cgul_list__pop_back(cgul_exception_t* cex,
                    cgul_list_t l);

/**
 * Move <code>n2</code> before <code>n1</code> in list <code>l</code>.
 * If <code>n1 == NULL</code>, move <code>n2</code> to before the
 * <code>NULL</code> at the end of the list.  <code>n1</code> and
 * <code>n2</code> must both already exist on the same list. <p>
 *
 * For example, to rotate the list wrapping the first element around
 * to the back of the list, do the following: <p>
 *
 * <pre>
 *     cgul_list__move_before_node(&local,
 *                                 lst,
 *                                 NULL,
 *                                 cgul_list__get_front(&local, lst));
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n1 insert before this node
 * @param[in] n2 node to insert
 */
extern CGUL_EXPORT void
cgul_list__move_before_node(cgul_exception_t* cex,
                            cgul_list_t l,
                            cgul_list_node_t n1,
                            cgul_list_node_t n2);

/**
 * Move <code>n2</code> after <code>n1</code> in list <code>l</code>.
 * If <code>n1 == NULL</code>, move <code>n2</code> to after the
 * <code>NULL</code> at the beginning of the list.  <code>n1</code>
 * and <code>n2</code> must both already exist on the same list. <p>
 *
 * For example, to rotate the list wrapping the last element around
 * to the front of the list, do the following: <p>
 *
 * <pre>
 *     cgul_list__move_after_node(&local,
 *                                lst,
 *                                NULL,
 *                                cgul_list__get_back(&local, lst));
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n1 insert after this node
 * @param[in] n2 node to insert
 */
extern CGUL_EXPORT void
cgul_list__move_after_node(cgul_exception_t* cex,
                           cgul_list_t l,
                           cgul_list_node_t n1,
                           cgul_list_node_t n2);

/**
 * Insert <code>value</code> before node <code>n</code> in list
 * <code>l</code>.  If <code>n == NULL</code>, then insert before the
 * <code>NULL</code> at the end of the list.  If successful, the
 * return value is the inserted node; otherwise, <code>NULL</code> is
 * returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n insert before this node
 * @param[in] value value to insert
 * @return node that holds <code>value</code>
 */
extern CGUL_EXPORT cgul_list_node_t
cgul_list__insert_before_node(cgul_exception_t* cex,
                              cgul_list_t l,
                              cgul_list_node_t n,
                              const void* value);

/**
 * Insert <code>value</code> after node <code>n</code> in list
 * <code>l</code>.  If <code>n == NULL</code>, then insert after the
 * <code>NULL</code> at the beginning of the list.  If successful, the
 * return value is the inserted node; otherwise, <code>NULL</code> is
 * returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n insert after this node
 * @param[in] value value to insert
 * @return node that holds <code>value</code>
 */
extern CGUL_EXPORT cgul_list_node_t
cgul_list__insert_after_node(cgul_exception_t* cex,
                             cgul_list_t l,
                             cgul_list_node_t n,
                             const void* value);

/**
 * Insert <code>value</code> into list <code>l</code> in sort order as
 * determined by the comparison function <code>f</code>.  If
 * successful, the return value is the inserted node; otherwise,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * This method assumes <code>l</code> is already sorted.  If not, call
 * <code>cgul_list__sort()</code> before calling this method. <p>
 *
 * This method is O(N).  Thus, this method is more efficient for a
 * single insert on an already sorted list than
 * <code>cgul_list__sort()</code> which is O(N log N), but calling
 * this method to perform N inserts is O(N^2) which is not as
 * efficient as N calls to <code>cgul_list__push_back()</code>
 * followed by a single call to <code>cgul_list__sort()</code>. <p>
 *
 * If you need to constantly maintain a sorted list,
 * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
 * better solutions. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] value value to insert
 * @param[in] f comparison function
 * @return node that holds <code>value</code>
 */
extern CGUL_EXPORT cgul_list_node_t
cgul_list__insert_sorted(cgul_exception_t* cex,
                         cgul_list_t l,
                         const void* value,
                         cgul_list__compare_t f);

/**
 * Sort the list <code>l</code> using the comparison function
 * <code>f</code>. <p>
 *
 * This method uses merge sort.  The main advantage of this method
 * over <code>cgul_merge_sort()</code> is that it can sort the list
 * in-line whereas <code>cgul_merge_sort</code> needs an extra copy of
 * the list. <p>
 *
 * This method is O(N log N).  Thus, this method is more efficient
 * when sorting a large number of items, but
 * <code>cgul_list__insert_sorted()</code> is O(N) so it is faster
 * when inserting just one item into an already sorted list. <p>
 *
 * If you need to constantly maintain a sorted list,
 * <code>cgul_rbtree</code> or <code>cgul_multimap</code> are usually
 * better solutions. <p>
 *
 * If you need to merge two sorted lists,
 * <code>cgul_list__merge()</code> is usually a better solution. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] f comparison function
 * @see cgul_merge_sort
 * @see cgul_list__merge
 */
extern CGUL_EXPORT void
cgul_list__sort(cgul_exception_t* cex,
                cgul_list_t l,
                cgul_list__compare_t f);

/**
 * Merge the sorted lists <code>l1</code> and <code>l2</code>
 * producing one large sorted list.  The sort order is determined by
 * the comparison function <code>f</code>.  When this method returns,
 * all the nodes will be in <code>l1</code> and <code>l2</code> will
 * be empty. <p>
 *
 * This method is extremely efficient doing its job in a single pass,
 * i.e., O(N).  This method should be the first thing that comes to
 * mind when you need to add a small, unsorted list to a large,
 * sorted list.  In this case, sort the small list using
 * <code>cgul_list__sort()</code> before calling this method to merge
 * the small list with the large list. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] l1 first list
 * @param[in] l2 second list
 * @param[in] f comparison function
 */
extern CGUL_EXPORT void
cgul_list__merge(cgul_exception_t* cex,
                 cgul_list_t l1,
                 cgul_list_t l2,
                 cgul_list__compare_t f);

/**
 * Reverse the order of the list.  It is usually more efficient,
 * however, to use either <code>cgul_list_node__get_prev()</code> to
 * iterate in reverse order or <code>cgul_list__push_front()</code> to
 * insert in reverse order.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 */
extern CGUL_EXPORT void
cgul_list__reverse(cgul_exception_t* cex,
                   cgul_list_t l);

/**
 * Remove all elements from the list. <p>
 *
 * Strictly as a convenience, this method is an exception to the rule
 * that cgul containers never free values.  If you pass in a
 * <code>values_cache</code> instance that is not <code>NULL</code>,
 * the values will be put back on their cache. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] values_cache values cache
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT void
cgul_list__clear(cgul_exception_t* cex,
                 cgul_list_t l,
                 cgul_cache_t values_cache);

/**
 * Remove the node <code>n</code> from the list <code>l</code>.  If
 * <code>value_out</code> is not <code>NULL</code>, the value held by
 * <code>n</code> is returned in <code>*value_out</code>. <p>
 *
 * It is almost always a mistake to naively call this method while
 * iterating over the tree.  The problem is that calling this
 * method invalidates the node making it impossible to call
 * <code>cgul_list_node__get_next()</code> afterward.  The
 * solution is simple.  Just call
 * <code>cgul_list_node__get_next()</code> before calling this
 * method:
 *
 * <pre>
 *
 *     curr = cgul_list__get_front(&local, l);
 *     for ( ; curr ; curr = next) {
 *         next = cgul_list_node__get_next(&local, curr);
 *         cgul_list__remove_node(&local, l, curr, &value_out);
 *         free(value_out);
 *     }
 *
 * </pre>
 *
 * Alternatively, you can use <code>cgul_list__remove_range()</code>,
 * <code>cgul_list__traverse()</code>, or
 * <code>cgul_list__traverse_range()</code>. <p>
 *
 * If you use one of the traversal methods, it is safe to call this
 * method naively.  For example, if your tree stores
 * <code>cgul_string</code> objects for its values, another way to
 * remove nodes is as follows: <p>
 *
 * <pre>
 *
 *     static int
 *     remove_node_cb(cgul_list_t l,
 *                    cgul_list_node_t n,
 *                    void* data)
 *     {
 *         cgul_exception_t local = NULL;
 *         cgul_string_t value_out = NULL;
 *         cgul_list__remove_node(&local, l, n, (void**)&value_out);
 *         cgul_string__delete(value_out);
 *         return 1;
 *     }
 *
 *     cgul_list__traverse(&local, l, &remove_node_cb, NULL);
 *
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n node to remove
 * @param[out] value_out value
 */
extern CGUL_EXPORT void
cgul_list__remove_node(cgul_exception_t* cex,
                       cgul_list_t l,
                       cgul_list_node_t n,
                       void** value_out);

/**
 * Remove the nodes between <code>n_begin</code> (inclusive) and
 * <code>n_end</code> (inclusive).  If <code>n_begin</code> is NULL,
 * start at the front of the list.  If <code>n_end</code> is NULL or
 * is not found, continue removing until the end of the list. <p>
 *
 * Strictly as a convenience, this method is an exception to the rule
 * that cgul containers never free values.  If you pass in a
 * <code>values_cache</code> instance that is not <code>NULL</code>,
 * the values will be put back on their cache. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] n_begin begin removing with and including this node
 * @param[in] n_end end removing with and including this node
 * @param[in] values_cache values cache
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT void
cgul_list__remove_range(cgul_exception_t* cex,
                        cgul_list_t l,
                        cgul_list_node_t n_begin,
                        cgul_list_node_t n_end,
                        cgul_cache_t values_cache);

/**
 * Remove all the nodes that have <code>value</code> for their value.
 * If at least one node was removed, this method returns
 * <code>1</code>; otherwise, it returns <code>0</code>. <p>
 *
 * Strictly as a convenience, this method is an exception to the rule
 * that cgul containers never free values.  If you pass in a
 * <code>values_cache</code> instance that is not <code>NULL</code>,
 * the values will be put back on their cache. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] value values to remove from the list
 * @param[in] values_cache values cache
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT int
cgul_list__remove_value(cgul_exception_t* cex,
                        cgul_list_t l,
                        const void* value,
                        cgul_cache_t values_cache);

/**
 * Return the size of the list, i.e., the number of elements in the
 * list.
 *
 * @param[in] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @return size of the list
 */
extern CGUL_EXPORT unsigned long int
cgul_list__get_size(cgul_exception_t* cex,
                    cgul_list_t l);

/**
 * Swap the underlying data for <code>l1</code> and <code>l2</code>.
 * For large lists, this should be much faster than trying to do the
 * same thing using removes and inserts.
 *
 * @param[in] cex c-style exception
 * @param[in] l1 first <code>cgul_list</code> instance
 * @param[in] l2 second <code>cgul_list</code> instance
 */
extern CGUL_EXPORT void
cgul_list__swap(cgul_exception_t* cex,
                cgul_list_t l1,
                cgul_list_t l2);

/**
 * This method performs a left fold of the list <code>l</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each value in the list starting at the front of the list
 * and iterating forward to the end of the list. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_list__foldl(cgul_exception_t* cex,
                 cgul_list_t l,
                 cgul_list__fold_t f,
                 void* data);

/**
 * This method performs a right fold of the list <code>l</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each value in the list starting at the back of the list
 * and iterating backward to the front of the list. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_list__foldr(cgul_exception_t* cex,
                 cgul_list_t l,
                 cgul_list__fold_t f,
                 void* data);

/**
 * Traverse all nodes passing each node to the function
 * <code>f</code>. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the list <code>l</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_list__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_list__traverse()</code> or
 * <code>cgul_list__traverse_range()</code> in order to iterate over
 * the list elements.  In fact, I would recommend that you use
 * <code>cgul_list_node__get_next()</code> for most of your iteration
 * needs.  If you need to remove nodes though, you should probably use
 * this method.  If you look at this method's source however, you will
 * see that the <code>cgul_list_node</code> class provides all the
 * public methods you need to safely remove nodes while you are
 * iterating over the list, but you do have to be careful.
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] f traversal callback function
 * @param[in] data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_list__traverse(cgul_exception_t* cex,
                    cgul_list_t l,
                    cgul_list__traverse_t f,
                    void* data);

/**
 * Traverse all nodes starting with <code>first</code> (inclusive) and
 * ending with <code>last</code> (inclusive) passing each node to the
 * function <code>f</code>.  If you know the first node, but are not
 * sure of the last node, just use <code>NULL</code> as the last node.
 * This will cause this method to iterate until it reaches the end of
 * the list.  You can than have <code>f</code> return 0 when it
 * determines that the last node has been reached. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the list <code>l</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_list__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * If <code>first</code> is <code>NULL</code>, iteration starts at the
 * beginning of the list.  If <code>last</code> is <code>NULL</code>,
 * iteration stops at the end of the list. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_list__traverse()</code> or
 * <code>cgul_list__traverse_range()</code> in order to iterate over
 * the list elements.  In fact, I would recommend that you use
 * <code>cgul_list_node__get_next()</code> for most of your
 * iteration needs.  If you need to remove nodes though, you should
 * probably use this method.  If you look at this method's source
 * however, you will see that the <code>cgul_list_node</code> class
 * provides all the public methods you need to safely remove nodes
 * while you are iterating over the list, but you do have to be
 * careful.
 *
 * @param[in,out] cex c-style exception
 * @param[in] l <code>cgul_list</code> instance
 * @param[in] first first node in range
 * @param[in] last last node in range
 * @param[in] f traversal callback function
 * @param[in] data client data passed to <code>f</code>
 * @see cgul_list__traverse
 */
extern CGUL_EXPORT void
cgul_list__traverse_range(cgul_exception_t* cex,
                          cgul_list_t l,
                          cgul_list_node_t first,
                          cgul_list_node_t last,
                          cgul_list__traverse_t f,
                          void* data);

CGUL_END_C

#endif  /*  CGUL_LIST_H  */
