#ifndef CGUL_MIME_H
#define CGUL_MIME_H

/**
 * Parser for multipart MIME messages.  This multipart MIME parser
 * should be sufficient for handling HTTP uploads.  It should scale
 * well being able to handle large, multipart MIME messages
 * efficiently.
 *
 * @file
 * @author Paul Serice
 * @brief parser for multipart MIME messages
 * @see cgul_base64
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/**
 * The sections of a multipart MIME file.
 */
typedef enum {

    /**
     * Invalid section type.
     */
    CGUL_MIME__SECTION_INVALID = 0,

    /**
     * Preamble section.  The "preamble" is defined in RFC 2046
     * section 5.1.1 as the part of the file "prior to the first
     * boundary delimiter line."  Note that when parsing an RFC 822
     * e-mail message, the e-mail headers will be part of the preamble
     * because the headers necessarily come before the first boundary
     * delimiter line.
     */
    CGUL_MIME__SECTION_PREAMBLE,

    /**
     * Header section.  This always indicates a header that
     * immediately follows a boundary delimiter line.  MIME allows for
     * empty headers in which case the length of the block will be
     * zero.
     */
    CGUL_MIME__SECTION_HEADER,

    /**
     * Body section.  This always indicates a body that immediately
     * follows a header that immediately follows a boundary delimiter
     * line.
     */
    CGUL_MIME__SECTION_BODY,

    /**
     * Epilogue section.  The "epilogue" is defined in RFC 2046
     * section 5.1.1 as the part of the file "following the final
     * boundary delimiter line."
     */
    CGUL_MIME__SECTION_EPILOGUE,

    /**
     * End of file.  The block associated with EOF is always empty.
     */
    CGUL_MIME__SECTION_EOF

} cgul_mime__section_t;

/**
 * This typedef is the interface for the callback function invoked by
 * <code>cgul_mime__parse()</code> to produce blocks from the original
 * MIME source so that the blocks can be parsed. <p>
 *
 * Functions that implement this interface should read bytes into the
 * block <code>block</code> that extends for <code>block_size</code>
 * bytes and return the number of bytes read into <code>block</code>.
 * If and only if EOF is reached, this function should return
 * <code>0</code>.  It is permissible to return a short read count
 * greater than zero.  If the parser still needs more data, it will
 * just invoke the callback again.  If an error occurs, this function
 * should throw an exception. <p>
 *
 * The value of <code>data</code> is the same as what was initially
 * passed into <code>cgul_mime__parse()</code> when the callback was
 * registered allowing the client to pass arbitrary data to the
 * producer. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] block block
 * @param[in] block_size block size
 * @param[in] data client data
 * @return number of bytes read into <code>block</code> or zero if EOF
 */
typedef size_t
(*cgul_mime__produce_t)(cgul_exception_t* cex,
                        char* block,
                        size_t block_size,
                        void* data);

/**
 * This typedef is the interface for the callback function invoked by
 * the following functions to consume the blocks generated by the
 * parser: <p>
 *
 * <pre>
 *     cgul_mime__parse()
 *     cgul_mime__parse_file()
 *     cgul_mime__parse_fname()
 *     cgul_mime__parse_memory()
 * </pre>
 *
 * Functions that implement this interface will be passed each block
 * <code>block</code> and its size in bytes <code>block_size</code>
 * from the MIME file as it is parsed along with the MIME section type
 * <code>section</code> from which the data originated.  The function
 * will also be passed the current part <code>part</code> of the
 * multipart MIME file that is being parsed.  The value for
 * <code>part</code> will be <code>0</code> for the preamble and the
 * epilogue, <code>1</code> for the first part, <code>2</code> for the
 * second part, and so on. <p>
 *
 * The value of <code>data</code> is the same as what was initially
 * passed in when the callback was registered allowing the client to
 * pass arbitrary data to the consumer.  If an error occurs, this
 * function should throw an exception. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] part MIME part of the multipart message
 * @param[in] section MIME section where the block originates
 * @param[in] block block
 * @param[in] block_size block size
 * @param[in] data client data
 * @return whether to continue
 */
typedef int
(*cgul_mime__consume_t)(cgul_exception_t* cex,
                        unsigned long int part,
                        cgul_mime__section_t section,
                        char* block,
                        size_t block_size,
                        void* data);

/**
 * This is the generic parsing function that works by parsing blocks
 * of data supplied by the <code>produce</code> function.  For each
 * block supplied by <code>produce</code>, it invokes the client's
 * callback function <code>consume</code> at least once passing in a
 * block that holds a subset of bytes from the current MIME message,
 * the size of the block, the MIME section type, and the client data
 * <code>consume_data</code>.  This function returns the total number
 * of MIME parts.  If an error occurs, <code>0</code> is returned, and
 * an exception is thrown. <p>
 *
 * The MIME parts are separated from each other by the MIME boundary
 * <code>boundary</code>.  Unfortunately, the boundary separators are
 * required to start with "--", but the "Content-Type" header
 * specifies the boundary without the implicit "--" prefix.  To make
 * it possible to directly pass the boundary read from the
 * "Content-Type" header into this function, <code>boundary</code>
 * must not include the implicit "--" prefix; instead, the "--" prefix
 * will automatically be inserted by this function. <p>
 *
 * According to the MIME standard, all EOL sequences should be "\r\n",
 * but as a practical matter, MIME messages often use the native EOL
 * sequence instead.  As a result, the client needs to pass in the
 * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
 * <code>NULL</code> this function will attempt to automatically
 * detect the EOL sequence by scanning the first 16K of the file.  If
 * it cannot automatically detect the EOL sequence, this function will
 * use "\r\n" as the EOL sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] boundary MIME boundary
 * @param[in] eol EOL sequence
 * @param[in] produce callback function to produce MIME blocks
 * @param[in] produce_data client data for <code>produce</code>
 * @param[in] consume callback function to consume parsed blocks
 * @param[in] consume_data client data for <code>consume</code>
 * @return total number of parts
 * @see cgul_mime__parse_memory()
 * @see cgul_mime__parse_file()
 */
extern CGUL_EXPORT unsigned long int
cgul_mime__parse(cgul_exception_t* cex,
                 const char* boundary,
                 const char* eol,
                 cgul_mime__produce_t produce,
                 void* produce_data,
                 cgul_mime__consume_t consume,
                 void* consume_data);

/**
 * This function is an adapter for <code>cgul_mime__parse()</code> for
 * parsing multipart MIME messages already resident in memory.  It
 * registers a special "producer" that allows the parser to directly
 * access the MIME message in memory rather than having to copy it
 * block-by-block into the parser's internal buffer. <p>
 *
 * This function works by parsing a multipart MIME message that is in
 * memory at <code>block</code> and extends for
 * <code>block_size</code> bytes.  The client's callback function
 * <code>consume</code> is invoked once for each MIME section passing
 * in a block that holds all the bytes from the current MIME section,
 * the size of the block, the MIME section type, and the client data
 * <code>consume_data</code>.  This function returns the total number
 * of MIME parts.  If an error occurs, <code>0</code> is returned, and
 * an exception is thrown. <p>
 *
 * Even though this function passes all the bytes from each MIME
 * section to the <code>consume</code> callback, it may be worthwhile
 * to write <code>consume</code> so that it can also be used with
 * <code>cgul_mime__parse_file()</code> which is similar but generally
 * requires multiple calls to <code>consume</code> in order to pass in
 * the same amount of data. <p>
 *
 * The MIME parts are separated from each other by the MIME boundary
 * <code>boundary</code>.  Unfortunately, the boundary separators are
 * required to start with "--", but the "Content-Type" header
 * specifies the boundary without the implicit "--" prefix.  To make
 * it possible to directly pass the boundary read from the
 * "Content-Type" header into this function, <code>boundary</code>
 * must not include the implicit "--" prefix; instead, the "--" prefix
 * will automatically be inserted by this function. <p>
 *
 * According to the MIME standard, all EOL sequences should be "\r\n",
 * but as a practical matter, MIME messages often use the native EOL
 * sequence instead.  As a result, the client needs to pass in the
 * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
 * <code>NULL</code> this function will attempt to automatically
 * detect the EOL sequence by scanning the first 16K of the file.  If
 * it cannot automatically detect the EOL sequence, this function will
 * use "\r\n" as the EOL sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] block block
 * @param[in] block_size block size
 * @param[in] boundary MIME boundary
 * @param[in] eol EOL sequence
 * @param[in] consume callback function to consume parsed blocks
 * @param[in] consume_data client data for <code>consume</code>
 * @return total number of parts
 * @see cgul_mime__split_memory()
 */
extern CGUL_EXPORT unsigned long int
cgul_mime__parse_memory(cgul_exception_t* cex,
                        char* block,
                        size_t block_size,
                        const char* boundary,
                        const char* eol,
                        cgul_mime__consume_t consume,
                        void* consume_data);

/**
 * This function is an adapter for <code>cgul_mime__parse()</code> for
 * parsing multipart MIME messages from a file.  It registers a
 * "producer" that copies the MIME message from file block-by-block
 * into the parser's internal buffer. <p>
 *
 * This function works by parsing a multipart MIME message in the file
 * <code>fin</code>.  For each block read from the file, it invokes
 * the client's callback function <code>consume</code> at least once
 * passing in a block that holds a subset of bytes from the current
 * MIME section, the size of the block, the MIME section type, and the
 * client data <code>consume_data</code>.  This function returns the
 * total number of MIME parts.  If an error occurs, <code>0</code> is
 * returned, and an exception is thrown. <p>
 *
 * The MIME parts are separated from each other by the MIME boundary
 * <code>boundary</code>.  Unfortunately, the boundary separators are
 * required to start with "--", but the "Content-Type" header
 * specifies the boundary without the implicit "--" prefix.  To make
 * it possible to directly pass the boundary read from the
 * "Content-Type" header into this function, <code>boundary</code>
 * must not include the implicit "--" prefix; instead, the "--" prefix
 * will automatically be inserted by this function. <p>
 *
 * According to the MIME standard, all EOL sequences should be "\r\n",
 * but as a practical matter, MIME messages often use the native EOL
 * sequence instead.  As a result, the client needs to pass in the
 * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
 * <code>NULL</code> this function will attempt to automatically
 * detect the EOL sequence by scanning the first 16K of the file.  If
 * it cannot automatically detect the EOL sequence, this function will
 * use "\r\n" as the EOL sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] fin input file
 * @param[in] boundary MIME boundary
 * @param[in] eol EOL sequence
 * @param[in] consume callback function to consume parsed blocks
 * @param[in] consume_data client data for <code>consume</code>
 * @return total number of parts
 * @see cgul_mime__split_file()
 */
extern CGUL_EXPORT unsigned long int
cgul_mime__parse_file(cgul_exception_t* cex,
                      FILE* fin,
                      const char* boundary,
                      const char* eol,
                      cgul_mime__consume_t consume,
                      void* consume_data);

/**
 * This function is an adapter for <code>cgul_mime__parse()</code> for
 * parsing multipart MIME messages from a file.  It registers a
 * "producer" that copies the MIME message from file block-by-block
 * into the parser's internal buffer. <p>
 *
 * This function works by parsing a multipart MIME message in the file
 * with name <code>fname</code>.  For each block read from the file,
 * it invokes the client's callback function <code>consume</code> at
 * least once passing in a block that holds a subset of bytes from the
 * current MIME section, the size of the block, the MIME section type,
 * and the client data <code>consume_data</code>.  This function
 * returns the total number of MIME parts.  If an error occurs,
 * <code>0</code> is returned, and an exception is thrown. <p>
 *
 * The MIME parts are separated from each other by the MIME boundary
 * <code>boundary</code>.  Unfortunately, the boundary separators are
 * required to start with "--", but the "Content-Type" header
 * specifies the boundary without the implicit "--" prefix.  To make
 * it possible to directly pass the boundary read from the
 * "Content-Type" header into this function, <code>boundary</code>
 * must not include the implicit "--" prefix; instead, the "--" prefix
 * will automatically be inserted by this function. <p>
 *
 * According to the MIME standard, all EOL sequences should be "\r\n",
 * but as a practical matter, MIME messages often use the native EOL
 * sequence instead.  As a result, the client needs to pass in the
 * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
 * <code>NULL</code> this function will attempt to automatically
 * detect the EOL sequence by scanning the first 16K of the file.  If
 * it cannot automatically detect the EOL sequence, this function will
 * use "\r\n" as the EOL sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname name of the input file
 * @param[in] boundary MIME boundary
 * @param[in] eol EOL sequence
 * @param[in] consume callback function to consume parsed blocks
 * @param[in] consume_data client data for <code>consume</code>
 * @return total number of parts
 * @see cgul_mime__split_fname()
 */
extern CGUL_EXPORT unsigned long int
cgul_mime__parse_fname(cgul_exception_t* cex,
                       const char* fname,
                       const char* boundary,
                       const char* eol,
                       cgul_mime__consume_t consume,
                       void* consume_data);

/**
 * This function is an adapter for
 * <code>cgul_mime__parse_memory()</code> that splits a multipart MIME
 * message in memory into multiple files on the host file system. <p>
 *
 * Split the multipart MIME message in memory starting at
 * <code>block</code> and extending for <code>block_size</code> bytes
 * into preamble, header, body, and epilogue sections.  The directory
 * where the files should be written is given by <code>dname</code>.
 * The name of the preamble is given by <code>pname</code>.  The names
 * for the header files will have the form <code>hname-%lu</code>.
 * The names for the body files will have the form
 * <code>bname-%lu</code> where "%lu" will be replaced by the part
 * number (starting with 1).  If any of <code>pname</code>,
 * <code>hname</code>, <code>bname</code>, or <code>ename</code> are
 * <code>NULL</code>, those files will not be written.  The name of
 * the epilogue file is given by <code>ename</code>.  This function
 * returns the total number of MIME parts.  If an error occurs,
 * <code>0</code> is returned, and an exception is thrown. <p>
 *
 * The MIME parts are separated from each other by the MIME boundary
 * <code>boundary</code>.  Unfortunately, the boundary separators are
 * required to start with "--", but the "Content-Type" header
 * specifies the boundary without the implicit "--" prefix.  To make
 * it possible to directly pass the boundary read from the
 * "Content-Type" header into this function, <code>boundary</code>
 * must not include the implicit "--" prefix; instead, the "--" prefix
 * will automatically be inserted by this function. <p>
 *
 * According to the MIME standard, all EOL sequences should be "\r\n",
 * but as a practical matter, MIME messages often use the native EOL
 * sequence instead.  As a result, the client needs to pass in the
 * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
 * <code>NULL</code> this function will attempt to automatically
 * detect the EOL sequence by scanning the MIME message.  If it cannot
 * automatically detect the EOL sequence, this function will use
 * "\r\n" as the EOL sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] block block
 * @param[in] block_size block size
 * @param[in] dname directory name where output files will be created
 * @param[in] pname name for the preamble file
 * @param[in] hname base name for header files
 * @param[in] bname base name for body files
 * @param[in] ename name for the epilogue file
 * @param[in] boundary MIME boundary
 * @param[in] eol EOL sequence
 * @return total number of parts
 * @see cgul_mime__parse_memory()
 */
extern CGUL_EXPORT unsigned long int
cgul_mime__split_memory(cgul_exception_t* cex,
                        char* block,
                        size_t block_size,
                        const char* dname,
                        const char* pname,
                        const char* hname,
                        const char* bname,
                        const char* ename,
                        const char* boundary,
                        const char* eol);

/**
 * This function is an adapter for
 * <code>cgul_mime__parse_file()</code> that splits a multipart MIME
 * message in a file into multiple files on the host file system. <p>
 *
 * Split the multipart MIME message in the file <code>fin</code> into
 * preamble, header, body, and epilogue sections.  The directory where
 * the files should be written is given by <code>dname</code>.  The
 * name of the preamble is given by <code>pname</code>.  The names for
 * the header files will have the form <code>hname-%lu</code>.  The
 * names for the body files will have the form <code>bname-%lu</code>
 * where "%lu" will be replaced by the part number (starting with 1).
 * If any of <code>pname</code>, <code>hname</code>,
 * <code>bname</code>, or <code>ename</code> are <code>NULL</code>,
 * those files will not be written.  The name of the epilogue file is
 * given by <code>ename</code>.  This function returns the total
 * number of MIME parts.  If an error occurs, <code>0</code> is
 * returned, and an exception is thrown. <p>
 *
 * The MIME parts are separated from each other by the MIME boundary
 * <code>boundary</code>.  Unfortunately, the boundary separators are
 * required to start with "--", but the "Content-Type" header
 * specifies the boundary without the implicit "--" prefix.  To make
 * it possible to directly pass the boundary read from the
 * "Content-Type" header into this function, <code>boundary</code>
 * must not include the implicit "--" prefix; instead, the "--" prefix
 * will automatically be inserted by this function. <p>
 *
 * According to the MIME standard, all EOL sequences should be "\r\n",
 * but as a practical matter, MIME messages often use the native EOL
 * sequence instead.  As a result, the client needs to pass in the
 * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
 * <code>NULL</code> this function will attempt to automatically
 * detect the EOL sequence by scanning the first 16K of the file.  If
 * it cannot automatically detect the EOL sequence, this function will
 * use "\r\n" as the EOL sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] fin MIME input file
 * @param[in] dname directory name where output files will be created
 * @param[in] pname name for the preamble file
 * @param[in] hname base name for header files
 * @param[in] bname base name for body files
 * @param[in] ename name for the epilogue file
 * @param[in] boundary MIME boundary
 * @param[in] eol EOL sequence
 * @return total number of parts
 * @see cgul_mime__parse_file()
 */
extern CGUL_EXPORT unsigned long int
cgul_mime__split_file(cgul_exception_t* cex,
                      FILE* fin,
                      const char* dname,
                      const char* pname,
                      const char* hname,
                      const char* bname,
                      const char* ename,
                      const char* boundary,
                      const char* eol);

/**
 * This function is an adapter for
 * <code>cgul_mime__parse_fname()</code> that splits a multipart MIME
 * message in a file into multiple files on the host file system. <p>
 *
 * Split the multipart MIME message in the file with name
 * <code>fname</code> into preamble, header, body, and epilogue
 * sections.  The directory where the files should be written is given
 * by <code>dname</code>.  The name of the preamble is given by
 * <code>pname</code>.  The names for the header files will have the
 * form <code>hname-%lu</code>.  The names for the body files will
 * have the form <code>bname-%lu</code> where "%lu" will be replaced
 * by the part number (starting with 1).  If any of
 * <code>pname</code>, <code>hname</code>, <code>bname</code>, or
 * <code>ename</code> are <code>NULL</code>, those files will not be
 * written.  The name of the epilogue file is given by
 * <code>ename</code>.  This function returns the total number of MIME
 * parts.  If an error occurs, <code>0</code> is returned, and an
 * exception is thrown. <p>
 *
 * The MIME parts are separated from each other by the MIME boundary
 * <code>boundary</code>.  Unfortunately, the boundary separators are
 * required to start with "--", but the "Content-Type" header
 * specifies the boundary without the implicit "--" prefix.  To make
 * it possible to directly pass the boundary read from the
 * "Content-Type" header into this function, <code>boundary</code>
 * must not include the implicit "--" prefix; instead, the "--" prefix
 * will automatically be inserted by this function. <p>
 *
 * According to the MIME standard, all EOL sequences should be "\r\n",
 * but as a practical matter, MIME messages often use the native EOL
 * sequence instead.  As a result, the client needs to pass in the
 * correct EOL sequence <code>eol</code>.  If <code>eol</code> is
 * <code>NULL</code> this function will attempt to automatically
 * detect the EOL sequence by scanning the first 16K of the file.  If
 * it cannot automatically detect the EOL sequence, this function will
 * use "\r\n" as the EOL sequence. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] fname name of MIME input file
 * @param[in] dname directory name where output files will be created
 * @param[in] pname name for the preamble file
 * @param[in] hname base name for header files
 * @param[in] bname base name for body files
 * @param[in] ename name for the epilogue file
 * @param[in] boundary MIME boundary
 * @param[in] eol EOL sequence
 * @return total number of parts
 * @see cgul_mime__parse_fname()
 */
extern CGUL_EXPORT unsigned long int
cgul_mime__split_fname(cgul_exception_t* cex,
                       const char* fname,
                       const char* dname,
                       const char* pname,
                       const char* hname,
                       const char* bname,
                       const char* ename,
                       const char* boundary,
                       const char* eol);

CGUL_END_C

#endif  /*  CGUL_MIME_H  */
