#ifndef CGUL_MULTIMAP_NODE_CXX_H
#define CGUL_MULTIMAP_NODE_CXX_H

/**
 * C++ wrapper for <code>cgul_multimap_node</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_multimap_node</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_multimap_node.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_multimap_node</code> objects.  The main purpose of this
 * class is to convert the C-style function calls and exception
 * handling in <code>cgul_multimap_node</code> into C++-style function
 * calls and exception handling.  This class must not be extended.
 *
 * @brief C++ bindings for <code>cgul_multimap_node</code>
 */
class cgul_multimap_node_cxx {

public:

    /*
     * IMPLEMENTATION NOTE:
     * ===================
     *
     * There are no real cgul_multimap_node_cxx objects.  Instead, this
     * class is carefully implemented so that "cgul_multimap_node"
     * objects (not "cgul_multimap_node_cxx" objects) can be directly
     * used to invoke the methods of this class.
     *
     * This works because this class has no associated data _and_ no
     * virtual methods which would cause a vtable lookup, and vtables
     * are typically embedded in the object itself.
     *
     * In summary, do _not_ add data or virtual functions to this
     * class.
     */

    friend class cgul_multimap_cxx;

    /**
     * Return the key stored in this node.  The key is used to
     * determine the sort order of the node.  If you want to set the
     * value of the key, there is no method to do so because changing
     * the value of a key requires changing the node's position in the
     * <code>cgul_multimap_cxx</code>.  Thus, if you need to change
     * the value of a key, you should delete the node from the
     * <code>cgul_multimap_cxx</code> and insert a new node with the
     * new value for the key.
     *
     * @return key
     */
    void*
    get_key() const
    {
        void* rv = NULL;
        cgul_multimap_node_t n = (cgul_multimap_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_multimap_node__get_key(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Return the value stored in this node.
     *
     * @return value
     */
    void*
    get_value() const
    {
        void* rv = NULL;
        cgul_multimap_node_t n = (cgul_multimap_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_multimap_node__get_value(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the value stored in this node.
     *
     * @param[in] value
     */
    void
    set_value(const void* value)
    {
        cgul_multimap_node_t n = (cgul_multimap_node_t)this;
        cgul_exception_t local = NULL;
        cgul_multimap_node__set_value(&local, n, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Return the previous node.  You can use this method to iterate over
     * the multimap.  When there are no more nodes, <code>NULL</code> is
     * returned.  You can use <code>get_back()</code> to
     * get the last node in a <code>cgul_multimap_cxx</code> in order to start
     * iterating.
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes. <p>
     *
     * The way the <code>older_is_greater</code> boolean works in this
     * implementation is that each key maps to a linked list of values.
     * The <code>compare</code> function only determines the relative
     * order of the keys.  To order the values, this implementation always
     * pushes newer values onto the front of the list.  Thus, the newest
     * value is always at the front of the list, and the oldest value is
     * always at the back of the list.  If <code>older_is_greater</code>
     * is true, the sort order is from front to back.  If
     * <code>older_is_greater</code> is false, the sort order is from back
     * to front. <p>
     *
     * For example, if you want to iterate over the values for a
     * particular key in the same order they were inserted, you would call
     * this method with <code>older_is_greater</code> set to
     * <code>0</code>.  This way, the older values are considered to be
     * less than the newer values. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @return previous node
     */
    cgul_multimap_node_cxx*
    get_prev(int older_is_greater) const
    {
        cgul_multimap_node_t rv = NULL;
        cgul_multimap_node_t n = (cgul_multimap_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_multimap_node__get_prev(&local, n, older_is_greater);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_multimap_node_cxx*)rv;
    }

    /**
     * Return the next node.  You can use this method to iterate over the
     * multimap.  When there are no more nodes, <code>NULL</code> is
     * returned.  You can use <code>get_front()</code> to
     * get the first node in a <code>cgul_multimap_cxx</code> in order to
     * start iterating.
     *
     * The primary sort criterion is the user's <code>compare</code>
     * function, of course, but the <code>older_is_greater</code> boolean
     * is used as a secondary sort criterion.  When a key maps to many
     * values, the <code>older_is_greater</code> boolean determines how to
     * sort the older nodes relative to the newer nodes. <p>
     *
     * The way the <code>older_is_greater</code> boolean works in this
     * implementation is that each key maps to a linked list of values.
     * The <code>compare</code> function only determines the relative
     * order of the keys.  To order the values, this implementation always
     * pushes newer values onto the front of the list.  Thus, the newest
     * value is always at the front of the list, and the oldest value is
     * always at the back of the list.  If <code>older_is_greater</code>
     * is true, the sort order is from front to back.  If
     * <code>older_is_greater</code> is false, the sort order is from back
     * to front. <p>
     *
     * For example, if you want to iterate over the values for a
     * particular key in the same order they were inserted, you would call
     * this method with <code>older_is_greater</code> set to
     * <code>0</code>.  This way, the older values are considered to be
     * less than the newer values. <p>
     *
     * @param[in] older_is_greater whether older nodes are greater
     * than newer nodes
     * @return next node
     */
    cgul_multimap_node_cxx*
    get_next(int older_is_greater) const
    {
        cgul_multimap_node_t rv = NULL;
        cgul_multimap_node_t n = (cgul_multimap_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_multimap_node__get_next(&local, n, older_is_greater);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_multimap_node_cxx*)rv;
    }

    /**
     * Return the next older node.  Together with
     * <code>get_younger()</code>, this method lets you traverse the
     * multimap in chronological order. If there is no older node,
     * this method returns <code>NULL</code>. <p>
     *
     * The following example shows how to iterate over the entire multimap
     * in reverse chronological order.  Notice that you have to start with
     * the youngest node when calling
     * <code>get_older()</code>:
     *
     * <pre>
     *     cgul_multimap_node_cxx* n = mm->get_youngest();
     *     for ( ; n ; n = n->get_older()) {
     *         ...
     *     }
     * </pre>
     *
     * @return older node
     */
    cgul_multimap_node_cxx*
    get_older() const
    {
        cgul_multimap_node_t rv = NULL;
        cgul_multimap_node_t n = (cgul_multimap_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_multimap_node__get_older(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_multimap_node_cxx*)rv;
    }

    /**
     * Return the next younger node.  Together with
     * <code>get_older()</code>, this method lets you traverse the
     * multimap in chronological order. If there is no younger node,
     * this method returns <code>NULL</code>. <p>
     *
     * The following example shows how to iterate over the entire multimap in
     * chronological order.  Notice that you have to start with the oldest
     * node when calling <code>get_younger()</code>:
     *
     * <pre>
     *     cgul_multimap_node_cxx* n = mm->get_oldest();
     *     for ( ; n ; n = n->get_younger()) {
     *         ...
     *     }
     * </pre>
     *
     * @return younger node
     */
    cgul_multimap_node_cxx*
    get_younger() const
    {
        cgul_multimap_node_t rv = NULL;
        cgul_multimap_node_t n = (cgul_multimap_node_t)this;
        cgul_exception_t local = NULL;
        rv = cgul_multimap_node__get_younger(&local, n);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return (cgul_multimap_node_cxx*)rv;
    }

    /**
     * Get the underlying <code>cgul_multimap_node</code> object.
     *
     * @return underlying object
     */
    cgul_multimap_node_t
    get_obj() const
    {
        /* Because of the special implementation of this class, "this"
         * is in fact the underlying object.  (See the comments at the
         * top of the file.) */
        return (cgul_multimap_node_t)this;
    }

private:

    /* Hide these. */
    cgul_multimap_node_cxx();
    cgul_multimap_node_cxx(const cgul_multimap_node_cxx&);
    ~cgul_multimap_node_cxx();
    cgul_multimap_node_cxx& operator=(const cgul_multimap_node_cxx&);

};

#endif  /*  CGUL_MULTIMAP_NODE_CXX_H  */
