#ifndef CGUL_PROGRESS_BAR_H
#define CGUL_PROGRESS_BAR_H

/**
 * This is a class for drawing text-based progress bars.  The
 * different styles of progress bar are defined by the
 * <code>cgul_progress_bar__style_t</code> typedef.
 *
 * @file
 * @author Paul Serice
 * @brief progress bar
 */

#include "cgul_common.h"
#include "cgul_exception.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_progress_bar</code> instance. */
typedef struct cgul_progress_bar* cgul_progress_bar_t;

/**
 * The style of progress bar to draw.
 */
typedef enum {

    /**
     * Simple style of progress bar.  It looks as shown below when
     * drawn at 100%.  This progress bar does not use any special
     * characters to draw.  Instead, it is just a steady progression
     * of characters until the entire bar is drawn.  The main
     * advantage of this progress bar is that it works over
     * connections with very low bandwidth.  It also means that when
     * the output from the progress bar is saved to a log file, what
     * you see is what you get. <p>
     *
     * <pre>
     *
     *     0...10...20...30...40...50...60...70...80...90...100%
     *
     * </pre>
     */
    CGUL_PROGRESS_BAR__SIMPLE,

    /**
     * Common style of progress bar.  This is the default progress
     * bar.  You will see similar progress bars in wget and BSD's ftp
     * client.  It looks as shown below when drawn at 100%.  This bar
     * requires a dedicated line so that it can overwrite itself for
     * each update.  Each update will change the numeric percentage
     * indicator, and the client can supply optional custom
     * <code>before</code> and <code>after</code> text that will also
     * be updated.  This bar can run in reverse if necessary.
     *
     * <pre>
     *
     *     before 100% [####################################] after
     *
     * </pre>
     */
    CGUL_PROGRESS_BAR__COMMON,

    /**
     * Invalid style.
     */
    CGUL_PROGRESS_BAR__INVALID

} cgul_progress_bar__style_t;

/**
 * Create a new <code>cgul_progress_bar</code> instance with a default
 * minimum value of 0 and a default maximum value of 100.  Also by
 * default, new instances of this class draw their progress indicator
 * to <code>stdout</code>.  If memory allocation fails,
 * <code>NULL</code> is returned, and an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_progress_bar</code> instance
 */
extern CGUL_EXPORT cgul_progress_bar_t
cgul_progress_bar__new(cgul_exception_t* cex);

/**
 * Delete the <code>bar</code> instance.  All internally allocated
 * resources will be freed.  The client must not use <code>bar</code>
 * after calling this method.
 *
 * @param[in] bar <code>cgul_progress_bar</code> instance
 */
extern CGUL_EXPORT void
cgul_progress_bar__delete(cgul_progress_bar_t bar);

/**
 * Reset <code>bar</code> so it can be reused.  This method does not
 * change the minimum or maximum values back to their default values.
 * It also does not change the current file where the progress
 * indicator will be drawn.  Calling this method is necessary only
 * when using <code>CGUL_PROGRESS_BAR__SIMPLE</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 */
extern CGUL_EXPORT void
cgul_progress_bar__reset(cgul_exception_t* cex,
                         cgul_progress_bar_t bar);

/**
 * Get the style of the progress bar.
 *
 * @param[in] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @return progress bar style
 */
extern CGUL_EXPORT cgul_progress_bar__style_t
cgul_progress_bar__get_style(cgul_exception_t* cex,
                             cgul_progress_bar_t bar);

/**
 * Set the style of the progress bar.  The default style is
 * <code>CGUL_PROGRESS_BAR__SIMPLE</code>.  If an invalid style is
 * specified, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @param[in] style progress bar style
 */
extern CGUL_EXPORT void
cgul_progress_bar__set_style(cgul_exception_t* cex,
                             cgul_progress_bar_t bar,
                             cgul_progress_bar__style_t style);

/**
 * Get the minimum value for the range of progress.
 *
 * @param[in] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @return minimum value
 */
extern CGUL_EXPORT double
cgul_progress_bar__get_min(cgul_exception_t* cex,
                           cgul_progress_bar_t bar);

/**
 * Set the minimum value for the range of progress.  By default, the
 * minimum value is 0.
 *
 * @param[in] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @param[in] min minimum value
 */
extern CGUL_EXPORT void
cgul_progress_bar__set_min(cgul_exception_t* cex,
                           cgul_progress_bar_t bar,
                           double min);

/**
 * Get the maximum value for the range of progress.
 *
 * @param[in] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @return maximum value
 */
extern CGUL_EXPORT double
cgul_progress_bar__get_max(cgul_exception_t* cex,
                           cgul_progress_bar_t bar);

/**
 * Set the maximum value for the range of progress.  By default, the
 * maximum value is 100.
 *
 * @param[in] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @param[in] max maximum value
 */
extern CGUL_EXPORT void
cgul_progress_bar__set_max(cgul_exception_t* cex,
                           cgul_progress_bar_t bar,
                           double max);

/**
 * Get the file where the progress indicator will be drawn.
 *
 * @param[in] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @return file
 */
extern CGUL_EXPORT FILE*
cgul_progress_bar__get_file(cgul_exception_t* cex,
                            cgul_progress_bar_t bar);

/**
 * Set the file where the progress indicator will be drawn.  By
 * default, the progress indicator is drawn to <code>stdout</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @param[in] f file
 */
extern CGUL_EXPORT void
cgul_progress_bar__set_file(cgul_exception_t* cex,
                            cgul_progress_bar_t bar,
                            FILE* f);

/**
 * Set the value used to indicated current progress.  This is the
 * method that makes the progress bar grow.  The value of
 * <code>progress</code> should be between the minimum and maximum
 * values set by <code>cgul_progress_bar__set_min()</code> and
 * <code>cgul_progress_bar__set_max()</code> respectively. <p>
 *
 * The actual percentage used when drawing the progress bar is rounded
 * to the nearest percentage with the exception that 100% is not used
 * until <code>progress</code> is greater than or equal to the maximum
 * value. <p>
 *
 * If <code>progress</code> is less than the minimum value for the
 * range, the minimum value will be used instead. If
 * <code>progress</code> is greater than the maximum value for the
 * range, the maximum value will be used instead. <p>
 *
 * If <code>before</code> or <code>after</code> is not
 * <code>NULL</code>, the value or values that are not
 * <code>NULL</code> will drawn in the correct location provided the
 * style of the progress bar allows it.  It is the responsibility of
 * the client to make sure that <code>before</code> and
 * <code>after</code> never get shorter; however, they can get
 * longer. <p>
 *
 * If an error occurs while drawing the progress bar, an exception is
 * throw.  If the minimum value is set greater than or equal to the
 * maximum value when this method is called, an exception is also
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] bar <code>cgul_progress_bar</code> instance
 * @param[in] progress progress
 * @param[in] before custom text to draw before the progress bar
 * @param[in] after custom text to draw after the progress bar
 */
extern CGUL_EXPORT void
cgul_progress_bar__update(cgul_exception_t* cex,
                          cgul_progress_bar_t bar,
                          double progress,
                          const char* before,
                          const char* after);

CGUL_END_C

#endif  /*  CGUL_PROGRESS_BAR_H  */
