#ifndef CGUL_PROGRESS_BAR_CXX_H
#define CGUL_PROGRESS_BAR_CXX_H

/**
 * C++ wrapper for <code>cgul_progress_bar</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_progress_bar</code>
 */

#include "cgul_progress_bar.h"
#include "cgul_exception_cxx.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_progress_bar</code> objects.  The main purpose of this
 * class is to convert the C-style function calls and exception
 * handling in <code>cgul_progress_bar</code> into C++-style function
 * calls and exception handling.
 *
 * @brief C++ bindings for <code>cgul_progress_bar</code>
 */
class cgul_progress_bar_cxx {

public:

    /**
     * Construct a new <code>cgul_progress_bar_cxx</code> object with
     * a default minimum value of 0 and a default maximum value of
     * 100.  Also by default, new instances of this class draw their
     * progress indicator to <code>stdout</code>.  If memory
     * allocation fails, an exception is thrown.
     */
    cgul_progress_bar_cxx() {
        cgul_exception_t local = NULL;
        obj = cgul_progress_bar__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method frees all internally allocated resources for the
     * object.
     */
    virtual
    ~cgul_progress_bar_cxx() {
        if (obj) {
            cgul_progress_bar__delete(obj);
        }
    }

    /**
     * Reset the object so it can be reused.  This method does not
     * change the minimum or maximum values back to their default values.
     * It also does not change the current file where the progress
     * indicator will be drawn.
     */
    virtual void
    reset() {
        cgul_exception_t local = NULL;
        cgul_progress_bar__reset(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the style of the progress bar.
     *
     * @return progress bar style
     */
    virtual cgul_progress_bar__style_t
    get_style() const {
        cgul_exception_t local = NULL;
        cgul_progress_bar__style_t rv =
            cgul_progress_bar__get_style(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the style of the progress bar.  The default style is
     * <code>CGUL_PROGRESS_BAR__SIMPLE</code>.  If an invalid style is
     * specified, an exception is thrown.
     *
     * @param[in] style progress bar style
     */
    virtual void
    set_style(cgul_progress_bar__style_t style) {
        cgul_exception_t local = NULL;
        cgul_progress_bar__set_style(&local, obj, style);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the minimum value for the range of progress.
     *
     * @return minimum value
     */
    virtual double
    get_min() const {
        cgul_exception_t local = NULL;
        double rv = cgul_progress_bar__get_min(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the minimum value for the range of progress.  By default, the
     * minimum value is 0.
     *
     * @param[in] min minimum value
     */
    virtual void
    set_min(double min) {
        cgul_exception_t local = NULL;
        cgul_progress_bar__set_min(&local, obj, min);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the maximum value for the range of progress.
     *
     * @return maximum value
     */
    virtual double
    get_max() const {
        cgul_exception_t local = NULL;
        double rv = cgul_progress_bar__get_max(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the maximum value for the range of progress.  By default, the
     * maximum value is 100.
     *
     * @param[in] max maximum value
     */
    virtual void
    set_max(double max) {
        cgul_exception_t local = NULL;
        cgul_progress_bar__set_max(&local, obj, max);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the file where the progress indicator will be drawn.
     *
     * @return file
     */
    virtual FILE*
    get_file() const {
        cgul_exception_t local = NULL;
        FILE* rv = cgul_progress_bar__get_file(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the file where the progress indicator will be drawn.  By
     * default, the progress indicator is drawn to <code>stdout</code>.
     *
     * @param[in] f file
     */
    virtual void
    set_file(FILE* f) {
        cgul_exception_t local = NULL;
        cgul_progress_bar__set_file(&local, obj, f);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Set the value used to indicated current progress.  This is the
     * method that makes the progress bar grow.  The value of
     * <code>progress</code> should be between the minimum and maximum
     * values set by <code>cgul_progress_bar__set_min()</code> and
     * <code>cgul_progress_bar__set_max()</code> respectively. <p>
     *
     * The actual percentage used when drawing the progress bar is rounded
     * to the nearest percentage with the exception that 100% is not used
     * until <code>progress</code> is greater than or equal to the maximum
     * value. <p>
     *
     * If <code>progress</code> is less than the minimum value for the
     * range, the minimum value will be used instead. If
     * <code>progress</code> is greater than the maximum value for the
     * range, the maximum value will be used instead. <p>
     *
     * If <code>before</code> or <code>after</code> is not
     * <code>NULL</code>, the value or values that are not
     * <code>NULL</code> will drawn in the correct location provided the
     * style of the progress bar allows it.  It is the responsibility of
     * the client to make sure that <code>before</code> and
     * <code>after</code> never get shorter; however, they can get
     * longer. <p>
     *
     * If an error occurs while drawing the progress bar, an exception is
     * throw.  If the minimum value is set greater than or equal to the
     * maximum value when this method is called, an exception is also
     * thrown. <p>
     *
     * @param[in] progress progress
     * @param[in] before custom text to draw before the progress bar
     * @param[in] after custom text to draw after the progress bar
     */
    virtual void
    update(double progress,
           const char* before = NULL,
           const char* after = NULL)
    {
        cgul_exception_t local = NULL;
        cgul_progress_bar__update(&local, obj, progress, before, after);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the underlying <code>cgul_progress_bar</code> object.
     *
     * @return underlying object
     */
    virtual cgul_progress_bar_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_progress_bar</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_progress_bar__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_progress_bar_t
    take_obj()
    {
        cgul_progress_bar_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_progress_bar_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_progress_bar__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /* Hide these. */
    cgul_progress_bar_cxx(const cgul_progress_bar_cxx&);
    cgul_progress_bar_cxx& operator=(const cgul_progress_bar_cxx&);

    /* Underlying <code>cgul_progress_bar</code> object. */
    cgul_progress_bar_t obj;

};

#endif  /*  CGUL_PROGRESS_BAR_CXX_H  */
