#ifndef CGUL_RANDOM_LFG32_H
#define CGUL_RANDOM_LFG32_H

/**
 * 32-bit lagged fibonacci pseudo-random number generator.  This PRNG
 * is a combination of a LFG with a period of (2^552 - 2^31) and a
 * Weyl generator with a period of (2^32 - 1).  The combined period is
 * a little less than 2^584.  This PRNG is the fastest in cgul.  It is
 * a good choice if <code>cgul_random_cmwc32</code> cannot be used for
 * some reason.  This PRNG requires about 2.1K to hold state
 * information.
 *
 * @file
 * @author Paul Serice
 * @brief 32-bit PRNG (LFG)
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_int.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_random_lfg32</code> instance. */
typedef struct cgul_random_lfg32* cgul_random_lfg32_t;

/**
 * This is a convenience function that creates a new
 * <code>cgul_random_lfg32</code> object seeding it with the value
 * returned by <code>time()</code>, i.e., the 32-bit seconds since the
 * epoch.  The caller is responsible for freeing the object by calling
 * <code>cgul_random_lfg32__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown.
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_random_lfg32</code> instance
 * @see cgul_random_lfg32__new_from_seed()
 * @see cgul_random_lfg32__new_from_seed_array()
 */
extern CGUL_EXPORT cgul_random_lfg32_t
cgul_random_lfg32__new(cgul_exception_t* cex);

/**
 * This is a convenience function that creates a new
 * <code>cgul_random_lfg32</code> object seeding it with
 * <code>seed</code>.  The caller is responsible for freeing the
 * object by calling <code>cgul_random_lfg32__delete()</code>.  If
 * memory cannot be allocated, <code>NULL</code> is returned, and an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] seed seed value
 * @return new <code>cgul_random_lfg32</code> instance
 * @see cgul_random_lfg32__new_from_seed_array()
 */
extern CGUL_EXPORT cgul_random_lfg32_t
cgul_random_lfg32__new_from_seed(cgul_exception_t* cex,
                                 cgul_uint32_t seed);

/**
 * This is the correct way to create a new
 * <code>cgul_random_lfg32</code> object by seeding it with the
 * <code>seeds</code> array.  It can be difficult, however, to acquire
 * the requisite entropy to fill the entire array.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_random_lfg32__delete()</code>.  If memory cannot be
 * allocated or if at least one seed value is not odd,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * The caller is responsible for making sure <code>seeds</code> has
 * the correct number of elements.  The easiest way to do this is by
 * using <code>cgul_random_lfg32__allocate_seed_array()</code> to
 * allocate the array and get its size. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] seeds array of 32-bit values used to seed the PRNG
 * @return new <code>cgul_random_lfg32</code> instance
 * @see cgul_random_lfg32__allocate_seed_array()
 */
extern CGUL_EXPORT cgul_random_lfg32_t
cgul_random_lfg32__new_from_seed_array(cgul_exception_t* cex,
                                       cgul_uint32_t* seeds);

/**
 * This method frees all internally allocated memory.  Do not attempt
 * to dereference <code>r</code> after calling this method.
 *
 * @param[in] r <code>cgul_random_lfg32</code> instance
 */
extern CGUL_EXPORT void
cgul_random_lfg32__delete(cgul_random_lfg32_t r);

/**
 * This function allocates and returns an array of 32-bit seed values
 * <code>v</code> where each element of the array is initialized to
 * zero.  The total number of 32-bit seed values in the array is
 * returned in <code>v_count</code>.  The caller should fill the array
 * with true random values where at least one random value is odd and
 * create a new PRNG seeded with the values by calling
 * <code>cgul_random_lfg32__new_from_seed_array()</code>.  The caller
 * is responsible for calling <code>free(*v)</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[out] v uninitialized array of 32-bit seed values
 * @param[out] v_count count of the number of 32-bit values in <code>v</code>
 */
extern CGUL_EXPORT void
cgul_random_lfg32__allocate_seed_array(cgul_exception_t* cex,
                                       cgul_uint32_t** v,
                                       unsigned int* v_count);

/**
 * Return the next element in the pseudo-random number sequence.  The
 * random numbers returned are 32 bits wide and should have a period a
 * little less than 2^584. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] r <code>cgul_random_lfg32</code> instance
 * @return next element in the random number sequence
 * @see cgul_random_lfg32__next_in_range()
 * @see cgul_random_lfg32__next_in_unit()
 */
extern CGUL_EXPORT cgul_uint32_t
cgul_random_lfg32__next(cgul_exception_t* cex,
                        cgul_random_lfg32_t r);

/**
 * If the value returned by <code>cgul_random_lfg32__next()</code> is
 * not the exact range that is needed, this method can limit the
 * return value to [0, range_max).  If <code>range_max</code> is zero,
 * an exception is thrown. <p>
 *
 * This function should avoid the very small round off error
 * associated with just scaling the value returned by
 * <code>cgul_random_lfg32__next_in_unit()</code>; however, this
 * function is much slower. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] r <code>cgul_random_lfg32</code> instance
 * @param[in] range_max exclusive limit on the maximum value returned
 * @return next element in the random number sequence
 */
extern CGUL_EXPORT cgul_uint32_t
cgul_random_lfg32__next_in_range(cgul_exception_t* cex,
                                 cgul_random_lfg32_t r,
                                 cgul_uint32_t range_max);

/**
 * Return the next element in the pseudo-random number sequence as
 * type <code>double</code> on the range [0 - 1). <p>
 *
 * @param[in] cex c-style exception
 * @param[in] r <code>cgul_random_lfg32</code> instance
 * @return next element in the random number sequence
 */
extern CGUL_EXPORT double
cgul_random_lfg32__next_in_unit(cgul_exception_t* cex,
                                cgul_random_lfg32_t r);

CGUL_END_C

#endif  /*  CGUL_RANDOM_LFG32_H  */
