#ifndef CGUL_STRING_CXX_H
#define CGUL_STRING_CXX_H

/**
 * C++ wrapper for <code>cgul_string</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_string</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_string.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_string</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_string</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_string</code>
 */
class cgul_string_cxx {

public:

    /**
     * Default constructor that creates a new instance with an empty
     * string.
     */
    cgul_string_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_string__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Copy constructor that creates a new instance by copying the
     * value from <code>rhs</code>.
     *
     * @param[in] rhs right-hand side
     */
    cgul_string_cxx(const cgul_string_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        obj = cgul_string__new_from_cstring(&local, rhs.get_value());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Copy constructor that creates a new instance by copying the
     * value from <code>rhs</code>.
     *
     * @param[in] rhs right-hand side
     */
    cgul_string_cxx(const char* rhs)
    {
        cgul_exception_t local = NULL;
        obj = cgul_string__new_from_cstring(&local, rhs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /** Destructor. */
    virtual
    ~cgul_string_cxx()
    {
        if (obj) {
            cgul_string__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Assign the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_string_cxx&
    operator=(const cgul_string_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        if (this != &rhs) {
            cgul_string__set_value(&local, obj, rhs.get_value());
            if (local) {
                throw cgul_exception_cxx(local);
            }
        }
        return *this;
    }

    /**
     * Assign the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_string_cxx&
    operator=(const char* rhs)
    {
        cgul_exception_t local = NULL;
        cgul_string__set_value(&local, obj, rhs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return *this;
    }

    /**
     * Append the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_string_cxx&
    operator+=(const cgul_string_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_string__append(&local, obj, rhs.get_value());
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return *this;
    }

    /**
     * Append the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_string_cxx&
    operator+=(const char* rhs)
    {
        cgul_exception_t local = NULL;
        cgul_string__append(&local, obj, rhs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return *this;
    }

    /**
     * Index operator that returns the character at
     * <code>index</code>.  If <code>index</code> is out of bounds, an
     * exception is thrown.
     */
    virtual char&
    operator[](size_t index)
    {
        char* value = NULL;
        cgul_exception_t local = NULL;

        /* Make sure "index" is within bounds.  Note that we use
         *
         *     index > length()
         *
         * instead of
         *
         *     index >= length()
         *
         * because we want the user to be able to access the trailing null
         * terminator.
         */
        if (index > cgul_string__get_length(&local, obj)) {
            local = cgul_exception__new("cgul_string_cxx::operator[]: "
                                        "index out of bounds: %lu",
                                        (unsigned long)index);
            throw cgul_exception_cxx(local);
        }

        /* Get the underlying character array. */
        value = cgul_string__get_value(&local, obj);

        /* Return the correct reference. */
        return value[index];
    }

    /**
     * Index operator that returns the character at
     * <code>index</code>.  If <code>index</code> is out of bounds, an
     * exception is thrown.  This is the overloaded
     * <code>operator[]()</code> that is necessary for use with
     * <code>const</code> objects.
     */
    virtual const char&
    operator[](size_t index) const
    {
        char* value = NULL;
        cgul_exception_t local = NULL;

        /* Make sure "index" is within bounds. */
        if (index >= cgul_string__get_length(&local, obj)) {
            local = cgul_exception__new("cgul_string_cxx::operator[]: "
                                        "index out of bounds: %lu",
                                        (unsigned long)index);
            throw cgul_exception_cxx(local);
        }

        /* Get the underlying character array. */
        value = cgul_string__get_value(&local, obj);

        /* Return the correct reference. */
        return value[index];
    }

    /**
     * This method allows instances of this class to be cast to
     * C-style strings.
     *
     * @return underlying c-style string
     */
    virtual
    operator char*() const
    {
        return get_value();
    }

    /**
     * This method compares this object with <code>rhs</code> using
     * the current locale.  It returns 1 if this object is equal to
     * <code>rhs</code>; otherwise, it returns 0.
     *
     * @param[in] rhs right-hand side
     * @return whether this object is equal to <code>rhs</code>
     */
    virtual int
    operator==(const cgul_string_cxx& rhs) const
    {
        return (this->compare(rhs) == 0);
    }

    /**
     * This method compares this object with <code>rhs</code> using
     * the current locale.  It returns 1 if this object is less than
     * <code>rhs</code>; otherwise, it returns 0.
     *
     * @param[in] rhs right-hand side
     * @return whether this object is less than <code>rhs</code>
     */
    virtual int
    operator<(const cgul_string_cxx& rhs) const
    {
        return (this->compare(rhs) < 0);
    }

    /**
     * This method compares this object with <code>rhs</code> using
     * the current locale.  It returns 1 if this object is less than
     * or equal to <code>rhs</code>; otherwise, it returns 0.
     *
     * @param[in] rhs right-hand side
     * @return whether this object is less than or equal to <code>rhs</code>
     */
    virtual int
    operator<=(const cgul_string_cxx& rhs) const
    {
        return (this->compare(rhs) <= 0);
    }

    /**
     * This method compares this object with <code>rhs</code> using
     * the current locale.  It returns 1 if this object is greater
     * than <code>rhs</code>; otherwise, it returns 0.
     *
     * @param[in] rhs right-hand side
     * @return whether this object is greater than <code>rhs</code>
     */
    virtual int
    operator>(const cgul_string_cxx& rhs) const
    {
        return (this->compare(rhs) > 0);
    }

    /**
     * This method compares this object with <code>rhs</code> using
     * the current locale.  It returns 1 if this object is greater than
     * or equal to <code>rhs</code>; otherwise, it returns 0.
     *
     * @param[in] rhs right-hand side
     * @return whether this object is greater than or equal to <code>rhs</code>
     */
    virtual int
    operator>=(const cgul_string_cxx& rhs) const
    {
        return (this->compare(rhs) >= 0);
    }

    /**
     * This method returns a pointer to the internal C-style string.
     * You are allowed to manually set individual characters in the
     * string.  Just remember to call
     * <code>externally_adjusted()</code> afterward.  Compare this
     * method to the <code>take_value()</code> method.
     *
     * @return underlying c-style string
     */
    virtual char*
    get_value() const
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_string__get_value(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method lets you take ownership of the underlying C-style
     * string which is returned.  Immediately upon taking ownership,
     * the internal state of this class is reset including resetting a
     * new underlying C-style string to be empty.  One use of this
     * function is to build up a string using
     * <code>cgul_string</code> that you then pass to
     * <code>cgul_hash</code> or <code>cgul_rbtree</code>.  Compare
     * this function with <code>get_value()</code>.
     *
     * @return underlying c-style string
     */
    virtual char*
    take_value()
    {
        char* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_string__take_value(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the value of the underlying c-style string to value.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] value new value
     */
    virtual void
    set_value(const char* value)
    {
        cgul_exception_t local = NULL;
        cgul_string__set_value(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method returns the character indicated by
     * <code>index</code> where <code>index</code> is zero-based.  If
     * <code>index</code> is out of bounds, an exception is thrown.
     *
     * @param[in] index index into the string
     * @return character at <code>index</code>
     */
    virtual char
    get_value_at(size_t index) const
    {
        char rv = '\0';
        cgul_exception_t local = NULL;
        rv = cgul_string__get_value_at(&local, obj, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method sets the character indicated by <code>index</code>
     * where <code>index</code> is zero-based to <code>c</code>.  If
     * <code>index</code> is out of bounds, an exception is thrown.
     *
     * @param[in] index index into the string
     * @param[in] c character
     */
    virtual void
    set_value_at(size_t index,
                 char c)
    {
        cgul_exception_t local = NULL;
        cgul_string__set_value_at(&local, obj, index, c);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method is equivalent to calling <code>set_value()</code>
     * where <code>value</code> is an empty string.
     */
    virtual void
    clear()
    {
        /* See the comments for this method for why we do not have to
         * pass in a c-style exception. */
        cgul_string__clear(obj);
    }

    /**
     * This method converts the string to a <code>signed char</code>.
     * Typically, "base" will be set to 0 to allow conversion for decimal,
     * hexadecimal, and octal based on the leading character in the
     * string.  The string cannot have any extraneous characters.  If it
     * does, the conversion will fail.  On success, the converted value is
     * returned; otherwise, 0 is returned and an exception is thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     * @see cgul_string__get_value_at()
     */
    virtual signed char
    to_schar(int base = 0) const
    {
        signed char rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__to_schar(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>unsigned char</code>.
     * Typically, <code>base</code> will be set to 0 to allow conversion
     * for decimal, hexadecimal, and octal based on the leading characters
     * in the string.  The string cannot have any extraneous characters.
     * If it does, the conversion will fail.  On success, the converted
     * value is returned; otherwise, 0 is returned and an exception is
     * thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     * @see cgul_string__get_value_at()
     */
    virtual unsigned char
    to_uchar(int base = 0) const
    {
        unsigned char rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__to_uchar(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to a <code>short</code>.
     * Typically, "base" will be set to 0 to allow conversion for decimal,
     * hexadecimal, and octal based on the leading character in the
     * string.  The string cannot have any extraneous characters.  If it
     * does, the conversion will fail.  On success, the converted value is
     * returned; otherwise, 0 is returned and an exception is thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual short int
    to_short(int base = 0) const
    {
        short int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__to_short(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>unsigned short</code>.
     * Typically, <code>base</code> will be set to 0 to allow conversion
     * for decimal, hexadecimal, and octal based on the leading characters
     * in the string.  The string cannot have any extraneous characters.
     * If it does, the conversion will fail.  On success, the converted
     * value is returned; otherwise, 0 is returned and an exception is
     * thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual unsigned short int
    to_ushort(int base = 0) const
    {
        unsigned short int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__to_ushort(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>int</code>.
     * Typically, <code>base</code> will be set to 0 to allow
     * conversion for decimal, hexadecimal, and octal based on the
     * leading character in the string.  The string cannot have any
     * extraneous characters.  If it does, the conversion fails, and
     * an exception is thrown.  An exception will also be thrown if
     * the converted value is out of bounds.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual int
    to_int(int base = 0) const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__to_int(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>unsigned</code>.
     * Typically, <code>base</code> will be set to 0 to allow
     * conversion for decimal, hexadecimal, and octal based on the
     * leading characters in the string.  The string cannot have any
     * extraneous characters.  If it does, the conversion fails, and
     * an exception is thrown.  An exception will also be thrown if
     * the converted value is out of bounds.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual unsigned int
    to_unsigned(int base = 0) const
    {
        unsigned rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__to_unsigned(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>long</code>.
     * Typically, <code>base</code> will be set to 0 to allow
     * conversion for decimal, hexadecimal, and octal based on the
     * leading characters in the string.  The string cannot have any
     * extraneous characters.  If it does, the conversion fails, and
     * an exception is thrown.  An exception will also be thrown if
     * the converted value is out of bounds.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual long int
    to_long(int base = 0) const
    {
        long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__to_long(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>unsigned
     * long</code>.  Typically, <code>base</code> will be set to 0 to
     * allow conversion for decimal, hexadecimal, and octal based on
     * the leading characters in the string.  The string cannot have
     * any extraneous characters.  If it does, the conversion fails,
     * and an exception is thrown.  An exception will also be thrown
     * if the converted value is out of bounds.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual unsigned long int
    to_unsigned_long(int base = 0) const
    {
        unsigned long rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__to_unsigned_long(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to a <code>double</code>.  The
     * string cannot have any extraneous characters.  If it does, the
     * conversion fails, and an exception is thrown.  An exception
     * will also be thrown if the converted value is out of bounds.
     *
     * @return converted value
     */
    virtual double
    to_double() const
    {
        double rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__to_double(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Appends <code>value</code> to the string.  If an error occurs,
     * an exception is thrown.
     *
     * @param[in] value value to append
     */
    virtual void
    append(const char* value)
    {
        cgul_exception_t local = NULL;
        cgul_string__append(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append a block of non-NUL characters.  The caller should verify
     * that there are no embedded NUL characters in the string.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] block block to append
     * @param[in] block_size size of block to append
     */
    virtual void
    append_block(const void* block,
                 size_t block_size)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_block(&local, obj, block, block_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This function is a thin wrapper around
     * <code>append_vsprintf()</code>.
     *
     * @param format format string
     * @param ... variable number of arguments that satisfy <code>format</code>
     * @return number of characters appended
     */
    virtual int
    append_sprintf(const char* format, ...)
#if defined(__GNUC__) && !defined(__MINGW_H)
        /* Because C++ passes "this" as arg1, "format" is actually arg2. */
        __attribute__ ((format (printf, 2, 3)))
#endif
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        va_list args;

        va_start(args, format);
        rv = append_vsprintf(format, args);
        va_end(args);

        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function substitutes the conversion specifications in the
     * <code>format</code> string and appends them to the underlying
     * object.  If an error occurs, an exception is thrown. <p>
     *
     * The format string is the same as what you pass into
     * <code>sprintf()</code> or <code>snprintf()</code>.  The following
     * length modifiers are honored: l, h, z. <p>
     *
     * Any string associated with a "%s" conversion specification is
     * assumed to be a UTF-8 string.  The main difference that this makes
     * is that, if a field width is specified, it specifies the number of
     * characters, not bytes, in the field. <p>
     *
     * The "%ls" conversion specification is honored, but instead of
     * converting a string composed of <code>wchar_t</code> characters to
     * a multi-byte string encoded according to the current locale, it
     * converts a string composed of Unicode UTF-32
     * <code>cgul_wchar_t</code> characters to a UTF-8 string.  The "%ls"
     * conversions also differ in that, if a precision is given, it
     * determines the number of characters appended not the number of
     * bytes. <p>
     *
     * The "%lc" conversion specification is honored, but like the "%ls"
     * conversion specification, it converts from UTF-32 to UTF-8
     * exclusively.  Also, it expects the associated argument to be
     * <code>cgul_wchar_t</code>, not <code>wchar_t</code>. <p>
     *
     * The GNU "%m" conversion specification is honored.  This conversion
     * specification appends the result of
     * <code>strerror(errno)</code>. <p>
     *
     * @param format format string
     * @param args variable number of arguments that satisfy <code>format</code>
     * @return number of characters appended
     */
    virtual int
    append_vsprintf(const char* format,
                    va_list args)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__append_vsprintf(&local, obj, format, args);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Append <code>c</code> to the string.
     *
     * @param[in] c character to append
     */
    virtual void
    append_char(char c)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_char(&local, obj, c);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append the UTF-16 character <code>utf16</code> as a UTF-8
     * sequence.  Because <code>utf16</code> could be a surrogate
     * pair, to avoid reading off the end of the sequence, the number
     * of <code>cgul_hchar_t</code> elements (i.e., 16-bit elements)
     * in the <code>utf16</code> buffer must be specified using
     * <code>utf16_length</code>.  If more than one UTF-16 character
     * is present in the <code>utf16</code> buffer, only the first
     * character is appended.
     *
     * @param[in] utf16 UTF-16 character to append
     * @param[in] utf16_length length of <code>utf16</code> in
     * <code>cgul_hchar_t</code> elements
     */
    virtual void
    append_hchar(cgul_hchar_t* utf16,
                 size_t utf16_length)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_hchar(&local, obj, utf16, utf16_length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append the UTF-32 character <code>utf32</code> to the string as
     * a UTF-8 sequence.
     *
     * @param[in] utf32 UTF-32 character to append
     */
    virtual void
    append_wchar(cgul_wchar_t utf32)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_wchar(&local, obj, utf32);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append <code>value</code> to the string in decimal format.
     *
     * @param[in] value to append
     */
    virtual void
    append_decimal(long int value)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_decimal(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append <code>value</code> to the string in decimal format.
     *
     * @param[in] value to append
     */
    virtual void
    append_unsigned_decimal(size_t value)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_unsigned_decimal(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append <code>value</code> to the string in octal format.
     *
     * @param[in] value to append
     */
    virtual void
    append_octal(size_t value)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_octal(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append <code>value</code> to the string in hexadecimal format
     * using "abcdef".
     *
     * @param[in] value to append
     * @param[in] caps whether to capitalize the hex digits
     */
    virtual void
    append_hexadecimal(size_t value,
                       int caps)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_hexadecimal(&local, obj, value, caps);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append <code>value</code> to the string.  "precision"
     * determines the number of significant digits that are printed.
     * Zeros after the decimal point at the end of number are not printed.
     *
     * @param[in] value to append
     * @param[in] precision precision
     */
    virtual void
    append_double(double value,
                  int precision)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_double(&local, obj, value, precision);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append <code>ptr</code> to the string.
     *
     * @param[in] ptr pointer to append
     */
    virtual void
    append_pointer(const void* ptr)
    {
        cgul_exception_t local = NULL;
        cgul_string__append_pointer(&local, obj, ptr);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Inserts <code>value</code> at <code>index</code> which can
     * range from 0 to the length (inclusive) of this string.  If
     * <code>index</code> is out of bounds or if memory cannot be
     * allocated, an exception is thrown.
     *
     * @param[in] index index where <code>value</code> is inserted
     * @param[in] value value to insert
     * @param[in] value_length length of value
     */
    virtual void
    insert(size_t index,
           const char* value,
           size_t value_length)
    {
        cgul_exception_t local = NULL;
        cgul_string__insert(&local, obj, index, value, value_length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Inserts <code>value</code> at <code>index</code> which can
     * range from 0 to the length (inclusive) of this string.  If
     * <code>index</code> is out of bounds or if memory cannot be
     * allocated, an exception is thrown.
     *
     * @param[in] index index where <code>value</code> is inserted
     * @param[in] value value to insert
     */
    virtual void
    insert_char(size_t index,
                char value)
    {
        cgul_exception_t local = NULL;
        cgul_string__insert_char(&local, obj, index, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Pad the current string such that it is at least
     * <code>width</code> characters wide using the <code>pad</code>
     * character.  The string will be right aligned.
     *
     * @param[in] width padding width
     * @param[in] pad character to use for padding
     */
    virtual void
    pad_right(size_t width,
              char pad)
    {
        cgul_exception_t local = NULL;
        cgul_string__pad_right(&local, obj, width, pad);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Pad the current string such that it is at least
     * <code>width</code> characters wide using the <code>pad</code>
     * character.  The string will be left aligned.
     *
     * @param[in] width padding width
     * @param[in] pad character to use for padding
     */
    virtual void
    pad_left(size_t width,
             char pad)
    {
        cgul_exception_t local = NULL;
        cgul_string__pad_left(&local, obj, width, pad);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Trim the string by removing leading and trailing white-space where
     * white-space is definded by <code>cgul_char::isspace()</code>.
     */
    virtual void
    trim()
    {
        cgul_exception_t local = NULL;
        cgul_string__trim(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Trim the string by removing leading white-space where white-space
     * is definded by <code>cgul_char::isspace()</code>.
     */
    virtual void
    trim_left()
    {
        cgul_exception_t local = NULL;
        cgul_string__trim_left(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Trim the string by removing trailing white-space where white-space
     * is definded by <code>cgul_char::isspace()</code>.
     */
    virtual void
    trim_right()
    {
        cgul_exception_t local = NULL;
        cgul_string__trim_right(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Reverse the order of the characters in the string.
     */
    virtual void
    reverse()
    {
        cgul_exception_t local = NULL;
        cgul_string__reverse(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method returns the length of the string.
     *
     * @return length of the string
     */
    virtual size_t
    get_length() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__get_length(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method can be used to set the length of the string.
     * It can only be used to make the string shorter.  If it is
     * used in an attempt to make the string longer, an exception
     * is thrown.  If the length of the string is successfully
     * changed, the previous length of the string is returned.
     *
     * @param[in] length new, shorter length of the string
     * @return previous length of the string
     */
    virtual size_t
    set_length(size_t length)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__set_length(&local, obj, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * If you manually alter the underlying buffer by modifying the
     * <code>char*</code> string returned by <code>get_value()</code>,
     * you'll need to manually have the class reset its internal
     * state.
     *
     * @see get_value
     */
    virtual void
    externally_adjusted()
    {
        cgul_exception_t local = NULL;
        cgul_string__externally_adjusted(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Trim the string by returning excess capacity to the heap.
     */
    virtual void
    trim_capacity()
    {
        cgul_exception_t local = NULL;
        cgul_string__trim_capacity(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the minimum capacity.
     *
     * @return minimum capacity
     */
    virtual size_t
    get_minimum_capacity() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__get_minimum_capacity(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the minimum capacity.  The default value is <code>128</code>
     * which means there is always capacity for 127 characters plus one
     * NUL-terminator.  Decreasing the capacity will decrease the amount
     * of space used and (typically) also decrease performance.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] minimum_capacity minimum capacity
     */
    virtual void
    set_minimum_capacity(size_t minimum_capacity)
    {
        cgul_exception_t local = NULL;
        cgul_string__set_minimum_capacity(&local, obj, minimum_capacity);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform a case-significant string comparison of this object and
     * <code>rhs</code> using the current locale.  Return -1, 0, or 1
     * if this object is less than, equal to, or greater than
     * <code>rhs</code> respectively. <p>
     *
     * Note that this method can throw an exception if either string
     * contains a character that is not part of the current locale's
     * collating sequence. <p>
     *
     * @param[in] rhs right-hand side
     * @return -1, 0, or 1 if <code>*this</code> is less than, equal
     * to, or greater than <code>rhs</code> respectively
     * @see cgul_char_cxx::strcmp()
     */
    virtual int
    compare(const cgul_string_cxx& rhs) const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__compare(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Perform a case-insensitive string comparison of this object
     * and <code>rhs</code> using the current locale.  Return -1, 0,
     * or 1 if this object is less than, equal to, or greater than
     * <code>rhs</code> respectively. <p>
     *
     * Note that this method can throw an exception if either string
     * contains a character that is not part of the current locale's
     * collating sequence. <p>
     *
     * @param[in] rhs right-hand side
     * @return -1, 0, or 1 if <code>*this</code> is less than, equal
     * to, or greater than <code>rhs</code> respectively
     */
    virtual int
    icompare(const cgul_string_cxx& rhs) const
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_string__icompare(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Swap the underlying data for this object and <code>rhs</code>.
     * For large strings, this should be much faster than trying to do
     * the same thing using <code>set_value()</code>.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    swap(cgul_string_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_string__swap(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }


    /**
     * Get the underlying <code>cgul_string</code> object.
     *
     * @return underlying object
     */
    virtual cgul_string_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_string</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_string__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_string_t
    take_obj()
    {
        cgul_string_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_string_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_string__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /** Underlying <code>cgul_string</code> object. */
    cgul_string_t obj;

};

#endif  /*  CGUL_STRING_CXX_H  */
