#ifndef CGUL_TRIE_H
#define CGUL_TRIE_H

/**
 * Prefix tree implemented as a Ternary Search Trie (TST).
 *
 * @file
 * @author Paul Serice
 * @brief prefix tree
 */

#include "cgul_common.h"
#include "cgul_exception.h"
#include "cgul_trie_node.h"

CGUL_BEGIN_C

/** Opaque pointer to a <code>cgul_trie</code> instance. */
typedef struct cgul_trie* cgul_trie_t;

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_trie__foldl_keys()</code>
 *     <code>cgul_trie__foldr_keys()</code>
 * </pre>
 *
 * The client must not alter <code>key</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] key key
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_trie__fold_key_t)(cgul_exception_t* cex,
                                     const char* key,
                                     void* data);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_trie__foldl_values()</code>
 *     <code>cgul_trie__foldr_values()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_trie__fold_value_t)(cgul_exception_t* cex,
                                       void* value,
                                       void* data);

/**
 * This typedef is the interface for the combining function used by
 * the following methods: <p>
 *
 * <pre>
 *     <code>cgul_trie__foldl_pairs()</code>
 *     <code>cgul_trie__foldr_pairs()</code>
 * </pre>
 *
 * The client must not alter <code>key</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] key key
 * @param[in] value value
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_trie__fold_pair_t)(cgul_exception_t* cex,
                                      const char* key,
                                      void* value,
                                      void* data);

/**
 * This typedef is the interface for the callback function used by the
 * following methods: <p>
 *
 * <pre>
 *     <code>cgul_trie__traverse()</code>
 *     <code>cgul_trie__traverse_range()</code>
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] n node
 * @param[in] data client data
 * @return whether to continue
 */
typedef int (*cgul_trie__traverse_t)(cgul_exception_t* cex,
                                     cgul_trie_t trie,
                                     cgul_trie_node_t n,
                                     void* data);

/**
 * Create a new <code>cgul_trie</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_trie__delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * The <code>cgul_trie</code> object that is returned does not allow
 * you to index into it as though it were an array.  This saves you
 * three <code>unsigned long</code> values for each node in the trie.
 * It also saves you about 15% for inserts and deletes.  If in doubt,
 * use this constructor. <p>
 *
 * This class does not take ownership of the inserted key/value pairs.
 * Instead, the client is responsible for freeing the key/value pairs
 * only after their nodes have been permanently removed from the
 * <code>cgul_trie</code>.  The <code>cgul_trie__traverse()</code>
 * method can be used to remove each node safely before calling
 * <code>cgul_trie__delete()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_trie</code> instance
 */
extern CGUL_EXPORT cgul_trie_t
cgul_trie__new(cgul_exception_t* cex);

/**
 * Create a new <code>cgul_trie</code> object.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_trie__delete()</code>.  If memory cannot be allocated,
 * <code>NULL</code> is returned, and an exception is thrown. <p>
 *
 * The <code>cgul_trie</code> object that is returned allows you to
 * index into it as though it were an array.  This costs you three
 * <code>unsigned long</code> values for each node in the trie.  It
 * also costs you about 15% for inserts and deletes.  You should only
 * use this constructor if you know you want to be able to call
 * <code>cgul_trie__find_at()</code> or
 * <code>cgul_trie__remove_at()</code>. <p>
 *
 * This class does not take ownership of the inserted key/value pairs.
 * Instead, the client is responsible for freeing the key/value pairs
 * only after their nodes have been permanently removed from the
 * <code>cgul_trie</code>.  The <code>cgul_trie__traverse()</code>
 * method can be used to remove each node safely before calling
 * <code>cgul_trie__delete()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_trie</code> instance
 */
extern CGUL_EXPORT cgul_trie_t
cgul_trie__new_with_indexing(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  This does not
 * include the key/value pairs stored in the trie.  The client is
 * responsible for freeing the key/value pairs when the client thinks
 * it is convenient and safe to do so; however, the client must
 * understand that freeing a key before removing its node invalidates
 * the data structure. <p>
 *
 * As a convenience, you may want to call
 * <code>cgul_trie__clear()</code> before calling this method in order
 * to properly put the keys and values back on their respective
 * <code>cgul_cache</code> objects (if <code>cgul_cache</code> objects
 * are being used) before you delete the trie.  If
 * <code>cgul_cache</code> objects are not being used, the client
 * needs to arrange some other mechanism to free the keys or
 * values. <p>
 *
 * Do not try to access <code>trie</code> after calling this method. <p>
 *
 * @param[in] trie <code>trie</code> instance
 * @see cgul_trie__free_keys()
 * @see cgul_trie__free_values()
 */
extern CGUL_EXPORT void
cgul_trie__delete(cgul_trie_t trie);

/**
 * This method calls <code>free()</code> on all the keys in the trie
 * <code>trie</code>.  Because this is such a common operation, it is
 * an exception to the rule that cgul containers never free keys.
 * This method should only ever be called immediately before calling
 * <code>cgul_trie__delete()</code> because it otherwise invalidates
 * the trie.  Because this function is basically an extension of
 * <code>cgul_trie__delete()</code>, it does not accept a
 * <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] trie <code>cgul_trie</code> instance
 */
extern CGUL_EXPORT void
cgul_trie__free_keys(cgul_trie_t trie);

/**
 * This method calls <code>free()</code> on all the values in the trie
 * <code>trie</code>.  Because this is such a common operation, it is
 * an exception to the rule that cgul containers never free values.
 * This method should only ever be called immediately before calling
 * <code>cgul_trie__delete()</code> because it otherwise invalidates
 * the trie.  Because this function is basically an extension of
 * <code>cgul_trie__delete()</code>, it does not accept a
 * <code>cgul_exception</code> object as an input parameter.
 *
 * @param[in] trie <code>cgul_trie</code> instance
 */
extern CGUL_EXPORT void
cgul_trie__free_values(cgul_trie_t trie);

/**
 * Get the size of the cache of nodes. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @return size of the cache of nodes
 */
extern CGUL_EXPORT unsigned long int
cgul_trie__get_cache_size(cgul_exception_t* cex,
                          cgul_trie_t trie);

/**
 * Set the size of the cache of nodes. <p>
 *
 * For efficiency, the <code>cgul_trie</code> can keep a cache of
 * nodes that can be reused.  In many situations, this can greatly
 * reduce the number of calls to <code>malloc()</code> which can
 * greatly improve performance and reduce memory fragmentation. <p>
 *
 * You can release all the cached nodes and disable caching of nodes
 * by setting the cache reserve to 0 and the cache size to 0 (in that
 * order).  This is the default. <p>
 *
 * If an error occurs (while enlarging the cache), an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] size of the cache of nodes
 */
extern CGUL_EXPORT void
cgul_trie__set_cache_size(cgul_exception_t* cex,
                          cgul_trie_t trie,
                          unsigned long int size);

/**
 * Get the reserve limit of the cache of nodes. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @return reserve limit of the cache of nodes
 */
extern CGUL_EXPORT unsigned long int
cgul_trie__get_cache_reserve(cgul_exception_t* cex,
                             cgul_trie_t trie);

/**
 * Set the reserve limit of the cache of nodes and guarantee that at
 * least that many nodes are allocated and waiting in reserve. <p>
 *
 * When you use this method to increase the reserve, you are
 * guaranteed that at least <code>reserve</code> count of nodes will
 * be allocated and reserved for future use. <p>
 *
 * This is useful because <code>cgul_trie__insert()</code> can
 * throw an exception if it cannot allocate a node to hold the
 * key/value pair.  Thus, by allocating and reserving a node now, we
 * know that <code>cgul_trie__insert()</code> will be able to run
 * without error later when the node is "unreserved".  Normally, this
 * is not an issue because you can almost always just insert at the
 * moment when it is requested, but if the <code>cgul_trie</code>
 * is busy at that moment (perhaps it is iterating over its nodes),
 * you'll have to queue insert (and remove) requests for later
 * execution. <p>
 *
 * To "unreserve" a node so that it can be used, you simply call this
 * method again with a value for <code>reserve</code> that is less
 * than the current reserve limit. <p>
 *
 * If an error occurs while allocating the reserved nodes, an
 * exception is thrown. <p>
 *
 * WARNING: If you reserve nodes without "unreserving" them, you will
 * introduce a memory leak that is difficult to track down because
 * this class is careful to free all of the cached nodes (including
 * the reserved nodes) when it is deleted, but the reserved nodes will
 * just sit around in memory until that time.  Thus, after reserving
 * nodes by calling this method, you must remember to timely set the
 * reserve level back to zero in order to make those nodes
 * available. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] reserve of the cache of nodes
 * @see cgul_cache__set_reserve()
 */
extern CGUL_EXPORT void
cgul_trie__set_cache_reserve(cgul_exception_t* cex,
                             cgul_trie_t trie,
                             unsigned long int reserve);

/**
 * Return 1 if the trie <code>trie</code> is empty; otherwise, return 0.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @return whether the trie is empty
 */
extern CGUL_EXPORT int
cgul_trie__is_empty(cgul_exception_t* cex,
                    cgul_trie_t trie);

/**
 * Balance the trie <code>trie</code>.  If an error occurs, an
 * exception is thrown. <p>
 *
 * Because the key/value pairs are reinserted randomly into the trie,
 * the resulting older/younger linked list (which tracks insertion
 * order) will generally not match the original insertion order.
 * However, if <code>maintain_relative_ages</code> is true, this
 * method will take the extra, relatively expensive step of adjusting
 * the older/younger list to match the original order. <p>
 *
 * Ternary Search Tries have worst-case search behavior of O(n).  This
 * problem occurs when the keys are inserted in sort order.  This
 * method can be used to reinsert all the keys at random which will
 * likely come very close to minimizing the depth of the trie which
 * will result in search behavior of O(log n).  This is an expensive
 * operation which should be called rarely over the lifetime of the
 * trie. <p>
 *
 * One plausible scenario where this method could be called is if all
 * the keys are read at once from a sorted list.  After inserting all
 * the keys, this method could be called to balance the trie.
 * (Another solution to this problem would be to insert all the keys
 * through <code>cgul_trie__insert_from_arrays()</code>.) <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] maintain_relative_ages whether to maintain relative node ages
 * @see cgul_trie__insert_from_arrays()
 */
extern CGUL_EXPORT void
cgul_trie__balance(cgul_exception_t* cex,
                   cgul_trie_t trie,
                   int maintain_relative_ages);

/**
 * Insert <code>key</code> into the trie <code>trie</code>.  If
 * <code>key</code> already exists, another node with the same key is
 * not created.  The new or pre-existing node is returned in
 * <code>n</code> if <code>n</code> is not <code>NULL</code>. <p>
 *
 * If a matching node is not found, <code>1</code> is returned because
 * a new node can and will be inserted.  If a matching node is found,
 * <code>0</code> is returned because the insertion failed because the
 * key already exists.<p>
 *
 * If an error occurs, 0 is returned, <code>*n</code> is set to
 * <code>NULL</code> (if <code>n</code> is not <code>NULL</code>), and
 * an exception is thrown. <p>
 *
 * It is important to understand that the <code>cgul_trie</code>
 * class (like all containers in the cgul library) does not take
 * ownership of <code>key</code>.  It also does not attempt to make a
 * copy of the thing pointed to by <code>key</code>.  This gives the
 * user complete control over the lifetime of the key and gives a real
 * performance boost in many cases; however, this means that
 * <code>key</code> must not be invalidated while it is still being
 * used by this class. <p>
 *
 * This method operates as it does so that you only have to search the
 * trie once when doing an insert.  Logically, you would usually like
 * to do a find to see if the trie holds your key.  If it does, you
 * use the value associated with that key in some way.  If it does
 * not, you then want to insert a new key/value pair into the trie.
 * The problem with divorcing the find from the insert is that it then
 * requires you to search the trie twice.  Once to see if the key
 * exists and once to do the insertion. <p>
 *
 * To prevent having to search the trie twice, you can just call this
 * method to handle both inserting and updating nodes.  It will search
 * the trie once, and it will always return a pointer to the correct
 * node.  You can then use <code>cgul_trie_node__get_value()</code>
 * or <code>cgul_trie_node__set_value()</code> to get or set the
 * value indexed by "key". <p>
 *
 * TIP: You might find it convenient to use <code>cgul_string</code>
 * to build up a string that you use as a key.  You can then use
 * <code>cgul_string__get_value()</code> [yes, get_value()] and pass
 * the string to <code>cgul_trie__insert()</code>.  If the insert
 * succeeds, you can inform the <code>cgul_string</code> object that
 * it is no longer the owner of the underlying C-style string by
 * calling <code>cgul_string__take_value()</code> [yes, take_value()].
 * An example is as follows: <p>
 *
 * <pre>
 *
 *     int insert_rv = 0;
 *     cgul_trie_node_t n = NULL;
 *
 *     char* w = cgul_string__get_value(word);
 *     insert_rv = cgul_trie__insert(trie, w, &n);
 *     if (*cex) {
 *         goto out;
 *     }
 *     if (insert_rv) {
 *         cgul_string__take_value(word);
 *         . . .
 *     } else {
 *         . . .
 *     }
 *
 * </pre>
 *
 * NOTE: Inserting a key/value pair does not cause this class to take
 * ownership of <code>key</code> or <code>value</code>.  The client
 * must make sure the values pointed to by <code>key</code> and
 * <code>value</code> are not invalidated during the life of the newly
 * created node.  If the client needs to delete the <code>key</code>
 * or the <code>value</code> or both, the client needs to manually
 * arrange for this. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] key key
 * @param[out] n node that already existed or was newly created
 * @return 1 if a new node was inserted; 0 if a node with the same key
 * already exists
 * @see cgul_trie__balance()
 * @see cgul_trie__insert_from_arrays()
 */
extern CGUL_EXPORT int
cgul_trie__insert(cgul_exception_t* cex,
                  cgul_trie_t trie,
                  const char* key,
                  cgul_trie_node_t* n);

/**
 * Insert into the trie <code>trie</code> the key/value pairs formed
 * by interleaving keys from the <code>keys</code> array with values
 * from the <code>values</code> array.  If <code>values</code> is
 * <code>NULL</code>, all values will be set to <code>NULL</code>.
 * The number of new nodes insert is returned.  If an error occurs, an
 * exception is thrown. <p>
 *
 * This method copies the <code>keys</code> and <code>values</code>
 * arrays and then performs the same shuffle on each copy before
 * inserting the shuffled key/value pairs.  It shuffles the key/value
 * pairs because inserting sorted keys into a Ternary Search Trie
 * leads to worst-case search behavior of O(N).  By shuffling the keys
 * first, the trie will very likely be balanced leading to O(log N)
 * search behavior. <p>
 *
 * Because the key/value pairs are inserted randomly into the trie,
 * the resulting older/younger linked list (which tracks insertion
 * order) will generally not match the order of the key/value pairs
 * from the <code>keys</code> and <code>values</code> arrays.
 * However, if <code>maintain_relative_ages</code> is true, this
 * method will take the extra, relatively expensive step of adjusting
 * the older/younger list to match the original order of the key/value
 * pairs. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys keys
 * @param[in] values values
 * @param[in] kv_size number of key/value pairs (i.e., size of the arrays)
 * @param[in] maintain_relative_ages whether to maintain relative node ages
 * @return number of new nodes inserted
 * @see balance()
 */
extern CGUL_EXPORT int
cgul_trie__insert_from_arrays(cgul_exception_t* cex,
                              cgul_trie_t trie,
                              const char** keys,
                              void** values,
                              unsigned long int kv_size,
                              int maintain_relative_ages);

/**
 * Find the node indexed by <code>key</code>.  If the key does not
 * exist, <code>NULL</code> is returned; otherwise, the node
 * associated with the key is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] key key
 * @return node associated with the key or <code>NULL</code>
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie__find(cgul_exception_t* cex,
                cgul_trie_t trie,
                const char* key);

/**
 * Find the node indexed by position.  This lets you use a
 * <code>cgul_trie</code> instances like you would an array.  Even in
 * the best-case scenario where indexing into <code>cgul_trie</code>
 * is a logarithmic function, it will still be slower than indexing
 * into an array, but you get all the advantages of a trie including
 * relatively fast insertion and deletion.  So, if you ever feel the
 * need to shift the elements of an array, this method might be for
 * you. <p>
 *
 * If <code>index</code> is out of range, <code>NULL</code> is
 * returned; otherwise, the node returned is the same node that would
 * be returned if you had an array <code>trie</code> that held all the
 * nodes in the trie in sorted order, and you asked for
 * <code>trie[index]</code>. <p>
 *
 * This method throws an exception if <code>trie</code> was not created
 * with <code>cguil_trie__new_with_indexing()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] index index into sorted keys of node to return
 * @return node associated with <code>index</code>
 * @see cgul_trie__find_rank()
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie__find_at(cgul_exception_t* cex,
                   cgul_trie_t trie,
                   unsigned long int index);

/**
 * Return whether the rank of the key <code>key</code> could be
 * determined.  If the rank could be determined it is returned in
 * <code>*rank</code> (if <code>rank</code> is not
 * <code>NULL</code>). <p>
 *
 * The rank is the number of keys less than <code>key</code>.  It can
 * be used as the index parameter to <code>cgul_trie__find_at()</code>
 * in order to lookup <code>key</code> by index. <p>
 *
 * This method throws an exception if <code>trie</code> was not
 * created with <code>cguil_trie__new_with_indexing()</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] key key
 * @param[out] rank rank of key
 * @return whether the key exists in the trie
 * @see cgul_trie__find_at()
 */
extern CGUL_EXPORT int
cgul_trie__find_rank(cgul_exception_t* cex,
                     cgul_trie_t trie,
                     const char* key,
                     unsigned long int* rank);

/**
 * Find the node having the largest key less than or equal to the
 * search key <code>search_key</code> and return it.  If such a key
 * does not exist, <code>NULL</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] search_key search key
 * @return floor node or <code>NULL</code>
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie__find_floor(cgul_exception_t* cex,
                      cgul_trie_t trie,
                      const char* search_key);

/**
 * Find the node having the smallest key greater than or equal to the
 * search key <code>search_key</code> and return it.  If such a key
 * does not exist, <code>NULL</code> is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] search_key search key
 * @return ceiling node or <code>NULL</code>
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie__find_ceiling(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        const char* search_key);

/**
 * When dealing with prefixes that generate a large number of matches,
 * it may be more responsive from the user's perspective to iterate
 * over the results.  To that end, this method returns the closed
 * range <code>[begin_node, end_node]</code> for the nodes in the trie
 * <code>trie</code> associated with keys that start with the prefix
 * <code>prefix</code>.  If any nodes exist in the range, this method
 * returns <code>1</code>; otherwise, this methods returns
 * <code>0</code>.  If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prefix prefix
 * @param[out] begin_node beginning node (inclusive) or <code>NULL</code>
 * @param[out] end_node ending node (inclusive) or <code>NULL</code>
 * @return whether any nodes were found
 * @see cgul_trie__find_half_opened_range_by_prefix()
 */
extern CGUL_EXPORT int
cgul_trie__find_closed_range_by_prefix(cgul_exception_t* cex,
                                       cgul_trie_t trie,
                                       const char* prefix,
                                       cgul_trie_node_t* begin_node,
                                       cgul_trie_node_t* end_node);

/**
 * When dealing with prefixes that generate a large number of matches,
 * it may be more responsive from the user's perspective to iterate
 * over the results.  To that end, this method returns the half-opened
 * range <code>[begin_node, end_node]</code> for the nodes in the trie
 * <code>trie</code> associated with keys that start with the prefix
 * <code>prefix</code>.  If any nodes exist in the range, this method
 * returns <code>1</code>; otherwise, this methods returns
 * <code>0</code>.  If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prefix prefix
 * @param[out] begin_node beginning node (inclusive) or <code>NULL</code>
 * @param[out] end_node ending node (exclusive) or <code>NULL</code>
 * @return whether any nodes were found
 * @see cgul_trie__find_closed_range_by_prefix()
 */
extern CGUL_EXPORT int
cgul_trie__find_half_opened_range_by_prefix(cgul_exception_t* cex,
                                            cgul_trie_t trie,
                                            const char* prefix,
                                            cgul_trie_node_t* begin_node,
                                            cgul_trie_node_t* end_node);

/**
 * All keys in the trie <code>trie</code> that start with the prefix
 * <code>prefix</code> are returned in the arrays <code>keys</code>
 * (if not <code>NULL</code>) and the number of keys is returned in
 * <code>keys_size</code> (if not <code>NULL</code>).  If any keys are
 * found, this method returns <code>1</code>; otherwise, this methods
 * returns <code>0</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * The client is responsible for calling <code>free()</code> on the
 * <code>keys</code> array but must not modify or delete any of the
 * individual keys. <p>
 *
 * The following example code shows how to print all the keys in the
 * trie that match a particular prefix and how to clean up afterward: <p>
 *
 * <pre>
 *
 *      void
 *      print_keys_by_prefix(cgul_exception_t* cex,
 *                           cgul_trie_t trie,
 *                           const char* prefix)
 *      {
 *          unsigned long int i = 0;
 *          const char** keys = NULL;
 *          unsigned long int keys_size = 0;
 *      
 *          assert(cex);
 *          if (*cex) {
 *              goto out;
 *          }
 *      
 *          // Iterate over all the keys that match the prefix.
 *          if (cgul_trie__find_keys_by_prefix(
 *                  cex, trie, prefix, &keys, &keys_size))
 *          {
 *              for (i = 0 ; i < keys_size ; ++i) {
 *                  printf("%s -> %s\n", prefix, keys[i]);
 *              }
 *          }
 *      
 *       out:
 *      
 *          // Clean up.
 *          if (keys) {
 *              free(keys);
 *          }
 *      
 *      }
 *
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prefix prefix
 * @param[out] keys keys found by prefix
 * @param[out] keys_size number of keys found
 * @return whether any keys were found
 * @see cgul_trie__find_nodes_by_prefix()
 * @see cgul_trie__find_closed_range_by_prefix()
 * @see cgul_trie__find_half_opened_range_by_prefix()
 */
extern CGUL_EXPORT int
cgul_trie__find_keys_by_prefix(cgul_exception_t* cex,
                               cgul_trie_t trie,
                               const char* prefix,
                               const char*** keys,
                               unsigned long int* keys_size);

/**
 * All values in the trie <code>trie</code> associated with keys that
 * start with the prefix <code>prefix</code> are returned in the
 * arrays <code>values</code> (if not <code>NULL</code>) and the
 * number of values is returned in <code>values_size</code> (if not
 * <code>NULL</code>).  If any values are found, this method returns
 * <code>1</code>; otherwise, this methods returns <code>0</code>.  If
 * an error occurs, an exception is thrown. <p>
 *
 * The client is responsible for calling <code>free()</code> on the
 * <code>values</code> array.  The client is free to modify but not
 * delete any of the individual values unless the value is also used
 * as a key for the trie.  If a value needs to be delete the client
 * should do so only after using
 * <code>cgul_trie_node__set_value()</code> to clear the value. <p>
 *
 * The following example code shows how to print all the values in the
 * trie associated with keys that match a particular prefix and how to
 * clean up afterward: <p>
 *
 * <pre>
 *
 *      void
 *      print_values_by_prefix(cgul_exception_t* cex,
 *                             cgul_trie_t trie,
 *                             const char* prefix)
 *      {
 *          unsigned long int i = 0;
 *          void** values = NULL;
 *          unsigned long int values_size = 0;
 *      
 *          assert(cex);
 *          if (*cex) {
 *              goto out;
 *          }
 *      
 *          // Iterate over all the values that match the prefix.
 *          if (cgul_trie__find_values_by_prefix(
 *                  cex, trie, prefix, &values, &values_size))
 *          {
 *              for (i = 0 ; i < values_size ; ++i) {
 *                  printf("%s -> %p\n", prefix, values[i]);
 *              }
 *          }
 *      
 *       out:
 *      
 *          // Clean up.
 *          if (values) {
 *              free(values);
 *          }
 *      
 *      }
 *
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prefix prefix
 * @param[out] values values found by prefix
 * @param[out] values_size number of values found
 * @return whether any values were found
 * @see cgul_trie__find_nodes_by_prefix()
 * @see cgul_trie__find_closed_range_by_prefix()
 * @see cgul_trie__find_half_opened_range_by_prefix()
 */
extern CGUL_EXPORT int
cgul_trie__find_values_by_prefix(cgul_exception_t* cex,
                                 cgul_trie_t trie,
                                 const char* prefix,
                                 void*** values,
                                 unsigned long int* values_size);

/**
 * All nodes in the trie <code>trie</code> associated with keys that
 * start with the prefix <code>prefix</code> are returned in the
 * arrays <code>nodes</code> (if not <code>NULL</code>) and the number
 * of nodes is returned in <code>nodes_size</code> (if not
 * <code>NULL</code>).  If any nodes are found, this method returns
 * <code>1</code>; otherwise, this methods returns <code>0</code>.  If
 * an error occurs, an exception is thrown. <p>
 *
 * The client is responsible for calling <code>free()</code> on the
 * <code>nodes</code> array.  The client must not delete any of the
 * nodes nor delete or modify any of the keys stored in the nodes, but
 * the client is free to modify but not delete any of the individual
 * values stored in the nodes unless the value is also used as a key
 * for the trie.  If a value needs to be delete the client should do
 * so only after using <code>cgul_trie_node__set_value()</code> to
 * clear the value. <p>
 *
 * The following example code shows how to print all the key/value
 * pairs in the trie where the keys match a particular prefix and how
 * to clean up afterward: <p>
 *
 * <pre>
 *
 *      void
 *      print_nodes_by_prefix(cgul_exception_t* cex,
 *                            cgul_trie_t trie,
 *                            const char* prefix)
 *      {
 *          unsigned long int i = 0;
 *          const char* key = NULL;
 *          void* value = NULL;
 *          cgul_trie_node_t* nodes = NULL;
 *          unsigned long int nodes_size = 0;
 *      
 *          assert(cex);
 *          if (*cex) {
 *              goto out;
 *          }
 *      
 *          // Iterate over all the nodes that match the prefix.
 *          if (cgul_trie__find_nodes_by_prefix(
 *                  cex, trie, prefix, &nodes, &nodes_size))
 *          {
 *              for (i = 0 ; i < nodes_size ; ++i) {
 *                  key = cgul_trie_node__get_key(cex, nodes[i]);
 *                  value = cgul_trie_node__get_value(cex, nodes[i]);
 *                  printf("%s -> (%s, %p)\n", prefix, key, value);
 *              }
 *          }
 *      
 *       out:
 *      
 *          // Clean up.
 *          if (nodes) {
 *              free(nodes);
 *          }
 *      
 *      }
 *
 * </pre>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] prefix prefix
 * @param[out] nodes nodes found by prefix
 * @param[out] nodes_size number of nodes found
 * @return whether any nodes were found
 * @see cgul_trie__find_closed_range_by_prefix()
 * @see cgul_trie__find_half_opened_range_by_prefix()
 */
extern CGUL_EXPORT int
cgul_trie__find_nodes_by_prefix(cgul_exception_t* cex,
                                cgul_trie_t trie,
                                const char* prefix,
                                cgul_trie_node_t** nodes,
                                unsigned long int* nodes_size);

/**
 * Return the key in the trie <code>trie</code> that is the longest
 * prefix of <code>complete_string</code>.  To match, the entire key
 * must be a prefix of <code>complete_string</code>.  If no keys are a
 * prefix, <code>NULL</code> is returned. <p>
 *
 * For example, assume "abcx" and "abcxyz123" are the only keys in the
 * trie and this method is called with <code>complete_string</code>
 * equal to "abcxyz", the longest key that is a prefix of the complete
 * string is "abcx".  The result is not "abcxyz123" because the entire
 * key must be a prefix of <code>complete_string</code>. <p>
 *
 * One practical example for the use of this function is if the trie
 * holds IP routing information for subnetworks.  To route a packet to
 * a particular IP address, you will likely want to use the route that
 * is the longest prefix of the IP address.  (See Wikipedia's "Longest
 * Prefix Match" entry.) <p>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] complete_string complete string
 * @return key that is the longest prefix of the complete string
 */
extern CGUL_EXPORT const char*
cgul_trie__find_longest_prefix_of(cgul_exception_t* cex,
                                  cgul_trie_t trie,
                                  const char* complete_string);

/**
 * Return the node holding the first key according to sort order.
 * This operation is *not* O(1) because it has to descend the trie
 * looking for the front node.  If the trie is empty,
 * <code>NULL</code> is returned. <p>
 *
 * The following example shows how to iterate over the entire trie in
 * sort order: <p>
 *
 * <pre>
 *     cgul_trie_node_t n = cgul_trie__get_front(cex, t);
 *     for ( ; n ; n = cgul_trie_node__get_next(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @return front node
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie__get_front(cgul_exception_t* cex,
                     cgul_trie_t trie);

/**
 * Return the node holding the last key according to sort order.  This
 * operation is *not* O(1) because it has to descend the trie looking
 * for the back node.  If the trie is empty, <code>NULL</code> is
 * returned. <p>
 *
 * The following example shows how to iterate over the entire trie in
 * reverse sort order: <p>
 *
 * <pre>
 *     cgul_trie_node_t n = cgul_trie__get_back(cex, t);
 *     for ( ; n ; n = cgul_trie_node__get_prev(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @return back node
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie__get_back(cgul_exception_t* cex,
                    cgul_trie_t trie);

/**
 * Return the oldest node according to chronological order (i.e., the
 * order in which the nodes are inserted).  This operation is O(1).
 * If the trie is empty, <code>NULL</code> is returned.  This class
 * does not have to search the trie in order to find the oldest node
 * because it efficiently keeps a direct pointer to the oldest node up
 * to date. <p>
 *
 * The following example shows how to iterate over the entire trie in
 * chronological order:
 *
 * <pre>
 *     cgul_trie_node_t n = cgul_trie__get_oldest(cex, t);
 *     for ( ; n ; n = cgul_trie_node__get_younger(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @return oldest node
 * @see cgul_trie_node__get_younger()
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie__get_oldest(cgul_exception_t* cex,
                      cgul_trie_t trie);

/**
 * Set the oldest node in the trie to be <code>n</code>.  Calling this
 * method has the potential to confuse iterators and should be handled
 * with roughly the same level of caution as calling
 * <code>cgul_trie__remove_node()</code>. <p>
 *
 * This method could be used, for example, if your code expires the
 * oldest node in the trie, and you want to force <code>n</code> to be
 * the next node to expire. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] n node
 * @see cgul_trie__set_youngest()
 */
extern CGUL_EXPORT void
cgul_trie__set_oldest(cgul_exception_t* cex,
                      cgul_trie_t trie,
                      cgul_trie_node_t n);

/**
 * Return the youngest node according to chronological order (i.e.,
 * the order in which the nodes are inserted).  This operation is
 * O(1).  If the trie is empty, <code>NULL</code> is returned.  This
 * class does not have to search the trie in order to find the
 * youngest node because it efficiently keeps a direct pointer to the
 * youngest node up to date. <p>
 *
 * The following example shows how to iterate over the entire trie in
 * reverse chronological order:
 *
 * <pre>
 *     cgul_trie_node_t n = cgul_trie__get_youngest(cex, t);
 *     for ( ; n ; n = cgul_trie_node__get_older(cex, n)) {
 *         ...
 *     }
 * </pre>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @return youngest node
 * @see cgul_trie_node__get_older()
 */
extern CGUL_EXPORT cgul_trie_node_t
cgul_trie__get_youngest(cgul_exception_t* cex,
                        cgul_trie_t trie);

/**
 * Set the youngest node in the trie to be <code>n</code>.  Calling
 * this method has the potential to confuse iterators and should be
 * handled with roughly the same level of caution as calling
 * <code>cgul_trie__remove_node()</code>. <p>
 *
 * This method could be used, for example, if your code expires the
 * least-recently used (LRU) node.  By calling
 * <code>cgul_trie__set_youngest()</code> each time a node is used,
 * the least-recently used node will be the oldest node in the
 * trie. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] n node
 * @see cgul_trie__set_oldest()
 */
extern CGUL_EXPORT void
cgul_trie__set_youngest(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        cgul_trie_node_t n);

/**
 * Remove the node in <code>trie</code> associated with
 * <code>key_in</code>.  The key and value pointers stored in the node
 * that is to be removed will be returned in <code>key_out</code> and
 * <code>value_out</code> if you pass in pointers that are not
 * <code>NULL</code>.  This method returns <code>1</code> if the node
 * was removed; it returns <code>0</code> otherwise. <p>
 *
 * It is almost always a mistake to naively call this method while
 * iterating over the trie.  See <code>cgul_trie__remove_node()</code>
 * for details. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] key_in key associated with the node to be removed
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_trie__remove(cgul_exception_t* cex,
                  cgul_trie_t trie,
                  const char* key_in,
                  char** key_out,
                  void** value_out);

/**
 * Remove <code>node</code> from the trie <code>trie</code>.  The key
 * and value pointers stored in the node that is to be removed will be
 * returned in <code>key_out</code> and <code>value_out</code> if you
 * pass in pointers that are not <code>NULL</code>. <p>
 *
 * It is almost always a mistake to naively call this method while
 * iterating over the trie.  The problem is that calling this method
 * invalidates the node making it impossible to call
 * <code>cgul_trie_node__get_next()</code> afterward.  The solution is
 * simple.  Just call <code>cgul_trie_node__get_next()</code> before
 * calling this method: <p>
 *
 * <pre>
 *
 *     curr = cgul_trie__find(&local, trie, key);
 *     for ( ; curr ; curr = next) {
 *         next = cgul_trie_node__get_next(&local, curr);
 *         cgul_trie__remove_node(&local,
 *                                trie,
 *                                curr,
 *                                &key_out,
 *                                &value_out);
 *         free(key_out);
 *         free(value_out);
 *     }
 *
 * </pre>
 *
 * Alternatively, you can use
 * <code>cgul_trie__remove_range()</code>,
 * <code>cgul_trie__traverse()</code>, or
 * <code>cgul_trie__traverse_range()</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] n node to be removed
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 */
extern CGUL_EXPORT void
cgul_trie__remove_node(cgul_exception_t* cex,
                       cgul_trie_t trie,
                       cgul_trie_node_t n,
                       char** key_out,
                       void** value_out);

/**
 * This method lets you remove a node based on its sorted order
 * <code>index</code>.  If the index is out of bounds, 0 is returned;
 * otherwise 1 is returned.  If you are interested in getting your
 * hands on the key and value pointers stored in the node that is to
 * be removed, they will be returned in <code>key_out</code> and
 * <code>value_out</code> if you pass in pointers that are not
 * <code>NULL</code>.  <code>index</code> is zero-based. <p>
 *
 * This method throws an exception if <code>trie</code> was not
 * created with <code>cguil_trie__new_with_indexing()</code>. <p>
 *
 * It is almost always a mistake to naively call this method while
 * iterating over the trie.  See <code>cgul_trie__remove_node()</code>
 * for details. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] index index of the sorted key associated with the node
 * to be removed
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_trie__remove_at(cgul_exception_t* cex,
                     cgul_trie_t trie,
                     unsigned long int index,
                     char** key_out,
                     void** value_out);

/**
 * Remove the first key/value pair as determined by sort order from
 * the trie <code>trie</code>.  The key and value pointers stored in
 * the node that is to be removed will be returned in
 * <code>key_out</code> and <code>value_out</code> if you pass in
 * pointers that are not <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_trie__remove_front(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        char** key_out,
                        void** value_out);

/**
 * Remove the last key/value pair as determined by sort order from the
 * trie <code>trie</code>.  The key and value pointers stored in the
 * node that is to be removed will be returned in <code>key_out</code>
 * and <code>value_out</code> if you pass in pointers that are not
 * <code>NULL</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[out] key_out pointer for the key stored in the node
 * @param[out] value_out pointer for the value stored in the node
 * @return whether the node was removed
 */
extern CGUL_EXPORT int
cgul_trie__remove_back(cgul_exception_t* cex,
                       cgul_trie_t trie,
                       char** key_out,
                       void** value_out);

/**
 * This method removes nodes in the range <code>first</code>
 * (inclusive) to <code>last</code> (inclusive).  Strictly as a
 * convenience, this method is an exception to the rule that cgul
 * containers never free keys or values.  If you pass in
 * <code>keys_cache</code> or <code>values_cache</code> instances that
 * are not <code>NULL</code>, the keys or values will be put back on
 * their respective caches. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] first first node in range
 * @param[in] last last node in range
 * @param[in] keys_cache keys cache
 * @param[in] values_cache values cache
 * @return whether any nodes were removed.
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT int
cgul_trie__remove_range(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        cgul_trie_node_t first,
                        cgul_trie_node_t last,
                        cgul_cache_t keys_cache,
                        cgul_cache_t values_cache);

/**
 * This method clears the trie by removing each node individually.
 * Strictly as a convenience, this method is an exception to the rule
 * that cgul containers never free keys or values.  If you pass in
 * <code>keys_cache</code> or <code>values_cache</code> instances that
 * are not <code>NULL</code>, the keys or values will be put back on
 * their respective caches. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] keys_cache keys cache
 * @param[in] values_cache values cache
 * @see cgul_trie__remove_range
 * @see cgul_cache__get_freer()
 */
extern CGUL_EXPORT void
cgul_trie__clear(cgul_exception_t* cex,
                 cgul_trie_t trie,
                 cgul_cache_t keys_cache,
                 cgul_cache_t values_cache);

/**
 * Return the number of key/value pairs stored in the trie
 * <code>trie</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>trie</code> instance
 * @return number of key/value pairs stored in the trie
 */
extern CGUL_EXPORT unsigned long int
cgul_trie__get_size(cgul_exception_t* cex,
                    cgul_trie_t trie);

/**
 * Swap the underlying data for <code>trie1</code> and
 * <code>trie2</code>.  For large tries, this should be much faster
 * than trying to do the same thing using removes and inserts.
 *
 * @param[in] cex c-style exception
 * @param[in] trie1 first <code>cgul_trie</code> instance
 * @param[in] trie2 second <code>cgul_trie</code> instance
 */
extern CGUL_EXPORT void
cgul_trie__swap(cgul_exception_t* cex,
                cgul_trie_t trie1,
                cgul_trie_t trie2);

/**
 * This method performs a left fold of the trie <code>trie</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key in the trie starting at the front of the trie and
 * iterating forward to the end of the trie. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the client data which is where the result of the
 * fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_trie__foldl_keys(cgul_exception_t* cex,
                      cgul_trie_t trie,
                      cgul_trie__fold_key_t f,
                      void* data);

/**
 * This method performs a right fold of the trie <code>trie</code>
 * with the combining function <code>f</code>.  <code>f</code> is
 * called once for each key in the trie starting at the back of the
 * trie and iterating backward to the front of the trie. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_trie__foldr_keys(cgul_exception_t* cex,
                      cgul_trie_t trie,
                      cgul_trie__fold_key_t f,
                      void* data);

/**
 * This method performs a left fold of the trie <code>trie</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each value in the trie starting at the front of the trie
 * and iterating forward to the end of the trie. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_trie__foldl_values(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        cgul_trie__fold_value_t f,
                        void* data);

/**
 * This method performs a right fold of the trie <code>trie</code>
 * with the combining function <code>f</code>.  <code>f</code> is
 * called once for each value in the trie starting at the back of the
 * trie and iterating backward to the front of the trie. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current value.  The third parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_trie__foldr_values(cgul_exception_t* cex,
                        cgul_trie_t trie,
                        cgul_trie__fold_value_t f,
                        void* data);

/**
 * This method performs a left fold of the trie <code>trie</code> with
 * the combining function <code>f</code>.  <code>f</code> is called
 * once for each key/value pair in the trie starting at the front of
 * the trie and iterating forward to the end of the trie.<p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the current value.  The fourth parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_trie__foldl_pairs(cgul_exception_t* cex,
                       cgul_trie_t trie,
                       cgul_trie__fold_pair_t f,
                       void* data);

/**
 * This method performs a right fold of the trie <code>trie</code>
 * with the combining function <code>f</code>.  <code>f</code> is
 * called once for each key/value pair in the trie starting at the
 * back of the trie and iterating backward to the front of the
 * trie. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the current key.  The third parameter passed into
 * <code>f</code> is the current value.  The fourth parameter passed
 * into <code>f</code> is the client data which is where the result of
 * the fold should be accumulated. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * iteration to continue. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] f combining function
 * @param[in] data client data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_trie__foldr_pairs(cgul_exception_t* cex,
                       cgul_trie_t trie,
                       cgul_trie__fold_pair_t f,
                       void* data);

/**
 * Traverse all nodes passing each node to the function
 * <code>f</code>. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the trie <code>trie</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_trie__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_trie__traverse()</code> or
 * <code>cgul_trie__traverse_range()</code> in order to iterate over
 * the trie elements.  In fact, I would recommend that you use
 * <code>cgul_trie_node__get_next()</code> for most of your iteration
 * needs.  If you need to remove nodes though, you should probably use
 * this method.  If you look at this method's source however, you will
 * see that the <code>cgul_trie_node</code> class provides all the
 * public methods you need to safely remove nodes while you are
 * iterating over the trie, but you do have to be careful. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] f traversal callback function
 * @param[in] data passed to <code>f</code>
 */
extern CGUL_EXPORT void
cgul_trie__traverse(cgul_exception_t* cex,
                    cgul_trie_t trie,
                    cgul_trie__traverse_t f,
                    void* data);

/**
 * Traverse all nodes starting with <code>first</code> (inclusive) and
 * ending with <code>last</code> (inclusive) passing each node to the
 * function <code>f</code>.  If you know the first node, but are not
 * sure of the last node, just use <code>NULL</code> as the last node.
 * This will cause this method to iterate until it reaches the end of
 * the trie.  You can than have <code>f</code> return 0 when it
 * determines that the last node has been reached. <p>
 *
 * The first parameter passed into <code>f</code> is the standard
 * c-style exception parameter.  It can be used by <code>f</code> to
 * throw an exception.  The second parameter passed into
 * <code>f</code> is the trie <code>trie</code>.  The third paramenter
 * passed into <code>f</code> is the node for this iteration.  The
 * fourth parameter passed into <code>f</code> is the client data
 * <code>data</code>. <p>
 *
 * <code>f</code> is provided with a safe context in which it can
 * remove the node that is passed into <code>f</code> by calling
 * <code>cgul_trie__remove_node()</code>. <p>
 *
 * <code>f</code> must return true after each iteration in order for
 * the traversal to continue. <p>
 *
 * If <code>first</code> is <code>NULL</code>, iteration starts at the
 * beginning of the trie.  If <code>last</code> is <code>NULL</code>,
 * iteration stops at the end of the trie. <p>
 *
 * NOTE: It is not strictly necessary that you use
 * <code>cgul_trie__traverse()</code> or
 * <code>cgul_trie__traverse_range()</code> in order to iterate over
 * the trie elements.  In fact, I would recommend that you use
 * <code>cgul_trie_node__get_next()</code> for most of your
 * iteration needs.  If you need to remove nodes though, you should
 * probably use this method.  If you look at this method's source
 * however, you will see that the <code>cgul_trie_node</code> class
 * provides all the public methods you need to safely remove nodes
 * while you are iterating over the trie, but you do have to be
 * careful. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] first first node in range
 * @param[in] last last node in range
 * @param[in] f traversal callback function
 * @param[in] data client data passed to <code>f</code>
 * @see cgul_trie__traverse
 */
extern CGUL_EXPORT void
cgul_trie__traverse_range(cgul_exception_t* cex,
                          cgul_trie_t trie,
                          cgul_trie_node_t first,
                          cgul_trie_node_t last,
                          cgul_trie__traverse_t f,
                          void* data);

CGUL_END_C

#endif  /*  CGUL_TRIE_H  */
