#ifndef CGUL_WCHAR_CXX_H
#define CGUL_WCHAR_CXX_H

/**
 * C++ wrapper for <code>cgul_wchar</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_wchar</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_wchar.h"

/**
 * This class provides the C++ bindings for C <code>cgul_wchar</code>
 * external functions.  The main purpose of this class is to convert
 * the C-style function calls and exception handling in
 * <code>cgul_wchar</code> into C++-style function calls and exception
 * handling.
 *
 * @brief C++ bindings for <code>cgul_wchar</code>
 */
class cgul_wchar_cxx {

public:

    /**
     * This function returns the number of 32-bit wide-characters in
     * <code>ws</code>.  The return value is determined by counting all
     * the wide characters that come before the trailing NUL wide
     * character.  The caller is responsible for insuring that
     * <code>ws</code> is NUL-terminated.
     *
     * @param[in] ws 32-bit wide-character string
     * @return number of 32-bit wide characters in <code>ws</code>
     */
    static size_t
    wcslen(const cgul_wchar_t* ws)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wchar__wcslen(&local, ws);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This function copies the 32-bit wide characters in <code>src</code>
     * to <code>dst</code>.  The caller is responsible for making sure
     * that <code>src</code> is NUL-terminated and that <code>dst</code>
     * is large enough to hold all the characters in <code>src</code>
     * including the trailing NUL character.
     *
     * @param[in] dst destination 32-bit wide-character string
     * @param[in] src source 32-bit wide-character string
     */
    static void
    wcscpy(cgul_wchar_t* dst,
           const cgul_wchar_t* src)
    {
        cgul_exception_t local = NULL;
        cgul_wchar__wcscpy(&local, dst, src);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform a case-significant, 32-bit, wide-character string
     * comparison of <code>ws1</code> and <code>ws2</code>.  Return
     * -1, 0, or 1 if <code>ws1</code> is less than, equal to, or
     * greater than <code>ws2</code> respectively. <p>
     *
     * For simplicity, this function compares strings stricly by the
     * ordinal value of each character which should be sufficient, for
     * example, when comparing strings when inserting them as keys
     * into a <code>cgul_rbtree</code>; however, the comparison
     * results are not likely to be the same as what
     * <code>strcoll()</code> would return if used on the UTF-8
     * version of the strings. <p>
     *
     * @param[in] ws1 left-hand side
     * @param[in] ws2 right-hand side
     * @return -1, 0, or 1 if <code>ws1</code> is less than, equal to, or
     * greater than <code>ws2</code> respectively
     * @see cgul_wstring__compare()
     */
    static int
    wcscmp(const cgul_wchar_t* ws1,
           const cgul_wchar_t* ws2)
    {
        return cgul_wchar__wcscmp(ws1, ws2);
    }

    /**
     * Return whether the 32-bit, wide character <code>wc</code> is
     * considered to be white-space.  To avoid locale dependencies,
     * white-space is defined as any character from the following list:
     * ' ', '\\t', '\\n', '\\r', '\\f', '\\v'
     *
     * @param[in] wc 32-bit, wide character
     * @return whether the character is white-space
     */
    static int
    isspace(cgul_wchar_t wc)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wchar__isspace(&local, wc);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

};

#endif  /*  CGUL_WCHAR_CXX_H  */
