#ifndef CGUL_WSTRING_CXX_H
#define CGUL_WSTRING_CXX_H

/**
 * C++ wrapper for <code>cgul_wstring</code>.
 *
 * @file
 * @author Paul Serice
 * @brief C++ wrapper for <code>cgul_wstring</code>
 */

#include "cgul_exception_cxx.h"
#include "cgul_wstring.h"

/**
 * This class provides the C++ bindings for C
 * <code>cgul_wstring</code> objects.  The main purpose of this class
 * is to convert the C-style function calls and exception handling in
 * <code>cgul_wstring</code> into C++-style function calls and
 * exception handling.
 *
 * @brief C++ bindings for <code>cgul_wstring</code>
 */
class cgul_wstring_cxx {

public:

    /**
     * Default constructor that creates a new instance with an empty
     * string.
     */
    cgul_wstring_cxx()
    {
        cgul_exception_t local = NULL;
        obj = cgul_wstring__new(&local);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Copy constructor that creates a new instance by copying the
     * value from <code>rhs</code>.
     *
     * @param[in] rhs right-hand side
     */
    cgul_wstring_cxx(const cgul_wstring_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        obj = cgul_wstring__new_from_wcstring(&local, rhs.get_value());
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Copy constructor that creates a new instance by copying the
     * value from <code>rhs</code>.
     *
     * @param[in] rhs right-hand side
     */
    cgul_wstring_cxx(const cgul_wchar_t* rhs)
    {
        cgul_exception_t local = NULL;
        obj = cgul_wstring__new_from_wcstring(&local, rhs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /** Destructor. */
    virtual
    ~cgul_wstring_cxx()
    {
        if (obj) {
            cgul_wstring__delete(obj);
            obj = NULL;
        }
    }

    /**
     * Assign the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_wstring_cxx&
    operator=(const cgul_wstring_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        if (this != &rhs) {
            cgul_wstring__set_value(&local, obj, rhs.get_value());
            if (local) {
                throw cgul_exception_cxx(local);
            }
        }
        return *this;
    }

    /**
     * Assign the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_wstring_cxx&
    operator=(const cgul_wchar_t* rhs)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__set_value(&local, obj, rhs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return *this;
    }

    /**
     * Append the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_wstring_cxx&
    operator+=(const cgul_wstring_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__append(&local, obj, rhs.get_value());
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return *this;
    }

    /**
     * Append the value from <code>rhs</code> to this object and
     * return this object.
     *
     * @param[in] rhs right-hand side
     * @return *this
     */
    virtual cgul_wstring_cxx&
    operator+=(const cgul_wchar_t* rhs)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__append(&local, obj, rhs);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return *this;
    }

    /**
     * Index operator that returns the wide character at
     * <code>index</code>.  If <code>index</code> is out of bounds, an
     * exception is thrown.
     */
    virtual cgul_wchar_t&
    operator[](size_t index)
    {
        cgul_wchar_t* value = NULL;
        cgul_exception_t local = NULL;

        /* Make sure "index" is within bounds.  Note that we use
         *
         *     index > length()
         *
         * instead of
         *
         *     index >= length()
         *
         * because we want the user to be able to access the trailing null
         * terminator.
         */
        if (index > cgul_wstring__get_length(&local, obj)) {
            local = cgul_exception__new("cgul_wstring_cxx::operator[]: "
                                        "index out of bounds: %lu",
                                        (unsigned long)index);
            throw cgul_exception_cxx(local);
        }

        /* Get the underlying character array. */
        value = cgul_wstring__get_value(&local, obj);

        /* Return the correct reference. */
        return value[index];
    }

    /**
     * Index operator that returns the wide character at
     * <code>index</code>.  If <code>index</code> is out of bounds, an
     * exception is thrown.  This is the overloaded
     * <code>operator[]()</code> that is necessary for use with
     * <code>const</code> objects.
     */
    virtual const cgul_wchar_t&
    operator[](size_t index) const
    {
        cgul_wchar_t* value = NULL;
        cgul_exception_t local = NULL;

        /* Make sure "index" is within bounds. */
        if (index >= cgul_wstring__get_length(&local, obj)) {
            local = cgul_exception__new("cgul_wstring_cxx::operator[]: "
                                        "index out of bounds: %lu",
                                        (unsigned long)index);
            throw cgul_exception_cxx(local);
        }

        /* Get the underlying character array. */
        value = cgul_wstring__get_value(&local, obj);

        /* Return the correct reference. */
        return value[index];
    }

    /**
     * This method allows instances of this class to be cast to
     * C-style strings.
     *
     * @return underlying c-style string
     */
    virtual
    operator cgul_wchar_t*() const
    {
        return get_value();
    }

    /**
     * This method returns a pointer to the internal C-style string.
     * You are allowed to manually set individual characters in the
     * string.  Just remember to call
     * <code>externally_adjusted()</code> afterward.  Compare this
     * method to the <code>take_value()</code> method.
     *
     * @return underlying c-style string
     */
    virtual cgul_wchar_t*
    get_value() const
    {
        cgul_wchar_t* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__get_value(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method lets you take ownership of the underlying C-style
     * string which is returned.  Immediately upon taking ownership,
     * the internal state of this class is reset including resetting a
     * new underlying C-style string to be empty.  One use of this
     * function is to build up a string using
     * <code>cgul_wstring</code> that you then pass to
     * <code>cgul_hash</code> or <code>cgul_rbtree</code>.  Compare
     * this function with <code>get_value()</code>.
     *
     * @return underlying c-style string
     */
    virtual cgul_wchar_t*
    take_value()
    {
        cgul_wchar_t* rv = NULL;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__take_value(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method returns the character indicated by
     * <code>index</code> where <code>index</code> is zero-based.  If
     * <code>index</code> is out of bounds, an exception is thrown.
     *
     * @param[in] index index into the string
     * @return character at <code>index</code>
     */
    virtual cgul_wchar_t
    get_value_at(size_t index) const
    {
        cgul_wchar_t rv = '\0';
        cgul_exception_t local = NULL;
        rv = cgul_wstring__get_value_at(&local, obj, index);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the value of the underlying c-style string to value.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] value new value
     */
    virtual void
    set_value(const cgul_wchar_t* value)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__set_value(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method is equivalent to calling <code>set_value()</code>
     * where <code>value</code> is an empty string.
     */
    virtual void
    clear()
    {
        /* See the comments for this method for why we do not have to
         * pass in a c-style exception. */
        cgul_wstring__clear(obj);
    }

    /**
     * This method converts the string to a <code>signed char</code>.
     * Typically, "base" will be set to 0 to allow conversion for decimal,
     * hexadecimal, and octal based on the leading character in the
     * string.  The string cannot have any extraneous characters.  If it
     * does, the conversion will fail.  On success, the converted value is
     * returned; otherwise, 0 is returned and an exception is thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     * @see get_value_at()
     */
    virtual signed char
    to_schar(int base = 0)
    {
        signed char rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__to_schar(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>unsigned char</code>.
     * Typically, <code>base</code> will be set to 0 to allow conversion
     * for decimal, hexadecimal, and octal based on the leading characters
     * in the string.  The string cannot have any extraneous characters.
     * If it does, the conversion will fail.  On success, the converted
     * value is returned; otherwise, 0 is returned and an exception is
     * thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     * @see get_value_at()
     */
    virtual unsigned char
    to_uchar(int base = 0)
    {
        unsigned char rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__to_uchar(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to a <code>short</code>.
     * Typically, "base" will be set to 0 to allow conversion for decimal,
     * hexadecimal, and octal based on the leading character in the
     * string.  The string cannot have any extraneous characters.  If it
     * does, the conversion will fail.  On success, the converted value is
     * returned; otherwise, 0 is returned and an exception is thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual short int
    to_short(int base = 0)
    {
        short int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__to_short(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>unsigned short</code>.
     * Typically, <code>base</code> will be set to 0 to allow conversion
     * for decimal, hexadecimal, and octal based on the leading characters
     * in the string.  The string cannot have any extraneous characters.
     * If it does, the conversion will fail.  On success, the converted
     * value is returned; otherwise, 0 is returned and an exception is
     * thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual unsigned short int
    to_ushort(int base = 0)
    {
        unsigned short int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__to_ushort(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>int</code>.  Typically,
     * "base" will be set to 0 to allow conversion for decimal,
     * hexadecimal, and octal based on the leading character in the
     * string.  The string cannot have any extraneous characters.  If it
     * does, the conversion will fail.  On success, the converted value is
     * returned; otherwise, 0 is returned and an exception is thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual int
    to_int(int base = 0)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__to_int(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to an <code>unsigned</code>.
     * Typically, <code>base</code> will be set to 0 to allow conversion
     * for decimal, hexadecimal, and octal based on the leading characters
     * in the string.  The string cannot have any extraneous characters.
     * If it does, the conversion will fail.  On success, the converted
     * value is returned; otherwise, 0 is returned and an exception is
     * thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual unsigned int
    to_unsigned(int base = 0)
    {
        unsigned int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__to_unsigned(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to a <code>long</code>.  Typically,
     * <code>base</code> will be set to 0 to allow conversion for decimal,
     * hexadecimal, and octal based on the leading characters in the
     * string.  The string cannot have any extraneous characters.  If it
     * does, the conversion will fail.  On success, the converted value is
     * returned; otherwise, 0 is returned and an exception is thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual long int
    to_long(int base = 0)
    {
        long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__to_long(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }
    
    /**
     * This method converts the string to an <code>unsigned long</code>.
     * Typically, <code>base</code> will be set to 0 to allow conversion
     * for decimal, hexadecimal, and octal based on the leading characters
     * in the string.  The string cannot have any extraneous characters.
     * If it does, the conversion will fail.  On success, the converted
     * value is returned; otherwise, 0 is returned and an exception is
     * thrown.
     *
     * @param[in] base base to use for the conversion
     * @return converted value
     */
    virtual unsigned long int
    to_unsigned_long(int base = 0)
    {
        unsigned long int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__to_unsigned_long(&local, obj, base);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method converts the string to a <code>double</code>.  The
     * string cannot have any extraneous characters.  If it does, the
     * conversion will fail.  On success, the converted value is returned;
     * otherwise, 0.0 is returned and an exception is thrown.
     *
     * @return converted value
     */
    virtual double
    to_double()
    {
        double rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__to_double(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Appends <code>value</code> to the string.  If an error occurs,
     * an exception is thrown.
     *
     * @param[in] value value to append
     */
    virtual void
    append(const cgul_wchar_t* value)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__append(&local, obj, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append a block of non-<code>CGUL_WCHAR__NUL</code> characters.
     * The caller should verify that there are no embedded NUL
     * characters in the string.  If an error occurs, an exception is
     * thrown.
     *
     * @param[in] block block to append
     * @param[in] block_size size of block to append
     */
    virtual void
    append_block(const void* block,
                 size_t block_size)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__append_block(&local, obj, block, block_size);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Append <code>c</code> to the string.
     *
     * @param[in] c wide character to append
     */
    virtual void
    append_char(cgul_wchar_t c)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__append_char(&local, obj, c);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Inserts <code>value</code> at <code>index</code> which can
     * range from 0 to the length (inclusive) of this string.  If
     * <code>index</code> is out of bounds or if memory cannot be
     * allocated, an exception is thrown.
     *
     * @param[in] index index where <code>value</code> is inserted
     * @param[in] value value to insert
     * @param[in] value_length length of value
     */
    virtual void
    insert(size_t index,
           const cgul_wchar_t* value,
           size_t value_length)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__insert(&local, obj, index, value, value_length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Inserts <code>value</code> at <code>index</code> which can
     * range from 0 to the length (inclusive) of this string.  If
     * <code>index</code> is out of bounds or if memory cannot be
     * allocated, an exception is thrown.
     *
     * @param[in] index index where <code>value</code> is inserted
     * @param[in] value value to insert
     */
    virtual void
    insert_char(size_t index,
                cgul_wchar_t value)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__insert_char(&local, obj, index, value);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Pad the current string such that it is at least
     * <code>width</code> characters wide using the <code>pad</code>
     * character.  The string will be right aligned.
     *
     * @param[in] width padding width
     * @param[in] pad character to use for padding
     */
    virtual void
    pad_right(size_t width,
              cgul_wchar_t pad)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__pad_right(&local, obj, width, pad);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Pad the current string such that it is at least
     * <code>width</code> characters wide using the <code>pad</code>
     * character.  The string will be left aligned.
     *
     * @param[in] width padding width
     * @param[in] pad character to use for padding
     */
    virtual void
    pad_left(size_t width,
             cgul_wchar_t pad)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__pad_left(&local, obj, width, pad);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Trim the string by removing leading and trailing white-space where
     * white-space is definded by <code>cgul_wchar::isspace()</code>.
     */
    virtual void
    trim()
    {
        cgul_exception_t local = NULL;
        cgul_wstring__trim(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Trim the string by removing leading white-space where white-space
     * is definded by <code>cgul_wchar::isspace()</code>.
     */
    virtual void
    trim_left()
    {
        cgul_exception_t local = NULL;
        cgul_wstring__trim_left(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Trim the string by removing trailing white-space where white-space
     * is definded by <code>cgul_wchar::isspace()</code>.
     */
    virtual void
    trim_right()
    {
        cgul_exception_t local = NULL;
        cgul_wstring__trim_right(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Reverse the order of the characters in the string.
     */
    virtual void
    reverse()
    {
        cgul_exception_t local = NULL;
        cgul_wstring__reverse(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * This method returns the length of the string.
     *
     * @return length of the string
     */
    virtual size_t
    get_length() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__get_length(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * This method can be used to set the length of the string.
     * It can only be used to make the string shorter.  If it is
     * used in an attempt to make the string longer, an exception
     * is thrown.  If the length of the string is successfully
     * changed, the previous length of the string is returned.
     *
     * @param[in] length new, shorter length of the string
     * @return previous length of the string
     */
    virtual size_t
    set_length(size_t length)
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__set_length(&local, obj, length);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * If you manually alter the underlying buffer by modifying the
     * <code>cgul_wchar_t*</code> string returned by
     * <code>get_value()</code>, you'll need to manually have the
     * class reset its internal state.
     *
     * @see get_value
     */
    virtual void
    externally_adjusted()
    {
        cgul_exception_t local = NULL;
        cgul_wstring__externally_adjusted(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Trim the string by returning excess capacity to the heap.
     */
    virtual void
    trim_capacity()
    {
        cgul_exception_t local = NULL;
        cgul_wstring__trim_capacity(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Get the minimum capacity.
     *
     * @return minimum capacity
     */
    virtual size_t
    get_minimum_capacity() const
    {
        size_t rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__get_minimum_capacity(&local, obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Set the minimum capacity.  The default value is <code>128</code>
     * which means there is always capacity for 127 characters plus one
     * NUL-terminator.  Decreasing the capacity will decrease the amount
     * of space used and (typically) also decrease performance.  If an
     * error occurs, an exception is thrown.
     *
     * @param[in] minimum_capacity minimum capacity
     */
    virtual void
    set_minimum_capacity(size_t minimum_capacity)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__set_minimum_capacity(&local, obj, minimum_capacity);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }

    /**
     * Perform a case-significant wide-character string comparison of
     * <code>*this</code> and <code>rhs</code>.  Return -1, 0, or 1 if
     * <code>*this</code> is less than, equal to, or greater than
     * <code>rhs</code> respectively. <p>
     *
     * For simplicity, this function compares strings stricly by the
     * ordinal value of each character which should be sufficient, for
     * example, when inserting entire <code>cgul_wstring_cxx</code>
     * objects as keys into a <code>cgul_rbtree_cxx</code> object;
     * however, the comparison results are not likely to be the same
     * as what <code>strcoll()</code> would return if used on the
     * UTF-8 version of the strings. <p>
     *
     * @note Use <code>cgul_wchar_cxx::wcscmp()</code> instead if
     * inserting raw <code>cgul_wchar</code> strings as keys into a
     * <code>cgul_rbtree_cxx</code> object. <p>
     *
     * @param[in] rhs right-hand side
     * @return -1, 0, or 1 if <code>*this</code> is less than, equal
     * to, or greater than <code>rhs</code> respectively
     * @see cgul_wchar_cxx::wcscmp()
     */
    virtual int
    compare(const cgul_wstring_cxx& rhs)
    {
        int rv = 0;
        cgul_exception_t local = NULL;
        rv = cgul_wstring__compare(&local, obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
        return rv;
    }

    /**
     * Swap the underlying data for this object and <code>rhs</code>.
     * For large strings, this should be much faster than trying to do
     * the same thing using <code>set_value()</code>.
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    swap(cgul_wstring_cxx& rhs)
    {
        cgul_exception_t local = NULL;
        cgul_wstring__swap(&local, this->obj, rhs.obj);
        if (local) {
            throw cgul_exception_cxx(local);
        }
    }


    /**
     * Get the underlying <code>cgul_wstring</code> object.
     *
     * @return underlying object
     */
    virtual cgul_wstring_t
    get_obj() const
    {
        return obj;
    }

    /**
     * Take the underlying <code>cgul_wstring</code> object.  This
     * means the underlying object will not be deleted when the
     * wrapper goes out of scope.  Also, because you have taken the
     * underlying object, no other methods should be called on this
     * wrapper's instance.  Lastly, after taking the underlying
     * object, it is the caller's responsibility to delete the
     * underlying object by calling
     * <code>cgul_wstring__delete()</code>.
     *
     * @return underlying object
     */
    virtual cgul_wstring_t
    take_obj()
    {
        cgul_wstring_t rv = obj;
        obj = NULL;
        return rv;
    }

    /**
     * Set the new underlying object to <code>rhs</code>.  This causes
     * the old underlying object to be deleted which invalidates any
     * outstanding pointers to or iterators for the old underlying
     * object. <p>
     *
     * This instance takes ownership of <code>rhs</code> which means
     * <code>rhs</code> will be automatically deleted when the C++
     * wrapper is deleted.  To prevent automatic deletion of
     * <code>rhs</code>, call <code>take_obj()</code> when the C++
     * wrapper is no longer needed. <p>
     *
     * @param[in] rhs right-hand side
     */
    virtual void
    set_obj(cgul_wstring_t rhs)
    {
        if (rhs && (rhs != obj)) {
            if (obj) {
                cgul_wstring__delete(obj);
            }
            obj = rhs;
        }
    }

private:

    /** Underlying <code>cgul_wstring</code> object. */
    cgul_wstring_t obj;

};

#endif  /*  CGUL_WSTRING_CXX_H  */
