#include "cgul_base64_p.h"
#include "cgul_libc.h"

/**
 * Map the lower six bits of the index into this array to the base64
 * alphabet.
 */
static const char base64_encode_map[] = {
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
    'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
    'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
    'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
    'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
    'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
    'w', 'x', 'y', 'z', '0', '1', '2', '3',
    '4', '5', '6', '7', '8', '9', '+', '/'
};

/**
 * I decided against having a "base64_decode_map[]" array for
 * portability reasons because not every OS uses ASCII making it
 * non-portable to map from 8-bit ASCII values back to the original
 * 6-bit binary.  Instead, I'll use a switch.
 *
 * If the value returned is BASE64_PAD_CHAR, your input to this
 * function is the pad character.
 *
 * If the value returned is BASE64_ILLEGAL_CHAR, your input to this
 * function is illegal.
 */
static base64_u8 base64_decode_at(const base64_u8 c)
{
    switch( c ) {
    case 'A': return 0;
    case 'B': return 1;
    case 'C': return 2;
    case 'D': return 3;
    case 'E': return 4;
    case 'F': return 5;
    case 'G': return 6;
    case 'H': return 7;
    case 'I': return 8;
    case 'J': return 9;
    case 'K': return 10;
    case 'L': return 11;
    case 'M': return 12;
    case 'N': return 13;
    case 'O': return 14;
    case 'P': return 15;
    case 'Q': return 16;
    case 'R': return 17;
    case 'S': return 18;
    case 'T': return 19;
    case 'U': return 20;
    case 'V': return 21;
    case 'W': return 22;
    case 'X': return 23;
    case 'Y': return 24;
    case 'Z': return 25;
    case 'a': return 26;
    case 'b': return 27;
    case 'c': return 28;
    case 'd': return 29;
    case 'e': return 30;
    case 'f': return 31;
    case 'g': return 32;
    case 'h': return 33;
    case 'i': return 34;
    case 'j': return 35;
    case 'k': return 36;
    case 'l': return 37;
    case 'm': return 38;
    case 'n': return 39;
    case 'o': return 40;
    case 'p': return 41;
    case 'q': return 42;
    case 'r': return 43;
    case 's': return 44;
    case 't': return 45;
    case 'u': return 46;
    case 'v': return 47;
    case 'w': return 48;
    case 'x': return 49;
    case 'y': return 50;
    case 'z': return 51;
    case '0': return 52;
    case '1': return 53;
    case '2': return 54;
    case '3': return 55;
    case '4': return 56;
    case '5': return 57;
    case '6': return 58;
    case '7': return 59;
    case '8': return 60;
    case '9': return 61;
    case '+': return 62;
    case '/': return 63;
    case '=': return BASE64_PAD_CHAR;
    default:
        break;
    }
    return BASE64_ILLEGAL_CHAR;
}

/**
 * Calculate the size of the buffer that is needed to hold the
 * encoding given <code>bin_buf_size</code> bytes of binary data to be
 * encoded.
 *
 * @param[in] cex c-style exception
 * @param[in] b64 base64 instance
 * @param[in] bin_buf_size size in bytes of the binary data to be encoded
 * @return size the encoding buffer needs to be
 */
static size_t
cgul_base64__calculate_encode_buf_size_max(cgul_exception_t* cex,
                                           cgul_base64_t b64,
                                           size_t bin_buf_size)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    assert(bin_buf_size > 0);

    /* First, determine how many 3-byte blocks of binary data are in
     * the input.  Each 3-byte block of binary data encodes to 4
     * characters of output.  Because this is integer arithmetic
     * dividing by 3 ignores any possible remainder.  So, we add one
     * to account for the remainder. */
    rv = (bin_buf_size / 3 + 1) * 4;

    /* Now, account for the number of embedded new lines.  Here,
     * integer division is sufficient because we do not embed a '\n'
     * character for the last line (i.e. the remainder). */
    if( b64->line_length_max > 0 ) {
        rv += rv / b64->line_length_max;
    }

    /* Add 1 for the trailing '\0'. */
    ++rv;

 out:

    return rv;
}

/**
 * Resize the buffer used for encoding if necessary.
 *
 * @param[in,out] cex c-style exception
 * @param[in] b64 base64 instance
 * @param[in] bin_buf_size size in bytes of the binary data to be encoded
 */
static void
cgul_base64__resize_encode_buffer_if_necessary(cgul_exception_t* cex,
                                               cgul_base64_t b64,
                                               size_t bin_buf_size)
{
    base64_u8* new_buf = NULL;
    size_t new_buf_size_max = (size_t)-1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    assert(bin_buf_size > 0);

    new_buf_size_max
        = cgul_base64__calculate_encode_buf_size_max(cex, b64, bin_buf_size);

    if( b64->ebuf_size_max < new_buf_size_max ) {

         new_buf = (base64_u8*)cgul_libc__realloc(cex,
                                                  b64->ebuf,
                                                  new_buf_size_max);
        if (*cex) {
            goto out;
        }

        /* Save the new buffer information. */
        b64->ebuf = new_buf;
        b64->ebuf_size_max = new_buf_size_max;
    }

 out:
    return;
}

/**
 * Resize the buffer used for decoding if necessary.
 *
 * @param[in,out] cex c-style exception
 * @param[in] b64 base64 instance
 * @param[in] b64_buf_size size in bytes of the base64 data to be decoded
 */
static void
cgul_base64__resize_decode_buffer_if_necessary(cgul_exception_t* cex,
                                               cgul_base64_t b64,
                                               size_t b64_buf_size)
{
    base64_u8* new_buf = NULL;
    size_t new_buf_size_max = (size_t)-1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    assert(b64_buf_size > 0);

    new_buf_size_max = (b64_buf_size / 4 + 1) * 3;

    if( b64->dbuf_size_max < new_buf_size_max ) {

         new_buf = (base64_u8*)cgul_libc__realloc(cex,
                                                  b64->dbuf,
                                                  new_buf_size_max);
        if (*cex) {
            goto out;
        }

        /* Save the new buffer information. */
        b64->dbuf = new_buf;
        b64->dbuf_size_max = new_buf_size_max;
    }

 out:
    return;
}

cgul_base64_t
cgul_base64__new(cgul_exception_t* cex)
{
    cgul_base64_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_base64_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up a partially constructed object. */
    if (*cex && rv) {
        cgul_base64__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_base64__delete(cgul_base64_t b64)
{
    if (b64) {
        if( b64->ebuf ) {
            free(b64->ebuf);
        }
        if( b64->dbuf ) {
            free(b64->dbuf);
        }
        free(b64);
    }
}

size_t
cgul_base64__get_line_length_max(cgul_exception_t* cex,
                                 cgul_base64_t b64)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    rv = b64->line_length_max;

 out:
    return rv;
}

void
cgul_base64__set_line_length_max(cgul_exception_t* cex,
                                 cgul_base64_t b64,
                                 size_t line_length_max)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    b64->line_length_max = line_length_max;

 out:
    return;
}


int
cgul_base64__get_strict_decode(cgul_exception_t* cex,
                               cgul_base64_t b64)
{
    int rv = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    rv = b64->strict_decode;

 out:
    return rv;
}

void
cgul_base64__set_strict_decode(cgul_exception_t* cex,
                               cgul_base64_t b64,
                               int strict_decode)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    b64->strict_decode = strict_decode;

 out:
    return;
}

void
cgul_base64__encode(cgul_exception_t* cex,
                    cgul_base64_t b64,
                    const void* buf,
                    size_t buf_size)
{
    base64_u8* buf_as_u8 = NULL;
    base64_u8 bin0 = 0;
    base64_u8 bin1 = 0;
    base64_u8 bin2 = 0;
    size_t i = 0;
    size_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    assert(buf);

    if( buf_size == 0 ) {
        goto out;
    }

    buf_as_u8 = (base64_u8*)buf;

    /* Resize the internal buffer if necessary. */
    cgul_base64__resize_encode_buffer_if_necessary(cex, b64, buf_size);
    if (*cex) {
        goto out;
    }

    /* Account for any remaining bytes that did not get processed the
     * last time cgul_base64__encode() was called. */
    if( b64->e_n_bytes_remaining == 2 ) {
        bin0 = b64->er0;
        bin1 = b64->er1;
        goto get_bin2;
    }
    if( b64->e_n_bytes_remaining == 1 ) {
        bin0 = b64->er0;
        goto get_bin1;
    }

    /* Encode buf. */
    for( ;; ) {

 /* get_bin0: */
        if( i < buf_size ) {
            bin0 = buf_as_u8[i++];
        } else {
            /* Out of Data. */
            b64->e_n_bytes_remaining = 0;
            break;
        }

    get_bin1:
        if( i < buf_size ) {
            bin1 = buf_as_u8[i++];
        } else {
            /* Out of Data. */
            b64->er0 = bin0;
            b64->e_n_bytes_remaining = 1;
            break;
        }

    get_bin2:
        if( i < buf_size ) {
            bin2 = buf_as_u8[i++];
        } else {
            /* Out of Data. */
            b64->er0 = bin0;
            b64->er1 = bin1;
            b64->e_n_bytes_remaining = 2;
            break;
        }

        /* 
         * Encode: The base64 encoding scheme outputs the 6 _most_
         * significant bits first.  Then the next 6 _most_ significant
         * bits.  Etc.  It is easiest to mentally picture the bits
         * flowing from left-to-right instead of the usual
         * right-to-left.
         */

        b64->ebuf[j++]
            = base64_encode_map[bin0 >> 2];

        b64->ebuf[j++]
            = base64_encode_map[ ((0x03 & bin0) << 4) | ((0xf0 & bin1) >> 4) ];

        b64->ebuf[j++]
            = base64_encode_map[ ((0x0f & bin1) << 2) | ((0xc0 & bin2) >> 6) ];

        b64->ebuf[j++]
            = base64_encode_map[0x3f & bin2];

        /* Add a new line if the user has requested it. */
        if( b64->line_length_max  ) {
            b64->line_length_curr += 4;
            if( b64->line_length_curr >= b64->line_length_max ) {
                b64->ebuf[j++] = '\n';
                b64->line_length_curr = 0;
            }
        }

    }

 out:

    if (b64->ebuf) {
        /* NULL terminate the string. */
        b64->ebuf[j] = '\0';
        
        /* Remember the size (excluding the NULL terminator) of the
         * internal buffer. */
        b64->ebuf_size = j;
    }

    return;
}

void
cgul_base64__encode_finish(cgul_exception_t* cex,
                           cgul_base64_t b64)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);

    /* Resize the internal buffer if necessary. */
    cgul_base64__resize_encode_buffer_if_necessary(cex, b64, 2);
    if (*cex) {
        goto out;
    }

    switch( b64->e_n_bytes_remaining ) {
    case 0:
        b64->ebuf[0] = '\0';
        b64->ebuf_size = 0;
        break;
    case 1:
        b64->ebuf[0] = base64_encode_map[b64->er0 >> 2];
        b64->ebuf[1] = base64_encode_map[ (0x03 & b64->er0) << 4 ];
        b64->ebuf[2] = '=';
        b64->ebuf[3] = '=';
        b64->ebuf[4] = '\0';
        b64->ebuf_size = 4;
        break;
    case 2:
        b64->ebuf[0] = base64_encode_map[b64->er0 >> 2];
        b64->ebuf[1] = base64_encode_map[ ((0x03 & b64->er0) << 4)
                                        | ((0xf0 & b64->er1) >> 4) ];
        b64->ebuf[2] = base64_encode_map[ (0x0f & b64->er1) << 2 ];
        b64->ebuf[3] = '=';
        b64->ebuf[4] = '\0';
        b64->ebuf_size = 4;
        break;
    default:
        *cex = cgul_exception__new("cgul_base64__encode_finish: "
                                   "Internal Error: "
                                   "Invalid number of bytes remaining.");
        goto out;
    }

    /* Clean up the internal data so that the next time
     * cgul_base64__encode() is called there will be no stale data. */
    b64->line_length_curr = 0;
    b64->e_n_bytes_remaining = 0;

 out:
    return;
}

char*
cgul_base64__get_encode_conversion(cgul_exception_t* cex,
                                   cgul_base64_t b64,
                                   size_t* buf_size)
{
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);

    if (buf_size) {
        *buf_size = b64->ebuf_size;
    }

    /* If the user never called cgul_base64__encode() and never called
     * cgul_base64__encode_finish(), we need to finish for the
     * caller so that we can at least return an empty string. */
    if (b64->ebuf == NULL) {
        cgul_base64__encode_finish(cex, b64);
    }

    /* Return the result of the conversion. */
    rv = (char*)b64->ebuf;

 out:
    return rv;
}

void
cgul_base64__decode(cgul_exception_t* cex,
                    cgul_base64_t b64,
                    const void* buf,
                    size_t buf_size)
{
    base64_u8* buf_as_u8 = (base64_u8*)buf;
    int bin0 = 0;
    int bin1 = 0;
    int bin2 = 0;
    int bin3 = 0;
    size_t i = 0;
    size_t j = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    assert(buf);

    if( buf_size == 0 ) {
        goto out;
    }

    /* Resize the internal buffer if necessary. */
    cgul_base64__resize_decode_buffer_if_necessary(cex, b64, buf_size);
    if (*cex) {
        goto out;
    }

    /* Account for any remaining bytes that did not get processed the
     * last time cgul_base64__decode() was called. */
    if( b64->d_n_bytes_remaining == 4 ) {
        /* It may seem odd that you have 4 input characters stored
         * because after every 4 input characters you can output 3
         * output characters . . . normally.  What happened here
         * though is that on a previous read, bin2 and bin3 were both
         * the pad characters, but they came at the very end of the
         * input buffer making it impossible to know if they are junk
         * or if they are the real pad characters at the end of the
         * file.
         *
         * If the user calls cgul_base64__decode_finish(), then you
         * know you are at the end of the buffer, but if you are here,
         * it means the user called cgul_base64__decode() again with
         * buf_size > 0 from which we can infer that the pad
         * characters stored in dr2 or dr3 or both are in fact junk
         * which we can simply discard. */
        if( b64->dr2 == BASE64_PAD_CHAR ) {
            bin0 = b64->dr0;
            bin1 = b64->dr1;
            goto get_bin2;
        }
        /* Only b64->dr3 is a pad character. */
        bin0 = b64->dr0;
        bin1 = b64->dr1;
        bin2 = b64->dr2;
        goto get_bin3;
    }
    if( b64->d_n_bytes_remaining == 3 ) {
        bin0 = b64->dr0;
        bin1 = b64->dr1;
        bin2 = b64->dr2;
        goto get_bin3;
    }
    if( b64->d_n_bytes_remaining == 2 ) {
        bin0 = b64->dr0;
        bin1 = b64->dr1;
        goto get_bin2;
    }
    if( b64->d_n_bytes_remaining == 1 ) {
        bin0 = b64->dr0;
        goto get_bin1;
    }

    /* Decode buf. */
    for( ;; ) {

    get_bin0:
        if( i < buf_size ) {
            bin0 = base64_decode_at(buf_as_u8[i++]);
            if( (bin0 == BASE64_ILLEGAL_CHAR) || (bin0 == BASE64_PAD_CHAR) ) {
                /* Invalid character in stream.  The pad character is also
                 * illegal at this point. */
                if( b64->strict_decode ) {
                    *cex = cgul_exception__new(
                               "cgul_base64__decode(): "
                               "Error: non-base64 character in stream.");
                    goto out;
                }
                goto get_bin0;
            }
        } else {
            /* Out of Data. */
            b64->d_n_bytes_remaining = 0;
            break;
        }

    get_bin1:
        if( i < buf_size ) {
            bin1 = base64_decode_at(buf_as_u8[i++]);
            if( (bin1 == BASE64_ILLEGAL_CHAR) || (bin1 == BASE64_PAD_CHAR) ) {
                /* Invalid character in stream.  The pad character is also
                 * illegal at this point. */
                if( b64->strict_decode ) {
                    *cex = cgul_exception__new(
                               "cgul_base64__decode(): "
                               "Error: non-base64 character in stream.");
                    goto out;
                }
                goto get_bin1;
            }
        } else {
            /* Out of Data */
            b64->dr0 = bin0;
            b64->d_n_bytes_remaining = 1;
            break;
        }

    get_bin2:
        if( i < buf_size ) {
            bin2 = base64_decode_at(buf_as_u8[i++]);
            if( bin2 == BASE64_ILLEGAL_CHAR ) {
                /* Invalid character in stream. */
                if( b64->strict_decode ) {
                    *cex = cgul_exception__new(
                               "cgul_base64__decode(): "
                               "Error: non-base64 character in stream.");
                    goto out;
                }
                goto get_bin2;
                /* NOTE: If bin2 == BASE64_PAD_CHAR, we wait until after
                 * reading bin3 to decide whether or not to ignore the pad
                 * character. */
            }
        } else {
            /* Out of Data. */
            b64->dr0 = bin0;
            b64->dr1 = bin1;
            b64->d_n_bytes_remaining = 2;
            break;
        }

        /*
         * This is the tricky one.
         */
    get_bin3:
        if( i < buf_size ) {
            bin3 = base64_decode_at(buf_as_u8[i++]);
            if( bin3 == BASE64_ILLEGAL_CHAR ) {
                /* Invalid character in stream. */
                if( b64->strict_decode ) {
                    *cex = cgul_exception__new(
                               "cgul_base64__decode(): "
                               "Error: non-base64 character in stream.");
                    goto out;
                }
                goto get_bin3;
            }
            if( (bin2 == BASE64_PAD_CHAR) && (bin3 != BASE64_PAD_CHAR) ) {
                /* bin2 is a pad character which means bin3 must also
                 * be a pad character but is not. */
                if( b64->strict_decode ) {
                    *cex = cgul_exception__new(
                               "cgul_base64__decode(): "
                               "Error: non-base64 character in stream.");
                    goto out;
                }
                /* If you are here, it is because we are not in strict
                 * mode meaning we are going to need to shift the non-pad
                 * character in bin3 to bin2 and then re-read bin3. */
                bin2 = bin3;
                goto get_bin3;
            }
            if( bin3 == BASE64_PAD_CHAR ) {
                /* A pad character in bin3 is legal only if it is the last
                 * character in the stream. */
                if( i < buf_size ) {
                    /* There are more characters in the stream.  So
                     * the pad character is junk. */
                    if( b64->strict_decode ) {
                        *cex = cgul_exception__new(
                                   "cgul_base64__decode(): "
                                   "Error: pad character not at "
                                   "end of in stream.");
                        goto out;
                    }
                    goto get_bin3;
                }
                /* Out of Data: If you are here, it is because the
                 * input buffer ended with a pad character.  Because
                 * the input buffer is exhausted, it is impossible to
                 * know whether we are just at the end of this buffer
                 * and more data is coming or whether this is actually
                 * the end of file.  We will store enough data now to
                 * make this determination the next time
                 * cgul_base64__decode() or cgul_base64__decode_finish()
                 * is called. */
                b64->dr0 = bin0;
                b64->dr1 = bin1;
                b64->dr2 = bin2;
                b64->dr3 = bin3;
                b64->d_n_bytes_remaining = 4;  /* Yes, 4! */
                break;
            }
        } else {
            /* Out of Data. */
            b64->dr0 = bin0;
            b64->dr1 = bin1;
            b64->dr2 = bin2;
            b64->d_n_bytes_remaining = 3;
            break;
        }

        /* Decode. */
        b64->dbuf[j++] = (bin0 << 2) | ((0x30 & bin1) >> 4);
        b64->dbuf[j++] = ((0x0f & bin1) << 4) | (0x3c & bin2) >> 2;
        b64->dbuf[j++] = ((0x03 & bin2) << 6) | bin3;

    }

 out:

    /* Remember the size of the internal buffer. */
    if (b64->dbuf) {
        b64->dbuf_size = j;
    }

    return;
}

void
cgul_base64__decode_finish(cgul_exception_t* cex,
                           cgul_base64_t b64)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);

    /* Resize the internal buffer if necessary. */
    cgul_base64__resize_decode_buffer_if_necessary(cex, b64, 4);
    if (*cex) {
        goto out;
    }

    switch( b64->d_n_bytes_remaining ) {
    case 0:
        b64->dbuf_size = 0;
        break;
    case 1:
        if( b64->strict_decode ) {
            *cex = cgul_exception__new("cgul_base64__decode_finish: "
                                       "Error: Unexpected end of data.");
            goto out;
        }
        /* I honestly don't know what to do here.  Dropping a pad
         * character is one thing, but dropping a byte with
         * significant bits is another.  I'll just assume the missing
         * bits were all zeros (as if b64->dr1 were zero) and that pad
         * characters would have followed in b64->dr2 and b64->dr3. */
        b64->dbuf[0] = b64->dr0 << 2;
        b64->dbuf_size = 1;
        break;
    case 2:
        if( b64->strict_decode ) {
            *cex = cgul_exception__new("cgul_base64__decode_finish: "
                                       "Error: Unexpected end of data.");
            goto out;
        }
        /* Assume b64->dr2 and b64->dr3 would have been both been
         * BASE64_PAD_CHARs. */
        b64->dbuf[0] = (b64->dr0 << 2) | ((0x30 & b64->dr1) >> 4);
        b64->dbuf_size = 1;
        break;
    case 3:
        if( b64->strict_decode ) {
            *cex = cgul_exception__new("cgul_base64__decode_finish: "
                                       "Error: Unexpected end of data.");
            goto out;
        }
        /* Assume b64->dr3 would have been a BASE64_PAD_CHAR. */
        if( b64->dr2 == BASE64_PAD_CHAR ) {
            /* Two pad characters. */
            b64->dbuf[0] = (b64->dr0 << 2) | ((0x30 & b64->dr1) >> 4);
            b64->dbuf_size = 1;
        } else {
            /* One pad character. */
            b64->dbuf[0] = (b64->dr0 << 2) | ((0x30 & b64->dr1) >> 4);
            b64->dbuf[1] = ((0x0f & b64->dr1) << 4) | (0x3c & b64->dr2) >> 2;
            b64->dbuf_size = 2;
        }
        break;
    case 4:
        /* This and "case 0:" are the only cases that do not trigger a
         * strict_decode error.  (All encode base64 should be a
         * multiple of 4.)  You reach this section of code when your
         * base64 input stream is properly padded.  The number of
         * bytes you output depends on the padding.  (For the cases
         * above, we just pretend that missing bytes at the end of the
         * data stream would have been padding characters.) */
        if( b64->dr2 == BASE64_PAD_CHAR ) {
            /* Two pad characters. */
            b64->dbuf[0] = (b64->dr0 << 2) | ((0x30 & b64->dr1) >> 4);
            b64->dbuf_size = 1;
        } else {
            /* One pad character. */
            b64->dbuf[0] = (b64->dr0 << 2) | ((0x30 & b64->dr1) >> 4);
            b64->dbuf[1] = ((0x0f & b64->dr1) << 4) | (0x3c & b64->dr2) >> 2;
            b64->dbuf_size = 2;
        }
        break;
    default:
        *cex = cgul_exception__new("cgul_base64__decode_finish: "
                                   "Internal Error: "
                                   "Invalid number of bytes remaining.");
        goto out;
    }

    /* Clean up the internal data so that the next time
     * cgul_base64__decode() is called there will be no stale data. */
    b64->d_n_bytes_remaining = 0;

 out:
    return;
}

char*
cgul_base64__get_decode_conversion(cgul_exception_t* cex,
                                   cgul_base64_t b64,
                                   size_t* buf_size)
{
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(b64);
    assert(buf_size);

    *buf_size = b64->dbuf_size;
    rv = (char*)b64->dbuf;

 out:
    return rv;
}
