#ifndef CGUL_BDF_GLYPH_P_H
#define CGUL_BDF_GLYPH_P_H

#include "cgul_common.h"
#include "cgul_bdf_glyph.h"

CGUL_BEGIN_C

/** Data encapsulated by the <code>cgul_bdf</code> class. */
struct cgul_bdf_glyph {

    /*
     * WARNING: cgul_bdf_glyph__copy() performs a deep copy on this
     * data structure.  If you add pointers to dynamically allocated
     * memory, remember to update the following methods:
     *
     *     cgul_bdf_glyph__copy()
     *     cgul_bdf_glyph__release_resources()
     */

    /**
     * The descriptive name for the glyph if it exists on the
     * STARTCHAR line.  If not, it will be <code>NULL</code>.
     */
    char* descriptive_name;

    /**
     * The code point for this glyph.  It gets its name from the
     * ENCODING line where it is found in the BDF font file.
     */
    cgul_wchar_t encoding;

    /** Scalable width in the x-direction. */
    unsigned int swidth_x;

    /** Scalable width in the y-direction. */
    unsigned int swidth_y;

    /**
     * Device width in the x-direction.  Typically, this is the value
     * you used to advance the origin after drawing this glyph.
     */
    unsigned int dwidth_x;

    /** Device width in the y-direction. */
    unsigned int dwidth_y;

    /** The width of the glyph's bounding box. */
    unsigned int bb_width;

    /** The height of the glyph's bounding box. */
    unsigned int bb_height;

    /** The x-offset of the lower-left corner of the glyph's bounding box. */
    int bb_x_offset;

    /** The y-offset of the lower-left corner of the glyph's bounding box. */
    int bb_y_offset;

    /**
     * The bitmap that describes the glyph.  The pixel in the
     * upper-left corner of the glyph's bounding box is represented as
     * the least-significant bit of <code>bitmap[0]</code>.  Each row
     * is padded out with zeros to the nearest byte.  You can
     * determine how many bytes per row by examining
     * <code>bb_width</code>.  If you need to set or get individual
     * pixels, call <code>set_dot()</code> or <code>get_dot()</code>.
     */
    unsigned char* bitmap;

    /**
     * Because the glyph is stored as a bitmap with 8 pixels per byte,
     * this member stores the width of the bitmap in bytes.
     */
    size_t bytes_wide;

    /**
     * The byte offset into the BDF file where the STARTCHAR line
     * begins.  This is used to generate the *.bdfi index file so that
     * glyphs can be loaded on demand.
     */
    unsigned long int file_offset;

};

/**
 * Set the descriptive name.  This is the name of the character that
 * appears on the STARTCHAR line.  If you want to remove the
 * descriptive name, pass in <code>NULL</code> instead which will
 * cause this method to clear the descriptive name; otherwise, this
 * method makes a copy of <code>descriptive_name</code>.  If an error
 * occurs during the copy, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] descriptive_name descriptive name
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_descriptive_name(cgul_exception_t* cex,
                                     cgul_bdf_glyph_t glyph,
                                     const char* descriptive_name);

/**
 * Set the encoding.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] encoding encoding
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_encoding(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             cgul_wchar_t encoding);

/**
 * Set the scalable width in the x-direction.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] swidth_x scalable width in the x-direction
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_swidth_x(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int swidth_x);

/**
 * Set the scalable width in the y-direction.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] swidth_y scalable width in the y-direction
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_swidth_y(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int swidth_y);

/**
 * Set the device width in the x-direction.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] dwidth_x device width in the x-direction
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_dwidth_x(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int dwidth_x);

/**
 * Set the device width in the y-direction.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] dwidth_y device width in the y-direction
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_dwidth_y(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int dwidth_y);

/**
 * Set the width of the glyph's bounding box.  Note that setting
 * the width of the glyph's bound box forces the underlying bitmap
 * data to be invalidated and reallocated in order to match the
 * new dimensions.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] bb_width width of the glyph's bounding box
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_bb_width(cgul_exception_t* cex,
                             cgul_bdf_glyph_t glyph,
                             unsigned int bb_width);

/**
 * Set the height of the glyph's bounding box.  Note that setting
 * the height of the glyph's bound box forces the underlying
 * bitmap data to be invalidated and reallocated in order to match
 * the new dimensions.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] bb_height height of the glyph's bounding box
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_bb_height(cgul_exception_t* cex,
                              cgul_bdf_glyph_t glyph,
                              unsigned int bb_height);

/**
 * Set the x-offset of the lower-left corner of the glyph's bounding box.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] bb_x_offset x-offset of the lower-left corner of the
 * glyph's bounding box
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_bb_x_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph,
                                int bb_x_offset);

/**
 * Set the y-offset of the lower-left corner of the glyph's bounding box.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] bb_y_offset y-offset of the lower-left corner of the
 * glyph's bounding box
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_bb_y_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph,
                                int bb_y_offset);

/**
 * This method sets the dot at <code>(x,y)</code> inside of the
 * bounding box.  If the dot is set to <code>true</code>, it should be
 * drawn when rendering the glyph.  If the dot is set to
 * <code>false</code>, it should not be drawn.  The x- and
 * y-coordinates are relative to the top-left corner of the glyph's
 * bounding box, not the top-left corner of the glyph itself.  Also
 * note that <code>x</code> and <code>y</code> values start with zero,
 * not one.
 *
 * @param[in,out] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] x x-coordinate
 * @param[in] y y-coordinate
 * @param[in] is_set whether the dot should be drawn
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_dot(cgul_exception_t* cex,
                        cgul_bdf_glyph_t glyph,
                        unsigned int x,
                        unsigned int y,
                        int is_set);

/**
 * Set the file offset of the STARTCHAR line that describes this glyph
 * in the BDF.  This information is used to generate the *.bdfi index
 * file so that the glyph can be loaded on demand.
 *
 * @param[in] cex c-style exception
 * @param[in] glyph <code>cgul_bdf_glyph</code> instance
 * @param[in] offset file offset
 */
extern CGUL_EXPORT void
cgul_bdf_glyph__set_file_offset(cgul_exception_t* cex,
                                cgul_bdf_glyph_t glyph,
                                unsigned long int offset);

CGUL_END_C

#endif  /*  CGUL_BDF_GLYPH_P_H  */
