#ifndef CGUL_BIG_INTEGER_P_H
#define CGUL_BIG_INTEGER_P_H

#include "cgul_common.h"
#include "cgul_big_integer.h"

CGUL_BEGIN_C

/**
 * cgul big integer (cbi) element type -- Just make sure that
 * <code>cbi_element_double_t</code> is twice the size of
 * <code>cbi_element_t</code>.
 */
#if defined(UULONG_MAX) && (ULONG_MAX < UULONG_MAX)
    typedef unsigned long int  cbi_element_t;
    typedef unsigned long long int cbi_element_double_t;
#elif defined(UULONG_MAX) && (UINT_MAX < UULONG_MAX)
    typedef unsigned int  cbi_element_t;
    typedef unsigned long long int cbi_element_double_t;
#elif (UINT_MAX < ULONG_MAX)
    typedef unsigned int  cbi_element_t;
    typedef unsigned long int cbi_element_double_t;
#elif (USHRT_MAX < ULONG_MAX)
    typedef unsigned short int cbi_element_t;
    typedef unsigned long int cbi_element_double_t;
#elif (UCHAR_MAX < ULONG_MAX)
    typedef unsigned char cbi_element_t;
    typedef unsigned long int cbi_element_double_t;
#else
#    error "Unable to determine cbi_element_t and cbi_element_double_t"
#endif


#define CBI_ELEMENT_BITS (CHAR_BIT * sizeof(cbi_element_t))
#define CBI_ELEMENT_DOUBLE_BITS (CHAR_BIT * sizeof(cbi_element_double_t))


/**
 * Data encapsulated by the <code>cgul_big_integer</code> class.
 */
struct cgul_big_integer {

    /**
     * The <code>value</code> array holds your arbitrary precision
     * integer.  It is resized on-the-fly as necessary.  The
     * least-significant bytes are in value[0].  The most-significant
     * bytes are in value[size - 1].
     */
    cbi_element_t* value;

    /**
     * <code>is_positive</code> stores the sign of the value.  If the
     * value is not zero, a value of 1 means the integer is positive,
     * and a value of 0 means the integer is negative.  When you
     * manipulate this value, you must be very careful to always
     * account for both "positive" and "negative" zero.
     */
    int is_positive;

    /**
     * <code>size</code> is the number of <code>cbi_element_t</code>
     * elements currently active in the <code>value</code> array.
     * <code>size</code> is always less than or equal to
     * <code>capacity</code>.  Most importantly, <code>size</code>
     * should always be as small as possible.  Many of the
     * <code>cgul_big_integer</code> methods depend on this
     * invariant.  Call <code>cgul_big_integer__minimize_size()</code>
     * to maintain this invariant.
     */
    size_t size;

    /**
     * The current capacity of the <code>value</code> array.  It is
     * the number of <code>cbi_element_t</code> elements that
     * <code>value</code> can hold.
     */
    size_t capacity;

    /**
     * When allocating new capacity, allocate at least
     * <code>minimum_capacity_increment</code> more
     * <code>cbi_element_t</code> elements.
     */
    size_t minimum_capacity_increment;
};

CGUL_END_C

#endif  /*  CGUL_BIG_INTEGER_P_H  */
