#include "cgul_bitwise_trie_p.h"
#include "cgul_bitwise_trie_node_p.h"
#include "cgul_libc.h"

/************************************************************************
 * General Comments
 ************************************************************************/

/*
 * 1) The root node is empty!  In other words, the root node does not
 * correspond to any index value and has the following form where the
 * "even_subtrie" is the subtrie composed of nodes that correspond to
 * all the even index values and "odd_subtrie" is the subtrie composed
 * of nodes that correspond to all the odd index values:
 *
 *                 +---------+
 *                 |  root   |
 *                 | (empty) |
 *                 +---------+
 *                /           \
 *               /             \
 *         <even_subtrie>   <odd_subtrie>
 *
 * 2) The rules for navigating the trie are relatively simple.  Start
 * with the least-significant bit (lsb).  If it is zero, go left.  If
 * it is one, go right.  Repeat for each bit to navigate to the next
 * level of trie.
 *
 * 3) To insert a node, follow the path described above until you
 * reach a subtrie that is NULL.  Replace the NULL value with the new
 * node.
 *
 * 4) To delete a node, just remove the node from the trie and replace
 * it with either the left or right child.  It doesn't matter which as
 * all children (left or right) are guaranteed to have the same prefix
 * as the node being removed.
 *
 * 5) To search for a node follow the path described above until a
 * node having the same index value is reach or until a NULL subtrie
 * is reached.
 *
 * 6) Note, the only guarantee this implementation provides regarding
 * the location of the node is that it is somewhere along the path
 * described above.  It could be the first node, the last node, or
 * somewhere in between.  This makes it easy to insert nodes in random
 * order without affecting the worst-case search time which is still
 * O(log N).  It also opens up the possibility of easily moving nodes
 * that are used with a higher frequency closer to the top of the
 * trie.
 */

/************************************************************************
 * Trie Utilities
 ************************************************************************/

/**
 * This method calls <code>cgul_cache__put()</code> on each node in
 * <code>t</code>.  It does not bother to remove the nodes from the
 * trie because this function should only be called from
 * <code>cgul_bitwise_trie__delete()</code>.  Because this method is only
 * called from the destructor, it is intentional that a
 * <code>cgul_exception</code> instance is not passed to this method.
 *
 * @param[in] t <code>cgul_bitwise_trie</code> instance
 */
static void
cgul_bitwise_trie__free_all_nodes(cgul_bitwise_trie_t t)
{
    cgul_bitwise_trie_node_t curr = NULL;
    cgul_bitwise_trie_node_t next = NULL;
    cgul_exception_t local = NULL;

    assert(t);

    /* Iterate over all the nodes in the trie using the embedded
     * older/younger linked list. */
    for (curr = t->youngest ; curr ; curr = next) {
        /* Get the next node while we still can. */
        next = curr->older;
        /* Put the current node back on its cache. */
        cgul_cache__put(&local, t->cache, curr);
    }

    /* Because the root node does not hold any values, it is not part
     * of oloder/younger linked list so it has to be freed separately.
     * (See comment at the top of this source file.) */
    if (t->root) {
        cgul_cache__put(&local, t->cache, t->root);
        t->root = NULL;
    }

    return;
}

/**
 * Detach a leaf node from the path that includes the node
 * <code>node</code>.  If <code>node</code> has no leaf nodes,
 * <code>node</code> is returned; otherwise, the trie will be
 * traversed by always going left if possible and going right
 * otherwise until a leaf node is found. <p>
 *
 * This method is used by <code>cgul_bitwise_trie__remove()</code> to find a
 * suitable leaf node that can be used to replace the internal node
 * <code>node</code>. <p>
 *
 * @param[in] cex c-style exception
 * @return detached leaf node
 */
static cgul_bitwise_trie_node_t
cgul_bitwise_trie__detach_leaf_node(cgul_exception_t* cex,
                                    cgul_bitwise_trie_node_t node)
{
    cgul_bitwise_trie_node_t rv = NULL;
    cgul_bitwise_trie_node_t parent = NULL;
    cgul_bitwise_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(node);

    /* If node is a leaf node, just return it. */
    if (!node->children[0] && !node->children[1]) {
        rv = node;
        goto out;
    }

    /* Initialize.  Because of the above if(), we know that parent is
     * not a leaf node so curr will not be NULL when the while() loop
     * is first entered. */
    parent = node;
    curr = parent->children[0] ? parent->children[0] : parent->children[1];
    assert(curr);

    /* Find a leaf node. */
    for (;;) {
        if (curr->children[0]) {
            parent = curr;
            curr = curr->children[0];
        } else if (curr->children[1]) {
            parent = curr;
            curr = curr->children[1];
        } else {
            /* We've reached the leaf node. */
            break;
        }
    }

    /* Detach the leaf node. */
    if (curr == parent->children[0]) {
        parent->children[0] = NULL;
    } else {
        assert(curr == parent->children[1]);
        parent->children[1] = NULL;
    }

    /* Success. */
    rv = curr;

 out:
    return rv;
}

/************************************************************************
 * Public Methods
 ************************************************************************/

cgul_bitwise_trie_t
cgul_bitwise_trie__new(cgul_exception_t* cex)
{
    cgul_bitwise_trie_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the trie. */
    rv = (cgul_bitwise_trie_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Allocate the cache from which nodes will be drawn. */
    rv->cache = cgul_cache__new(
                    cex,
                    (cgul_cache__constructor_t)&cgul_bitwise_trie_node__new,
                    (cgul_cache__destructor_t)&cgul_bitwise_trie_node__delete,
                    (cgul_cache__restructor_t)&cgul_bitwise_trie_node__initialize);
    if (*cex) {
        goto out;
    }

    /* Allocate the root node which only has left and right children.
     * It never stores a value associated with an index. */
    rv->root = (cgul_bitwise_trie_node_t)cgul_cache__get(cex, rv->cache);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (*cex && rv) {
        cgul_bitwise_trie__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_bitwise_trie__delete(cgul_bitwise_trie_t t)
{
    if (t) {

        /* cache */
        if (t->cache) {
            /* Delete all the nodes.  This might move some of the
             * nodes to the cache.  Thus, this must be done before
             * deleting the cache. */
            cgul_bitwise_trie__free_all_nodes(t);
            cgul_cache__delete(t->cache);
        }

        /* Delete the object itself. */
        free(t);
    }
}

void
cgul_bitwise_trie__free_values(cgul_bitwise_trie_t t)
{
    cgul_bitwise_trie_node_t curr = NULL;

    if (t) {
        /* Iterate over each node. */
        for (curr = t->oldest ; curr ; curr = curr->younger) {
            /* Free the value. */
            free((void*)curr->value);
        }
    }
}

unsigned long int
cgul_bitwise_trie__get_cache_size(cgul_exception_t* cex,
                                  cgul_bitwise_trie_t t)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = cgul_cache__get_size(cex, t->cache);

 out:
    return rv;
}

void
cgul_bitwise_trie__set_cache_size(cgul_exception_t* cex,
                                  cgul_bitwise_trie_t t,
                                  unsigned long int size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    cgul_cache__set_size(cex, t->cache, size);

 out:
    return;
}

unsigned long int
cgul_bitwise_trie__get_cache_reserve(cgul_exception_t* cex,
                                     cgul_bitwise_trie_t t)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = cgul_cache__get_reserve(cex, t->cache);

 out:
    return rv;
}

void
cgul_bitwise_trie__set_cache_reserve(cgul_exception_t* cex,
                                     cgul_bitwise_trie_t t,
                                     unsigned long int reserve)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    cgul_cache__set_reserve(cex, t->cache, reserve);

 out:
    return;
}

int
cgul_bitwise_trie__is_empty(cgul_exception_t* cex,
                            cgul_bitwise_trie_t t)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = (t->size == 0);

 out:
    return rv;
}

void
cgul_bitwise_trie__assign(cgul_exception_t* cex,
                          cgul_bitwise_trie_t t,
                          unsigned long int index,
                          void* ptr)
{
    unsigned long int path = 0;
    cgul_bitwise_trie_node_t parent = NULL;
    cgul_bitwise_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    /* The path through the trie is uniquely described by the bits in
     * the index. */
    path = index;

    /* Search the path for an existing node having the same index.
     * Remember that the root node always exists and does NOT map to
     * any index but does have valid child subtries. */
    parent = t->root;
    for (curr = t->root->children[path & 0x01];
         curr;
         parent = curr, curr = curr->children[(path >>= 1) & 0x01])
    {
        /* Break if an existing node with the same index is found. */
        if (curr->index == index) {
            break;
        }
    }

    if (!curr) {

        /* If the node does not exist, create it. */
        curr = (cgul_bitwise_trie_node_t)cgul_cache__get(cex, t->cache);
        if (*cex) {
            goto out;
        }

        /* Initialize it and update the older/younger linked list. */
        curr->index = index;
        if (!t->oldest) {
            t->oldest = curr;
        }
        if (t->youngest) {
            t->youngest->younger = curr;
            curr->older = t->youngest;
        }
        t->youngest = curr;

        /* Update the parent. */
        parent->children[path & 0x01] = curr;

        /* Increment the size of the trie. */
        ++t->size;
    }

    /* Assign the value. */
    curr->value = ptr;

    /* Success.  The trie now owns the node. */
    curr = NULL;

 out:

    /* Clean up. */
    if (curr) {
        cgul_cache__put(cex, t->cache, curr);
    }

    return;
}

void*
cgul_bitwise_trie__at(cgul_exception_t* cex,
                      cgul_bitwise_trie_t t,
                      unsigned long int index,
                      int* exists)
{
    void* rv = NULL;
    unsigned long int path = 0;
    cgul_bitwise_trie_node_t curr = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Note that "exists" is allowed to be NULL. */
    assert(t);

    /* The path through the trie is uniquely described by the bits in
     * the index. */
    path = index;

    /* Search the path described by the bits in "index" for an
     * existing node having the same index.  Remember that the root
     * node always exists and does NOT map to any index but does have
     * valid child subtries.  Also remember that the node could be
     * anywhere along the path depending on the order of
     * insertions. */
    for (curr = t->root->children[path & 0x01];
         curr;
         curr = curr->children[(path >>= 1) & 0x01])
    {
        /* Break if an existing node with the same index is found. */
        if (curr->index == index) {
            break;
        }
    }

    if (curr) {
        /* Success. */
        rv = (void*)curr->value;
        if (exists) {
            *exists = 1;
        }
    } else {
        /* Failure. */
        rv = NULL;
        if (exists) {
            *exists = 0;
        }
    }

 out:
    return rv;
}

int
cgul_bitwise_trie__remove(cgul_exception_t* cex,
                          cgul_bitwise_trie_t t,
                          unsigned long int index,
                          void** value_out)
{
    int rv = 0;
    unsigned long int path = 0;
    cgul_bitwise_trie_node_t parent = NULL;
    cgul_bitwise_trie_node_t curr = NULL;
    cgul_bitwise_trie_node_t leaf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Note that "value_out" is allowed to be NULL. */
    assert(t);

    /* Clear the user's output parameter. */
    if (value_out) {
        *value_out = NULL;
    }

    /* The path through the trie is uniquely described by the bits in
     * the index. */
    path = index;

    /* Search the path for an existing node having the same index.
     * Remember that the root node always exists and does NOT map to
     * any index but does have valid child subtries. */
    parent = t->root;
    for (curr = t->root->children[path & 0x01];
         curr;
         parent = curr, curr = curr->children[(path >>= 1) & 0x01])
    {
        /* Break if an existing node with the same index is found. */
        if (curr->index == index) {
            break;
        }
    }

    /* If the node was found, remove it. */
    if (curr) {

        /*
         * Update the trie.
         */

        if (!curr->children[0] && !curr->children[1]) {
            /* If this is a leaf node, just propagate the NULL up. */
            parent->children[path & 0x01] = NULL;
        } else {
            /* At least one child exists.  Detach a leaf node and use
             * it to replace curr.  Note that any node along the same
             * path as curr shares a common prefix with curr which
             * makes this possible. */
            cgul_bitwise_trie_node_t child =
                curr->children[0] ? curr->children[0] : curr->children[1];
            leaf = cgul_bitwise_trie__detach_leaf_node(cex, child);
                       

            /* Reattach the leaf node so that it replaces curr.  You
             * have to be careful here to avoid making leaf its own
             * child which can happen when you are replacing the
             * parent of leaf with leaf. */
            parent->children[path & 0x01] = leaf;
            if (curr->children[0] == leaf) {
                leaf->children[0] = NULL;
            } else {
                leaf->children[0] = curr->children[0];
            }
            if (curr->children[1] == leaf) {
                leaf->children[1] = NULL;
            } else {
                leaf->children[1] = curr->children[1];
            }
        }

        /*
         * Update the older/younger linked list.
         */
        if (curr->younger) {
            curr->younger->older = curr->older;
        }
        if (curr->older) {
            curr->older->younger = curr->younger;
        }
        if (t->oldest == curr) {
            t->oldest = curr->younger;
        }
        if (t->youngest == curr) {
            t->youngest = curr->older;
        }

        /* Decrement the size. */
        --t->size;

        /* Pass the value pointer back to the user. */
        if (value_out) {
            *value_out = (void*)curr->value;
        }

        /* Free the node. */
        cgul_cache__put(cex, t->cache, curr);
        curr = NULL;

        /* Success. */
        rv = 1;
    }

 out:
    return rv;
}

cgul_bitwise_trie_node_t
cgul_bitwise_trie__get_oldest(cgul_exception_t* cex,
                              cgul_bitwise_trie_t t)
{
    cgul_bitwise_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->oldest;

 out:
    return rv;
}

void
cgul_bitwise_trie__set_oldest(cgul_exception_t* cex,
                              cgul_bitwise_trie_t t,
                              cgul_bitwise_trie_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(t->oldest);
    assert(n);

    /* There is nothing to do if n is already the oldest. */
    if (n == t->oldest) {
        goto out;
    }

    /* If n is currently the youngest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new youngest node. */
    if (n == t->youngest) {
        t->youngest = n->older;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the oldest node.  We assert t->oldest above because
     * any non-empty trie has an oldest node, and because n is a node
     * in t, t is non-empty. */
    t->oldest->older = n;
    n->younger = t->oldest;
    n->older = NULL;

    /* Remember that n is now the oldest node. */
    t->oldest = n;

 out:
    return;
}

cgul_bitwise_trie_node_t
cgul_bitwise_trie__get_youngest(cgul_exception_t* cex,
                                cgul_bitwise_trie_t t)
{
    cgul_bitwise_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->youngest;

 out:
    return rv;
}

void
cgul_bitwise_trie__set_youngest(cgul_exception_t* cex,
                                cgul_bitwise_trie_t t,
                                cgul_bitwise_trie_node_t n)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);
    assert(t->youngest);
    assert(n);

    /* There is nothing to do if n is already the youngest. */
    if (n == t->youngest) {
        goto out;
    }

    /* If n is currently the oldest, it will not be when this method
     * returns.  In this case, we need to remember which node is the
     * new oldest node. */
    if (n == t->oldest) {
        t->oldest = n->younger;
    }

    /* Carefully remove n from the list. */
    if (n->older) {
        n->older->younger = n->younger;
    }
    if (n->younger) {
        n->younger->older = n->older;
    }

    /* Insert n as the youngest node.  We assert t->youngest above because
     * any non-empty trie has an youngest node, and because n is a node
     * in t, t is non-empty. */
    t->youngest->younger = n;
    n->older = t->youngest;
    n->younger = NULL;

    /* Remember that n is now the youngest node. */
    t->youngest = n;

 out:
    return;
}

unsigned long int
cgul_bitwise_trie__get_size(cgul_exception_t* cex,
                            cgul_bitwise_trie_t t)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t);

    rv = t->size;

 out:
    return rv;
}

void
cgul_bitwise_trie__swap(cgul_exception_t* cex,
                        cgul_bitwise_trie_t t1,
                        cgul_bitwise_trie_t t2)
{
    struct cgul_bitwise_trie tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(t1);
    assert(t2);

    /* Nothing to swap if they point to the same objects. */
    if (t1 == t2) {
        goto out;
    }

    /* A shallow copy, by design, is exactly what is needed. */
    tmp = *t1;
    *t1 = *t2;
    *t2 = tmp;

 out:
    return;
}
