#include "cgul_bitwise_trie_node_p.h"
#include "cgul_libc.h"

/*
 * This method is used as the "restructor" for the "cgul_cache" of
 * nodes in the "cgul_bitwise_trie".  "restructor" methods are defined in
 * "cgul_cache" to not throw exceptions.
 *
 * While the signature for this method does not exactly match the
 * signature defined for "restructor" methods, it can be safely
 * typecast to match.
 */
void
cgul_bitwise_trie_node__initialize(cgul_bitwise_trie_node_t n)
{
    assert(n);
    memset(n, 0, sizeof(*n));
}

cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__new(cgul_exception_t* cex)
{
    cgul_bitwise_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_bitwise_trie_node_t)cgul_libc__malloc(cex, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_bitwise_trie_node__initialize(rv);

 out:

    /* Delete partially constructed objects. */
    if (*cex && rv) {
        cgul_bitwise_trie_node__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_bitwise_trie_node__delete(cgul_bitwise_trie_node_t n)
{
    /* If you alter this, you will also likely need to alter
     * cgul_bitwise_trie_node_with_indexing__delete() which inherits from
     * this class. */
    if (n) {
        free(n);
    }
}

unsigned long int
cgul_bitwise_trie_node__get_index(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->index;

 out:
    return rv;
}

void*
cgul_bitwise_trie_node__get_value(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = (void*)n->value;

 out:
    return rv;
}

void
cgul_bitwise_trie_node__set_index(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n,
                                  unsigned long int index)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    n->index = index;

 out:
    return;
}

void
cgul_bitwise_trie_node__set_value(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n,
                                  const void* value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    n->value = value;

 out:
    return;
}

cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__get_left_child(cgul_exception_t* cex,
                                       cgul_bitwise_trie_node_t n)
{
    cgul_bitwise_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->children[0];

 out:
    return rv;
}

cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__get_right_child(cgul_exception_t* cex,
                                        cgul_bitwise_trie_node_t n)
{
    cgul_bitwise_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->children[1];

 out:
    return rv;
}

cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__get_older(cgul_exception_t* cex,
                                  cgul_bitwise_trie_node_t n)
{
    cgul_bitwise_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->older;

 out:
    return rv;
}

cgul_bitwise_trie_node_t
cgul_bitwise_trie_node__get_younger(cgul_exception_t* cex,
                                    cgul_bitwise_trie_node_t n)
{
    cgul_bitwise_trie_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->younger;

 out:
    return rv;
}
