#include "cgul_crlf_file_p.h"
#include "cgul_get_line.h"
#include "cgul_libc.h"
#include "cgul_vector.h"

#define CGUL_CRLF_FILE__DEFAULT_BUFSIZE (16384)

/**
 * This is a helper for <code>cgul_crlf_file__get_line()</code> that
 * gets the next line of text when the data source is a block-buffered
 * file or a memory-based file (which can easily be treated as
 * block-buffered by simply using a <code>memcpy()</code>).
 *
 * @param[in,out] cex c-style exception
 * @param[in] crlf_file <code>cgul_crlf_file</code> instance
 * @return next line
 */
static const char*
cgul_crlf_file__get_line__block_buffered(cgul_exception_t* cex,
                                         cgul_crlf_file_t crlf_file)
{
    const char* rv = NULL;
    size_t rcount = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Try to get the next line from the underlying crlf object. */
    rv = cgul_crlf__get_line(cex, crlf_file->crlf);

    /* We have a line so just return. */
    if (rv) {
        goto out;
    }

    /*
     * We are here because the underlying crlf object has exhausted
     * the current block so we need to feed it the next block.
     */

    /* We need to be able to handle very long lines of text.  So, we
     * keep looping until the underlying cgul_crlf object reports
     * that it has a line or we reach EOF. */
    for (;;) {

        /* Read the next block into the buffer. */
        switch (crlf_file->source_type) {

        /* block-buffered file */
        case CGUL_CRLF_FILE_SOURCE__BLOCK_BUFFERED_FILE:
            rcount = cgul_libc__fread(cex,
                                      crlf_file->buf,
                                      1,
                                      crlf_file->bsize,
                                      crlf_file->f);
            if (*cex) {
                goto out;
            }
            break;

        /* memory */
        case CGUL_CRLF_FILE_SOURCE__MEMORY:
            assert(crlf_file->moffset <= crlf_file->msize);
            rcount = crlf_file->bsize;
            if (rcount + crlf_file->moffset > crlf_file->msize) {
                rcount = crlf_file->msize - crlf_file->moffset;
            }
            memcpy(crlf_file->buf, crlf_file->mem + crlf_file->moffset, rcount);
            crlf_file->moffset += rcount;
            break;

        /* error */
        default:
            *cex = cgul_exception__new("cgul_crlf_file__get_line: "
                                       "not block-buffered data source: %d",
                                       crlf_file->source_type);
            goto out;
        }

        /* Check for EOF. */
        if (rcount == 0) {
            break;
        }

        /* Feed the buffer to the cgul_crlf object. */
        cgul_crlf__convert(cex,
                           crlf_file->crlf,
                           (char*)crlf_file->buf,
                           (unsigned long)rcount);
        if (*cex) {
            goto out;
        }

        /* Try to get the next line of text.  Note, that this does not
         * necessarily return a line of text.  For example, we just
         * read in the last line of a file which was missing the final
         * EOL character or we just read a line of text that is larger
         * than our buffer. */
        rv = cgul_crlf__get_line(cex, crlf_file->crlf);

        /* We have a line so just return. */
        if (rv) {
            goto out;
        }

    }

    /* If we still do not have a line of text, return the remainder
     * and set "is_finished" to true to avoid doing this again. */
    if (rv == NULL) {
        rv = cgul_crlf__get_remainder(cex, crlf_file->crlf);
        crlf_file->is_finished = 1;
    }

 out:
    return rv;
}

cgul_crlf_file_t
cgul_crlf_file__new(cgul_exception_t* cex)
{
    cgul_crlf_file_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate memory for the new object. */
    rv = (cgul_crlf_file_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* The source type is unknown. */
    rv->source_type = CGUL_CRLF_FILE_SOURCE__UNKNOWN;

    /* Create the underlying cgul_crlf object used to convert
     * block-buffered files. */
    rv->crlf = cgul_crlf__new(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the buffer we feed to crlf. */
    rv->buf = (unsigned char*)
                  cgul_libc__malloc(
                      cex,
                      CGUL_CRLF_FILE__DEFAULT_BUFSIZE);
    if (*cex) {
        goto out;
    }

    /* Remember the size of the buffer. */
    rv->bsize = CGUL_CRLF_FILE__DEFAULT_BUFSIZE;

    /* Allocate the string used when cgul_get_line_relaxed() is called. */
    rv->s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up if an exception occurs. */
    if (*cex && rv) {
        cgul_crlf_file__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_crlf_file_t
cgul_crlf_file__new_from_fname(cgul_exception_t* cex,
                               const char* fname,
                               int is_block_buffered)
{
    cgul_crlf_file_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default object. */
    rv = cgul_crlf_file__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open "fname". */
    cgul_crlf_file__open_fname(cex, rv, fname, is_block_buffered);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially instantiated objects. */
    if (*cex && rv) {
        cgul_crlf_file__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_crlf_file_t
cgul_crlf_file__new_from_file(cgul_exception_t* cex,
                              FILE* f,
                              int is_block_buffered)
{
    cgul_crlf_file_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default object. */
    rv = cgul_crlf_file__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open "f". */
    cgul_crlf_file__open_file(cex, rv, f, is_block_buffered);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially instantiated objects. */
    if (*cex && rv) {
        cgul_crlf_file__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_crlf_file_t
cgul_crlf_file__new_from_memory(cgul_exception_t* cex,
                                const char* mem,
                                size_t msize)
{
    cgul_crlf_file_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default object. */
    rv = cgul_crlf_file__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open "mem". */
    cgul_crlf_file__open_memory(cex, rv, mem, msize);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially instantiated objects. */
    if (*cex && rv) {
        cgul_crlf_file__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_crlf_file__delete(cgul_crlf_file_t crlf_file)
{
    cgul_exception_t local = NULL;

    if (crlf_file) {

        /* Close the underlying file. */
        cgul_crlf_file__close(&local, crlf_file);
        if (local) {
            /* Printing a warning is the best we can do here because
             * we are inside the destructor. */
            const char* msg = cgul_exception__get_message(local);
            fprintf(stderr, "*** Warning: cgul_crlf_file__delete: %s\n", msg);
            cgul_exception__delete(local);
            local = NULL;
        }

        /* Delete the string used when calling cgul_get_line_relaxed(). */
        if (crlf_file->s) {
            cgul_string__delete(crlf_file->s);
            crlf_file->s = NULL;
        }

        /* Delete the buffer. */
        if (crlf_file->buf) {
            free(crlf_file->buf);
            crlf_file->buf = NULL;
        }

        /* Delete the underlying crlf object. */
        if (crlf_file->crlf) {
            cgul_crlf__delete(crlf_file->crlf);
            crlf_file->crlf = NULL;
        }

        /* Free the crlf_file object itself. */
        free(crlf_file);
    }
}

void
cgul_crlf_file__open_fname(cgul_exception_t* cex,
                           cgul_crlf_file_t crlf_file,
                           const char* fname,
                           int is_block_buffered)
{
    int is_fname_duplicated = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Sanity check. */
    if (!fname) {
        *cex = cgul_exception__new("cgul_crlf_file__open_fname: "
                                   "file name is NULL");
        goto out;
    }

    /* Close the current file before attempting to open the new file. */
    cgul_crlf_file__close(cex, crlf_file);

    /* Duplicate the file name. */
    assert(crlf_file->fname == NULL);
    crlf_file->fname = cgul_libc__strdup(cex, fname);
    if (*cex) {
        goto out;
    }
    is_fname_duplicated = 1;

    /* Remember the source type. */
    if (is_block_buffered) {
        crlf_file->source_type = CGUL_CRLF_FILE_SOURCE__BLOCK_BUFFERED_FILE;
    } else {
        crlf_file->source_type = CGUL_CRLF_FILE_SOURCE__CHAR_BUFFERED_FILE;
    }

    /* Open the underlying file. */
    crlf_file->f = cgul_libc__fopen(cex, fname, "rb");
    if (*cex) {
        goto out;
    }

    /* Remember that we own f. */
    crlf_file->we_own_f = 1;

 out:

    /* Clean up. */
    if (*cex && is_fname_duplicated) {
        free(crlf_file->fname);
        crlf_file->fname = NULL;
        crlf_file->source_type = CGUL_CRLF_FILE_SOURCE__UNKNOWN;
    }

    return;
}

void
cgul_crlf_file__open_file(cgul_exception_t* cex,
                          cgul_crlf_file_t crlf_file,
                          FILE* f,
                          int is_block_buffered)
{
    int is_fname_duplicated = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Sanity check. */
    if (!f) {
        *cex = cgul_exception__new("cgul_crlf_file__open_file: "
                                   "file is NULL");
        goto out;
    }

    /* Close the current file before attempting to open the new file. */
    cgul_crlf_file__close(cex, crlf_file);

    /* Duplicate the file name. */
    assert(crlf_file->fname == NULL);
    crlf_file->fname = cgul_libc__strdup(cex, "FILE");
    if (*cex) {
        goto out;
    }
    is_fname_duplicated = 1;

    /* Remember the source type. */
    if (is_block_buffered) {
        crlf_file->source_type = CGUL_CRLF_FILE_SOURCE__BLOCK_BUFFERED_FILE;
    } else {
        crlf_file->source_type = CGUL_CRLF_FILE_SOURCE__CHAR_BUFFERED_FILE;
    }

    /* Set the file to use. */
    crlf_file->f = f;

    /* Remember that we do not own f. */
    crlf_file->we_own_f = 0;

 out:

    /* Clean up. */
    if (*cex && is_fname_duplicated) {
        free(crlf_file->fname);
        crlf_file->fname = NULL;
        crlf_file->source_type = CGUL_CRLF_FILE_SOURCE__UNKNOWN;
    }

    return;
}

void
cgul_crlf_file__open_memory(cgul_exception_t* cex,
                            cgul_crlf_file_t crlf_file,
                            const char* mem,
                            size_t msize)
{
    int is_fname_duplicated = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Sanity check. */
    if (!mem) {
        *cex = cgul_exception__new("cgul_crlf_file__open_memory: "
                                   "input buffer is NULL");
        goto out;
    }

    /* Close the current file before attempting to open the new file. */
    cgul_crlf_file__close(cex, crlf_file);

    /* Duplicate the file name. */
    assert(crlf_file->fname == NULL);
    crlf_file->fname = cgul_libc__strdup(cex, "MEMORY");
    if (*cex) {
        goto out;
    }
    is_fname_duplicated = 1;

    /* Remember the source type. */
    crlf_file->source_type = CGUL_CRLF_FILE_SOURCE__MEMORY;

    /* Set the memory buffer and its size. */
    crlf_file->mem = mem;
    crlf_file->msize = msize;

    /* Initialize the offset to point to the start of the memory buffer. */
    crlf_file->moffset = 0;

 out:

    /* Clean up. */
    if (*cex && is_fname_duplicated) {
        free(crlf_file->fname);
        crlf_file->fname = NULL;
        crlf_file->source_type = CGUL_CRLF_FILE_SOURCE__UNKNOWN;
    }

    return;
}

void
cgul_crlf_file__close(cgul_exception_t* cex,
                      cgul_crlf_file_t crlf_file)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Reset the underlying cgul_crlf object. */
    if (crlf_file->crlf) {
        cgul_crlf__reset(cex, crlf_file->crlf, 0);
    }

    /* Close the file but only if we own it. */
    if (crlf_file->f) {
        if (crlf_file->we_own_f) {
            fclose(crlf_file->f);
        }
        crlf_file->f = NULL;
    }

    /* Free the file name. */
    if (crlf_file->fname) {
        free(crlf_file->fname);
        crlf_file->fname = NULL;
    }

    /* Reset the string used when calling cgul_get_line_relaxed(). */
    cgul_string__set_value(cex, crlf_file->s, "");
    assert(*cex == NULL);

    /* Reset most other member variables. */
    crlf_file->source_type = CGUL_CRLF_FILE_SOURCE__UNKNOWN;
    crlf_file->we_own_f = 0;
    crlf_file->mem = NULL;
    crlf_file->msize = 0;
    crlf_file->moffset = 0;
    crlf_file->is_finished = 0;

 out:
    return;
}

const char*
cgul_crlf_file__get_fname(cgul_exception_t* cex,
                          cgul_crlf_file_t crlf_file)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Make sure the source type has been specified. */
    if (crlf_file->source_type == CGUL_CRLF_FILE_SOURCE__UNKNOWN) {
        *cex = cgul_exception__new("cgul_crlf_file__get_fname: "
                                   "unspecified data source");
        goto out;
    }

    rv = crlf_file->fname;

 out:
    return rv;
}

const void*
cgul_crlf_file__get_file(cgul_exception_t* cex,
                         cgul_crlf_file_t crlf_file)
{
    const void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Return the file depending on the source type. */
    switch (crlf_file->source_type) {
    case CGUL_CRLF_FILE_SOURCE__BLOCK_BUFFERED_FILE:
    case CGUL_CRLF_FILE_SOURCE__CHAR_BUFFERED_FILE:
        rv = (void*)crlf_file->f;
        break;
    case CGUL_CRLF_FILE_SOURCE__MEMORY:
        rv = (void*)crlf_file->mem;
        break;
    default:
        *cex = cgul_exception__new("cgul_crlf_file__get_file: "
                                   "unspecified data source");
        goto out;
    }

 out:
    return rv;
}

int
cgul_crlf_file__get_strip_utf8_bom(cgul_exception_t* cex,
                                   cgul_crlf_file_t crlf_file)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Forward to the cgul_crlf object. */
    rv = cgul_crlf__get_strip_utf8_bom(cex, crlf_file->crlf);

 out:
    return rv;
}

void
cgul_crlf_file__set_strip_utf8_bom(cgul_exception_t* cex,
                                   cgul_crlf_file_t crlf_file,
                                   int strip_utf8_bom)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Forward to the cgul_crlf object. */
    cgul_crlf__set_strip_utf8_bom(cex, crlf_file->crlf, strip_utf8_bom);

 out:
    return;
}

void
cgul_crlf_file__get_lines(cgul_exception_t* cex,
                          cgul_crlf_file_t crlf_file,
                          char*** lines,
                          unsigned long int* line_count)
{
    const char* line = NULL;
    char* s = NULL;
    cgul_vector_t vlines = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /*
     * Legend:
     *
     *   1)      s - Dynamic copy of "line".
     *   1)   line - Static line of text from the file.  Read-Only!
     *   2)  lines - Result being returned to the user.
     *   2) vlines - Vector of lines of text used to generate "lines".
     */

    /* Assume failure. */
    if (lines) {
        *lines = NULL;
    }
    if (line_count) {
        *line_count = 0;
    }

    /* Allocate. */
    vlines = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Copy each line of text to the "vlines" vector. */
    for (;;) {

        /* Get the next line. */
        line = cgul_crlf_file__get_line(cex, crlf_file);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (!line) {
            break;
        }

        /* Save a copy of the line to the "vlines" vector. */
        s = cgul_libc__strdup(cex, line);
        cgul_vector__push_back(cex, vlines, s);
        if (*cex) {
            goto out;
        }

        /* Transfer ownership of the "s" to the "vlines" vector. */
        s = NULL;
    }

    /* Success. */
    if (line_count) {
        /* Get the line count before take the array. */
        *line_count = cgul_vector__get_size(cex, vlines);
    }
    if (lines) {
        /* Transfer ownership of each line to "lines". */
        *lines = (char**)cgul_vector__take_array(cex, vlines);
    }

 out:

    /* Clean up. */
    if (s) {
        free(s);
    }
    if (vlines) {
        cgul_vector__free_values(vlines);
        cgul_vector__delete(vlines);
    }

    return;
}

const char*
cgul_crlf_file__get_line(cgul_exception_t* cex,
                         cgul_crlf_file_t crlf_file)
{
    const char* rv = NULL;
    int get_line_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Make sure the source type has been specified. */
    if (crlf_file->source_type == CGUL_CRLF_FILE_SOURCE__UNKNOWN) {
        *cex = cgul_exception__new("cgul_crlf_file__get_line: "
                                   "unspecified data source");
        goto out;
    }

    /* Nothing to do if EOF was already reached. */
    if (crlf_file->is_finished) {
        goto out;
    }

    /* Get the next line depending on the source type. */
    switch (crlf_file->source_type) {

    /* memory or block-buffered file */
    case CGUL_CRLF_FILE_SOURCE__MEMORY:
    case CGUL_CRLF_FILE_SOURCE__BLOCK_BUFFERED_FILE:
        rv = cgul_crlf_file__get_line__block_buffered(cex, crlf_file);
        if (*cex) {
            goto out;
        }
        break;

    /* character-buffered file */
    case CGUL_CRLF_FILE_SOURCE__CHAR_BUFFERED_FILE:
        cgul_string__set_value(cex, crlf_file->s, "");
        get_line_rv = cgul_get_line_relaxed(cex, crlf_file->f, crlf_file->s);
        if (*cex) {
            goto out;
        }
        rv = get_line_rv ? cgul_string__get_value(cex, crlf_file->s) : NULL;
        break;

    /* error */
    default:
        *cex = cgul_exception__new("cgul_crlf_file__get_line: "
                                   "unexpected source type: %d",
                                   crlf_file->source_type);
        goto out;
    }

    /* EOF */
    if (rv == NULL) {
        crlf_file->is_finished = 1;
    }

 out:
    return rv;
}

unsigned long
cgul_crlf_file__get_line_count(cgul_exception_t* cex,
                               cgul_crlf_file_t crlf_file)
{
    unsigned long rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Forward to the cgul_crlf object. */
    rv = cgul_crlf__get_line_count(cex, crlf_file->crlf);

 out:
    return rv;
}

unsigned long
cgul_crlf_file__get_line_offset(cgul_exception_t* cex,
                                cgul_crlf_file_t crlf_file)
{
    unsigned long rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Forward to the cgul_crlf object. */
    rv = cgul_crlf__get_line_offset(cex, crlf_file->crlf);

 out:
    return rv;
}

void
cgul_crlf_file__fseek(cgul_exception_t* cex,
                      cgul_crlf_file_t crlf_file,
                      unsigned long offset,
                      int whence)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Make sure the file is open. */
    if (!crlf_file->f) {
        *cex = cgul_exception__new("cgul_crlf_file__fseek: "
                                   "file not open");
        goto out;
    }

    /* Reposition the underlying file. */
    cgul_libc__fseek(cex, crlf_file->f, offset, whence);
    if (*cex) {
        goto out;
    }

    /* Reset the underlying cgul_crlf object. */
    cgul_crlf__reset(cex, crlf_file->crlf, offset);

    /* We are not finished any more. */
    crlf_file->is_finished = 0;

 out:
    return;
}

void
cgul_crlf_file__rewind(cgul_exception_t* cex,
                       cgul_crlf_file_t crlf_file)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Make sure the file is open. */
    if (!crlf_file->f) {
        *cex = cgul_exception__new("cgul_crlf_file__rewind: "
                                   "file not open");
        goto out;
    }

    /* Rewind the underlying file. */
    cgul_libc__rewind(cex, crlf_file->f);
    if (*cex) {
        goto out;
    }

    /* Reset the underlying cgul_crlf object. */
    cgul_crlf__reset(cex, crlf_file->crlf, 0);

    /* We are not finished any more. */
    crlf_file->is_finished = 0;

 out:
    return;
}

unsigned long
cgul_crlf_file__get_buffer_size(cgul_exception_t* cex,
                                cgul_crlf_file_t crlf_file)
{
    unsigned long rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = crlf_file->bsize;

 out:
    return rv;
}

void
cgul_crlf_file__set_buffer_size(cgul_exception_t* cex,
                                cgul_crlf_file_t crlf_file,
                                unsigned long bsize)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Sanity check bsize. */
    if (bsize == 0) {
        goto out;
    }

    /* realloc() the buffer. */
    crlf_file->buf = (unsigned char*)cgul_libc__realloc(cex,
                                                        crlf_file->buf,
                                                        bsize);
    if (*cex) {
        goto out;
    }

    /* Remember the size of the new buffer. */
    crlf_file->bsize = bsize;

 out:
    return;
}
