#ifndef CGUL_CRLF_P_H
#define CGUL_CRLF_P_H

#include "cgul_crlf.h"
#include "cgul_string.h"
#include "cgul_vector.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_crlf</code> class.
 */
struct cgul_crlf {

    /**
     * If we run out of data while checking for the "\r\n"
     * combination, we flag the condition here.
     */
    int skip_leading_new_line;

    /**
     * Whether we should strip the leading UTF-8 byte order mark (BOM)
     * if it is present.  It is used in conjunction with
     * <code>test_utf8_bom</code>.
     */
    int strip_utf8_bom;

    /**
     * Whether <code>cgul_crlf__get_line()</code> should test
     * <code>strip_utf8_bom</code> in order to strip the leading UTF-8
     * byte-order mark (BOM).  This is necessary, because only the
     * leading BOM should be removed.  The other BOMs that are
     * internal to the text file are significant.
     */
    int test_utf8_bom;

    /**
     * The count of the total number of lines returned by
     * <code>cgul_crlf__get_line()</code> so far.  It is
     * one-based. <p>
     *
     * No attempt is made to prevent <code>line_count</code> from
     * overflowing.  I do this partly for speed and partly so that
     * this class can be used on very large files when knowing the
     * line count is not important. <p>
     *
     * Note that calling <code>cgul_crlf_reset()</code> resets the
     * line count to zero.  This can be used by the client to
     * implement a line counter that does not overflow. <p>
     */
    unsigned long line_count;

    /**
     * The block offset for the current block.  This is the zero-based
     * number of bytes from the start of the stream to the start of
     * the current block.  The block_offset is used to calculate the
     * values stored in the <code>line_offsets</code> array and in
     * <code>remainder_offset</code>.  These values can be used by the
     * client to return to the current line by calling
     * <code>fseek()</code>.  We do not attempt to prevent this value
     * from overflowing.  This is partly for speed and partly because
     * <code>fseek()</code> only takes a <code>long</code> anyway.
     */
    unsigned long int block_offset;

    /**
     * Index into <code>lines</code> for the current block.  It is
     * reset whenever a new block is converted, and it is incremented
     * whenever <code>cgul_crlf__get_line()</code> is called.
     */
    unsigned long int next_line_index;

    /**
     * Maximum allowed index into <code>lines</code> and
     * <code>line_offsets</code> for the current block.
     */
    unsigned long int max_line_index;

    /**
     * Array of pointers to new lines.
     */
    char** lines;

    /**
     * Array of pointers to the offsets of new lines.
     */
    unsigned long int* line_offsets;

    /**
     * The size of the <code>lines</code> and
     * <code>line_offsets</code> arrays.
     */
    unsigned long int array_size;

    /**
     * This holds the text of the last line in the block that is not
     * terminated with an EOL marker.
     */
    cgul_string_t remainder;

    /**
     * Value of the remainder offset.
     */
    unsigned long int remainder_offset;

    /**
     * Whether <code>cgul_crlf__get_remainder()</code> has been
     * called.  This is needed so that we increment line_count only
     * once even if the user calls
     * <code>cgul_crlf__get_remainder()</code> more than once.
     */
    int has_get_remainder_been_called;

    /**
     * If their is a remainder, this is the combination of the
     * remainder from the last go around and what is first in the new
     * buffer this go around.  We need an extra string to hold this
     * because <code>remainder</code> is going to be overwritten with
     * the remainder for the new block.
     */
    cgul_string_t first_line;
};

CGUL_END_C

#endif  /*  CGUL_CRLF_P_H  */
