#include "cgul_csv_p.h"
#include "cgul_get_line.h"
#include "cgul_libc.h"

/**
 * This method frees the fields contained in the
 * <code>csv->fields</code> vector and then clears the vector.
 *
 * @param[in] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 */
static void
cgul_csv__clear_fields(cgul_exception_t* cex,
                       cgul_csv_t csv)
{
    unsigned long int i = 0;
    unsigned long int i_max = 0;
    void** fields_array = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);
    assert(csv->fields);

    /* Get the internal "fields" array. */
    fields_array = cgul_vector__get_array(cex, csv->fields);

    /* Get the size of the array. */
    i_max = cgul_vector__get_size(cex, csv->fields);

    /* Iterate over the array freeing each pointer. */
    for (i = 0 ; i < i_max ; ++i) {
        if (fields_array[i]) {
            free(fields_array[i]);
        }
    }

    /* Clear the array to reflect that all elements have been removed. */
    cgul_vector__clear(cex, csv->fields);

 out:
    return;
}

/**
 * This method frees the headers contained in the
 * <code>csv->headers</code> tree clearing the tree in the process.
 *
 * @param[in] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 */
static void
cgul_csv__clear_headers(cgul_exception_t* cex,
                        cgul_csv_t csv)
{
    void* key = NULL;
    void* value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    /* Remove each node from the tree. */
    while (cgul_rbtree__remove_front(cex, csv->headers, &key, &value)) {
        free(key);
        free(value);
    }

 out:
    return;
}

/**
 * This method maps each header found in the first row of the csv file
 * to its positional index.  The mapping is accomplished by adding one
 * key/value pair for each header to the <code>csv->headers</code>
 * tree where the key is the header and the value is its positional
 * index.  This allows <code>cgul_csv__lookup_by_header()</code> to
 * function.
 *
 * This method is called immediately after
 * <code>cgul_csv__load_next()</code> has finished loading the first
 * row in the csv file.  It is not called for subsequent rows.  This
 * method assumes the first row holds the headers for each column in
 * the csv file.  It does not matter if this assumption is false
 * because <code>cgul_csv__lookup_by_header</code> will not be used in
 * that case any way. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] csv <code>cgul_csv</code> instance
 */
static void
cgul_csv__map_headers_to_columns(cgul_exception_t* cex,
                                 cgul_csv_t csv)
{
    int insert_rv = 0;
    unsigned long int i = 0;
    unsigned long int i_max = 0;
    const char* header = NULL;
    char* key = NULL;
    unsigned long int* value = NULL;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    /* Get the number of headers. */
    i_max = cgul_vector__get_size(cex, csv->fields);

    /* Iterate over the headers. */
    for (i = 0 ; i < i_max ; ++i) {

        /* Reinitialize for this loop.  This is needed for proper
         * clean up. */
        insert_rv = 0;
        key = NULL;
        value = NULL;

        /* Get the header. */
        header = (const char*)cgul_vector__at(cex, csv->fields, i);

        /* "header" will be freed when the next row is loaded so we
         * need a copy to store in the "headers" tree. */
        key = cgul_libc__strdup(cex, header);
        if (*cex) {
            goto out;
        }

        /* Allocate an unsigned long to hold the positional index. */
        value = (unsigned long int*)cgul_libc__calloc(cex, 1, sizeof(*value));
        if (*cex) {
            goto out;
        }

        /* Assign the positional index. */
        *value = i;

        /* Try to insert the header into the "headers" tree. */
        insert_rv = cgul_rbtree__insert(cex, csv->headers, key, &node);
        if (*cex) {
            goto out;
        }

        if (!insert_rv) {
            /* The header already exists!  This is a problem if the
             * first row in the csv file actually contains headers,
             * but it is expected if the first row only contains data.
             * Because there is no good way to automatically
             * distinguish between a row that contains headers and a
             * row that contains data, cgul_csv__lookup_by_header()
             * always just returns the value for the first column. */
            free(key);
            free(value);
            continue;
        }

        /* Map the header to its positional index. */
        cgul_rbtree_node__set_value(cex, node, value);
    }

 out:

    /* Clean up. */
    if (*cex) {
        if (key) {
            free(key);
        }
        if (value) {
            free(value);
        }
    }

    return;
}

cgul_csv_t
cgul_csv__new(cgul_exception_t* cex)
{
    cgul_csv_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create the new csv object. */
    rv = (cgul_csv_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Create the vector that holds the fields for each record. */
    rv->fields = cgul_vector__new(cex);
    if (*cex) {
        goto out;
    }

    /* Create the tree that maps each header to its column. */
    rv->headers = cgul_rbtree__new(cex, (cgul_rbtree__compare_t)strcmp);
    if (*cex) {
        goto out;
    }

    /* Use a comma as the default separator. */
    rv->separator = ',';

 out:

    /* Clean up if an exception is thrown. */
    if (*cex && rv) {
        cgul_csv__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_csv_t
cgul_csv__new_from_fname(cgul_exception_t* cex,
                         const char* fname,
                         int is_block_buffered)
{
    cgul_csv_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(fname);

    /* Create a default csv object. */
    rv = cgul_csv__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the file. */
    cgul_csv__open_fname(cex, rv, fname, is_block_buffered);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up if an exception is thrown. */
    if (*cex && rv) {
        cgul_csv__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_csv_t
cgul_csv__new_from_file(cgul_exception_t* cex,
                        FILE* f,
                        int is_block_buffered)
{
    cgul_csv_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(f);

    /* Create a default csv object. */
    rv = cgul_csv__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the file. */
    cgul_csv__open_file(cex, rv, f, is_block_buffered);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up if an exception is thrown. */
    if (*cex && rv) {
        cgul_csv__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_csv_t
cgul_csv__new_from_memory(cgul_exception_t* cex,
                          const char* buf,
                          size_t buf_size)
{
    cgul_csv_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(buf);

    /* Create a default csv object. */
    rv = cgul_csv__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the memory. */
    cgul_csv__open_memory(cex, rv, buf, buf_size);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up if an exception is thrown. */
    if (*cex && rv) {
        cgul_csv__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_csv_t
cgul_csv__new_from_crlf_file(cgul_exception_t* cex,
                             cgul_crlf_file_t crlf_file)
{
    cgul_csv_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(crlf_file);

    /* Create a default csv object. */
    rv = cgul_csv__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open "crlf_file". */
    cgul_csv__open_crlf_file(cex, rv, crlf_file);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up if an exception is thrown. */
    if (*cex && rv) {
        cgul_csv__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_csv__delete(cgul_csv_t csv)
{
    cgul_exception_t local = NULL;

    if (csv) {

        /* Close the CSV source type and free fname. */
        cgul_csv__close(&local, csv);

        /* Delete the fields vector. */
        if (csv->fields) {
            /* Remove the fields stored in the vector. */
            cgul_csv__clear_fields(&local, csv);
            /* Delete the vector itself. */
            cgul_vector__delete(csv->fields);
            csv->fields = NULL;
        }

        /* Delete the headers tree. */
        if (csv->headers) {
            /* Remove the headers from the tree. */
            cgul_csv__clear_headers(&local, csv);
            /* Delete the tree itself. */
            cgul_rbtree__delete(csv->headers);
            csv->headers = NULL;
        }

        /* Free csv itself. */
        free(csv);
    }
}

void
cgul_csv__open_fname(cgul_exception_t* cex,
                     cgul_csv_t csv,
                     const char* fname,
                     int is_block_buffered)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);
    assert(fname);

    /* Close the current file before attempting to open the new file. */
    cgul_csv__close(cex, csv);

    /* Allocate a new cgul_crlf_file object. */
    csv->f = cgul_crlf_file__new_from_fname(cex, fname, is_block_buffered);
    if (*cex) {
        goto out;
    }

    /* Remember that we own f. */
    csv->we_own_f = 1;

 out:
    return;
}

void
cgul_csv__open_file(cgul_exception_t* cex,
                    cgul_csv_t csv,
                    FILE* f,
                    int is_block_buffered)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);
    assert(f);

    /* Close the current file before attempting to open the new file. */
    cgul_csv__close(cex, csv);

    /* Allocate a new cgul_crlf_file object. */
    csv->f = cgul_crlf_file__new_from_file(cex, f, is_block_buffered);
    if (*cex) {
        goto out;
    }

    /* Remember that we own f. */
    csv->we_own_f = 1;

 out:
    return;
}

void
cgul_csv__open_memory(cgul_exception_t* cex,
                      cgul_csv_t csv,
                      const char* buf,
                      size_t buf_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);
    assert(buf);

    /* Close the current file before attempting to open the new file. */
    cgul_csv__close(cex, csv);

    /* Allocate a new cgul_crlf_file object. */
    csv->f = cgul_crlf_file__new_from_memory(cex, buf, buf_size);
    if (*cex) {
        goto out;
    }

    /* Remember that we own f. */
    csv->we_own_f = 1;

 out:
    return;
}

void
cgul_csv__open_crlf_file(cgul_exception_t* cex,
                         cgul_csv_t csv,
                         cgul_crlf_file_t crlf_file)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);
    assert(crlf_file);

    /* Close the current file before attempting to open the new file. */
    cgul_csv__close(cex, csv);

    /* Remember the file. */
    csv->f = crlf_file;

    /* Remember that we do not own f. */
    csv->we_own_f = 0;

 out:
    return;
}

void
cgul_csv__close(cgul_exception_t* cex,
                cgul_csv_t csv)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    /* Remove the fields stored in the vector. */
    if (csv->fields) {
        cgul_csv__clear_fields(cex, csv);
    }

    /* Remove the headers from the tree. */
    if (csv->headers) {
        cgul_csv__clear_headers(cex, csv);
    }

    /* Delete f if we own it. */
    if (csv->f && csv->we_own_f) {
        cgul_crlf_file__delete(csv->f);
    }
    csv->f = NULL;
    csv->we_own_f = 0;

    /* Reset other variables. */
    csv->are_headers_mapped = 0;
    csv->line_count = 0;

 out:
    return;
}

char
cgul_csv__get_separator(cgul_exception_t* cex,
                        cgul_csv_t csv)
{
    char rv = '\0';

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    rv = csv->separator;

 out:
    return rv;
}

void
cgul_csv__set_separator(cgul_exception_t* cex,
                        cgul_csv_t csv,
                        char separator)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    csv->separator = separator;

 out:
    return;
}

int
cgul_csv__load_next(cgul_exception_t* cex,
                    cgul_csv_t csv)
{
    int rv = 0;
    int quoted = 0;
    size_t i = 0;
    size_t line_length = 0;
    size_t current_field_length = 0;
    const char* line = NULL;
    const char* field = NULL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    /* Make sure a CSV source has been opened. */
    if (!csv->f) {
        *cex = cgul_exception__new("cgul_csv__load_next: "
                                   "CSV file not open");
        goto out;
    }

    /*
     * NOTE: This function is messy.  There isn't too much that can be
     * done about it because there are many asymmetries in processing
     * CSV files like not all fields ending with a separator
     * (typically a comma) and not all records ending on the same line
     * where they start.  Also, there are two modes of reading fields
     * from the file.  So, you have to be ready for literal fields and
     * quoted fields, and you have to be able to unescape the escaped
     * double quotes.  Lastly, you have to be able to report errors.
     */

    /* Allocate the cgul_string we will use to build the string for
     * each field. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Clear the fields from the previous record. */
    cgul_csv__clear_fields(cex, csv);

    /* CSV records can span across multiple lines when fields have
     * embedded line breaks.  So, we keep iterating over each line
     * until we have read in the entire record. */
    for (;;) {

        /* Get the next line. */
        ++csv->line_count;
        line = cgul_crlf_file__get_line(cex, csv->f);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (line == NULL) {
            if (quoted) {
                cgul_exception_t local = NULL;
                unsigned long int field_number =
                    cgul_vector__get_size(&local, csv->fields) + 1;
                *cex = cgul_exception__new(
                           "%s:%lu: "
                           "quote for field %lu not terminated",
                           cgul_crlf_file__get_fname(cex, csv->f),
                           (csv->line_count - 1),
                           field_number);
                goto out;
            }
            goto out;
        }

        /* Calculate the line length. */
        line_length = strlen(line);

        /* Iterate over each character in the line splitting it into fields. */
        for (i = 0 ; i < line_length ; ++i) {

            /* Get the string length of the current field. */
            current_field_length = cgul_string__get_length(cex, s);

            /* If this is the first character of the field, it could
             * be a double quote character which indicates that the
             * field is quoted.  We also check to make sure quoting
             * has not yet been enabled because when we consume a
             * leading double quote and "continue" below the length of
             * "s" will still be zero the next time through. */
            if (!quoted && (current_field_length == 0)) {
                if (line[i] == '"') {
                    /* Enable quoting and consume this character. */
                    quoted = 1;
                    continue;
                } else {
                    /* Disable quoting and fall through. */
                    quoted = 0;
                }
            }

            /* It is illegal to embed double quotes in an literal
             * (i.e., unquoted) field. */
            if (!quoted && (current_field_length > 0) && (line[i] == '"')) {
                cgul_exception_t local = NULL;
                unsigned long int field_number =
                    cgul_vector__get_size(&local, csv->fields) + 1;
                *cex = cgul_exception__new(
                           "%s:%lu: "
                           "illegal quote embedded in literal field %lu",
                           cgul_crlf_file__get_fname(cex, csv->f),
                           csv->line_count,
                           field_number);
                goto out;
            }

            /* If the field is quoted, it may contain embedded double
             * quote characters.  Also, this could be the double quote
             * character that indicates the end of the field. */
            if (quoted && (line[i] == '"')) {

                /* Two double quote characters in a row means inside a
                 * quoted field is the escape sequence for a
                 * single double quote character. */
                if (line[i+1] == '"') {
                    /* Just skip the first double quote character and
                     * fall through to normal process of the second
                     * doulble quote character. */
                    ++i;

                /* A double quote character followed by the separator
                 * (typically a comma) means we have reached the end
                 * of the quoted field.  So, we consume the trailing
                 * double quote, disable quote processing for the
                 * field, and continue with normal processing. */
                } else if (line[i+1] == csv->separator) {
                    ++i;
                    quoted = 0;

                /* A double quote followed by EOL also ends a quoted
                 * field.  We still consume the double quote, but
                 * here, we continue at the top of the loop.  So that
                 * normal EOL processing can occur. */
                } else if (line[i+1] == '\0') {
                    ++i;
                    quoted = 0;
                    continue;

                /* Illegal sequence. */
                } else {
                    cgul_exception_t local = NULL;
                    unsigned long int field_number =
                        cgul_vector__get_size(&local, csv->fields) + 1;
                    *cex = cgul_exception__new(
                               "%s:%lu: "
                               "illegal double quote in field %lu",
                               cgul_crlf_file__get_fname(cex, csv->f),
                               csv->line_count,
                               field_number);
                    goto out;
                }
            }

            /*
             * If we have reached the end of the field, save it.
             *
             * Note that this conditional is entered even for fields
             * that are quoted because we set "quoted" to 0 (above)
             * when we detect the end of the quoted field falling
             * through to process the separator (typically a comma)
             * here.
             */
            if (!quoted && (line[i] == csv->separator)) {
                field = cgul_string__take_value(cex, s);
                cgul_vector__push_back(cex, csv->fields, field);
                if (*cex) {
                    goto out;
                }
                continue;
            }

            /* Just add this character to the current field. */
            cgul_string__append_char(cex, s, line[i]);
        }

        /* If the line ended in the middle of a quoted field, that
         * field has an embedded new-line character.  We need add a
         * corresponding new-line character to the field and get the
         * rest of the record from the next line. */
        if (quoted) {
            cgul_string__append_char(cex, s, '\n');
            continue;
        }

        /* We have reached the end of the record. */
        break;

    }

    /* The last field comes at the end of the line (without a trailing
     * separator). */
    field = cgul_string__take_value(cex, s);
    cgul_vector__push_back(cex, csv->fields, field);
    if (*cex) {
        goto out;
    }

    /* The first column in the csv file, often holds the headers.  We
     * need to map the headers to their columns so
     * cgul_csv__lookup_by_header() can function. */
    if (!csv->are_headers_mapped) {
        cgul_csv__map_headers_to_columns(cex, csv);
        csv->are_headers_mapped = 1;
    }

    /* Success. */
    rv = 1;

 out:

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}

unsigned long int
cgul_csv__get_number_of_fields(cgul_exception_t* cex,
                               cgul_csv_t csv)
{
    unsigned long int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    /* Make sure a CSV source has been opened. */
    if (!csv->f) {
        *cex = cgul_exception__new("cgul_csv__get_number_of_fields: "
                                   "CSV file not open");
        goto out;
    }

    rv = cgul_vector__get_size(cex, csv->fields);

 out:
    return rv;
}

const char*
cgul_csv__lookup(cgul_exception_t* cex,
                 cgul_csv_t csv,
                 unsigned long int field)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    /* Make sure a CSV source has been opened. */
    if (!csv->f) {
        *cex = cgul_exception__new("cgul_csv__lookup: "
                                   "CSV file not open");
        goto out;
    }

    /* Get the string for this field. */
    rv = (const char*)cgul_vector__at(cex, csv->fields, field);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

const char*
cgul_csv__lookup_by_header(cgul_exception_t* cex,
                           cgul_csv_t csv,
                           const char* header)
{
    unsigned long int* field = 0;
    const char* rv = NULL;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Make sure a CSV source has been opened. */
    if (!csv->f) {
        *cex = cgul_exception__new("cgul_csv__lookup_by_header: "
                                   "CSV file not open");
        goto out;
    }

    /* Sanity check. */
    if (!header) {
        goto out;
    }

    /* Find mapping for this header. */
    node = cgul_rbtree__find(cex, csv->headers, header);
    if (node) {
        /* Get the positional index. */
        field = (unsigned long int*)cgul_rbtree_node__get_value(cex, node);

        /* Perform the lookup. */
        rv = cgul_csv__lookup(cex, csv, *field);
    }

 out:
    return rv;
}

char*
cgul_csv__take_value(cgul_exception_t* cex,
                     cgul_csv_t csv,
                     unsigned long int field)
{
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    /* Make sure a CSV source has been opened. */
    if (!csv->f) {
        *cex = cgul_exception__new("cgul_csv__take_value: "
                                   "CSV file not open");
        goto out;
    }

    /* Get the string for this field. */
    rv = (char*)cgul_vector__at(cex, csv->fields, field);
    if (*cex) {
        goto out;
    }

    /* Transfer ownership to the caller by setting our pointer to
     * NULL.  This will prevent cgul_csv__clear_fields() from trying
     * to delete it. */
    cgul_vector__assign(cex, csv->fields, field, NULL);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

char*
cgul_csv__take_value_by_header(cgul_exception_t* cex,
                               cgul_csv_t csv,
                               const char* header)
{
    unsigned long int* field = 0;
    char* rv = NULL;
    cgul_rbtree_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Make sure a CSV source has been opened. */
    if (!csv->f) {
        *cex = cgul_exception__new("cgul_csv__take_value_by_header: "
                                   "CSV file not open");
        goto out;
    }

    /* Sanity check. */
    if (!header) {
        goto out;
    }

    /* Find mapping for this header. */
    node = cgul_rbtree__find(cex, csv->headers, header);
    if (node) {
        /* Get the positional index. */
        field = (unsigned long int*)cgul_rbtree_node__get_value(cex, node);

        /* Perform the lookup. */
        rv = cgul_csv__take_value(cex, csv, *field);
    }

 out:
    return rv;
}

cgul_vector_t
cgul_csv__get_vector(cgul_exception_t* cex,
                     cgul_csv_t csv)
{
    cgul_vector_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    /* Make sure a CSV source has been opened. */
    if (!csv->f) {
        *cex = cgul_exception__new("cgul_csv__get_vector: "
                                   "CSV file not open");
        goto out;
    }

    rv = csv->fields;

 out:
    return rv;
}

cgul_rbtree_t
cgul_csv__get_headers(cgul_exception_t* cex,
                      cgul_csv_t csv)
{
    cgul_rbtree_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(csv);

    /* Make sure a CSV source has been opened. */
    if (!csv->f) {
        *cex = cgul_exception__new("cgul_csv__get_headers: "
                                   "CSV file not open");
        goto out;
    }

    rv = csv->headers;

 out:
    return rv;
}

void
cgul_csv__quote(cgul_exception_t* cex,
                const char* bare_field,
                cgul_string_t quoted_field)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(bare_field);
    assert(quoted_field);

    /*
     * For simplicity, we quote the entire field.
     */

    /* Insert opening double quote. */
    cgul_string__append_char(cex, quoted_field, '"');

    /* Iterate over the characters in "bare_field". */
    for (i = 0 ; bare_field[i] != '\0' ; ++i) {
        /* Escape embedded double quotes. */
        if (bare_field[i] == '"') {
            cgul_string__append_char(cex, quoted_field, '"');
        }
        /* Insert the character. */
        cgul_string__append_char(cex, quoted_field, bare_field[i]);
    }

    /* Insert closing double quote. */
    cgul_string__append_char(cex, quoted_field, '"');

 out:
    return;
}
