#include "cgul_libc.h"
#include "cgul_log2.h"
#include "cgul_merge_sort.h"

/**
 * This typedef defines the interface for the functions that implement
 * sorting of pointers.  It is used so that
 * <code>cgul_merge_sort__common()</code> can be passed which
 * implementation to use in order to sort the array of pointers it
 * generates.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of pointers to elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] cf comparison function
 */
typedef void
cgul_merge_sort__sort_pointers_t(cgul_exception_t* cex,
                                 void** v,
                                 size_t v_count,
                                 cgul_merge_sort__compare_t cf);

/**
 * If your array is already composed of pointers, you can directly
 * call this function.  <code>v</code> will be sorted using
 * <code>cf</code> the same as described for
 * <code>cgul_merge_sort()</code> except insertion sort will be
 * used. <p>
 *
 * An advantage of using insertion sort is that it can sort in-line
 * without having to create an extra copy of the original array.  This
 * makes insertion sort more efficient when sorting lots of small
 * arrays.  The code below takes advantage of this by switching to
 * insertion sort when the size of the partitions is small. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] v array of pointers to elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] cf comparison function
 */
static void
cgul_merge_sort__insertion_sort_pointers(cgul_exception_t* cex,
                                         void** v,
                                         size_t v_count,
                                         cgul_merge_sort__compare_t cf)
{
    size_t i = 0;
    size_t j = 0;
    void* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* The outer loop works left-to-right picking the next value to insert. */
    for (i = 1 ; i < v_count ; ++i) {

        /* Remember the value to insert so we can overwrite its slot
         * when shifting the array to make room for the value elsewhere. */
        tmp = v[i];

        /* The inner loop works right-to-left shifting right the
         * larger values (that have already been sorted) in order to
         * make room for the smaller value. */
        for (j = i ; j > 0 ; --j) {

            /* Compare the selected value with the value in this slot. */
            if (cf(tmp, v[j - 1]) >= 0) {
                /* Insert the selected value in the next slot if the
                 * selected value is greater than the value in this slot. */
                v[j] = tmp;
                break;
            }

            /* Otherwise, shift the value in this slot right. */
            v[j] = v[j - 1];

            /* Handle the boundary case where we have shifted
             * everything right because "tmp" is smaller than
             * everything else and belongs at the front of the array.
             * The for() loop necessarily terminates after this
             * because j will be equal to zero on the next
             * iteration. */
            if (j - 1 == 0) {
                v[j - 1] = tmp;
            }
        }
    }

 out:
    return;
}

/**
 * This is a helper function for
 * <code>cgul_merge_sort_pointers()</code>.  It recursively calls
 * itself in order to sort each partition of the array.  If an error
 * occurs, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * For efficiency, this method has a cut-off to insertion sort for
 * small arrays.  For insertion sort to be efficient, it must do its
 * work directly on <code>v</code>.  Thus, if insertion sort was used,
 * the return value will be <code>v</code> itself, but if the array is
 * not small, a new array will be allocated and returned that holds
 * the result of the merge.  This means the caller must carefully
 * check to determine if the return value is <code>v</code> or not and
 * act accordingly.  If the return value is not <code>v</code>, the
 * caller is responsible for calling <code>free()</code> on it. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of pointers to elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] cf comparison function
 * @return newly allocated array of sorted pointers
 */
static void**
cgul_merge_sort_pointers__recursive(cgul_exception_t* cex,
                                    void** v,
                                    size_t v_count,
                                    cgul_merge_sort__compare_t cf)
{
    size_t i = 0;
    size_t j = 0;
    size_t curr = 0;
    size_t i_max = 0;
    size_t j_max = 0;
    void** h1 = NULL;
    void** h2 = NULL;
    void** rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(cf);

    /* Base case.  Switch to insertion sort once the size of the array
     * is small enough for insertion sort to run efficiently. */
    if (v_count <= 16) {
        cgul_merge_sort__insertion_sort_pointers(cex, v, v_count, cf);
        rv = v;
        goto out;
    }

    /* Split v into two halves. */
    i_max = (v_count >> 1);
    j_max = v_count - i_max;

    /* Recursively merge sort the two halves of v. */
    h1 = cgul_merge_sort_pointers__recursive(cex, v, i_max, cf);
    h2 = cgul_merge_sort_pointers__recursive(cex, &v[i_max], j_max, cf);
    if (*cex) {
        goto out;
    }

    /* Allocate a new array to hold the sorted pointers. */
    rv = (void**)cgul_libc__malloc(cex, v_count * sizeof(*v));
    if (*cex) {
        goto out;
    }

    /* Merge the two halves of v that we just sorted individually. */
    while ((i < i_max) && (j < j_max)) {
        /* Compare the next two values.  The equal part of <= makes
         * the sort stable. */
        if (cf(h1[i], h2[j]) <= 0) {
            rv[curr++] = h1[i++];
        } else {
            rv[curr++] = h2[j++];
        }
    }

    /* Copy the remaining bytes to rv. */
    while (i < i_max) {
        rv[curr++] = h1[i++];
    }
    while (j < j_max) {
        rv[curr++] = h2[j++];
    }

 out:

    /* Clean up. */
    if (h2 && (h2 != &v[i_max])) {
        free(h2);
    }
    if (h1 && (h1 != v)) {
        free(h1);
    }
    if (*cex && rv && (rv != v)) {
        free(rv);
        rv = NULL;
    }

    return rv;
}

/**
 * This is the code common to <code>cgul_merge_sort()</code> and
 * <code>cgul_merge_sort_iteratively()</code>.  It sets up an array of
 * pointers into <code>v</code> and then calls
 * <code>sort_pointers</code> forwarding the array to be sorted
 * <code>v</code>, the number of elements in the array
 * <code>v_count</code>, the size of each element in the array
 * <code>element_size</code>, and the comparison function
 * <code>cf</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v array of elements to sort
 * @param[in] v_count number of elements in <code>v</code>
 * @param[in] element_size size of each element in <code>v</code>
 * @param[in] cf comparison function
 * @param[in] sf sort function
 */
static void
cgul_merge_sort__common(cgul_exception_t* cex,
                        void* v,
                        size_t v_count,
                        size_t element_size,
                        cgul_merge_sort__compare_t cf,
                        cgul_merge_sort__sort_pointers_t sf)
{
    void* scratch = NULL;
    void** pointers_array = NULL;
    char* tmp = NULL;
    size_t i = 0;
    size_t v_size = v_count * element_size;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(cf);

    /* Sanity checks. */
    if ((v_count <= 1) || (element_size <= 0)) {
        return;
    }

    /*
     * The general idea is to sort pointers that point into a copy of
     * "v" under the assumption that copying pointers during a sort
     * will be more efficient than copying entire structures during
     * the sort.  After sorting the pointers, we will copy individual
     * elements from the copy back into the original "v" based upon
     * the sorted pointers.
     */

    /* Create an exact copy of "v" and put it in "scratch". */
    scratch = cgul_libc__malloc(cex, v_size);
    if (*cex) {
        goto out;
    }
    memcpy(scratch, v, v_size);

    /* Generate an array of pointers into the scratch area. */
    pointers_array = (void**)cgul_libc__malloc(cex, sizeof(void*) * v_count);
    if (*cex) {
        goto out;
    }
    for (i = 0, tmp = (char*)scratch;
         i < v_count;
         ++i, tmp += element_size)
    {
        pointers_array[i] = tmp;
    }

    /* Sort the pointers. */
    sf(cex, pointers_array, v_count, cf);
    if (*cex) {
        goto out;
    }

    /* Copy the elements back into "v" based upon the sorted pointers. */
    for (i = 0, tmp = (char*)v ;
         i < v_count ;
         ++i, tmp += element_size)
    {
        memcpy(tmp, pointers_array[i], element_size);
    }

 out:

    /* Clean up. */
    if (pointers_array) {
        free(pointers_array);
    }
    if (scratch) {
        free(scratch);
    }
}

void
cgul_merge_sort(cgul_exception_t* cex,
                void* v,
                size_t v_count,
                size_t element_size,
                cgul_merge_sort__compare_t cf)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_merge_sort__common(cex,
                            v,
                            v_count,
                            element_size,
                            cf,
                            cgul_merge_sort_pointers);

 out:
    return;
}

void
cgul_merge_sort_iteratively(cgul_exception_t* cex,
                            void* v,
                            size_t v_count,
                            size_t element_size,
                            cgul_merge_sort__compare_t cf)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    cgul_merge_sort__common(cex,
                            v,
                            v_count,
                            element_size,
                            cf,
                            cgul_merge_sort_pointers_iteratively);

 out:
    return;
}

void
cgul_merge_sort_pointers(cgul_exception_t* cex,
                         void** v,
                         size_t v_count,
                         cgul_merge_sort__compare_t cf)
{
    size_t i = 0;
    void** v_sorted = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Nothing to do if there are one or zero elements in the array. */
    if (v_count <= 1) {
        return;
    }

    /*
     * NOTE:
     * ====
     *
     * In the general case the following call, will return a new
     * array, but for small arrays, it has a cut-off to insertion sort
     * which works directly on v resulting in (v_sorted == v).
     */

    /* Recursively sort the array. */
    v_sorted = cgul_merge_sort_pointers__recursive(cex, v, v_count, cf);
    if (*cex) {
        goto out;
    }

    /* Copy the sorted points from v_sorted back to v. */
    if (v_sorted != v) {
        for (i = 0 ; i < v_count ; ++i) {
            v[i] = v_sorted[i];
        }
    }

 out:

    /* Clean up. */
    if (v_sorted && (v_sorted != v)) {
        free(v_sorted);
    }

    return;
}

void
cgul_merge_sort_pointers_iteratively(cgul_exception_t* cex,
                                     void** v,
                                     size_t v_count,
                                     cgul_merge_sort__compare_t cf)
{
    /*
     * Variable declarations.
     */

    /* Whether to use insertion sort for the first pass. */
    int insertion_sort_needed = 0;
    /* Total character count of v. */
    size_t v_size = sizeof(*v) * v_count;
    /* We need a scratch buffer for doing the merge sort. */
    void** scratch = NULL;
    /* We need to swap which is the source and which is the
     * destination after each pass. */
    void** src = NULL;
    void** dst = NULL;
    void** tmp = NULL;
    /* "interval_length" is the number sorted elements in each
     * interval for this pass.  After each pass, this doubles (of
     * course). */
    size_t interval_length = 0;
    /* Each pass needs two iterators "i" and "j" one for each
     * interval.  "i_max" and "j_max" are one less than the maximum
     * allowed values for "i" and "j" for their respective
     * intervals. */
    size_t i = 0;
    size_t i_max = 0;
    size_t j = 0;
    size_t j_max = 0;
    /* We need a pointer into "dst" to know where to copy the current
     * value. */
    size_t curr = 0;
    size_t total_passes = 0;

    /*
     * WARNING:
     *
     * This is the iterative implementation of merge sort.  It is not
     * as fast as the recursive implementation so it is not the
     * default.  You should probably looking at the following
     * functions instead:
     *
     *     cgul_merge_sort_pointers()
     *     cgul_merge_sort_pointers__recursive()
     */

    /*
     * Sorting starts here.
     */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(cf);
    assert(v);

    /* Nothing to do if there are one or zero elements in the array. */
    if (v_count <= 1) {
        return;
    }

    scratch = (void**)cgul_libc__malloc(cex, v_size);
    if (*cex) {
        goto out;
    }

    /* Determine the total number of passes which is ceil(log2(v_count)). */
    total_passes = cgul_log2__ceil(cex, v_count) + 1;

    /* If total_passes is odd, use insertion sort for the first pass.
     * Unlike merge sort, insertion sort works in-line.  Merge sort,
     * on the other hand, needs an extra buffer, and when the total
     * number of passes is odd, merge sort needs to copy the values in
     * the extra buffer back to the user's buffer just before
     * returning.  The extra copy can be avoided when the total number
     * of pass is odd by using insertion sort for the first pass. */
    if (total_passes & 0x01) {
        insertion_sort_needed = 1;
    }

    /* Set up the pointers to the source and destination buffers. */
    src = v;
    dst = scratch;

    while (interval_length < v_count) {

        /* Determine the length of the interval for this pass. */
        interval_length = CGUL_MAX(1, interval_length << 1);

        /* Initialize iteration variables for this pass. */
        i = 0;
        i_max = CGUL_MIN(interval_length, v_count);
        j = i_max;
        j_max = CGUL_MIN(j + interval_length, v_count);
        curr = 0;

        /* Perform this pass. */
        while (curr < v_count) {

            /* If necessary, use insertion sort on these two
             * _intervals_.  See comments above. */
            if (insertion_sort_needed) {
                cgul_merge_sort__insertion_sort_pointers(
                    cex, &src[i], j_max - i, cf);
                curr = j_max;

            } else {

                /* Merge these two _intervals_. */
                while (curr < j_max) {
                    /* Check to see if we have no more i values.  If so,
                     * just copy the remaining j values to dst. */
                    if (i >= i_max) {
                        while (j < j_max) {
                            dst[curr++] = src[j++];
                        }
                        break;
                    }
                    /* Check to see if we have no more j values.  If so,
                     * just copy the remaining i values to dst. */
                    if (j >= j_max) {
                        while (i < i_max) {
                            dst[curr++] = src[i++];
                        }
                        break;
                    }
                    /* Compare the i value to the j value and copy the one
                     * that is less than to dst.  In order to have a
                     * stable sort, we choose the i value when the i and j
                     * values are equal. */
                    if (cf(src[i], src[j]) <= 0) {
                        dst[curr++] = src[i++];
                    } else {
                        dst[curr++] = src[j++];
                    }
                }
            }

            /* Move to the next interval. */
            i = j_max;
            i_max = CGUL_MIN(i + interval_length, v_count);
            j = i_max;
            j_max = CGUL_MIN(j + interval_length, v_count);
        }

        /* Swap the scratch buffers for the next pass if this wasn't
         * an insertion-sort pass; otherwise, make sure we only have
         * one insertion sort pass. */
        if (insertion_sort_needed) {
            insertion_sort_needed = 0;
        } else {
            tmp = src;
            src = dst;
            dst = tmp;
        }

    }

    /* Copy the final sorted scratch back to "v" if necessary. */
    if (src != v) {
        /* If the "insertion_sort_needed" boolean is working
         * correctly, we should never have to perform this extra copy. */
        assert(0);
        memcpy(v, src, v_size);
    }

 out:

    if (scratch) {
        free(scratch);
    }
}
