#include "cgul_microxml_p.h"
#include "cgul_block_reader_file.h"
#include "cgul_block_reader_memory.h"
#include "cgul_char.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"
#include "cgul_unicode.h"
#include "cgul_wchar.h"
#include "cgul_wstring.h"

/**
 * Default maximum entity length.
 */
static size_t CGUL_MICROXML__DEFAULT_ENTITY_LENGTH_MAX = 16384;

/**
 * Default text size.
 */
static size_t CGUL_MICROXML__DEFAULT_TEXT_LENGTH_MAX = 16384;

/**
 * Default maximum comment length.
 */
static size_t CGUL_MICROXML__DEFAULT_COMMENT_LENGTH_MAX = 16384;

/**
 * Default size for <code>strings_cache</code>.
 */
static size_t CGUL_MICROXML__DEFAULT_STRINGS_CACHE_SIZE = 128;

/**
 * Default size for <code>hashes_cache</code>.
 */
static size_t CGUL_MICROXML__DEFAULT_HASHES_CACHE_SIZE = 16;

/**
 * Default size for the <code>open_elements_cache</code>.
 */
static size_t CGUL_MICROXML__DEFAULT_OPEN_ELEMENTS_CACHE_SIZE = 128;

/* Manifest constants for Unicode characters. */
#define CGUL_MICROXML_CONSTANT__EXCLAM ((cgul_wchar_t)0x21)
#define CGUL_MICROXML_CONSTANT__DOUBLE_QUOTE ((cgul_wchar_t)0x22)
#define CGUL_MICROXML_CONSTANT__POUND ((cgul_wchar_t)0x23)
#define CGUL_MICROXML_CONSTANT__AMPERSAND ((cgul_wchar_t)0x26)
#define CGUL_MICROXML_CONSTANT__SINGLE_QUOTE ((cgul_wchar_t)0x27)
#define CGUL_MICROXML_CONSTANT__HYPHEN ((cgul_wchar_t)0x2d)
#define CGUL_MICROXML_CONSTANT__DOT ((cgul_wchar_t)0x2e)
#define CGUL_MICROXML_CONSTANT__SLASH ((cgul_wchar_t)0x2f)
#define CGUL_MICROXML_CONSTANT__DIGIT_0 ((cgul_wchar_t)0x30)
#define CGUL_MICROXML_CONSTANT__DIGIT_9 ((cgul_wchar_t)0x39)
#define CGUL_MICROXML_CONSTANT__COLON ((cgul_wchar_t)0x3a)
#define CGUL_MICROXML_CONSTANT__SEMICOLON ((cgul_wchar_t)0x3b)
#define CGUL_MICROXML_CONSTANT__LESS_THAN ((cgul_wchar_t)0x3c)
#define CGUL_MICROXML_CONSTANT__EQUAL ((cgul_wchar_t)0x3d)
#define CGUL_MICROXML_CONSTANT__GREATER_THAN ((cgul_wchar_t)0x3e)
#define CGUL_MICROXML_CONSTANT__UPPER_A ((cgul_wchar_t)0x41)
#define CGUL_MICROXML_CONSTANT__UPPER_F ((cgul_wchar_t)0x46)
#define CGUL_MICROXML_CONSTANT__UPPER_L ((cgul_wchar_t)0x4c)
#define CGUL_MICROXML_CONSTANT__UPPER_M ((cgul_wchar_t)0x4d)
#define CGUL_MICROXML_CONSTANT__UPPER_X ((cgul_wchar_t)0x58)
#define CGUL_MICROXML_CONSTANT__UPPER_Z ((cgul_wchar_t)0x5a)
#define CGUL_MICROXML_CONSTANT__UNDERSCORE ((cgul_wchar_t)0x5f)
#define CGUL_MICROXML_CONSTANT__LOWER_A ((cgul_wchar_t)0x61)
#define CGUL_MICROXML_CONSTANT__LOWER_F ((cgul_wchar_t)0x66)
#define CGUL_MICROXML_CONSTANT__LOWER_L ((cgul_wchar_t)0x6c)
#define CGUL_MICROXML_CONSTANT__LOWER_M ((cgul_wchar_t)0x6d)
#define CGUL_MICROXML_CONSTANT__LOWER_X ((cgul_wchar_t)0x78)
#define CGUL_MICROXML_CONSTANT__LOWER_Z ((cgul_wchar_t)0x7a)
#define CGUL_MICROXML_CONSTANT__BOM ((cgul_wchar_t)0xfeff)

/* Manifest constants for Unicode strings. */
static cgul_wchar_t CGUL_MICROXML_CONSTANT__AMP_CHAR_REF_STRING[] = {
    0x26, 0x61, 0x6d, 0x70, 0x3b, 0x00
};
static cgul_wchar_t CGUL_MICROXML_CONSTANT__LT_CHAR_REF_STRING[] = {
    0x26, 0x6c, 0x74, 0x3b, 0x00
};
static cgul_wchar_t CGUL_MICROXML_CONSTANT__GT_CHAR_REF_STRING[] = {
    0x26, 0x67, 0x74, 0x3b, 0x00
};
static cgul_wchar_t CGUL_MICROXML_CONSTANT__QUOT_CHAR_REF_STRING[] = {
    0x26, 0x71, 0x75, 0x6f, 0x74, 0x3b, 0x00
};
static cgul_wchar_t CGUL_MICROXML_CONSTANT__APOS_CHAR_REF_STRING[] = {
    0x26, 0x61, 0x70, 0x6f, 0x73, 0x3b, 0x00
};

/*
 * Forward declarations.
 */

static void
cgul_microxml__read_state__document_start(cgul_exception_t* cex,
                                          cgul_microxml_t uxml);

static void
cgul_microxml__read_state__prolog(cgul_exception_t* cex,
                                  cgul_microxml_t uxml);

static void
cgul_microxml__read_state__main_loop(cgul_exception_t* cex,
                                     cgul_microxml_t uxml);

static void
cgul_microxml__read_state__inject_element_end_token(cgul_exception_t* cex,
                                                    cgul_microxml_t uxml);

static void
cgul_microxml__read_state__epilog(cgul_exception_t* cex,
                                  cgul_microxml_t uxml);

static void
cgul_microxml__read_state__document_end(cgul_exception_t* cex,
                                        cgul_microxml_t uxml);

static void
cgul_microxml__read_state__final(cgul_exception_t* cex,
                                 cgul_microxml_t uxml);

static void
cgul_microxml__clear_open_elements(cgul_exception_t* cex,
                                   cgul_microxml_t uxml,
                                   cgul_vector_t open_elements);

static void
cgul_microxml__clear_element_attributes(cgul_exception_t* cex,
                                        cgul_microxml_t uxml,
                                        cgul_hash_t attributes);

static const char*
cgul_microxml__intern_symbol(cgul_exception_t* cex,
                             cgul_microxml_t uxml,
                             cgul_string_t s);

static void
cgul_microxml__clear_symbols(cgul_exception_t* cex,
                             cgul_microxml_t uxml,
                             cgul_hash_t symbols);

/**
 * Return whether the wide character <code>wc</code> is a latin
 * upper-case character (i.e., [A-Z]).
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> is a latin upper-case character
 */
#define CGUL_MICROXML_CTYPE__IS_LATIN_UPPER_CASE_CHARACTER(wc) \
    ((CGUL_MICROXML_CONSTANT__UPPER_A <= (wc)) &&              \
     ((wc) <= CGUL_MICROXML_CONSTANT__UPPER_Z))

/**
 * Return whether the wide character <code>wc</code> is a latin
 * lower-case character (i.e., [a-z]).
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> is a latin lower-case character
 */
#define CGUL_MICROXML_CTYPE__IS_LATIN_LOWER_CASE_CHARACTER(wc) \
    ((CGUL_MICROXML_CONSTANT__LOWER_A <= (wc)) &&              \
     ((wc) <= CGUL_MICROXML_CONSTANT__LOWER_Z))

/**
 * Return whether the wide character <code>wc</code> is a decimal
 * digit (i.e., [0-9]).
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> is a decimal digit
 */
#define CGUL_MICROXML_CTYPE__IS_DECIMAL_DIGIT(wc) \
    ((CGUL_MICROXML_CONSTANT__DIGIT_0 <= (wc)) && \
     ((wc) <= CGUL_MICROXML_CONSTANT__DIGIT_9))

/**
 * Return whether the wide character <code>wc</code> is a hexadecimal
 * digit (i.e., [0-9]).
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> is a hexadecimal digit
 */
#define CGUL_MICROXML_CTYPE__IS_HEXADECIMAL_DIGIT(wc)       \
    (((CGUL_MICROXML_CONSTANT__DIGIT_0 <= (wc)) &&          \
      ((wc) <= CGUL_MICROXML_CONSTANT__DIGIT_9)) ||         \
     ((CGUL_MICROXML_CONSTANT__UPPER_A <= (wc)) &&          \
      ((wc) <= CGUL_MICROXML_CONSTANT__UPPER_F)) ||         \
     ((CGUL_MICROXML_CONSTANT__LOWER_A <= (wc)) &&          \
      ((wc) <= CGUL_MICROXML_CONSTANT__LOWER_F)))

/**
 * Return whether the wide character <code>wc</code> satisfies the
 * "nonCharacterCodePoint" production in the MicroXML specification.
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> is a non-character code point
 */
#define CGUL_MICROXML_CTYPE__IS_NONCHARACTER_CODE_POINT(wc) \
    (((0x00fdd0 <= (wc)) && ((wc) <= 0x00fdef)) ||          \
     ((0x00fffe <= (wc)) && ((wc) <= 0x00ffff)) ||          \
     ((0x01fffe <= (wc)) && ((wc) <= 0x01ffff)) ||          \
     ((0x02fffe <= (wc)) && ((wc) <= 0x02ffff)) ||          \
     ((0x03fffe <= (wc)) && ((wc) <= 0x03ffff)) ||          \
     ((0x04fffe <= (wc)) && ((wc) <= 0x04ffff)) ||          \
     ((0x05fffe <= (wc)) && ((wc) <= 0x05ffff)) ||          \
     ((0x06fffe <= (wc)) && ((wc) <= 0x06ffff)) ||          \
     ((0x07fffe <= (wc)) && ((wc) <= 0x07ffff)) ||          \
     ((0x08fffe <= (wc)) && ((wc) <= 0x08ffff)) ||          \
     ((0x09fffe <= (wc)) && ((wc) <= 0x09ffff)) ||          \
     ((0x0afffe <= (wc)) && ((wc) <= 0x0affff)) ||          \
     ((0x0bfffe <= (wc)) && ((wc) <= 0x0bffff)) ||          \
     ((0x0cfffe <= (wc)) && ((wc) <= 0x0cffff)) ||          \
     ((0x0dfffe <= (wc)) && ((wc) <= 0x0dffff)) ||          \
     ((0x0efffe <= (wc)) && ((wc) <= 0x0effff)) ||          \
     ((0x0ffffe <= (wc)) && ((wc) <= 0x0fffff)) ||          \
     ((0x10fffe <= (wc)) && ((wc) <= 0x10ffff)))

/**
 * Return whether the wide character <code>wc</code> satisfies the
 * "surrogateCodePoint" production in the MicroXML specification.
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> is a surrogate code point
 */
#define CGUL_MICROXML_CTYPE__IS_SURROGATE_CODE_POINT(wc) \
    ((0xd800 <= (wc)) && ((wc) <= 0xdfff))

/**
 * Return whether the wide character <code>wc</code> satisfies the
 * "controlCodePoint" production in the MicroXML specification.
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> is a control code point
 */
#define CGUL_MICROXML_CTYPE__IS_CONTROL_CODE_POINT(wc) \
    (((wc) <= 0x1f) || ((0x7f <= (wc)) && ((wc) <= 0x9f)))

/**
 * Return whether the wide character <code>wc</code> satisfies the
 * "forbiddenCodePoint" production in the MicroXML specification.
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> is a forbidden code point
 */
#define CGUL_MICROXML_CTYPE__IS_FORBIDDEN_CODE_POINT( wc)   \
    (CGUL_MICROXML_CTYPE__IS_NONCHARACTER_CODE_POINT(wc) || \
     CGUL_MICROXML_CTYPE__IS_SURROGATE_CODE_POINT(wc) ||    \
     CGUL_MICROXML_CTYPE__IS_CONTROL_CODE_POINT(wc))

/**
 * Return whether the wide character <code>wc</code> satisfies the "s"
 * production (i.e, white space) in the MicroXML specification.
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> is white space
 */
#define CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(wc) \
    (((wc) == 0x9) || ((wc) == 0xa) || ((wc) == 0x20))

/**
 * Return whether the wide character <code>wc</code> satisfies the
 * "char" production in the MicroXML specification.
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> matches the "char" production
 */
#define CGUL_MICROXML_CTYPE__IS_CHAR(wc)                \
    (CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(wc) ||         \
     (((wc) <= 0x10ffff) &&                             \
      !CGUL_MICROXML_CTYPE__IS_CONTROL_CODE_POINT(wc)))

/**
 * Return whether the wide character <code>wc</code> satisfies the
 * "dataChar" production (aka "attributeValueChar" production) in the
 * MicroXML specification.
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> matches the "char" production
 */
#define CGUL_MICROXML_CTYPE__IS_DATA_CHAR(wc)           \
    (CGUL_MICROXML_CTYPE__IS_CHAR(wc) &&                \
     ((wc) != CGUL_MICROXML_CONSTANT__LESS_THAN) &&     \
     ((wc) != CGUL_MICROXML_CONSTANT__AMPERSAND) &&     \
     ((wc) != CGUL_MICROXML_CONSTANT__GREATER_THAN))

/**
 * Return whether the wide character <code>wc</code> satisfies the
 * "nameStartChar" production in the MicroXML specification.
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> matches the "nameStartChar" production
 */
#define CGUL_MICROXML_CTYPE__IS_NAME_START_CHAR(wc)             \
    (CGUL_MICROXML_CTYPE__IS_LATIN_UPPER_CASE_CHARACTER(wc) ||  \
     CGUL_MICROXML_CTYPE__IS_LATIN_LOWER_CASE_CHARACTER(wc) ||  \
     ((wc) == CGUL_MICROXML_CONSTANT__UNDERSCORE) ||            \
     ((0xc0 <= (wc)) && ((wc) <= 0xd6)) ||                      \
     ((0xd8 <= (wc)) && ((wc) <= 0xf6)) ||                      \
     ((0xf8 <= (wc)) && ((wc) <= 0x2ff)) ||                     \
     ((0x370 <= (wc)) && ((wc) <= 0x37d)) ||                    \
     ((0x37f <= (wc)) && ((wc) <= 0x1fff)) ||                   \
     ((0x200c <= (wc)) && ((wc) <= 0x200d)) ||                  \
     ((0x2070 <= (wc)) && ((wc) <= 0x218f)) ||                  \
     ((0x2c00 <= (wc)) && ((wc) <= 0x2fef)) ||                  \
     ((0x3001 <= (wc)) && ((wc) <= 0xd7ff)) ||                  \
     (((0xf900 <= (wc)) && ((wc) <= 0xeffff)) &&                \
      !CGUL_MICROXML_CTYPE__IS_NONCHARACTER_CODE_POINT(wc)))

/**
 * Return whether the wide character <code>wc</code> satisfies the
 * "nameChar" production in the MicroXML specification.
 *
 * @param[in] wc wide character
 * @return whether <code>wc</code> matches the "nameChar" production
 */
#define CGUL_MICROXML_CTYPE__IS_NAME_CHAR(wc)           \
    (CGUL_MICROXML_CTYPE__IS_NAME_START_CHAR(wc) ||     \
     CGUL_MICROXML_CTYPE__IS_DECIMAL_DIGIT(wc) ||       \
     ((wc) == CGUL_MICROXML_CONSTANT__HYPHEN) ||        \
     ((wc) == CGUL_MICROXML_CONSTANT__DOT) ||           \
     ((wc) == 0xb7) ||                                  \
     ((0x0300 <= (wc)) && ((wc) <= 0x036F)) ||          \
     ((0x203F <= (wc)) && ((wc) <= 0x2040)))

/**
 * Transition the state machine based on the new current token
 * <code>uxml->current_token</code> and various other state held in
 * the <code>cgul_microxml</code> instance <code>uxml</code>.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__state_machine__transition(cgul_exception_t* cex,
                                         cgul_microxml_t uxml)
{
    size_t open_elements_count = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Transition the state machine based (mostly) on the previous
     * state and the new current token. */
    switch (uxml->read_state) {

    /* INVALID */
    case CGUL_MICROXML__READ_STATE__INVALID: {
        *cex = cgul_exception__new(
            "cgul_microxml__state_machine__transition: internal error: "
            "invalid state: CGUL_MICROXML__READ_STATE__INVALID");
        goto out;
    }

    /* DOCUMENT_START */
    case CGUL_MICROXML__READ_STATE__DOCUMENT_START: {
        /* TOKEN_DOCUMENT_START */
        if (uxml->current_token == CGUL_MICROXML__TOKEN_DOCUMENT_START) {
            uxml->read_state = CGUL_MICROXML__READ_STATE__PROLOG;
        } else {
            *cex = cgul_exception__new(
                "cgul_microxml__state_machine__transition: internal error: "
                "invalid state transition for "
                "CGUL_MICROXML__READ_STATE__DOCUMENT_START");
            goto out;
        }
        break;
    }

    /* PROLOG */
    case CGUL_MICROXML__READ_STATE__PROLOG: {
        /* TOKEN_COMMENT */
        if (uxml->current_token == CGUL_MICROXML__TOKEN_COMMENT) {
            uxml->read_state = CGUL_MICROXML__READ_STATE__PROLOG;
        /* TOKEN_TEXT (white space only) */
        } else if ((uxml->is_white_space) &&
                   (uxml->current_token == CGUL_MICROXML__TOKEN_TEXT))
        {
            uxml->read_state = CGUL_MICROXML__READ_STATE__PROLOG;
        /* TOKEN_ELEMENT_START */
        } else if (uxml->current_token == CGUL_MICROXML__TOKEN_ELEMENT_START) {
            if (uxml->is_empty_element) {
                uxml->read_state =
                    CGUL_MICROXML__READ_STATE__INJECT_ELEMENT_END_TOKEN;
            } else {
                uxml->read_state = CGUL_MICROXML__READ_STATE__MAIN_LOOP;
            }
        } else {
            *cex = cgul_exception__new(
                "cgul_microxml__state_machine__transition: internal error: "
                "invalid state transition for "
                "CGUL_MICROXML__READ_STATE__PROLOG");
            goto out;
        }
        break;
    }

    /* MAIN_LOOP */
    case CGUL_MICROXML__READ_STATE__MAIN_LOOP: {
        /* TOKEN_TEXT */
        if (uxml->current_token == CGUL_MICROXML__TOKEN_TEXT) {
            uxml->read_state = CGUL_MICROXML__READ_STATE__MAIN_LOOP;
        /* TOKEN_COMMENT */
        } else if (uxml->current_token == CGUL_MICROXML__TOKEN_COMMENT) {
            uxml->read_state = CGUL_MICROXML__READ_STATE__MAIN_LOOP;
        /* TOKEN_ELEMENT_START */
        } else if (uxml->current_token == CGUL_MICROXML__TOKEN_ELEMENT_START) {
            if (uxml->is_empty_element) {
                uxml->read_state =
                    CGUL_MICROXML__READ_STATE__INJECT_ELEMENT_END_TOKEN;
            } else {
                uxml->read_state = CGUL_MICROXML__READ_STATE__MAIN_LOOP;
            }
        /* TOKEN_ELEMENT_END */
        } else if (uxml->current_token == CGUL_MICROXML__TOKEN_ELEMENT_END) {
            /* Get the number of open elements that remain. */
            open_elements_count = cgul_vector__get_size(
                cex, uxml->open_elements);
            if (open_elements_count > 0) {
                /* There are still some open elements.  Remain in main loop. */
                uxml->read_state = CGUL_MICROXML__READ_STATE__MAIN_LOOP;
            } else if (cgul_stream_reader__eof(cex, uxml->sr)) {
                /* No more open elements and EOF.  Close the document. */
                uxml->read_state = CGUL_MICROXML__READ_STATE__DOCUMENT_END;
            } else {
                /* No more open elements and not EOF.  Read the epilog. */
                uxml->read_state = CGUL_MICROXML__READ_STATE__EPILOG;
            }
        } else {
            *cex = cgul_exception__new(
                "cgul_microxml__state_machine__transition: internal error: "
                "invalid state transition for "
                "CGUL_MICROXML__READ_STATE__MAIN_LOOP");
            goto out;
        }
        break;
    }

    /* INJECT_ELEMENT_END_TOKEN */
    case CGUL_MICROXML__READ_STATE__INJECT_ELEMENT_END_TOKEN: {
        /* TOKEN_ELEMENT_END */
        if (uxml->current_token == CGUL_MICROXML__TOKEN_ELEMENT_END) {
            /* Get the number of open elements that remain. */
            open_elements_count = cgul_vector__get_size(
                cex, uxml->open_elements);
            if (open_elements_count > 0) {
                /* There are still some open elements.  Remain in main loop. */
                uxml->read_state = CGUL_MICROXML__READ_STATE__MAIN_LOOP;
            } else if (cgul_stream_reader__eof(cex, uxml->sr)) {
                /* No more open elements and EOF.  Close the document. */
                uxml->read_state = CGUL_MICROXML__READ_STATE__DOCUMENT_END;
            } else {
                /* No more open elements and not EOF.  Read the epilog. */
                uxml->read_state = CGUL_MICROXML__READ_STATE__EPILOG;
            }
        }
        break;
    }

    /* EPILOG */
    case CGUL_MICROXML__READ_STATE__EPILOG: {
        /* TOKEN_COMMENT */
        if (uxml->current_token == CGUL_MICROXML__TOKEN_COMMENT) {
            if (cgul_stream_reader__eof(cex, uxml->sr)) {
                /* EOF */
                uxml->read_state = CGUL_MICROXML__READ_STATE__DOCUMENT_END;
            } else {
                /* Not EOF */
                uxml->read_state = CGUL_MICROXML__READ_STATE__EPILOG;
            }
        /* TOKEN_TEXT (white space only) */
        } else if ((uxml->is_white_space) &&
                   (uxml->current_token == CGUL_MICROXML__TOKEN_TEXT))
        {
            if (cgul_stream_reader__eof(cex, uxml->sr)) {
                /* EOF */
                uxml->read_state = CGUL_MICROXML__READ_STATE__DOCUMENT_END;
            } else {
                /* Not EOF */
                uxml->read_state = CGUL_MICROXML__READ_STATE__EPILOG;
            }
        }
        break;
    }

    /* DOCUMENT_END */
    case CGUL_MICROXML__READ_STATE__DOCUMENT_END: {
        uxml->read_state = CGUL_MICROXML__READ_STATE__FINAL;
        break;
    }

    /* FINAL */
    case CGUL_MICROXML__READ_STATE__FINAL: {
        uxml->read_state = CGUL_MICROXML__READ_STATE__FINAL;
        break;
    }

    /* default */
    default:
        *cex = cgul_exception__new(
            "cgul_microxml__state_machine__transition: internal error: "
            "invalid state: %d", (int)uxml->read_state);
        goto out;
    }

 out:
    return;
}

/**
 * Push the open element <code>uxml->element_symbol</code> onto the
 * open elements stack <code>uxml->open_elements</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__push_open_element(cgul_exception_t* cex,
                                 cgul_microxml_t uxml)
{
    open_element_t open_element = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Allocate. */
    open_element = (open_element_t)cgul_cache__get(
        cex, uxml->open_elements_cache);
    if (*cex) {
        goto out;
    }

    /* Initialize. */
    open_element->element_symbol = uxml->element_symbol;
    open_element->line_start = uxml->token_line_start;
    open_element->column_start = uxml->token_column_start;

    /* Push the open element structure onto the open elements stack. */
    cgul_vector__push_back(cex, uxml->open_elements, open_element);
    if (*cex) {
        goto out;
    }

    /* The open elements stack now owns the open element structure. */
    open_element = NULL;

 out:

    /* Clean up. */
    if (open_element) {
        cgul_cache__put(&local, uxml->open_elements_cache, open_element);
    }

    return;
}

/**
 * Verify that the end tag name <code>uxml->element_symbol</code>
 * matches the start tag name at the top of the open elements stack
 * <code>uxml->open_elements</code>.  As a side-effect, this method
 * pops the top open element from the list of open elements.  If the
 * tag names do not match, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__verify_matching_tags(cgul_exception_t* cex,
                                    cgul_microxml_t uxml)
{
    open_element_t expected_element = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Get the information for the matching opening tag. */
    expected_element =
        (open_element_t)cgul_vector__pop_back(cex, uxml->open_elements);
    if (!expected_element) {
        *cex = cgul_exception__new(
            "cgul_microxml__verify_matching_tags: internal error: "
            "open elements stack is empty!");
        goto out;
    }

    assert(expected_element);
    assert(expected_element->element_symbol);

    /* The name of the closing tag should exactly match the name of
     * the opening tag so it doesn't matter if we compare the UTF-8
     * encoded versions of the names.  Because these are interned
     * symbols, pointer comparison is sufficient; we do not need to do
     * a full string comparison. */
    if (expected_element->element_symbol != uxml->element_symbol) {
        /* Error. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: "
            "end tag </%s> does not match start tag <%s> at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            uxml->token_line_start,
            uxml->token_column_start,
            uxml->element_symbol,
            expected_element->element_symbol,
            expected_element->line_start,
            expected_element->column_start);
        goto out;
    }

 out:

    /* Clean up. */
    if (expected_element) {
        cgul_cache__put(&local, uxml->open_elements_cache, expected_element);
    }

    return;
}

/**
 * Expect a named character reference.  If successful, the referenced
 * character is appended to <code>s</code> as UTF-8 (if <code>s</code>
 * is not <code>NULL</code>).  If an error occurs, an exception is
 * thrown. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned
 * just after the <code>'&amp;'</code> character that starts the
 * character reference.  On return, <code>uxml->current_char</code> is
 * positioned on the <code>';'</code> character that closes the
 * character reference. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in,out] s UTF-8 string to which the accepted characters are appended
 */
static void
cgul_microxml__expect_named_char_ref(cgul_exception_t* cex,
                                     cgul_microxml_t uxml,
                                     cgul_string_t s)
{
    size_t i = 0;
    size_t line_start = 0;
    size_t column_start = 0;
    size_t entity_utf32_length = 0;
    cgul_wchar_t* name_utf32 = NULL;
    struct {
        /* Character reference string like "&lt;". */
        cgul_wchar_t* char_ref_string;
        /* Referenced character like '<'. */
        cgul_wchar_t referenced_char;
    } table[] = {
        {CGUL_MICROXML_CONSTANT__AMP_CHAR_REF_STRING,
         CGUL_MICROXML_CONSTANT__AMPERSAND},
        {CGUL_MICROXML_CONSTANT__LT_CHAR_REF_STRING,
         CGUL_MICROXML_CONSTANT__LESS_THAN},
        {CGUL_MICROXML_CONSTANT__GT_CHAR_REF_STRING,
         CGUL_MICROXML_CONSTANT__GREATER_THAN},
        {CGUL_MICROXML_CONSTANT__QUOT_CHAR_REF_STRING,
         CGUL_MICROXML_CONSTANT__DOUBLE_QUOTE},
        {CGUL_MICROXML_CONSTANT__APOS_CHAR_REF_STRING,
         CGUL_MICROXML_CONSTANT__SINGLE_QUOTE},
    };
    size_t table_size = sizeof(table) / sizeof(table[0]);

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Remember where this character reference starts. */
    line_start =
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr);
    column_start =
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr) - 1;

    /* Initialize the uxml->entity_utf32 accumulator with an '@'
     * character (which we have already skipped past). */
    cgul_wstring__set_length(cex, uxml->entity_utf32, 0);
    cgul_wstring__append_char(
        cex, uxml->entity_utf32, CGUL_MICROXML_CONSTANT__AMPERSAND);

    /* Append the reset of the name for the character reference
     * (including the trailing semi-colon) to uxml->entity_utf32. */
    for (;;) {

        /* Append this character. */
        cgul_wstring__append_char(cex, uxml->entity_utf32, uxml->current_char);
        if (*cex) {
            goto out;
        }

        /* Make sure the entity does not grow too large unless the
         * maximum entity length has been disabled by setting it to 0. */
        if (uxml->entity_length_max > 0) {
            entity_utf32_length =
                cgul_wstring__get_length(cex, uxml->entity_utf32);
            if (entity_utf32_length > uxml->entity_length_max) {
                /* Error. */
                *cex = cgul_exception__new(
                   "%s:%zu:%zu: entity size limit exceeded "
                   "for entity starting at (%zu, %zu)",
                   cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                   cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                   cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                   line_start,
                   column_start);
                goto out;
            }
        }

        /* Break on the trailing semi-colon. */
        if (uxml->current_char == CGUL_MICROXML_CONSTANT__SEMICOLON) {
            break;
        }

        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unexpected end of file processing "
                "character reference starting at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                line_start,
                column_start);
            goto out;
        }

    }

    /* Create an alias for the name of the character reference. */
    name_utf32 = cgul_wstring__get_value(cex, uxml->entity_utf32);

    /* Iterate over each character reference in the table looking for
     * a match for name_utf32. */
    for (i = 0 ; i < table_size ; ++i) {

        /* Compare with the name of the character reference we just read. */
        if (cgul_wchar__wcscmp(name_utf32, table[i].char_ref_string) == 0) {

            /* Append the referenced character to s. */
            if (s) {
                cgul_string__append_wchar(cex, s, table[i].referenced_char);
                if (*cex) {
                    goto out;
                }
            }

            /* Found it. */
            goto out;
        }

    }

    /*
     * Error.
     */

    /* Convert the name of the character reference from UTF-32 to UTF-8. */
    cgul_string__set_length(cex, uxml->entity_utf8, 0);
    cgul_unicode__wcstombs(
        cex,
        cgul_wstring__get_value(cex, uxml->entity_utf32),
        cgul_wstring__get_length(cex, uxml->entity_utf32),
        /* skip_bom */ 1,
        uxml->entity_utf8);
    if (*cex) {
        goto out;
    }
    
    /* Throw exception. */
    *cex = cgul_exception__new(
        "%s:%zu:%zu: unknown named character reference \"%s\"",
        cgul_stream_reader__get_presentation_name(cex, uxml->sr),
        line_start,
        column_start,
        cgul_string__get_value(cex, uxml->entity_utf8));
    goto out;

 out:
    return;
}

/**
 * Expect a numeric character reference.  If successful, the
 * referenced character is appended to <code>s</code> as UTF-8 (if
 * <code>s</code> is not <code>NULL</code>).  If an error occurs, an
 * exception is thrown. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the <code>'#'</code> character in the numeric character reference
 * escape sequence <code>&amp;#</code>.  On return,
 * <code>uxml->current_char</code> is positioned on the
 * <code>';'</code> character that closes the character reference. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in,out] s UTF-8 string to which the accepted characters are appended
 */
static void
cgul_microxml__expect_numeric_char_ref(cgul_exception_t* cex,
                                       cgul_microxml_t uxml,
                                       cgul_string_t s)
{
    size_t line_start = 0;
    size_t column_start = 0;
    size_t entity_utf8_length = 0;
    cgul_wchar_t utf32 = CGUL_WCHAR__NUL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Remember the starting line and column. */
    line_start =
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr);
    column_start =
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr) - 1;

    /* Clear the accumulator. */
    cgul_string__set_length(cex, uxml->entity_utf8, 0);

    /* Read the next character. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected end of file while processing "
            "numerical character reference starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr),
            line_start,
            column_start);
        goto out;
    }

    /* Move passed the 'x' in "&#x". */
    if (uxml->current_char != CGUL_MICROXML_CONSTANT__LOWER_X) {
        *cex = cgul_exception__new(
            "%s:%zu:%zu: expected 'x' in \"&#x\" sequence "
            "because MicroXML only supports hexadecimal "
            "numerical character references",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            line_start,
            column_start);
        goto out;
    }

    /* Iterate over each hexadecimal digit. */
    for (;;) {

        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unexpected end of file processing "
                "numeric character reference starting at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                line_start,
                column_start);
            goto out;
        }

        /* Append each hexadecimal digit to the accumulator and continue. */
        if (CGUL_MICROXML_CTYPE__IS_HEXADECIMAL_DIGIT(uxml->current_char)) {
            cgul_string__append_wchar(
                cex, uxml->entity_utf8, uxml->current_char);
            if (*cex) {
                goto out;
            }
            continue;
        }

        /* Make sure the entity does not grow too large unless the
         * maximum entity length has been disabled by setting it to 0. */
        if (uxml->entity_length_max > 0) {
            entity_utf8_length =
                cgul_string__get_length(cex, uxml->entity_utf8);
            if (entity_utf8_length > uxml->entity_length_max) {
                /* Error. */
                *cex = cgul_exception__new(
                   "%s:%zu:%zu: entity size limit exceeded "
                   "for entity starting at (%zu, %zu)",
                   cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                   cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                   cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                   line_start,
                   column_start);
                goto out;
            }
        }

        /* Break when the trailing semi-colon is found. */
        if (uxml->current_char == CGUL_MICROXML_CONSTANT__SEMICOLON) {
            break;
        }

        /* Error. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: invalid character (&#x%04zX) in "
            "numeric character reference starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
            (size_t)uxml->current_char,
            line_start,
            column_start);
        goto out;
    }

    /* Make sure a number was specified. */
    if (cgul_string__get_length(cex, uxml->entity_utf8) == 0) {
        /* Error. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: no number specified for numeric character reference",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            line_start,
            column_start);
        goto out;
    }

    /* Convert the hexadecimal string to UTF-32. */
    utf32 = (cgul_wchar_t)cgul_string__to_unsigned_long(
                              cex, uxml->entity_utf8, /* base */ 16);
    if (*cex) {

        /* Quash the exception. */
        cgul_exception__delete(*cex);
        *cex = NULL;

        /* Provide better error message. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: numeric character reference (&#x%s;) exceeds limits",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            line_start,
            column_start,
            cgul_string__get_value(cex, uxml->entity_utf8));
        goto out;
    }

    /* Range check the character reference for an improved error message. */
    if (utf32 > 0x10ffff) {
        *cex = cgul_exception__new(
            "%s:%zu:%zu: numeric character reference (&#x%s;) is out of bounds",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            line_start,
            column_start,
            cgul_string__get_value(cex, uxml->entity_utf8));
        goto out;
    }

    /* MicroXML requires the referenced character to match the "char"
     * production.*/
    if (!CGUL_MICROXML_CTYPE__IS_CHAR(utf32)) {
        /* Error. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: numeric character reference (&#x%s;) "
            "is not a valid character",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            line_start,
            column_start,
            cgul_string__get_value(cex, uxml->entity_utf8));
        goto out;
    }

    /* Append the UTF-32 character to s. */
    if (s) {
        cgul_string__append_wchar(cex, s, utf32);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

/**
 * Expect a character reference.  If successful, the referenced
 * character is appended to <code>s</code> as UTF-8 (if <code>s</code>
 * is not <code>NULL</code>).  If an error occurs, an exception is
 * thrown. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the <code>'&amp;'</code> character that starts the character
 * reference.  On return, <code>uxml->current_char</code> is
 * positioned on the <code>';'</code> character that closes the
 * character reference. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in,out] s UTF-8 string to which the accepted characters are appended
 */
static void
cgul_microxml__expect_char_ref(cgul_exception_t* cex,
                               cgul_microxml_t uxml,
                               cgul_string_t s)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* uxml->current_char should be positioned on the '&' character. */
    assert(uxml->current_char == CGUL_MICROXML_CONSTANT__AMPERSAND);

    /* Read the next character. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected end of file processing "
            "character reference staring at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr),
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* If "&#" has been seen, expect a numeric character reference. */
    if (uxml->current_char == CGUL_MICROXML_CONSTANT__POUND) {
        cgul_microxml__expect_numeric_char_ref(cex, uxml, s);
        if (*cex) {
            goto out;
        }
    } else {
        /* Otherwise, expect a named character reference. */
        cgul_microxml__expect_named_char_ref(cex, uxml, s);
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

/**
 * Expect a comment starting from the first character in the comment
 * (i.e., the first character after <code>&lt;!--</code>.  If
 * successful, <code>uxml->comment</code> is set to the comment as
 * UTF-8.  If an error occurs, an exception is thrown. <p>
 *
 * Like the other methods that comprise the recursive descent parser,
 * this method throws exceptions when EOF is reached unexpectedly;
 * however, it does *not* throw an exception if EOF is reached when
 * reading comments in the epilog of the MicroXML document because EOF
 * is allowed in this case, and the state machine will transition to a
 * safe state where <code>uxml->current_char</code> is no longer
 * accessed.  (See
 * <code>cgul_microxml__state_machine__transition()</code>.) <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned
 * just after the double dash in the markup that starts the comment
 * (<code>&lt;!--</code>).  On return, <code>uxml->current_char</code>
 * is positioned just after the markup that closes the comment
 * (<code>--&gt;</code>). <p>
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @see cgul_microxml__expect_comment_starting_from_exclam()
 */
static void
cgul_microxml__expect_comment(cgul_exception_t* cex,
                              cgul_microxml_t uxml)
{
    size_t length = 0;
    size_t contiguous_dashes_seen = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Clear the comment. */
    cgul_string__set_length(cex, uxml->comment, 0);

    /* Iterate over each comment character. */
    for (;;) {

        /*
         * To avoid unnecessary backtracking, we are going to append
         * all dashes to uxml->comment just like any other comment
         * character, but we will have to do a little post-processing
         * to remove valid and invalid "--" sequences (in constant time).
         */

        /* Test for "-->" (valid) and "--" (invalid). */
        if (contiguous_dashes_seen == 2) {

            /* Remove the previous two dashes. */
            length = cgul_string__get_length(cex, uxml->comment);
            assert(length >= 2);
            cgul_string__set_length(cex, uxml->comment, length - 2);

            /* Valid "-->" sequence. */
            if (uxml->current_char == CGUL_MICROXML_CONSTANT__GREATER_THAN)
            {
                /* Success. */
                break;

            }
            /* Invalid embedded "--" sequence. */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: comments cannot contain "
                "embedded \"--\" sequences "
                "for comment starting at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(
                    cex, uxml->sr),
                cgul_stream_reader__get_line_number_previous(
                    cex, uxml->sr),
                cgul_stream_reader__get_column_number_previous(
                    cex, uxml->sr) - 2,
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }

        /* Test for invalid comment characters. */
        if (!CGUL_MICROXML_CTYPE__IS_CHAR(uxml->current_char)) {
            /* Invalid character. */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: "
                "invalid comment character in comment starting at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
                cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }

        /* Unconditionally accept the valid comment character.  We
         * remove valid and invalid "--" sequences above. */
        cgul_string__append_wchar(cex, uxml->comment, uxml->current_char);
        if (*cex) {
            goto out;
        }

        /* Keep track of the number of contiguous dashes seen. */
        if (uxml->current_char == CGUL_MICROXML_CONSTANT__HYPHEN) {
            ++contiguous_dashes_seen;
        } else {
            contiguous_dashes_seen = 0;
        }

        /* Make sure the comment does not grow too large unless the
         * maximum comment length has been disabled by setting it to 0. */
        if (uxml->comment_length_max > 0) {
            length = cgul_string__get_length(cex, uxml->comment);
            if ((length - contiguous_dashes_seen) > uxml->comment_length_max) {
                /* Error. */
                *cex = cgul_exception__new(
                   "%s:%zu:%zu: comment size limit exceeded "
                   "for comment starting at (%zu, %zu)",
                   cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                   cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                   cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                   uxml->token_line_start,
                   uxml->token_column_start);
                goto out;
            }
        }

        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF => Unterminated comment. */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unterminated comment",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }

    }

    /* Read passed the "-->" closing sequence. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF -- See main comments!  Short version is that EOF is
           allowed to immediately follow comments in the epilog. */
        if (uxml->read_state != CGUL_MICROXML__READ_STATE__EPILOG) {
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unexpected end of file "
                "because some elements are still open",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr));
            goto out;
        }
    }

 out:
    return;
}

/**
 * Expect a comment starting from the exclamation mark in
 * <code>&lt;!</code>.  If successful, <code>uxml->comment</code> is
 * set to the comment as UTF-8.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method consumes the remaing <code>--</code> characters in the
 * markup that starts the comment and then calls
 * <code>cgul_microxml__expect_comment()</code>. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned
 * just after the exclamation mark in the markup that starts the
 * comment (<code>&lt;!--</code>).  On return,
 * <code>uxml->current_char</code> is positioned just after the markup
 * that closes the comment (<code>--&gt;</code>). <p>
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__expect_comment_starting_from_exclam(cgul_exception_t* cex,
                                                   cgul_microxml_t uxml)
{
    int i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* uxml->current_char should be on the exclamation mark. */
    assert(uxml->current_char == CGUL_MICROXML_CONSTANT__EXCLAM);

    /* Found "<!" so far.  Looking for "<!--" to start a comment. */
    for (i = 0 ; i < 2 ; ++i) {
        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unexpected end of file for comment at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }
        /* Test for '-' character. */
        if (uxml->current_char != CGUL_MICROXML_CONSTANT__HYPHEN) {
            *cex = cgul_exception__new(
                "%s:%zu:%zu: invalid markup starting at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
                cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }
    }

    /* Found "<!--".  Now, position uxml->current_char just passed
     * this sequence. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected end of file for comment at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr),
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Expect the comment. */
    cgul_microxml__expect_comment(cex, uxml);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Expect white space characters appending them to <code>s</code> as
 * UTF-8 (if <code>s</code> is not <code>NULL</code>).  If an error
 * occurs, an exception is thrown. <p>
 *
 * Note that numeric character references that resolve to the MicroXML
 * white space characters do *not* count as white space for the
 * purposes of this method because this is how the MicroXML
 * specification defines white space. <p>
 *
 * To facilitate allowing the user to control the number of characters
 * returned by any one text token, this method accepts the
 * <code>char_count</code> parameter which is passed in with the
 * number of Unicode characters in <code>s</code> and is updated when
 * this method returns to reflect the total number of Unicode
 * characters in <code>s</code> including the Unicode characters
 * appended to <code>s</code> by this method.  If
 * <code>char_count</code> is <code>NULL</code>, the initial number of
 * characters in <code>s</code> is assumed to be zero. <p>
 *
 * If <code>char_count_max</code> is <code>0</code>, there is no limit
 * (other than available memory) on the number of characters appended
 * to <code>s</code> by this method; otherwise, it is the total number
 * of Unicode characters that <code>s</code> can hold.  So if
 * <code>char_count_max</code> is positive, the maximum number of
 * characters that this method will append to <code>s</code> is
 * <code>char_count_max - char_count</code>.  If
 * <code>char_count</code> is greater than
 * <code>char_count_max</code>, this method will not append any
 * characters to <code>s</code>. <p>
 *
 * Like the other methods that comprise the recursive descent parser,
 * this method throws exceptions when EOF is reached unexpectedly;
 * however, it does *not* throw an exception if EOF is reached when
 * reading white space in the epilog of the MicroXML document because
 * EOF is allowed in this case, and the state machine will transition
 * to a safe state where <code>uxml->current_char</code> is no longer
 * accessed.  (See
 * <code>cgul_microxml__state_machine__transition()</code>.) <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the first expected white space character.  On return,
 * <code>uxml->current_char</code> is positioned on the first
 * non-white space character. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in,out] s UTF-8 string to which the accepted characters are appended
 * @param[in,out] char_count Unicode character count (not byte count)
 */
static void
cgul_microxml__expect_white_space(cgul_exception_t* cex,
                                  cgul_microxml_t uxml,
                                  cgul_string_t s,
                                  size_t* char_count,
                                  size_t char_count_max)
{
    size_t tmp_char_count = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* The method should be called with uxml->current_char already set
     * to the first white space character. */
    assert(CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char));

    /* Initialize tmp_char_count to hold the number of characters
     * already in s. */
    if (char_count) {
        tmp_char_count = *char_count;
    }

    /* Iterate over each white space character.  If char_count_max is
     * zero, we accumulate as many characters as physical memory will
     * allow; otherwise, we accumulate characters until char_count_max
     * is reached. */
    while ((char_count_max == 0) || (tmp_char_count < char_count_max)) {

        /* Append this white space character to s. */
        if (s) {
            cgul_string__append_wchar(cex, s, uxml->current_char);
            if (*cex) {
                goto out;
            }
        }

        /* Increment the character count. */
        tmp_char_count = cgul_defensive__add_size_type(cex, tmp_char_count, 1);
        if (*cex) {
            goto out;
        }

        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF -- See main comments!  Short version is that EOF is
               allowed to immediately follow white space in the epilog. */
            if (uxml->read_state != CGUL_MICROXML__READ_STATE__EPILOG) {
                *cex = cgul_exception__new(
                  "%s:%zu:%zu: unexpected end of file reading white space "
                  "for token starting at (%zu, %zu)",
                  cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                  cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                  cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                  uxml->token_line_start,
                  uxml->token_column_start);
                goto out;
            }
            break;
        }

        /* Break on the first non-white space character. */
        if (!CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char)) {
            break;
        }

    }

    /* Success. */
    if (char_count) {
        *char_count = tmp_char_count;
    }

 out:
    return;
}

/**
 * Expect text.  If successful, <code>uxml->text</code> is set to the
 * text as UTF-8.  If an error occurs, an exception is thrown. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the first data character or character reference (e.g. &#x1234;).
 * On return, <code>uxml->current_char</code> is positioned just after
 * the last character of text. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__expect_text(cgul_exception_t* cex,
                           cgul_microxml_t uxml)
{
    size_t count = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* uxml->current_char should be a data character or a character
     * reference. */
    assert((uxml->current_char == CGUL_MICROXML_CONSTANT__AMPERSAND) ||
           (CGUL_MICROXML_CTYPE__IS_DATA_CHAR(uxml->current_char)));

    /* Clear the text string. */
    cgul_string__set_length(cex, uxml->text, 0);

    /* Assume the text is not pure white space (which is the case for
     * the empty text string). */
    uxml->is_white_space = 0;

    /* Accept leading white space. */
    if (CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char)) {
        cgul_microxml__expect_white_space(
            cex, uxml, uxml->text, &count, uxml->text_length_max);
        if (*cex) {
            goto out;
        }
        /* We have only seen white space characters so far. */
        uxml->is_white_space = 1;
    }

    /*
     * Iterate over the remaining characters.  The iteration is
     * unlimited if (uxml->text_length_max == 0); otherwise, it is
     * limited to uxml->text_length_max Unicode characters.
     *
     * NOTE: On the first pass, the character has to be a non-white
     * space character because the call above to accept_white_space()
     * stopped on a non-white space character (assuming the comment
     * size limit has not been reached).
     */
    while ((uxml->text_length_max == 0) || (count < uxml->text_length_max)) {

        /* Accept any data character. */
        if (CGUL_MICROXML_CTYPE__IS_DATA_CHAR(uxml->current_char)) {
            /* Append the character to uxml->text. */
            cgul_string__append_wchar(cex, uxml->text, uxml->current_char);
            if (*cex) {
                goto out;
            }
            /* Remember that non-white space was found.  See NOTE above. */
            uxml->is_white_space = 0;
            /* Increment the character count. */
            count = cgul_defensive__add_size_type(cex, count, 1);
            if (*cex) {
                goto out;
            }

        /* Accept any character reference. */
        } else if (uxml->current_char == CGUL_MICROXML_CONSTANT__AMPERSAND) {
            /* Append the referenced character to uxml->text. */
            cgul_microxml__expect_char_ref(cex, uxml, uxml->text);
            if (*cex) {
                goto out;
            }
            /* Remember that non-white space was found.  See NOTE above. */
            uxml->is_white_space = 0;
            /* Increment the character count. */
            count = cgul_defensive__add_size_type(cex, count, 1);
            if (*cex) {
                goto out;
            }

        /* Anything else ends the text section. */
        } else {
            break;
        }

        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF -- We are in this loop because there is some
             * non-white space characters, and a MicroXML file cannot
             * end on non-white space text. */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unexpected end of file "
                "reading text that starts at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }

    }

 out:
    return;
}

/**
 * Expect a MicroXML name.  If successful, <code>name</code> is
 * appended to the accepted name as UTF-8 (if <code>name</code> is not
 * <code>NULL</code>).  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method throws an exception when the accepted name begins with
 * (('X'|'x')('M'|'m')('L'|'l')) because the MicroXML specification
 * reserves these names for standarization by the W3c. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the name start character.  On return
 * <code>uxml->current_char</code> is positioned just after the
 * name. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in,out] name name as UTF-8
 */
static void
cgul_microxml__expect_name(cgul_exception_t* cex,
                           cgul_microxml_t uxml,
                           cgul_string_t name)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* uxml->current_char should be positioned at the start of the name. */
    assert(CGUL_MICROXML_CTYPE__IS_NAME_START_CHAR(uxml->current_char));

    if (name) {
        /* Append the name start character. */
        cgul_string__append_wchar(cex, name, uxml->current_char);
        if (*cex) {
            goto out;
        }
    }

    /* Iterate over the rest of the characters in the name. */
    for (;;) {

        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unexpected end of file reading name "
                "for token starting at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }

        /* If it is part of the name append it to s. */
        if (CGUL_MICROXML_CTYPE__IS_NAME_CHAR(uxml->current_char)) {
            if (name) {
                cgul_string__append_wchar(cex, name, uxml->current_char);
                if (*cex) {
                    goto out;
                }
            }
            continue;
        }

        /* Special check for colons embedded in an element name.  This
         * could be a common error when trying to parse general XML
         * which has namespaces and the colon plays the role of
         * namespace separator. */
        if (uxml->current_char == CGUL_MICROXML_CONSTANT__COLON) {
            *cex = cgul_exception__new(
                "%s:%zu:%zu: colons are not allowed in MicroXML names "
                "for token starting at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
                cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }

        /* End of Name */
        break;
    }

#if 0
    /*
     * NOTE:
     * ====
     *
     * Section 3 of "Namespaces in XML 1.0 (Third Edition)" currently
     * at http://www.w3.org/TR/xml-names/ has the following to say
     * about namespace prefixes that start with "xml":
     *
     *     All other prefixes beginning with the three-letter sequence
     *     x, m, l, in any case combination, are reserved.  This means
     *     that:
     *
     *         - users SHOULD NOT use them except as defined by later
     *           specifications
     *
     *         - processors MUST NOT treat them as fatal errors.
     *
     * Section 3.6 of the MicroXML specification says the following:
     *
     *     Names beginning with a match to ["xml"] are reserved for
     *     standardization by the W3C.
     *
     * Thus, the section below which checks for names starting with
     * "xml" is commented out.
     */
    {
        size_t length = 0;
        const unsigned char* buf = NULL;

        /* Throw an exception if the name starts with the reserved
         * (('X'|'x')('M'|'m')('L'|'l')) sequence. */
        length = cgul_string__get_length(cex, name);
        buf = (unsigned char*)cgul_string__get_value(cex, name);
        if ((length >= 3) &&
            ((buf[0] == CGUL_MICROXML_CONSTANT__UPPER_X) ||
             (buf[0] == CGUL_MICROXML_CONSTANT__LOWER_X)) &&
            ((buf[1] == CGUL_MICROXML_CONSTANT__UPPER_M) ||
             (buf[1] == CGUL_MICROXML_CONSTANT__LOWER_M)) &&
            ((buf[2] == CGUL_MICROXML_CONSTANT__UPPER_L) ||
             (buf[2] == CGUL_MICROXML_CONSTANT__LOWER_L)))
        {
            *cex = cgul_exception__new(
                "%s:%zu:%zu: \"xml\" name prefix for \"%s\" "
                "is reserved for standardization by the W3C",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                uxml->token_line_start,
                uxml->token_column_start,
                buf);
            goto out;
        }
    }
#endif

 out:
    return;
}

/**
 * Expect an attribute value.  If successful, <code>value</code> is
 * set to the attribute value as UTF-8 (with the leading and trailing
 * quotation marks (single or double quotes) removed).  If an error
 * occurs, an exception is thrown. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the starting quotation mark that quotes the attribute value.  On
 * return, <code>uxml->current_char</code> is positioned on the first
 * character after the closing quotation mark. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[out] value attribute value
 */
static void
cgul_microxml__expect_attribute_value(cgul_exception_t* cex,
                                      cgul_microxml_t uxml,
                                      cgul_string_t value)
{
    cgul_wchar_t qmark = CGUL_WCHAR__NUL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* uxml->current_char should be positioned on the quotation mark
     * that quotes the attribute value. */
    assert((uxml->current_char != CGUL_MICROXML_CONSTANT__DOUBLE_QUOTE) ||
           (uxml->current_char != CGUL_MICROXML_CONSTANT__SINGLE_QUOTE));

    /* Clear the attribute value. */
    cgul_string__set_length(cex, value, 0);

    /* Remember if single or double quotes are being used. */
    qmark = uxml->current_char;

    /* Iterate over each quoted character. */
    for (;;) {

        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unexpected end of file processing "
                "attribute value for element <%s> starting at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                uxml->element_symbol,
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }

        /* Stop when the matching quotation mark is found. */
        if (uxml->current_char == qmark) {
            break;
        }

        /* Accept any character reference and continue. */
        if (uxml->current_char == CGUL_MICROXML_CONSTANT__AMPERSAND) {
            cgul_microxml__expect_char_ref(cex, uxml, value);
            if (*cex) {
                goto out;
            }
            continue;
        }

        /* Accept any data character and continue. */
        if (CGUL_MICROXML_CTYPE__IS_DATA_CHAR(uxml->current_char)) {
            cgul_string__append_wchar(cex, value, uxml->current_char);
            if (*cex) {
                goto out;
            }
            continue;
        }

        /* Error. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: invalid character (&#%04zX;) in attribute value "
            "for the element <%s> starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
            (size_t)uxml->current_char,
            uxml->element_symbol,
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Read past the closing quotation mark. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected end of file processing "
            "attribute value for element <%s> staring at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr),
            uxml->element_symbol,
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

 out:
    return;
}

/**
 * Expect an attribute.  If successful, <code>key</code> and
 * <code>value</code> are set to the attribute name and value
 * respectively as UTF-8.  If an error occurs, an exception is
 * thrown. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the first character of the attribute name.  On return,
 * <code>uxml->current_char</code> is position just after the last
 * character in the attribute. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[out] key attribute name
 * @param[out] value attribute value
 */
static void
cgul_microxml__expect_attribute(cgul_exception_t* cex,
                                cgul_microxml_t uxml,
                                cgul_string_t key,
                                cgul_string_t value)
{
    size_t line_start = 0;
    size_t column_start = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* uxml->current_char should be positioned on the first character
     * of the attribute's name. */
    assert(CGUL_MICROXML_CTYPE__IS_NAME_START_CHAR(uxml->current_char));

    /* Remember where this attribute name starts. */
    line_start =
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr);
    column_start =
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr);

    /* Clear the attribute name (i.e., key) and value. */
    cgul_string__set_length(cex, key, 0);
    cgul_string__set_length(cex, value, 0);

    /* Expect the attribute name. */
    cgul_microxml__expect_name(cex, uxml, key);
    if (*cex) {
        goto out;
    }

    /* MicroXML expressly disallows attributes named "xmlns". */
    if (cgul_char__strcmp("xmlns", cgul_string__get_value(cex, key)) == 0) {
        *cex = cgul_exception__new(
            "%s:%zu:%zu: MicroXML disallows attributes named \"xmlns\" "
            "for element <%s> starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            line_start,
            column_start,
            uxml->element_symbol,
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Accept white space. */
    if (CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char)) {
        cgul_microxml__expect_white_space(cex, uxml, NULL, NULL, 0);
        if (*cex) {
            goto out;
        }
    }

    /* Expect an '=' sign. */
    if (uxml->current_char != CGUL_MICROXML_CONSTANT__EQUAL) {
        *cex = cgul_exception__new(
            "%s:%zu:%zu: missing '=' sign for attribute "
            "for element <%s> starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
            uxml->element_symbol,
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Read passed the '=' sign. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected end of file processing attribute "
            "for element <%s> that starts at <%zu, %zu>",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr),
            uxml->element_symbol,
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Accept white space. */
    if (CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char)) {
        cgul_microxml__expect_white_space(cex, uxml, NULL, NULL, 0);
        if (*cex) {
            goto out;
        }
    }

    /* Expect the '"' or "'" character that starts an attribute value. */
    if ((uxml->current_char != CGUL_MICROXML_CONSTANT__DOUBLE_QUOTE) &&
        (uxml->current_char != CGUL_MICROXML_CONSTANT__SINGLE_QUOTE))
    {
        *cex = cgul_exception__new(
            "%s:%zu:%zu: attribute value must be quoted "
            "for element <%s> starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
            uxml->element_symbol,
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Expect the attribute value. */
    cgul_microxml__expect_attribute_value(cex, uxml, value);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Expect an attribute list in the <code>uxml->sr</code> stream.  The
 * attribute list can be empty.  If successful, the
 * <code>uxml->element_attributes</code> map is set such that each
 * attribute contributes one key/value pair with each key and each
 * value as UTF-8.  The keys are interned in the symbol table
 * <code>uxml->symbols</code>.  The values are not interned.  If an
 * error occurs, an exception is thrown. <p>
 *
 * This method consumes all leading and trailing white space
 * surrounding the attribute list. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the first character after the element name.  On return,
 * <code>uxml->current_char</code> is positioned on the first
 * non-white space character after the (possibly empty) attributes
 * list. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__expect_attribute_list(cgul_exception_t* cex,
                                     cgul_microxml_t uxml)
{
    size_t line_start = 0;
    size_t column_start = 0;
    const char* key_symbol = NULL;
    cgul_string_t key = NULL;
    cgul_string_t value = NULL;
    cgul_hash_node_t node = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Allocate. */
    key = (cgul_string_t)cgul_cache__get(cex, uxml->strings_cache);
    if (*cex) {
        goto out;
    }

    /* Clear output. */
    cgul_microxml__clear_element_attributes(
        cex, uxml, uxml->element_attributes);

    /* Leading white space is mandatory to have a non-empty attribute list. */
    if (CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char)) {
        cgul_microxml__expect_white_space(cex, uxml, NULL, NULL, 0);
        if (*cex) {
            goto out;
        }
    } else {
        /* An empty attribute list is ok. */
        goto out;
    }

    /* Iterate over the key/value attribute pairs. */
    for (;;) {

        assert(!value);

        /* Allocate. */
        value = (cgul_string_t)cgul_cache__get(cex, uxml->strings_cache);
        if (*cex) {
            goto out;
        }

        /* Make sure that uxml->current_char is position on the start
         * of an attribute name. */
        if (!CGUL_MICROXML_CTYPE__IS_NAME_START_CHAR(uxml->current_char)) {
            /* End of Attributes */
            break;
        }

        /* Remember the starting line and column numbers for this attribute. */
        line_start =
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr);
        column_start =
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr);

        /* Expect the attribute (i.e., one key/value pair). */
        cgul_microxml__expect_attribute(cex, uxml, key, value);
        if (*cex) {
            goto out;
        }

        /* Intern the attribute name (i.e., key). */
        key_symbol = cgul_microxml__intern_symbol(cex, uxml, key);
        if (*cex) {
            goto out;
        }

        /* Try to insert the attribute name into the attributes hash. */
        if (!cgul_hash__insert(
                 cex, uxml->element_attributes, key_symbol, &node))
        {
            if (*cex) {
                goto out;
            }
            /* Duplicate attribute name. */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: duplicate attribute name \"%s\" for "
                "element <%s> starting at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                line_start,
                column_start,
                key_symbol,
                uxml->element_symbol,
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }

        /* Pair the attribute value with the attribute name. */
        cgul_hash_node__set_value(cex, node, value);

        /* Transfer ownership of the attribute value to the hash.
         * (The keys in uxml->element_attributes are owned by the
         * symbol table uxml->symbols.) */
        value = NULL;

        /* If the current character is not the '>' or '/>' delimiter
         * that closes the tag, it must be white space that separates
         * attributes or that immediately precedes the '>' delimiter
         * that closes the tag. */
        if ((uxml->current_char != CGUL_MICROXML_CONSTANT__GREATER_THAN) &&
            (uxml->current_char != CGUL_MICROXML_CONSTANT__SLASH))
        {
            if (CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char)) {
                cgul_microxml__expect_white_space(cex, uxml, NULL, NULL, 0);
                if (*cex) {
                    goto out;
                }
            } else {
                /* Error. */
                *cex = cgul_exception__new(
                  "%s:%zu:%zu: white space, '>', or \"/>\" expected "
                  "in attribute list but character (#x%04zX) found "
                  "for element <%s> starting at (%zu, %zu)",
                  cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                  cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
                  cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
                  (size_t)uxml->current_char,
                  uxml->element_symbol,
                  uxml->token_line_start,
                  uxml->token_column_start);
                goto out;
            }
        }
    }

 out:

    /* Clean up. */
    if (value) {
        cgul_cache__put(&local, uxml->strings_cache, value);
    }
    if (key) {
        cgul_cache__put(&local, uxml->strings_cache, key);
    }

    return;
}

/**
 * Expect an element start in the <code>uxml->sr</code> stream.  If
 * successful, <code>uxml->element_symbol</code>,
 * <code>uxml->element_attributes</code>, and
 * <code>uxml->is_element_empty</code> will be set.  Also, the
 * interned element name and its line number and column number are
 * pushed onto the <code>uxml->open_elements</code> stack.  If an
 * error occurs, an exception is thrown. <p>
 *
 * Like the other methods that comprise the recursive descent parser,
 * this method throws exceptions when EOF is reached unexpectedly;
 * however, it does *not* throw an exception if EOF is reached when
 * moving beyond the last character in the end token for the *empty*
 * *root* element because EOF is allowed in this case, and the state
 * machine will transition to a safe state where
 * <code>uxml->current_char</code> is no longer accessed.  (See
 * <code>cgul_microxml__state_machine__transition()</code>.) <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the name start character for the element name (i.e., the first
 * character after the <code>'&lt;'</code> that starts the tag).  On
 * return, <code>uxml->current_char</code> is positioned just after
 * the <code>'&gt;'</code> character that closes the start tag. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__expect_element_start(cgul_exception_t* cex,
                                    cgul_microxml_t uxml)
{
    size_t open_elements_count = 0;
    cgul_string_t element_name = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* uxml->current_char should be positioned on the name start character. */
    assert(CGUL_MICROXML_CTYPE__IS_NAME_START_CHAR(uxml->current_char));

    /* Allocate. */
    element_name = (cgul_string_t)cgul_cache__get(cex, uxml->strings_cache);
    if (*cex) {
        goto out;
    }

    /* Clear the return values. */
    uxml->element_symbol = NULL;
    cgul_microxml__clear_element_attributes(
        cex, uxml, uxml->element_attributes);
    uxml->is_empty_element = 0;

    /* Expect the name for this element. */
    cgul_microxml__expect_name(cex, uxml, element_name);
    if (*cex) {
        goto out;
    }

    /* Intern the element name. */
    uxml->element_symbol =
        cgul_microxml__intern_symbol(cex, uxml, element_name);
    if (*cex) {
        goto out;
    }

    /* Expect the (possibly empty) attributes list (consuming leading
     * *and* trailing white space). */
    cgul_microxml__expect_attribute_list(cex, uxml);
    if (*cex) {
        goto out;
    }

    /* Try to accept the '/' character in "/>". */
    if (uxml->current_char == CGUL_MICROXML_CONSTANT__SLASH) {
        /* This must be an empty element. */
        uxml->is_empty_element = 1;
        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unexpected end of file processing "
                "start tag <%s> at (%zu, %zu)",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                uxml->element_symbol,
                uxml->token_line_start,
                uxml->token_column_start);
            goto out;
        }
    }

    /* Expect the '>' character in either "/>" (we just read the '/'
     * character) or just ">". */
    if (uxml->current_char == CGUL_MICROXML_CONSTANT__GREATER_THAN) {

        /* Push the open element onto the open elements stack. */
        cgul_microxml__push_open_element(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* Move to just after the end of this tag. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF -- See main comments!  Short version is that EOF is
             * allowed immediately following an *empty* *root* element
             * like <foo/> for example. */
            open_elements_count =
                cgul_vector__get_size(cex, uxml->open_elements);
            /* If (open_elements_count > 1), this is not the root
             * element (remember that we just pushed this element onto
             * the uxml->open_elements stack). */
            if ((open_elements_count > 1) || !uxml->is_empty_element) {
                *cex = cgul_exception__new(
                   "%s:%zu:%zu: unexpected end of file processing "
                   "start tag <%s> at (%zu, %zu)",
                   cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                   cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                   cgul_stream_reader__get_column_number_current(cex, uxml->sr),
                   uxml->element_symbol,
                   uxml->token_line_start,
                   uxml->token_column_start);
                goto out;
            }
        }

    } else {

        /* Error. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected character (#x%04zX) in attribute list "
            "for element <%s> starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
            (size_t)uxml->current_char,
            uxml->element_symbol,
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;

    }

 out:

    /* Clean up. */
    if (element_name) {
        cgul_cache__put(&local, uxml->strings_cache, element_name);
    }

    return;
}

/**
 * Expect the end tag for an element.  On success,
 * <code>uxml->element_symbol</code> is set to the name of element,
 * and the corresponding open tag is popped from the
 * <code>uxml->open_elements</code> stack and verified as matching
 * this end tag.  If an error occurs, an exception is thrown. <p>
 *
 * Like the other methods that comprise the recursive descent parser,
 * this method throws exceptions when EOF is reached unexpectedly;
 * however, it does *not* throw an exception if EOF is reached when
 * moving beyond the last character in the end token for the *root*
 * element because EOF is allowed in this case, and the state machine
 * will transition to a safe state where
 * <code>uxml->current_char</code> is no longer accessed.  (See
 * <code>cgul_microxml__state_machine__transition()</code>.) <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the <code>'/'</code> character in the <code>"&lt;/"</code> sequence
 * that starts the end tag.  On return,
 * <code>uxml->current_char</code> is positioned just after the
 * <code>'&gt;'</code> character that closes the end tag or at the end
 * of file (as explained above). <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__expect_element_end(cgul_exception_t* cex,
                                  cgul_microxml_t uxml)
{
    size_t open_elements_count = 0;
    cgul_string_t element_name = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Allocate. */
    element_name = (cgul_string_t)cgul_cache__get(cex, uxml->strings_cache);
    if (*cex) {
        goto out;
    }

    /* Clear the element symbol. */
    uxml->element_symbol = NULL;

    /* Empty elements do not have explicit end tags. */
    uxml->is_empty_element = 0;

    /* Read the next character. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected end of file processing "
            "end tag that starts at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr),
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* The next character should start the element name. */
    if (!CGUL_MICROXML_CTYPE__IS_NAME_START_CHAR(uxml->current_char)) {
        /* Error. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: no element name for closing tag at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Expect the element name. */
    cgul_microxml__expect_name(cex, uxml, element_name);
    if (*cex) {
        goto out;
    }

    /* Intern the element name. */
    uxml->element_symbol =
        cgul_microxml__intern_symbol(cex, uxml, element_name);
    if (*cex) {
        goto out;
    }

    /* Accept optional white space. */
    if (CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char)) {
        cgul_microxml__expect_white_space(cex, uxml, NULL, NULL, 0);
        if (*cex) {
            goto out;
        }
    }

    /* Expect the closing delimiter ">". */
    if (uxml->current_char != CGUL_MICROXML_CONSTANT__GREATER_THAN) {
        /* Error. */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: end tag </%s> at (%zu, %zu) not closed",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
            uxml->element_symbol,
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Verify that the end tag name matches the the start tag name. */
    cgul_microxml__verify_matching_tags(cex, uxml);
    if (*cex) {
        goto out;
    }

    /* Move passed the closing delimiter ">". */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF -- See main comments! */
        open_elements_count = cgul_vector__get_size(cex, uxml->open_elements);
        if (open_elements_count > 0) {
            /* Error. */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: unexpected end of file "
                "because some elements are still open",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr));
            goto out;
        }
    }

 out:

    /* Clean up. */
    if (element_name) {
        cgul_cache__put(&local, uxml->strings_cache, element_name);
    }

    return;
}

/**
 * Expect markup in the form of a start tag or end tag for an element
 * or comment.  If an error occurs, an exception is thrown. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the <code>'&lt;'</code> character that starts the markup.  On
 * return, <code>uxml->current_char</code> is positioned just after
 * the <code>'&gt;'</code> character that closes the markup. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__expect_markup(cgul_exception_t* cex,
                             cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Read the next character. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected end of file "
            "processing markup that starts at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr),
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    if (uxml->current_char == CGUL_MICROXML_CONSTANT__SLASH) {

        /* Seen "</" so far.  This should be an end tag for an element. */
        cgul_microxml__expect_element_end(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* Emit the ELEMENT_END token. */
        uxml->current_token = CGUL_MICROXML__TOKEN_ELEMENT_END;

        /* Transition to the next state. */
        cgul_microxml__state_machine__transition(cex, uxml);
        if (*cex) {
            goto out;
        }

    } else if (uxml->current_char == CGUL_MICROXML_CONSTANT__EXCLAM) {

        /* Seen "<!" so far.  This could be a comment, DTD, or CDATA. */
        ///cgul_microxml__expect_markup_for_exclam(cex, uxml);
        cgul_microxml__expect_comment_starting_from_exclam(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* Emit the ELEMENT_START token. */
        uxml->current_token = CGUL_MICROXML__TOKEN_COMMENT;

        /* Transition to the next state. */
        cgul_microxml__state_machine__transition(cex, uxml);
        if (*cex) {
            goto out;
        }

    ///} else if (uxml->current_char == CGUL_MICROXML_CONSTANT__QUESTION_MARK) {
    ///
    ///    /* Seen "<?" so far.  This could be an XML declaration or PI. */
    ///    cgul_microxml__expect_markup_for_question_mark(cex, uxml);
    ///    if (*cex) {
    ///        goto out;
    ///    }

    } else if (CGUL_MICROXML_CTYPE__IS_NAME_START_CHAR(uxml->current_char)) {

        /* Seen "<" followed by the start of an element name. */
        cgul_microxml__expect_element_start(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* Emit the ELEMENT_START token. */
        uxml->current_token = CGUL_MICROXML__TOKEN_ELEMENT_START;

        /* Transition to the next state. */
        cgul_microxml__state_machine__transition(cex, uxml);
        if (*cex) {
            goto out;
        }

    } else {

        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: invalid markup starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;

    }

 out:
    return;
}

/**
 * Helper for <code>cgul_microxml__read_next_token()</code>
 * responsible for handling the
 * <code>CGUL_MICROXML__READ_STATE__DOCUMENT_START</code> state.  If
 * an error occurs, an exception is thrown. <p>
 *
 * The main job of this method simply is to consume the optional
 * Unicode byte-order mark (BOM), emit the
 * <code>CGUL_MICROXML__TOKEN_DOCUMENT_START</code> token, and
 * transitions to the <code>CGUL_MICROXML__READ_STATE__PROLOG</code>
 * state. <p>
 *
 * Call this method with the stream reader <code>uxml->sr</code>
 * positioned at the beginning of the file. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_state__document_start(cgul_exception_t* cex,
                                          cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Update the starting line and column numbers. */
    uxml->token_line_start =
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr);
    uxml->token_column_start =
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr);

    /* Read the first character which could be the Unicode Byte-Order
     * Mark (BOM) which will not cause uxml->sr to advance the line
     * and column numbers. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: empty document",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr));
        goto out;
    }

    /* Skip Unicode byte-order mark (BOM) if present. */
    if (uxml->current_char == CGUL_MICROXML_CONSTANT__BOM) {
        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
            if (*cex) {
                goto out;
            }
            /* EOF */
            *cex = cgul_exception__new(
                "%s:%zu:%zu: empty document",
                cgul_stream_reader__get_presentation_name(cex, uxml->sr),
                cgul_stream_reader__get_line_number_current(cex, uxml->sr),
                cgul_stream_reader__get_column_number_current(cex, uxml->sr));
            goto out;
        }
    }

    /* Emit the DOCUMENT_START token. */
    uxml->current_token = CGUL_MICROXML__TOKEN_DOCUMENT_START;

    /* Transition to the next state. */
    cgul_microxml__state_machine__transition(cex, uxml);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Helper for <code>cgul_microxml__read_state__prolog()</code>
 * responsible for handling the markup allowed in the MicroXML prolog.
 * If an error occurs, an exception is thrown. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the <code>'&lt;'</code> character that starts each markup tag.  On
 * return, <code>uxml->current_char</code> is positioned just after
 * the <code>'&gt;'</code> character that closes the markup tag. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_state__prolog__expect_markup(cgul_exception_t* cex,
                                                 cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Read the next character. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected end of file "
            "for prolog markup starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr),
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Test for '!' character. */
    if (uxml->current_char == CGUL_MICROXML_CONSTANT__EXCLAM) {
        /* Found "<!".  The only allowed markup in the prolog that
         * starts with "<!"  is a comment. */
        cgul_microxml__expect_comment_starting_from_exclam(cex, uxml);
        if (*cex) {
            goto out;
        }
        /* Set the current token. */
        uxml->current_token = CGUL_MICROXML__TOKEN_COMMENT;
        /* Transition to the next state. */
        cgul_microxml__state_machine__transition(cex, uxml);
        if (*cex) {
            goto out;
        }
        /* Success. */
        goto out;
    }

    /* Test for the start of an element name. */
    if (CGUL_MICROXML_CTYPE__IS_NAME_START_CHAR(uxml->current_char)) {
        /* Expect the root element. */
        cgul_microxml__expect_element_start(cex, uxml);
        /* Set the current token. */
        uxml->current_token = CGUL_MICROXML__TOKEN_ELEMENT_START;
        /* Transition to the next state. */
        cgul_microxml__state_machine__transition(cex, uxml);
        if (*cex) {
            goto out;
        }
        /* Success. */
        goto out;
    }

    /* Error. */
    *cex = cgul_exception__new(
        "%s:%zu:%zu: unexpected markup before root element "
        "starting at (%zu, %zu)",
        cgul_stream_reader__get_presentation_name(cex, uxml->sr),
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
        uxml->token_line_start,
        uxml->token_column_start);
    goto out;

 out:
    return;
}

/**
 * Helper for <code>cgul_microxml__read_next_token()</code>
 * responsible for handling the
 * <code>CGUL_MICROXML__READ_STATE__PROLOG</code> state.  If an error
 * occurs, an exception is thrown. <p>
 *
 * The main job of this method is to read and tokenize white space or
 * comments that are allowed to precede the root MicroXML element.
 * Then, it transitions either to itself or to the
 * <code>CGUL_MICROXML__READ_STATE__MAIN_LOOP</code> state. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the first character in the file (excluding the Unicode Byte-Order
 * Mark (BOM)). <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_state__prolog(cgul_exception_t* cex,
                                  cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Update the starting line and column numbers. */
    uxml->token_line_start =
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr);
    uxml->token_column_start =
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr);

    /* Test for '<' character. */
    if (uxml->current_char == CGUL_MICROXML_CONSTANT__LESS_THAN) {
        /* Expect markup. */
        cgul_microxml__read_state__prolog__expect_markup(cex, uxml);
        goto out;
    }

    /* The only other thing that can be in the prolog is white space. */
    if (CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char)) {
        /* Expect white space saving it to uxml->text. */
        cgul_string__set_length(cex, uxml->text, 0);
        cgul_microxml__expect_white_space(
            cex, uxml, uxml->text, NULL, uxml->text_length_max);
        if (*cex) {
            goto out;
        }
        /* The text is pure white space. */
        uxml->is_white_space = 1;
        /* Set the current token. */
        uxml->current_token = CGUL_MICROXML__TOKEN_TEXT;
        /* Transition to the next state. */
        cgul_microxml__state_machine__transition(cex, uxml);
        if (*cex) {
            goto out;
        }
        /* Success. */
        goto out;
    }

    /* Error. */
    *cex = cgul_exception__new(
        "%s:%zu:%zu: unexpected character in prolog",
        cgul_stream_reader__get_presentation_name(cex, uxml->sr),
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr));
    goto out;

 out:
    return;
}

/**
 * Helper for <code>cgul_microxml__read_next_token()</code>
 * responsible for handling the
 * <code>CGUL_MICROXML__READ_STATE__MAIN_LOOP</code> state. If an
 * error occurs, an exception is thrown. <p>
 *
 * Between the start tag and end tag for the root element, any
 * well-formed content is allowed.  The main job of this method is
 * read and tokenize that content.  Then, it transitions either to
 * itself or to
 * <code>CGUL_MICROXML__READ_STATE__INJECT_ELEMENT_END_TOKEN</code> or
 * to <code>CGUL_MICROXML__READ_STATE__EPILOG</code>. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the start of the next token.  On return,
 * <code>uxml->current_char</code> is positioned just after the token
 * that was read or at the end of the file (EOF). <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_state__main_loop(cgul_exception_t* cex,
                                     cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Update the line and column numbers. */
    uxml->token_line_start =
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr);
    uxml->token_column_start =
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr);

    /* Check for markup: element start, element end, comments. */
    if (uxml->current_char == CGUL_MICROXML_CONSTANT__LESS_THAN) {
        cgul_microxml__expect_markup(cex, uxml);
        goto out;
    }

    /* Check for text: character references (e.g., &#quot;) or a data
     * characters. */
    if ((uxml->current_char == CGUL_MICROXML_CONSTANT__AMPERSAND) ||
        (CGUL_MICROXML_CTYPE__IS_DATA_CHAR(uxml->current_char)))
    {
        /* Expect text. */
        cgul_microxml__expect_text(cex, uxml);
        if (*cex) {
            goto out;
        }
        /* Emit the TEXT token. */
        uxml->current_token = CGUL_MICROXML__TOKEN_TEXT;
        /* Transition to the next state. */
        cgul_microxml__state_machine__transition(cex, uxml);
        goto out;
    }

    /* Error. */
    *cex = cgul_exception__new(
        "%s:%zu:%zu: invalid character (&#x%04zX) in content",
        cgul_stream_reader__get_presentation_name(cex, uxml->sr),
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr),
        (size_t)uxml->current_char);
    goto out;

 out:
    return;
}

/**
 * Helper for <code>cgul_microxml__read_next_token()</code>
 * responsible for handling the
 * <code>CGUL_MICROXML__READ_STATE__INJECT_ELEMENT_END_TOKEN</code>
 * state.  As a side-effect, this method modifies
 * <code>uxml->current_token</code> and <code>uxml->read_state</code>.
 * If an error occurs, an exception is thrown. <p>
 *
 * The main job of this method is to inject a
 * <code>CGUL_MICROXML__TOKEN_ELEMENT_END</code> tokenize into the
 * token stream to for empty tags that use the special empty-tag
 * syntax.  Then, it transitions to either the
 * <code>CGUL_MICROXML__READ_STATE__MAIN_LOOP</code> or
 * <code>CGUL_MICROXML__READ_STATE__EPILOG</code> state. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_state__inject_element_end_token(cgul_exception_t* cex,
                                                    cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /*
     * This method is used to inject ELEMENT_END tokens for empty
     * elements that use the special empty-element syntax (e.g., <foo/>).
     */

    /* Verify that the start and end tags match. */
    cgul_microxml__verify_matching_tags(cex, uxml);
    if (*cex) {
        goto out;
    }

    /* Emit the ELEMENT_END token. */
    uxml->current_token = CGUL_MICROXML__TOKEN_ELEMENT_END;

    /* Transition to the next state. */
    cgul_microxml__state_machine__transition(cex, uxml);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Helper for <code>cgul_microxml__read_state__epilog()</code>
 * responsible for handling the markup allowed in the MicroXML epilog.
 * If an error occurs, an exception is thrown. <p>
 *
 * Call this method with <code>uxml->current_char</code> positioned on
 * the <code>'&lt;'</code> character that starts each markup tag.  On
 * return, <code>uxml->current_char</code> is positioned just after
 * the <code>'&gt;'</code> character that closes the markup tag. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_state__epilog__expect_markup(cgul_exception_t* cex,
                                                 cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Read the next character. */
    if (!cgul_stream_reader__read(cex, uxml->sr, &uxml->current_char)) {
        if (*cex) {
            goto out;
        }
        /* EOF */
        *cex = cgul_exception__new(
            "%s:%zu:%zu: unexpected end of file "
            "for epilog markup starting at (%zu, %zu)",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_current(cex, uxml->sr),
            cgul_stream_reader__get_column_number_current(cex, uxml->sr),
            uxml->token_line_start,
            uxml->token_column_start);
        goto out;
    }

    /* Test for '!' character. */
    if (uxml->current_char == CGUL_MICROXML_CONSTANT__EXCLAM) {
        /* Found "<!".  The only allowed markup in the epilog that
         * starts with "<!"  is a comment. */
        cgul_microxml__expect_comment_starting_from_exclam(cex, uxml);
        if (*cex) {
            goto out;
        }
        /* Set the current token. */
        uxml->current_token = CGUL_MICROXML__TOKEN_COMMENT;
        /* Transition to the next state. */
        cgul_microxml__state_machine__transition(cex, uxml);
        if (*cex) {
            goto out;
        }
        /* Success. */
        goto out;
    }

    /* Error */
    *cex = cgul_exception__new(
        "%s:%zu:%zu: unexpected markup in epilog",
        cgul_stream_reader__get_presentation_name(cex, uxml->sr),
        uxml->token_line_start,
        uxml->token_column_start);
    goto out;

 out:
    return;
}

/**
 * Helper for <code>cgul_microxml__read_next_token()</code>
 * responsible for handling the
 * <code>CGUL_MICROXML__READ_STATE__EPILOG</code> state.  As a
 * side-effect, this method modifies <code>uxml->current_token</code>
 * and <code>uxml->read_state</code>.  If an error occurs, an
 * exception is thrown. <p>
 *
 * The main job of this method is to read and tokenize white space or
 * comments (if any) that are allowed to follow the root MicroXML
 * element.  Then, it transitions either to itself or to the
 * <code>CGUL_MICROXML__READ_STATE__DOCUMENT_END</code> state. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_state__epilog(cgul_exception_t* cex,
                                  cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Update the starting line and column numbers. */
    uxml->token_line_start =
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr);
    uxml->token_column_start =
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr);

    /* Test for '<' character. */
    if (uxml->current_char == CGUL_MICROXML_CONSTANT__LESS_THAN) {
        /* Expect markup. */
        cgul_microxml__read_state__epilog__expect_markup(cex, uxml);
        goto out;
    }

    /* The only other thing that can be in the epilog is white space. */
    if (CGUL_MICROXML_CTYPE__IS_WHITE_SPACE(uxml->current_char)) {
        /* Expect white space saving it to uxml->text. */
        cgul_string__set_length(cex, uxml->text, 0);
        cgul_microxml__expect_white_space(
            cex, uxml, uxml->text, NULL, uxml->text_length_max);
        if (*cex) {
            goto out;
        }
        /* The text is pure white space. */
        uxml->is_white_space = 1;
        /* Set the current token. */
        uxml->current_token = CGUL_MICROXML__TOKEN_TEXT;
        /* Transition to the next state. */
        cgul_microxml__state_machine__transition(cex, uxml);
        if (*cex) {
            goto out;
        }
        /* Success. */
        goto out;
    }

    /* Error. */
    *cex = cgul_exception__new(
        "%s:%zu:%zu: unexpected character in epilog",
        cgul_stream_reader__get_presentation_name(cex, uxml->sr),
        cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
        cgul_stream_reader__get_column_number_previous(cex, uxml->sr));
    goto out;

 out:
    return;
}

/**
 * Helper for <code>cgul_microxml__read_next_token()</code>
 * responsible for handling the
 * <code>CGUL_MICROXML__READ_STATE__DOCUMENT_END</code> state.  As a
 * side-effect, this method modifies <code>uxml->current_token</code>
 * and <code>uxml->read_state</code>.  If an error occurs, an
 * exception is thrown. <p>
 *
 * The main job of this method simply is to verify that the end of
 * file has been reached and then emit the
 * <code>CGUL_MICROXML__TOKEN_DOCUMENT_END</code> token.  Then, it
 * transitions to the <code>CGUL_MICROXML__READ_STATE__FINAL</code>
 * state. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_state__document_end(cgul_exception_t* cex,
                                        cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Emit the DOCUMENT_END token. */
    uxml->current_token = CGUL_MICROXML__TOKEN_DOCUMENT_END;

    /* Transition to the next state. */
    cgul_microxml__state_machine__transition(cex, uxml);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Helper for <code>cgul_microxml__read_next_token()</code>
 * responsible for handling the
 * <code>CGUL_MICROXML__READ_STATE__FINAL</code> state.  As a
 * side-effect, this method modifies <code>uxml->current_token</code>
 * and <code>uxml->read_state</code>.  If an error occurs, an
 * exception is thrown. <p>
 *
 * The main job of this method simply throw an exception if the user
 * tries to get more tokens after the MicroXML document has been
 * exhausted.  This method does not transition out of the
 * <code>CGUL_MICROXML__READ_STATE__FINAL</code> state. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_state__final(cgul_exception_t* cex,
                                 cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if (uxml->read_state == CGUL_MICROXML__READ_STATE__FINAL) {
        *cex = cgul_exception__new(
            "%s:%zu:%zu: document exhausted",
            cgul_stream_reader__get_presentation_name(cex, uxml->sr),
            cgul_stream_reader__get_line_number_previous(cex, uxml->sr),
            cgul_stream_reader__get_column_number_previous(cex, uxml->sr));
        goto out;
    }

 out:
    return;
}

/**
 * This method reads the next token without skipping white space or
 * comments.  If no input source is selected,
 * <code>uxml->current_token</code> is set to
 * <code>CGUL_MICROXML__TOKEN_INVALID</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__read_next_token__absolute(cgul_exception_t* cex,
                                         cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Assume failure. */
    uxml->current_token = CGUL_MICROXML__TOKEN_INVALID;

    /* Make sure an input source has been selected. */
    if (!uxml->sr) {
        goto out;
    }

    /* Pick up where we left off based on the current read state.
     * Exceptions are propagated after the switch() statement. */
    switch (uxml->read_state) {
    case CGUL_MICROXML__READ_STATE__INVALID:
        *cex = cgul_exception__new(
                   "cgul_microxml__read_next_token__absolute: internal error: "
                   "invalid state");
        goto out;
    case CGUL_MICROXML__READ_STATE__DOCUMENT_START:
        cgul_microxml__read_state__document_start(cex, uxml);
        break;
    case CGUL_MICROXML__READ_STATE__PROLOG:
        cgul_microxml__read_state__prolog(cex, uxml);
        break;
    case CGUL_MICROXML__READ_STATE__MAIN_LOOP:
        cgul_microxml__read_state__main_loop(cex, uxml);
        break;
    case CGUL_MICROXML__READ_STATE__INJECT_ELEMENT_END_TOKEN:
        cgul_microxml__read_state__inject_element_end_token(cex, uxml);
        break;
    case CGUL_MICROXML__READ_STATE__EPILOG:
        cgul_microxml__read_state__epilog(cex, uxml);
        break;
    case CGUL_MICROXML__READ_STATE__DOCUMENT_END:
        cgul_microxml__read_state__document_end(cex, uxml);
        break;
    case CGUL_MICROXML__READ_STATE__FINAL:
        cgul_microxml__read_state__final(cex, uxml);
        break;
    default:
        *cex = cgul_exception__new(
                   "cgul_microxml__read_next_token__absolute: internal error: "
                   "unexpected token: %d", (int)uxml->read_state);
        goto out;
    }
    
    /* Propagate exceptions. */
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Clear the open elements stack <code>open_elements</code>.  This
 * method frees all the values on the stack without freeing the stack
 * itself.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] open_elements open elements stack
 */
static void
cgul_microxml__clear_open_elements(cgul_exception_t* cex,
                                   cgul_microxml_t uxml,
                                   cgul_vector_t open_elements)
{
    open_element_t open_element = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(open_elements);

    /* Free all the open element structures on the stack. */
    for (;;) {

        /* Pop the next open element structure. */
        open_element =
            (open_element_t)cgul_vector__pop_back(cex, open_elements);
        if (!open_element) {
            break;
        }

        /* Put the open element back on its cache. */
        cgul_cache__put(cex, uxml->open_elements_cache, open_element);
    }

 out:
    return;
}

/**
 * Clear the token attributes <code>attributes</code>.  This method
 * frees all the values in the hash table without freeing the keys or
 * the hash table itself.  The keys are not freed because they are
 * owned by the symbol table <code>symbols</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] attributes attributes
 */
static void
cgul_microxml__clear_element_attributes(cgul_exception_t* cex,
                                        cgul_microxml_t uxml,
                                        cgul_hash_t attributes)
{
    cgul_string_t value = NULL;
    cgul_hash_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!attributes) {
        goto out;
    }

    /* We do *not* need to free the keys because they are owned by the
     * symbol table uxml->symbols; however, values are cgul_string
     * objects that belong on the uxml->strings_cache. */
    node = cgul_hash__get_front(cex, attributes);
    for ( ; node ; node = cgul_hash_node__get_next(cex, node)) {
        value = (cgul_string_t)cgul_hash_node__get_value(cex, node);
        cgul_cache__put(cex, uxml->strings_cache, value);
    }

    /* Clear the hash to no longer reference the key/value pairs. */
    cgul_hash__clear(cex, attributes);

 out:
    return;
}

/**
 * Return the symbol for the string <code>s</code>.  If a symbol does
 * not yet exist, one is created.  If an error occurs, an exception is
 * thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> uxml
 * @param[in] s string to intern
 */
static const char*
cgul_microxml__intern_symbol(cgul_exception_t* cex,
                             cgul_microxml_t uxml,
                             cgul_string_t s)
{
    const char* rv = NULL;
    int insert_rv = 0;
    char* symbol = NULL;
    const char* cstring = NULL;
    cgul_hash_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);
    assert(s);

    /* Try to find the symbol. */
    cstring = cgul_string__get_value(cex, s);
    node = cgul_hash__find(cex, uxml->symbols, cstring);
    if (*cex) {
        goto out;
    }

    /* Insert a new symbol if necessary. */
    if (!node) {

        /* Duplicate the string to avoid damaging "s". */
        symbol = cgul_libc__strdup(cex, cstring);
        if (*cex) {
            goto out;
        }

        /* Insert the symbol (as the key). */
        insert_rv = cgul_hash__insert(cex, uxml->symbols, symbol, &node);
        if (*cex) {
            goto out;
        }
        if (!insert_rv) {
            *cex = cgul_exception__new(
                "cgul_microxml__intern_symbol: internal error: "
                "duplicate symbol detected!");
            goto out;
        }

        /* The symbol table now owns "symbol". */
        symbol = NULL;

    }

    /* Success. */
    rv = (const char*)cgul_hash_node__get_key(cex, node);

 out:

    /* Clean up. */
    if (symbol) {
        free(symbol);
    }

    return rv;
}

/**
 * Clear the symbol table <code>symbols</code>.  This method frees all
 * the keys in the hash table without freeing the hash table itself.
 * The values are not freed because they no values are stored in the
 * symbol table.
 *
 * @param[in] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 * @param[in] attributes attributes
 */
static void
cgul_microxml__clear_symbols(cgul_exception_t* cex,
                             cgul_microxml_t uxml,
                             cgul_hash_t symbols)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(symbols);

    /* Free the keys. */
    cgul_hash__free_keys(symbols);

    /* Clear the symbol table to no longer reference the key/value pairs. */
    cgul_hash__clear(cex, symbols);

 out:
    return;
}

/**
 * Reinitialize the <code>cgul_microxml</code> object
 * <code>uxml</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] uxml <code>cgul_microxml</code> instance
 */
static void
cgul_microxml__reinitialize(cgul_exception_t* cex,
                            cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* read_state */
    uxml->read_state = CGUL_MICROXML__READ_STATE__DOCUMENT_START;

    /* current_token */
    uxml->current_token = CGUL_MICROXML__TOKEN_INVALID;

    /* token_line_start */
    uxml->token_line_start = 0;

    /* token_column_start */
    uxml->token_column_start = 0;

    /* current_char */
    uxml->current_char = CGUL_WCHAR__NUL;

    /* symbols */
    if (uxml->symbols) {
        cgul_microxml__clear_symbols(cex, uxml, uxml->symbols);
    } else {
        uxml->symbols = cgul_hash__new(cex);
        if (*cex) {
            goto out;
        }
    }

    /* strings_cache */
    if (!uxml->strings_cache) {
        /* Allocate. */
        uxml->strings_cache = cgul_cache__new(cex,
            (cgul_cache__constructor_t)cgul_string__new,
            (cgul_cache__destructor_t)cgul_string__delete,
            (cgul_cache__restructor_t)cgul_string__clear);
        if (*cex) {
            goto out;
        }
        /* Initialize. */
        cgul_cache__set_size(cex,
                             uxml->strings_cache,
                             CGUL_MICROXML__DEFAULT_STRINGS_CACHE_SIZE);
    }

    /* hashes_cache */
    if (!uxml->hashes_cache) {
        /* Allocate. */
        uxml->hashes_cache = cgul_cache__new(cex,
            (cgul_cache__constructor_t)cgul_hash__new,
            (cgul_cache__destructor_t)cgul_hash__delete,
            (cgul_cache__restructor_t)NULL);
        if (*cex) {
            goto out;
        }
        /* Initialize. */
        cgul_cache__set_size(cex,
                             uxml->hashes_cache,
                             CGUL_MICROXML__DEFAULT_HASHES_CACHE_SIZE);
    }

    /* entity_length_max -- initialized in constructor */

    /* entity_utf8 */
    if (uxml->entity_utf8) {
        cgul_string__set_length(cex, uxml->entity_utf8, 0);
    } else {
        uxml->entity_utf8 = cgul_string__new(cex);
        cgul_string__set_minimum_capacity(
            cex, uxml->entity_utf8, uxml->entity_length_max);
        if (*cex) {
            goto out;
        }
    }

    /* entity_utf32 */
    if (uxml->entity_utf32) {
        cgul_wstring__set_length(cex, uxml->entity_utf32, 0);
    } else {
        uxml->entity_utf32 = cgul_wstring__new(cex);
        cgul_wstring__set_minimum_capacity(
            cex, uxml->entity_utf32, uxml->entity_length_max);
        if (*cex) {
            goto out;
        }
    }

    /* open_elements_cache */
    if (!uxml->open_elements_cache) {
        /* Allocate. */
        uxml->open_elements_cache = cgul_cache__new(cex,
           (cgul_cache__constructor_t)cgul_microxml__open_element__new,
           (cgul_cache__destructor_t)cgul_microxml__open_element__delete,
           (cgul_cache__restructor_t)cgul_microxml__open_element__reinitialize);
        if (*cex) {
            goto out;
        }
        /* Initialize. */
        cgul_cache__set_size(cex,
            uxml->open_elements_cache,
            CGUL_MICROXML__DEFAULT_OPEN_ELEMENTS_CACHE_SIZE);
    }

    /* open_elements */
    if (uxml->open_elements) {
        cgul_microxml__clear_open_elements(cex, uxml, uxml->open_elements);
    } else {
        uxml->open_elements = cgul_vector__new(cex);
        if (*cex) {
            goto out;
        }
    }

    /* element_symbol */
    uxml->element_symbol = NULL;

    /* element_attributes */
    if (uxml->element_attributes) {
        cgul_microxml__clear_element_attributes(
            cex, uxml, uxml->element_attributes);
    } else {
        uxml->element_attributes =
            (cgul_hash_t)cgul_cache__get(cex, uxml->hashes_cache);
        if (*cex) {
            goto out;
        }
    }

    /* is_empty_element */
    uxml->is_empty_element = 0;

    /* text */
    if (uxml->text) {
        cgul_string__set_length(cex, uxml->text, 0);
    } else {
        uxml->text = cgul_string__new(cex);
        cgul_string__set_minimum_capacity(
            cex, uxml->text, uxml->text_length_max);
        if (*cex) {
            goto out;
        }
    }

    /* text_length_max -- initialized in constructor */

    /* is_white_space */
    uxml->is_white_space = 0;

    /* skip_white_space -- initialized in constructor */

    /* comment */
    if (uxml->comment) {
        cgul_string__set_length(cex, uxml->comment, 0);
    } else {
        uxml->comment = cgul_string__new(cex);
        cgul_string__set_minimum_capacity(
            cex, uxml->comment, uxml->comment_length_max);
        if (*cex) {
            goto out;
        }
    }

    /* comment_length_max -- initialized in constructor */

    /* skip_comments -- initialized in constructor */

    /* skip_prolog_or_epilog_tokens -- initialized in constructor */

    /* sr */
    if (uxml->sr && uxml->we_own_sr) {
        cgul_stream_reader__delete(uxml->sr);
    }
    uxml->sr = NULL;

    /* we_own_sr */
    uxml->we_own_sr = 0;

    /* br */
    if (uxml->br) {
        cgul_block_reader__delete(uxml->br);
    }
    uxml->br = NULL;

 out:
    return;
}

/* See comments in cgul_microxml_p.h. */
const char*
cgul_microxml__token_to_string(cgul_exception_t* cex,
                               cgul_microxml__token_t token)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    switch (token) {
    case CGUL_MICROXML__TOKEN_INVALID:
        rv = "CGUL_MICROXML__TOKEN_INVALID";
        break;
    case CGUL_MICROXML__TOKEN_DOCUMENT_START:
        rv = "CGUL_MICROXML__TOKEN_DOCUMENT_START";
        break;
    case CGUL_MICROXML__TOKEN_DOCUMENT_END:
        rv = "CGUL_MICROXML__TOKEN_DOCUMENT_END";
        break;
    case CGUL_MICROXML__TOKEN_ELEMENT_START:
        rv = "CGUL_MICROXML__TOKEN_ELEMENT_START";
        break;
    case CGUL_MICROXML__TOKEN_ELEMENT_END:
        rv = "CGUL_MICROXML__TOKEN_ELEMENT_END";
        break;
    case CGUL_MICROXML__TOKEN_TEXT:
        rv = "CGUL_MICROXML__TOKEN_TEXT";
        break;
    case CGUL_MICROXML__TOKEN_COMMENT:
        rv = "CGUL_MICROXML__TOKEN_COMMENT";
        break;
    default:
        *cex = cgul_exception__new(
            "cgul_microxml__get_token_as_string: "
            "unexpected token type: %lu",
            (unsigned long)token);
        goto out;
    }

 out:
    return rv;
}

/* See comments in cgul_microxml_p.h. */
cgul_microxml__token_t
cgul_microxml__token_from_string(cgul_exception_t* cex,
                                 const char* s)
{
    cgul_microxml__token_t rv = CGUL_MICROXML__TOKEN_INVALID;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (strcmp(s, "CGUL_MICROXML__TOKEN_INVALID") == 0) {
        rv = CGUL_MICROXML__TOKEN_INVALID;
    } else if (strcmp(s, "CGUL_MICROXML__TOKEN_DOCUMENT_START") == 0) {
        rv = CGUL_MICROXML__TOKEN_DOCUMENT_START;
    } else if (strcmp(s, "CGUL_MICROXML__TOKEN_DOCUMENT_END") == 0) {
        rv = CGUL_MICROXML__TOKEN_DOCUMENT_END;
    } else if (strcmp(s, "CGUL_MICROXML__TOKEN_ELEMENT_START") == 0) {
        rv = CGUL_MICROXML__TOKEN_ELEMENT_START;
    } else if (strcmp(s, "CGUL_MICROXML__TOKEN_ELEMENT_END") == 0) {
        rv = CGUL_MICROXML__TOKEN_ELEMENT_END;
    } else if (strcmp(s, "CGUL_MICROXML__TOKEN_TEXT") == 0) {
        rv = CGUL_MICROXML__TOKEN_TEXT;
    } else if (strcmp(s, "CGUL_MICROXML__TOKEN_COMMENT") == 0) {
        rv = CGUL_MICROXML__TOKEN_COMMENT;
    } else {
        *cex = cgul_exception__new("cgul_microxml__token_from_string: "
                                   "unexpected toke type: %s",
                                   s);
        goto out;
    }

 out:
    return rv;
}

/** See comments in cgul_microxml_p.h. */
open_element_t
cgul_microxml__open_element__new(cgul_exception_t* cex)
{
    open_element_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    rv = (open_element_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_microxml__open_element__delete(rv);
        rv = NULL;
    }

    return rv;
}

/** See comments in cgul_microxml_p.h. */
void
cgul_microxml__open_element__delete(open_element_t open_element)
{
    if (open_element) {

        /* name -- a symbol so it should not be deleted */

        /* Free the structure itself. */
        free(open_element);
    }
}

void
cgul_microxml__open_element__reinitialize(open_element_t open_element)
{
    /* Clear the structure. */
    memset(open_element, 0, sizeof(*open_element));
}

char*
cgul_microxml__escape(cgul_exception_t* cex,
                      const char* s)
{
    char* rv = NULL;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!s) {
        goto out;
    }

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate over the characters in the string.  Even though the
     * input string is UTF-8 encoded, the code below works because
     * UTF-8 was designed to be skimmed. */
    for ( ; *s ; ++s) {
        switch (*s) {
        case '&':
            cgul_string__append(cex, tmp, "&amp;");
            break;
        case '<':
            cgul_string__append(cex, tmp, "&lt;");
            break;
        case '>':
            cgul_string__append(cex, tmp, "&gt;");
            break;
        case '"':
            cgul_string__append(cex, tmp, "&quot;");
            break;
        case '\'':
            cgul_string__append(cex, tmp, "&apos;");
            break;
        default:
            cgul_string__append_char(cex, tmp, *s);
        }
        if (*cex) {
            goto out;
        }
    }

    /* Take the escaped string so it can be returned to the caller. */
    rv = cgul_string__take_value(cex, tmp);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

cgul_microxml_t
cgul_microxml__new(cgul_exception_t* cex)
{
    cgul_microxml_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate. */
    rv = (cgul_microxml_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* entity_length_max */
    rv->entity_length_max = CGUL_MICROXML__DEFAULT_ENTITY_LENGTH_MAX;

    /* text_length_max */
    rv->text_length_max = CGUL_MICROXML__DEFAULT_TEXT_LENGTH_MAX;

    /* comment_length_max */
    rv->comment_length_max = CGUL_MICROXML__DEFAULT_COMMENT_LENGTH_MAX;

    /* skip_comments */
    rv->skip_comments = 1;

    /* skip_prolog_or_epilog_tokens */
    rv->skip_prolog_or_epilog_tokens = 1;

    /* Initialize the instance .*/
    cgul_microxml__reinitialize(cex, rv);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially initialized instances. */
    if (*cex && rv) {
        cgul_microxml__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_microxml_t
cgul_microxml__new_from_fname(cgul_exception_t* cex,
                              const char* fname)
{
    cgul_microxml_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default "cgul_microxml" instance. */
    rv = cgul_microxml__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the file. */
    cgul_microxml__open_fname(cex, rv, fname);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially initialized instances. */
    if (*cex && rv) {
        cgul_microxml__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_microxml_t
cgul_microxml__new_from_file(cgul_exception_t* cex,
                             FILE* f)
{
    cgul_microxml_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default "cgul_microxml" instance. */
    rv = cgul_microxml__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the file. */
    cgul_microxml__open_file(cex, rv, f);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially initialized instances. */
    if (*cex && rv) {
        cgul_microxml__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_microxml_t
cgul_microxml__new_from_memory(cgul_exception_t* cex,
                               const char* buffer,
                               size_t buffer_size)
{
    cgul_microxml_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default "cgul_microxml" instance. */
    rv = cgul_microxml__new(cex);
    if (*cex) {
        goto out;
    }

    /* Open the memory. */
    cgul_microxml__open_memory(cex, rv, buffer, buffer_size);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially initialized instances. */
    if (*cex && rv) {
        cgul_microxml__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_microxml_t
cgul_microxml__new_from_stream_reader(cgul_exception_t* cex,
                                      cgul_stream_reader_t sr)
{
    cgul_microxml_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a default object. */
    rv = cgul_microxml__new(cex);
    if (*cex) {
        goto out;
    }

    /* Try to set the stream reader. */
    cgul_microxml__set_stream_reader(cex, rv, sr);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed instances. */
    if (*cex && rv) {
        cgul_microxml__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_microxml__delete(cgul_microxml_t uxml)
{
    cgul_exception_t local = NULL;

    if (uxml) {

        /* sr */
        if (uxml->sr && uxml->we_own_sr) {
            cgul_stream_reader__delete(uxml->sr);
        }

        /* br */
        if (uxml->br) {
            cgul_block_reader__delete(uxml->br);
        }

        /* comment */
        if (uxml->comment) {
            cgul_string__delete(uxml->comment);
        }

        /* text */
        if (uxml->text) {
            cgul_string__delete(uxml->text);
        }

        /* element_attributes */
        if (uxml->element_attributes) {
            /* Does not throw an exception. */
            cgul_microxml__delete_attributes(
                &local, uxml, uxml->element_attributes);
        }

        /* open_elements */
        if (uxml->open_elements) {
            /* Does not throw an exception. */
            cgul_microxml__clear_open_elements(
                &local, uxml, uxml->open_elements);
            cgul_vector__delete(uxml->open_elements);
        }

        /* open_elements_cache */
        if (uxml->open_elements_cache) {
            cgul_cache__delete(uxml->open_elements_cache);
        }

        /* entity_utf32 */
        if (uxml->entity_utf32) {
            cgul_wstring__delete(uxml->entity_utf32);
        }

        /* entity_utf8 */
        if (uxml->entity_utf8) {
            cgul_string__delete(uxml->entity_utf8);
        }

        /* hashes_cache */
        if (uxml->hashes_cache) {
            cgul_cache__delete(uxml->hashes_cache);
        }

        /* strings_cache */
        if (uxml->strings_cache) {
            cgul_cache__delete(uxml->strings_cache);
        }

        /* symbols */
        if (uxml->symbols) {
            cgul_hash__free_keys(uxml->symbols);
            cgul_hash__delete(uxml->symbols);
        }

        /* Free the object itself. */
        free(uxml);

    }
}

void
cgul_microxml__open_fname(cgul_exception_t* cex,
                          cgul_microxml_t uxml,
                          const char* fname)
{
    cgul_stream_reader_t tmp_sr = NULL;
    cgul_block_reader_file_t tmp_brfile = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Reinitialize the object which has the side-effect of resetting
     * uxml->sr and the internal uxml->br (if necessary). */
    cgul_microxml__reinitialize(cex, uxml);
    if (*cex) {
        goto out;
    }

    /*
     * Pre-allocate the new stream reader and block reader.
     */

    /* Allocate the stream reader. */
    tmp_sr = cgul_stream_reader__new(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate a new block reader for the file. */
    tmp_brfile = cgul_block_reader_file__new_from_fname(cex, fname);
    if (*cex) {
        goto out;
    }

    /* Add the block reader to the stream reader. */
    cgul_stream_reader__add_block_reader(
        cex,
        tmp_sr,
        cgul_block_reader_file__as_block_reader(cex, tmp_brfile));
    if (*cex) {
        goto out;
    }

    /*
     * Transfer the pre-allocated objects to the "uxml" object.
     */

    /* Transfer ownership of "tmp_sr" to this class. */
    uxml->sr = tmp_sr;
    uxml->we_own_sr = 1;
    tmp_sr = NULL;

    /* Transfer ownership of "tmp_brfile" to this class.  (We always
     * own uxml->br if it is not NULL.) */
    uxml->br = cgul_block_reader_file__as_block_reader(cex, tmp_brfile);
    tmp_brfile = NULL;

 out:

    /* Clean up. */
    if (tmp_sr) {
        cgul_stream_reader__delete(tmp_sr);
    }
    if (tmp_brfile) {
        cgul_block_reader_file__delete(tmp_brfile);
    }

    return;
}

void
cgul_microxml__open_file(cgul_exception_t* cex,
                         cgul_microxml_t uxml,
                         FILE* f)
{
    cgul_stream_reader_t tmp_sr = NULL;
    cgul_block_reader_file_t tmp_brfile = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Reinitialize the object which has the side-effect of resetting
     * uxml->sr and the internal uxml->br (if necessary). */
    cgul_microxml__reinitialize(cex, uxml);
    if (*cex) {
        goto out;
    }

    /*
     * Pre-allocate the new stream reader and block reader.
     */

    /* Allocate the stream reader. */
    tmp_sr = cgul_stream_reader__new(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate a new block reader for the file. */
    tmp_brfile = cgul_block_reader_file__new_from_file(cex, f);
    if (*cex) {
        goto out;
    }

    /* Add the block reader to the stream reader. */
    cgul_stream_reader__add_block_reader(
        cex,
        tmp_sr,
        cgul_block_reader_file__as_block_reader(cex, tmp_brfile));
    if (*cex) {
        goto out;
    }

    /*
     * Transfer the pre-allocated objects to the "uxml" object.
     */

    /* Transfer ownership of "tmp_sr" to this class. */
    uxml->sr = tmp_sr;
    uxml->we_own_sr = 1;
    tmp_sr = NULL;

    /* Transfer ownership of "tmp_brfile" to this class.  (We always
     * own uxml->br if it is not NULL.) */
    uxml->br = cgul_block_reader_file__as_block_reader(cex, tmp_brfile);
    tmp_brfile = NULL;

 out:

    /* Clean up. */
    if (tmp_sr) {
        cgul_stream_reader__delete(tmp_sr);
    }
    if (tmp_brfile) {
        cgul_block_reader_file__delete(tmp_brfile);
    }

    return;
}

void
cgul_microxml__open_memory(cgul_exception_t* cex,
                           cgul_microxml_t uxml,
                           const char* buffer,
                           size_t buffer_size)
{
    cgul_stream_reader_t tmp_sr = NULL;
    cgul_block_reader_memory_t tmp_brmemory = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Reinitialize the object which has the side-effect of resetting
     * uxml->sr and the internal uxml->br (if necessary). */
    cgul_microxml__reinitialize(cex, uxml);
    if (*cex) {
        goto out;
    }

    /*
     * Pre-allocate the new stream reader and block reader.
     */

    /* Allocate the stream reader. */
    tmp_sr = cgul_stream_reader__new(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate a new block reader for the file. */
    tmp_brmemory =
        cgul_block_reader_memory__new_from_memory(cex,
                                                  (void*)buffer,
                                                  buffer_size);
    if (*cex) {
        goto out;
    }

    /* Add the block reader to the stream reader. */
    cgul_stream_reader__add_block_reader(
        cex,
        tmp_sr,
        cgul_block_reader_memory__as_block_reader(cex, tmp_brmemory));
    if (*cex) {
        goto out;
    }

    /*
     * Transfer the pre-allocated objects to the "uxml" object.
     */

    /* Transfer ownership of "tmp_sr" to this class. */
    uxml->sr = tmp_sr;
    uxml->we_own_sr = 1;
    tmp_sr = NULL;

    /* Transfer ownership of "tmp_brmemory" to this class.  (We always
     * own uxml->br if it is not NULL.) */
    uxml->br = cgul_block_reader_memory__as_block_reader(cex, tmp_brmemory);
    tmp_brmemory = NULL;

 out:

    /* Clean up. */
    if (tmp_sr) {
        cgul_stream_reader__delete(tmp_sr);
    }
    if (tmp_brmemory) {
        cgul_block_reader_memory__delete(tmp_brmemory);
    }

    return;
}

void
cgul_microxml__close(cgul_exception_t* cex,
                     cgul_microxml_t uxml)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Reinitialize. */
    cgul_microxml__reinitialize(cex, uxml);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

cgul_stream_reader_t
cgul_microxml__get_stream_reader(cgul_exception_t* cex,
                                 cgul_microxml_t uxml)
{
    cgul_stream_reader_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return the stream reader to the user. */
    rv = uxml->sr;

 out:
    return rv;
}

void
cgul_microxml__set_stream_reader(cgul_exception_t* cex,
                                 cgul_microxml_t uxml,
                                 cgul_stream_reader_t sr)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Reinitialize the object which has the side-effect of clearing
     * uxml->sr and deleting the internal uxml->br (if necessary). */
    cgul_microxml__reinitialize(cex, uxml);
    if (*cex) {
        goto out;
    }

    /* Set the stream reader. */
    uxml->sr = sr;

 out:
    return;
}

const char*
cgul_microxml__get_presentation_name(cgul_exception_t* cex,
                                     cgul_microxml_t uxml)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Make sure the user has selected an input source before
     * attempting to set its presentation name. */
    if (!uxml->sr) {
        *cex = cgul_exception__new("cgul_microxml__get_presentation_name: "
                                   "no input source selected");
        goto out;
    }

    /* Return the presentation name. */
    rv = cgul_stream_reader__get_presentation_name(cex, uxml->sr);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

void
cgul_microxml__set_presentation_name(cgul_exception_t* cex,
                                     cgul_microxml_t uxml,
                                     const char* presentation_name)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Make sure the user has selected an input source before
     * attempting to set its presentation name. */
    if (!uxml->sr) {
        *cex = cgul_exception__new("cgul_microxml__get_presentation_name: "
                                   "no input source selected");
        goto out;
    }

    /* Set the presentation name. */
    cgul_stream_reader__set_presentation_name(cex, uxml->sr, presentation_name);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

size_t
cgul_microxml__get_maximum_entity_length(cgul_exception_t* cex,
                                         cgul_microxml_t uxml)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return the maximum entity length to the user. */
    rv = uxml->entity_length_max;

 out:
    return rv;
}

void
cgul_microxml__set_maximum_entity_length(cgul_exception_t* cex,
                                         cgul_microxml_t uxml,
                                         size_t entity_length_max)
{
    size_t capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Set the maximum entity length. */
    uxml->entity_length_max = entity_length_max;

    /* Determine the capacity for the entity strings clipping the
     * minimum capacity at the default maximum entity length to avoid
     * allocating too much memory.  (It is ok for "capacity" to be
     * zero; in this case, the strings will be set to the cgul_string
     * default capacity.) */
    capacity = CGUL_MIN(entity_length_max,
                        CGUL_MICROXML__DEFAULT_ENTITY_LENGTH_MAX);

    /* Set the capacity for uxml->entity_utf8. */
    cgul_string__set_minimum_capacity(cex, uxml->entity_utf8, capacity);
    if (*cex) {
        goto out;
    }

    /* Set the capacity for uxml->entity_utf32. */
    cgul_wstring__set_minimum_capacity(cex, uxml->entity_utf32, capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

size_t
cgul_microxml__get_line_number(cgul_exception_t* cex,
                               cgul_microxml_t uxml)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if (!uxml->sr) {
        *cex = cgul_exception__new("cgul_microxml__get_line_number_current: "
                                   "no input source selected");
        goto out;
    }

    /* Return the line number for the current token. */
    rv = uxml->token_line_start;

 out:
    return rv;
}

size_t
cgul_microxml__get_column_number(cgul_exception_t* cex,
                                 cgul_microxml_t uxml)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if (!uxml->sr) {
        *cex = cgul_exception__new("cgul_microxml__get_column_number_current: "
                                   "no input source selected");
        goto out;
    }

    /* Return the column number for the current token. */
    rv = uxml->token_column_start;

 out:
    return rv;
}

void
cgul_microxml__read_next_token(cgul_exception_t* cex,
                               cgul_microxml_t uxml)
{
    int token_is_in_prolog_or_epilog = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Read the next token skipping white space and comments as necessary. */
    for (;;) {

        /* For a token to be in the prolog or epilog, the read state
         * needs to be PROLOG or EPILOG before the read. */
        token_is_in_prolog_or_epilog = 
            ((uxml->read_state == CGUL_MICROXML__READ_STATE__PROLOG) ||
             (uxml->read_state == CGUL_MICROXML__READ_STATE__EPILOG));

        /* Read the next token *without* skipping white space or comments. */
        cgul_microxml__read_next_token__absolute(cex, uxml);
        if (*cex) {
            goto out;
        }

        /* For a token to be in the prolog or epilog, the read state
         * needs to be PROLOG, EPILOG, or DOCUMENT_END after the read. */
        token_is_in_prolog_or_epilog &= 
            ((uxml->read_state == CGUL_MICROXML__READ_STATE__PROLOG) ||
             (uxml->read_state == CGUL_MICROXML__READ_STATE__EPILOG) ||
             (uxml->read_state == CGUL_MICROXML__READ_STATE__DOCUMENT_END));

        /* Check for tokens to skip in the prolog and epilog. */
        if (token_is_in_prolog_or_epilog) {

            /* If we are skipping prolog or epilog tokens, try again. */
            if (uxml->skip_prolog_or_epilog_tokens) {
                continue;
            }

        /* Check for tokens to skip that are inside the root element. */
        } else {

            /* If the token is text and it is white space and we are
             * skipping white space, try again. */
            if((uxml->current_token == CGUL_MICROXML__TOKEN_TEXT) &&
               uxml->is_white_space &&
               uxml->skip_white_space)
            {
                continue;
            }

            /* If the token is a comment and we are skipping comments, try
             * again. */
            if((uxml->current_token == CGUL_MICROXML__TOKEN_COMMENT) &&
               uxml->skip_comments)
            {
                continue;
            }

        }

        /* Success. */
        break;
    }

 out:
    return;
}

int
cgul_microxml__is_document_start(cgul_exception_t* cex,
                                 cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return whether the current token indicates the start of the document. */
    rv = (uxml->current_token == CGUL_MICROXML__TOKEN_DOCUMENT_START);

 out:
    return rv;
}

int
cgul_microxml__is_document_end(cgul_exception_t* cex,
                               cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return whether the current token indicates the end of the document. */
    rv = (uxml->current_token == CGUL_MICROXML__TOKEN_DOCUMENT_END);

 out:
    return rv;
}

int
cgul_microxml__is_element_start(cgul_exception_t* cex,
                                cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return whether the current token indicates an element start tag. */
    rv = (uxml->current_token == CGUL_MICROXML__TOKEN_ELEMENT_START);

 out:
    return rv;
}

int
cgul_microxml__is_element_end(cgul_exception_t* cex,
                              cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return whether the current token indicates an element end tag. */
    rv = (uxml->current_token == CGUL_MICROXML__TOKEN_ELEMENT_END);

 out:
    return rv;
}

int
cgul_microxml__is_text(cgul_exception_t* cex,
                       cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return whether the current token indicates text. */
    rv = (uxml->current_token == CGUL_MICROXML__TOKEN_TEXT);

 out:
    return rv;
}

int
cgul_microxml__is_comment(cgul_exception_t* cex,
                          cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return whether the current token indicates a comment. */
    rv = (uxml->current_token == CGUL_MICROXML__TOKEN_COMMENT);

 out:
    return rv;
}

const char*
cgul_microxml__get_element_name(cgul_exception_t* cex,
                                cgul_microxml_t uxml)
{
    const char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if ((uxml->current_token != CGUL_MICROXML__TOKEN_ELEMENT_START) &&
        (uxml->current_token != CGUL_MICROXML__TOKEN_ELEMENT_END))
    {
        *cex = cgul_exception__new(
            "cgul_microxml__get_element_name: "
            "not valid for current token: %s",
            cgul_microxml__token_to_string(cex, uxml->current_token));
        goto out;
    }

    /* Return the name to the user. */
    rv = uxml->element_symbol;

 out:
    return rv;
}

cgul_string_t
cgul_microxml__get_attribute_value(cgul_exception_t* cex,
                                   cgul_microxml_t uxml,
                                   const char* attribute_name)
{
    cgul_string_t rv = NULL;
    cgul_hash_node_t node = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if (!attribute_name) {
        goto out;
    }

    /* Find the key/value pair in the hash table. */
    node = cgul_hash__find(cex, uxml->element_attributes, attribute_name);
    if (!node) {
        goto out;
    }

    /* Success. */
    rv = (cgul_string_t)cgul_hash_node__get_value(cex, node);

 out:
    return rv;
}

cgul_hash_t
cgul_microxml__get_attributes(cgul_exception_t* cex,
                              cgul_microxml_t uxml)
{
    cgul_hash_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if (uxml->current_token != CGUL_MICROXML__TOKEN_ELEMENT_START) {
        *cex = cgul_exception__new(
            "cgul_microxml__get_attributes: "
            "not valid for current token: %s",
            cgul_microxml__token_to_string(cex, uxml->current_token));
        goto out;
    }

    /* Return the attributes to the user. */
    rv = uxml->element_attributes;

 out:
    return rv;
}

cgul_hash_t
cgul_microxml__take_attributes(cgul_exception_t* cex,
                               cgul_microxml_t uxml)
{
    cgul_hash_t rv = NULL;
    cgul_hash_t tmp = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if (uxml->current_token != CGUL_MICROXML__TOKEN_ELEMENT_START) {
        *cex = cgul_exception__new(
            "cgul_microxml__take_attributes: "
            "not valid for current token: %s",
            cgul_microxml__token_to_string(cex, uxml->current_token));
        goto out;
    }

    /* Pre-allocate the new hash. */
    tmp = (cgul_hash_t)cgul_cache__get(cex, uxml->hashes_cache);
    if (*cex) {
        goto out;
    }

    /* Return the attributes to the user transfering ownership of the
     * hash itself and the values stored in the hash.  Because the
     * keys are interned strings, they are still owned by the symbol
     * table uxml->symbols. */
    rv = uxml->element_attributes;
    if (*cex) {
        goto out;
    }

    /* Take ownership of the pre-allocated hash. */
    uxml->element_attributes = tmp;
    tmp = NULL;

 out:

    /* Clean up. */
    if (tmp) {
        cgul_cache__put(&local, uxml->hashes_cache, tmp);
    }

    return rv;
}

void
cgul_microxml__delete_attributes(cgul_exception_t* cex,
                                 cgul_microxml_t uxml,
                                 cgul_hash_t attributes)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (!attributes) {
        goto out;
    }

    /* Clear the token attributes (i.e., all key/value pairs).  Note
     * that we must do this manually because no "restructor" has been
     * set for the uxml->hashes_cache. */
    cgul_microxml__clear_element_attributes(cex, uxml, attributes);

    /* Free the hash table itself. */
    cgul_cache__put(cex, uxml->hashes_cache, attributes);

 out:
    return;
}

int
cgul_microxml__is_empty_element(cgul_exception_t* cex,
                                cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if ((uxml->current_token != CGUL_MICROXML__TOKEN_ELEMENT_START) &&
        (uxml->current_token != CGUL_MICROXML__TOKEN_ELEMENT_END))
    {
        *cex = cgul_exception__new(
            "cgul_microxml__is_empty_element: "
            "not valid for current token: %s",
            cgul_microxml__token_to_string(cex, uxml->current_token));
        goto out;
    }

    /* Return whether the start tag was closed using the empty-element
     * syntax (e.g., <foo/>). */
    rv = uxml->is_empty_element;

 out:
    return rv;
}

cgul_string_t
cgul_microxml__get_text(cgul_exception_t* cex,
                        cgul_microxml_t uxml)
{
    cgul_string_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if (uxml->current_token != CGUL_MICROXML__TOKEN_TEXT) {
        *cex = cgul_exception__new(
            "cgul_microxml__get_text: "
            "not valid for current token: %s",
            cgul_microxml__token_to_string(cex, uxml->current_token));
        goto out;
    }

    /* Return the text to the user. */
    rv = uxml->text;

 out:
    return rv;
}

int
cgul_microxml__is_white_space(cgul_exception_t* cex,
                              cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if (uxml->current_token != CGUL_MICROXML__TOKEN_TEXT) {
        *cex = cgul_exception__new(
            "cgul_microxml__is_white_space: "
            "not valid for current token: %s",
            cgul_microxml__token_to_string(cex, uxml->current_token));
        goto out;
    }

    /* Return whether the text is pure white space. */
    rv = uxml->is_white_space;

 out:
    return rv;
}

int
cgul_microxml__get_skip_white_space(cgul_exception_t* cex,
                                    cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return whether to skip text that is pure white space. */
    rv = uxml->skip_white_space;

 out:
    return rv;
}

void
cgul_microxml__set_skip_white_space(cgul_exception_t* cex,
                                    cgul_microxml_t uxml,
                                    int skip)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Set whether to skip text that is pure white space. */
    uxml->skip_white_space = skip;

 out:
    return;
}

size_t
cgul_microxml__get_maximum_text_length(cgul_exception_t* cex,
                                       cgul_microxml_t uxml)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return the maximum size to which text is allowed to grow before
     * being fully tokenized. */
    rv = uxml->text_length_max;

 out:
    return rv;
}

void
cgul_microxml__set_maximum_text_length(cgul_exception_t* cex,
                                       cgul_microxml_t uxml,
                                       size_t text_length_max)
{
    size_t capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Set the maximum size to which text is allowed to grow before
     * being fully tokenized. */
    uxml->text_length_max = text_length_max;

    /* Determine the capacity for the text strings clipping the
     * minimum capacity at the default maximum text length to avoid
     * allocating too much memory.  (It is ok for "capacity" to be
     * zero; in this case, the strings will be set to the cgul_string
     * default capacity.) */
    capacity = CGUL_MIN(text_length_max,
                        CGUL_MICROXML__DEFAULT_TEXT_LENGTH_MAX);

    /* Set the capacity for uxml->text. */
    cgul_string__set_minimum_capacity(cex, uxml->text, capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

cgul_string_t
cgul_microxml__get_comment(cgul_exception_t* cex,
                           cgul_microxml_t uxml)
{
    cgul_string_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Sanity check. */
    if (uxml->current_token != CGUL_MICROXML__TOKEN_COMMENT) {
        *cex = cgul_exception__new(
            "cgul_microxml__get_comment: "
            "not valid for current token: %s",
            cgul_microxml__token_to_string(cex, uxml->current_token));
        goto out;
    }

    /* Return the comment to the user. */
    rv = uxml->comment;

 out:
    return rv;
}

int
cgul_microxml__get_skip_comments(cgul_exception_t* cex,
                                 cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return whether to skip comments. */
    rv = uxml->skip_comments;

 out:
    return rv;
}

void
cgul_microxml__set_skip_comments(cgul_exception_t* cex,
                                 cgul_microxml_t uxml,
                                 int skip)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Set whether to skip comments. */
    uxml->skip_comments = skip;

 out:
    return;
}

size_t
cgul_microxml__get_maximum_comment_length(cgul_exception_t* cex,
                                          cgul_microxml_t uxml)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return the maximum comment length to the user. */
    rv = uxml->comment_length_max;

 out:
    return rv;
}

void
cgul_microxml__set_maximum_comment_length(cgul_exception_t* cex,
                                          cgul_microxml_t uxml,
                                          size_t comment_length_max)
{
    size_t capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Set the maximum comment length. */
    uxml->comment_length_max = comment_length_max;

    /* Determine the capacity for the comment string clipping the
     * minimum capacity at the default maximum comment length to avoid
     * allocating too much memory.  (It is ok for "capacity" to be
     * zero; in this case, the string will be set to the cgul_string
     * default capacity.) */
    capacity = CGUL_MIN(comment_length_max,
                        CGUL_MICROXML__DEFAULT_COMMENT_LENGTH_MAX);

    /* Set the capacity for uxml->comment. */
    cgul_string__set_minimum_capacity(cex, uxml->comment, capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

int
cgul_microxml__get_skip_prolog_or_epilog_tokens(cgul_exception_t* cex,
                                                cgul_microxml_t uxml)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Return whether to skip prolog and epilog tokens. */
    rv = uxml->skip_prolog_or_epilog_tokens;

 out:
    return rv;
}

void
cgul_microxml__set_skip_prolog_or_epilog_tokens(cgul_exception_t* cex,
                                                cgul_microxml_t uxml,
                                                int skip)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(uxml);

    /* Set whether to skip prolog and epilog tokens. */
    uxml->skip_prolog_or_epilog_tokens = skip;

 out:
    return;
}
