#include "cgul_multimap_node_p.h"
#include "cgul_libc.h"

/*
 * This method is used as the "restructor" for the "cgul_cache" of
 * nodes in the "cgul_multimap".  "restructor" methods are defined in
 * "cgul_cache" to not throw exceptions.
 *
 * While the signature for this method does not exactly match the
 * signature defined for "restructor" methods, it can be safely
 * typecast to match.
 */
void
cgul_multimap_node__initialize(cgul_multimap_node_t n)
{
    assert(n);
    memset(n, 0, sizeof(*n));
}

cgul_multimap_node_t
cgul_multimap_node__new(cgul_exception_t* cex)
{
    cgul_multimap_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_multimap_node_t)cgul_libc__malloc(cex, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_multimap_node__initialize(rv);

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_multimap_node__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_multimap_node__delete(cgul_multimap_node_t n)
{
    if (n) {
        free(n);
    }
}

void*
cgul_multimap_node__get_key(cgul_exception_t* cex,
                            cgul_multimap_node_t n)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);
    assert(n->tree_node);

    rv = cgul_rbtree_node__get_key(cex, n->tree_node);

 out:
    return rv;
}

void*
cgul_multimap_node__get_value(cgul_exception_t* cex,
                              cgul_multimap_node_t n)
{
    void* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = (void*)n->value;

 out:
    return rv;
}

void
cgul_multimap_node__set_value(cgul_exception_t* cex,
                              cgul_multimap_node_t n,
                              const void* value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    n->value = value;

 out:
    return;
}

cgul_multimap_node_t
cgul_multimap_node__get_prev(cgul_exception_t* cex,
                             cgul_multimap_node_t n,
                             int older_is_greater)
{
    cgul_list_t l = NULL;
    cgul_list_node_t prev_list_node = NULL;
    cgul_rbtree_node_t prev_rbtree_node = NULL;
    cgul_multimap_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    /* How we iterate over the list (not the red-black tree) is
     * determined by older_is_greater. */
    if (older_is_greater) {

        /* Try to get the previous node in the current list. */
        prev_list_node = cgul_list_node__get_prev(cex, n->list_node);
        if (prev_list_node == NULL) {
            /* We have to move to the previous node in the red-black tree. */
            prev_rbtree_node = cgul_rbtree_node__get_prev(cex, n->tree_node);
            if (prev_rbtree_node) {
                /* Get the list in the next red-black tree node. */
                l = (cgul_list_t)cgul_rbtree_node__get_value(cex,
                                                             prev_rbtree_node);
                assert(l);
                /* Get the last node in the list. */
                prev_list_node = cgul_list__get_back(cex, l);
            }
        }

    } else {

        /* Try to get the previous node in the current list.  We use
         * "get_next()" because the use set older_is_greater to false
         * which means older nodes are now considered to be less than
         * younger nodes and the older nodes are always at the back of
         * the list. */
        prev_list_node = cgul_list_node__get_next(cex, n->list_node);
        if (prev_list_node == NULL) {
            /* We have to move to the previous node in the red-black
             * tree.  Note that older_is_greater has no affect on
             * this.  It is merely a secondary sort criterion.  The
             * primary criterion is the user's "compare" function. */
            prev_rbtree_node = cgul_rbtree_node__get_prev(cex, n->tree_node);
            if (prev_rbtree_node) {
                /* Get the list in the next red-black tree node. */
                l = (cgul_list_t)cgul_rbtree_node__get_value(cex,
                                                             prev_rbtree_node);
                assert(l);
                /* Get the first node (i.e., the youngest node) in the
                 * list.  We do this because the younger nodes are
                 * considered to be greater. */
                prev_list_node = cgul_list__get_front(cex, l);
            }
        }

    }

    /* Get the multimap node from the list node. */
    if (prev_list_node) {
        rv = (cgul_multimap_node_t)cgul_list_node__get_value(cex,
                                                             prev_list_node);
    }

 out:
    return rv;
}

cgul_multimap_node_t
cgul_multimap_node__get_next(cgul_exception_t* cex,
                             cgul_multimap_node_t n,
                             int older_is_greater)
{
    cgul_list_t l = NULL;
    cgul_list_node_t next_list_node = NULL;
    cgul_rbtree_node_t next_rbtree_node = NULL;
    cgul_multimap_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    /* How we iterate over the list (not the red-black tree) is
     * determined by older_is_greater. */
    if (older_is_greater) {

        /* Try to get the next node in the current list. */
        next_list_node = cgul_list_node__get_next(cex, n->list_node);
        if (next_list_node == NULL) {
            /* We have to move to the next node in the red-black tree. */
            next_rbtree_node = cgul_rbtree_node__get_next(cex, n->tree_node);
            if (next_rbtree_node) {
                /* Get the list in the next red-black tree node. */
                l = (cgul_list_t)cgul_rbtree_node__get_value(cex,
                                                             next_rbtree_node);
                assert(l);
                /* Get the first node in the list. */
                next_list_node = cgul_list__get_front(cex, l);
            }
        }

    } else {

        /* Try to get the next node in the current list.  We use
         * "get_prev()" because the use set older_is_greater to false
         * which means younger nodes are now considered to be greater
         * than older nodes and the younger nodes are always at the
         * front of the list. */
        next_list_node = cgul_list_node__get_prev(cex, n->list_node);
        if (next_list_node == NULL) {
            /* We have to move to the next node in the red-black tree.
             * Note that older_is_greater has no affect on this.  It
             * is merely a secondary sort criterion.  The primary
             * criterion is the user's "compare" function. */
            next_rbtree_node = cgul_rbtree_node__get_next(cex, n->tree_node);
            if (next_rbtree_node) {
                /* Get the list in the next red-black tree node. */
                l = (cgul_list_t)cgul_rbtree_node__get_value(cex,
                                                             next_rbtree_node);
                assert(l);
                /* Get the last node (i.e., the oldest node) in the
                 * list.  We do this because the younger nodes are
                 * considered to be greater. */
                next_list_node = cgul_list__get_back(cex, l);
            }
        }

    }

    /* Get the multimap node from the list node. */
    if (next_list_node) {
        rv = (cgul_multimap_node_t)cgul_list_node__get_value(cex,
                                                             next_list_node);
    }

 out:
    return rv;
}

cgul_multimap_node_t
cgul_multimap_node__get_older(cgul_exception_t* cex,
                              cgul_multimap_node_t n)
{
    cgul_multimap_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->older;

 out:
    return rv;
}

cgul_multimap_node_t
cgul_multimap_node__get_younger(cgul_exception_t* cex,
                                cgul_multimap_node_t n)
{
    cgul_multimap_node_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(n);

    rv = n->younger;

 out:
    return rv;
}
