#include "cgul_random_bbs_p.h"
#include "cgul_libc.h"
#include "cgul_random_lcg256.h"

/**
 * The number of "active" bits in <code>r->curr</code>.  They are
 * active in the sense that they are returned to the user.  Note that
 * <code>r->curr</code> (and <code>r->m</code>) actually have more
 * bits, but only the least-significant
 * <code>CGUL_RANDOM_BBS__BITS</code> of <code>r->curr</code> will be
 * returned to the user per iteration. <p>
 *
 * The security of the BBS generator depends on this value being
 * small.  It seems that the most bits that can be generated per
 * iteration is O(log log m).  Typically, just the least significant
 * bit is used. <p>
 */
#define CGUL_RANDOM_BBS__BITS 1

/**
 * Initialize all values of <code>r</code> except the seed value.
 *
 * @param[in] cex c-style exception
 * @param[in] r <code>cgul_random_bbs</code> instance
 */
static void
cgul_random_bbs__initialize_all_except_seed(cgul_exception_t* cex,
                                            cgul_random_bbs_t r)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* tmp is used to hold temporary results. */
    r->tmp = cgul_big_integer__new(cex);

    /* m is the modulus in the Blum-Blum-Shum formula
     *
     *     next = (curr * curr) % m
     *
     * It is the product of two special primes p and q where both p
     * and q have to have the following properties (where tmp1 and
     * tmp2 are also prime):
     *
     *     1)    p = 2 * tmp1 + 1
     *        tmp1 = 2 * tmp2 + 1
     *
     *     2) p congruent to 3 mod 4
     *
     * You can use the "calc" script in ~/src/calc/blum_blum_shub to
     * generate the special prime factors p and q.  To generate
     * different primes, you'll need to change the seed values for the
     * script as explained in the script.
     *
     * NOTE: r->m is not prime; it is the product of p and q.
     */
    r->m = cgul_big_integer__new_from_string(
               cex,
               "0xe5ed43b0761da5f1ed037643ef5b4bde5817af3e7a6fb22b7695333e760147cca7aea041");

    /* For many operations, we need to know the exclusive, upper bound
     * on the pseudo-random numbers returned by this class. */
    r->max = cgul_big_integer__new_from_long(cex, 1);
    cgul_big_integer__shift_left(cex, r->max, r->max, CGUL_RANDOM_BBS__BITS);

    /* Mask used to limit the number of bits in r->curr that are used.
     * This is just (r->mask = r->max - 1) which sets all the "active"
     * bits to one. */
    r->mask = cgul_big_integer__new_from_long(cex, 1);
    cgul_big_integer__subtract(cex, r->mask, r->max, r->mask);

    /* Calculate m^2 so we know how large to scale m_reciprocal. */
    cgul_big_integer__multiply(cex, r->tmp, r->m, r->m);

    /* Calculate the reciprocal of m. */
    r->m_reciprocal = cgul_big_integer__new(cex);
    r->m_scale = cgul_big_integer__get_inversion_scale(cex, r->tmp, r->m);
    cgul_big_integer__invert(cex, r->m_reciprocal, r->m, r->m_scale);

    /* Calculate max^2 so we know how large to scale max_reciprocal. */
    cgul_big_integer__multiply(cex, r->tmp, r->max, r->max);

    /* Calculate the reciprocal of max. */
    r->max_reciprocal = cgul_big_integer__new(cex);
    r->max_scale = cgul_big_integer__get_inversion_scale(cex, r->tmp, r->max);
    cgul_big_integer__invert(cex, r->max_reciprocal, r->max, r->max_scale);

 out:
    return;
}

/**
 * One of the characteristics of the Blum-Blum-Shub generator is that
 * it is possible to have very short cycle lengths.  Some nice
 * mathematicians have gone to the trouble to prescribe a method for
 * selecting <code>r->m</code> such that the cycle lengths should be
 * either sufficient or degenerate.  This method reseeds
 * <code>r</code> until it finds a non-degenerate cycle.  The vast
 * majority of cycles are supposed to be non-degenerate. <p>
 *
 * This method uses <code>cgul_random_lcg</code> which is a linear
 * congruential generator to find an acceptable seed value. </code>
 *
 * @param[in,out] cex c-style exception
 * @param[in] r <code>cgul_random_bbs</code> instance
 */
static void
cgul_random_bbs__reseed(cgul_exception_t* cex,
                        cgul_random_bbs_t r)
{
    cgul_random_lcg256_t lcg = NULL;
    cgul_big_integer_t prev = NULL;
    cgul_big_integer_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);

    /* Allocate. */
    lcg = cgul_random_lcg256__new_from_seed(cex, r->curr);
    prev = cgul_big_integer__new(cex);
    tmp = cgul_big_integer__new(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate until a non-degenerate cycle is found. */
    for (;;) {

        /* Reseed the BBS generator using the next iteration of the
         * LCG generator. */
        cgul_random_lcg256__next_in_range(cex, lcg, r->curr, r->m);

        /* Have the BBS generator iterate twice discarding the bits
         * that are generated but remember "curr". */
        cgul_random_bbs__next(cex, r, tmp);
        cgul_big_integer__assign(cex, prev, r->curr);
        cgul_random_bbs__next(cex, r, tmp);

        /* Check for errors. */
        if (*cex) {
            goto out;
        }

        /* If the previous "curr" and the current "curr" are not
         * equal, the current cycle is not degenerate so it should be
         * sufficient. */
        if (cgul_big_integer__compare(cex, prev, r->curr) != 0) {
            break;
        }

    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_big_integer__delete(tmp);
    }
    if (prev) {
        cgul_big_integer__delete(prev);
    }
    if (lcg) {
        cgul_random_lcg256__delete(lcg);
    }

    return;
}

cgul_random_bbs_t
cgul_random_bbs__new(cgul_exception_t* cex)
{
    cgul_random_bbs_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_random_bbs_t)cgul_libc__malloc(cex, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_random_bbs__initialize_all_except_seed(cex, rv);

    /* Seed with the time (in seconds).  It's a common thing to do.  I
     * would like better resolution than just one second though. */
    rv->curr = cgul_big_integer__new_from_long(cex, (long)time(NULL));

    /* Reseed to make sure rv is on a sufficiently long cycle. */
    cgul_random_bbs__reseed(cex, rv);

 out:

    /* Clean up any partially initialized objects. */
    if (*cex && rv) {
        cgul_random_bbs__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_random_bbs_t
cgul_random_bbs__new_from_seed(cgul_exception_t* cex,
                               cgul_big_integer_t seed)
{
    cgul_random_bbs_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_random_bbs_t)cgul_libc__malloc(cex, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_random_bbs__initialize_all_except_seed(cex, rv);
    rv->curr = cgul_big_integer__new(cex);
    cgul_big_integer__assign(cex, rv->curr, seed);

    /* Reseed to make sure rv is on a sufficiently long cycle. */
    cgul_random_bbs__reseed(cex, rv);

 out:

    /* Clean up any partially initialized objects. */
    if (*cex && rv) {
        cgul_random_bbs__delete(rv);
        rv = NULL;
    }

    return rv;
}


void
cgul_random_bbs__delete(cgul_random_bbs_t r)
{
    if (r) {
        if (r->curr) {
            cgul_big_integer__delete(r->curr);
        }
        if (r->tmp) {
            cgul_big_integer__delete(r->tmp);
        }
        if (r->max_reciprocal) {
            cgul_big_integer__delete(r->max_reciprocal);
        }
        if (r->max) {
            cgul_big_integer__delete(r->max);
        }
        if (r->mask) {
            cgul_big_integer__delete(r->mask);
        }
        if (r->m_reciprocal) {
            cgul_big_integer__delete(r->m_reciprocal);
        }
        if (r->m) {
            cgul_big_integer__delete(r->m);
        }
        free(r);
    }
}

void
cgul_random_bbs__next(cgul_exception_t* cex,
                      cgul_random_bbs_t r,
                      cgul_big_integer_t next)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);
    assert(next);

    /*
     * The Blum-Blum-Shub iteration formula is
     *
     *     next = (curr * curr) % m
     *
     * but you have to use r->mask to limit the number of bits
     * returned because r->m is not a power of 2.  There are theorems
     * that say this is ok so long as you take the least-significant
     * bits and do not exceed one less than the number of significant
     * bits in m.
     */

    cgul_big_integer__multiply(cex, r->curr, r->curr, r->curr);
    cgul_big_integer__divide_newton(cex,
                                    r->tmp,
                                    r->curr,
                                    r->curr,
                                    r->m,
                                    r->m_reciprocal,
                                    r->m_scale);
    cgul_big_integer__and(cex, next, r->curr, r->mask);

 out:
    return;
}

void
cgul_random_bbs__next_in_range(cgul_exception_t* cex,
                               cgul_random_bbs_t r,
                               cgul_big_integer_t next,
                               cgul_big_integer_t user_range_max)
{
    cgul_big_integer_t prng_range_max = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);
    assert(next);
    assert(user_range_max);

    /* Set "next" equal to the next pseudo-random number. */
    cgul_random_bbs__next(cex, r, next);
    if (*cex) {
        goto out;
    }

    /* Test if the user wants an even larger pseudo-random number. */
    if (cgul_big_integer__compare(cex, user_range_max, r->max) > 0) {

        /* Allocate. */
        prng_range_max = cgul_big_integer__new(cex);
        if (*cex) {
            goto out;
        }

        /* r->max is only accurate if the PRNG is iterated once, but
         * we are about to iterate more than once.  So it is important
         * to keep track of the new maximum. */
        cgul_big_integer__assign(cex, prng_range_max, r->max);
        if (*cex) {
            goto out;
        }

        /* While user_range_max is greater than prng_range_max, keep
         * generating more bits.  It is ok if we generate more bits
         * than requested by user_range_max because they will be
         * mapped to that range (below). */
        while (cgul_big_integer__compare(cex,
                                         user_range_max,
                                         prng_range_max) > 0)
        {
            /* Get the next set of bits, but put them in r->tmp. */
            cgul_random_bbs__next(cex, r, r->tmp);

            /* Shift "next" to make room for the next set of bits. */
            cgul_big_integer__shift_left(cex,
                                         next,
                                         next,
                                         CGUL_RANDOM_BBS__BITS);

            /* Add the bits in r->tmp to "next". */
            cgul_big_integer__xor(cex, next, next, r->tmp);

            /* Keep track of the new maximum range. */
            cgul_big_integer__shift_left(cex,
                                         prng_range_max,
                                         prng_range_max,
                                         CGUL_RANDOM_BBS__BITS);
            if (*cex) {
                goto out;
            }
        }
    }

    /*
     * Map "next" to the range [0 - user_range_max).  The if() selects
     * between using r->max and prng_range_max when performing the
     * mapping.  The difference is that r->max is accurate for the
     * common case where user_range_max is less than r->max;
     * otherwise, prng_range_max is accurate.
     *
     * Typically, this would be something like the following:
     *
     *                        next
     *     user_range_max * -------
     *                       r->max
     *
     * The problem with the equation above is that (next / r->max) is
     * less than one which is in a range that cgul_big_integer cannot
     * handle; instead, we just change the order of operations so that
     * we are always dealing with integer values:
     *
     *     user_range_max * next
     *     ---------------------
     *             r->max
     *
     * The code below might benefit from just returning (next %
     * user_range_max), but the typical way of returning a value in a
     * range is inherited from LCG which does not have random
     * low-order bits which heavily influences the result of the mod
     * operation hence the use of the scaling operation.  This
     * shouldn't be a problem with Blum-Blum-Shum, but I'm sticking
     * with what I think works in case there is some other reason for
     * doing it like this of which I'm not aware.
     */
    if (prng_range_max) {
        cgul_big_integer__multiply(cex, next, next, user_range_max);
        cgul_big_integer__divide(cex, next, r->tmp, next, prng_range_max);
    } else {
        cgul_big_integer__multiply(cex, next, next, user_range_max);
        cgul_big_integer__divide_newton(cex,
                                        next,
                                        r->tmp,
                                        next,
                                        r->max,
                                        r->max_reciprocal,
                                        r->max_scale);
    }

 out:

    /* Clean up. */
    if (prng_range_max) {
        cgul_big_integer__delete(prng_range_max);
    }

    return;
}
