#include "cgul_random_lcg256_p.h"
#include "cgul_libc.h"

/**
 * The maximum number of bits that can be generated in one iteration.
 */
#define CGUL_RANDOM_LCG256__BITS 256

/**
 * Initialize all values of <code>r</code> except the seed value.
 *
 * @param[in] cex c-style exception
 * @param[in] r <code>cgul_random_lcg256</code> instance
 */
static void
cgul_random_lcg256__initialize_all_except_seed(cgul_exception_t* cex,
                                               cgul_random_lcg256_t r)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    /* tmp is used to hold temporary results. */
    r->tmp = cgul_big_integer__new(cex);

    /* one is a constant. */
    r->one = cgul_big_integer__new_from_long(cex, 1L);

    /* b is one digit smaller than m and ends in x21, where x is even.
     * Note that "x21, where x is even" is describing a number given
     * in decimal format, not in hexadecimal format.  According to
     * other sources, the main requirements for b are as follows:
     *
     *     1) "b - 1" is divisible by all prime factors of m.
     *     2) "b - 1" is a multiple of 4 if m is a multiple of 4.
     *
     * It should be easy to determine whether the requirements listed
     * above hold because m is chosen below to be an exact power of 2
     * meanings its prime factors are only 2. */
    r->b = cgul_big_integer__new_from_string(
               cex,
               "93334463374286093114556012971841503630383314309757101832787784093259065030821");

    /* m should be large.  Here it is one larger than the maximum
     * value that can be stored in CGUL_RANDOM_LCG256__BITS bits. */
    r->m = cgul_big_integer__new_from_long(cex, 1);
    cgul_big_integer__shift_left(cex, r->m, r->m, CGUL_RANDOM_LCG256__BITS);

    /* The mask is just r->m with all significant bits set to one. */
    r->mask = cgul_big_integer__new(cex);
    cgul_big_integer__assign(cex, r->mask, r->m);
    cgul_big_integer__subtract(cex, r->mask, r->mask, r->one);

    /* Calculate m^2 so we know how large to scale m_reciprocal. */
    cgul_big_integer__multiply(cex, r->tmp, r->m, r->m);

    /* Calculate the reciprocal of m. */
    r->m_reciprocal = cgul_big_integer__new(cex);
    r->m_scale = cgul_big_integer__get_inversion_scale(cex, r->tmp, r->m);
    cgul_big_integer__invert(cex, r->m_reciprocal, r->m, r->m_scale);

 out:
    return;
}

cgul_random_lcg256_t
cgul_random_lcg256__new(cgul_exception_t* cex)
{
    cgul_random_lcg256_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_random_lcg256_t)cgul_libc__malloc(cex, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_random_lcg256__initialize_all_except_seed(cex, rv);

    /* Seed with the time (in seconds).  It's a common thing to do.  I
     * would like better resolution than just one second though. */
    rv->curr = cgul_big_integer__new_from_long(cex, (long)time(NULL));

 out:

    /* Clean up any partially initialized objects. */
    if (*cex && rv) {
        cgul_random_lcg256__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_random_lcg256_t
cgul_random_lcg256__new_from_seed(cgul_exception_t* cex,
                                  cgul_big_integer_t seed)
{
    cgul_random_lcg256_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_random_lcg256_t)cgul_libc__malloc(cex, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_random_lcg256__initialize_all_except_seed(cex, rv);
    rv->curr = cgul_big_integer__new(cex);
    cgul_big_integer__assign(cex, rv->curr, seed);

 out:

    /* Clean up any partially initialized objects. */
    if (*cex && rv) {
        cgul_random_lcg256__delete(rv);
        rv = NULL;
    }

    return rv;
}


void
cgul_random_lcg256__delete(cgul_random_lcg256_t r)
{
    if (r) {
        if (r->tmp) {
            cgul_big_integer__delete(r->tmp);
        }
        if (r->one) {
            cgul_big_integer__delete(r->one);
        }
        if (r->m_reciprocal) {
            cgul_big_integer__delete(r->m_reciprocal);
        }
        if (r->mask) {
            cgul_big_integer__delete(r->mask);
        }
        if (r->m) {
            cgul_big_integer__delete(r->m);
        }
        if (r->b) {
            cgul_big_integer__delete(r->b);
        }
        if (r->curr) {
            cgul_big_integer__delete(r->curr);
        }
        free(r);
    }
}

void
cgul_random_lcg256__next(cgul_exception_t* cex,
                         cgul_random_lcg256_t r,
                         cgul_big_integer_t next)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);
    assert(next);

    /* This algorithm is the "Linear Congruential Method" described in
     * "Algorithms in C++" by Robert Sedgewick.  Knuth also has a
     * discussion of this algorithm.
     *
     * Because r->m is chosen to be a power of 2, there is no need to
     * do a full division in order to determine mod r->m.  Instead,
     * just mask off the high-order bits. */
    cgul_big_integer__multiply(cex, r->curr, r->curr, r->b);
    cgul_big_integer__add(cex, r->curr, r->curr, r->one);
#if 1
    cgul_big_integer__and(cex, r->curr, r->curr, r->mask);
#else
    cgul_big_integer__divide_newton(cex,
                                    r->tmp,
                                    r->curr,
                                    r->curr,
                                    r->m,
                                    r->m_reciprocal,
                                    r->m_scale);
#endif
    cgul_big_integer__assign(cex, next, r->curr);

 out:
    return;
}

void
cgul_random_lcg256__next_in_range(cgul_exception_t* cex,
                                  cgul_random_lcg256_t r,
                                  cgul_big_integer_t next,
                                  cgul_big_integer_t user_range_max)
{
    cgul_big_integer_t prng_range_max = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);
    assert(next);
    assert(user_range_max);

    /* Set "next" equal to the next pseudo-random number. */
    cgul_random_lcg256__next(cex, r, next);
    if (*cex) {
        goto out;
    }

    /* Test if the user wants an even larger pseudo-random number. */
    if (cgul_big_integer__compare(cex, user_range_max, r->m) > 0) {

        /* r->m is only accurate if the PRNG is iterated once, but
         * we are about to iterate more than once.  So it is important
         * to keep track of the new maximum. */
        prng_range_max = cgul_big_integer__new_from_long(cex, 1);
        if (*cex) {
            goto out;
        }

        /* For LCG, do not blindly add more bits from each new
         * iteration because the low-order bits are not random.
         * Instead, shift out the lower half of the bits.  (See the
         * doxygen comments for cgul_random_lcg256__next().) */
        cgul_big_integer__shift_right(
            cex, next, next, CGUL_RANDOM_LCG256__BITS >> 1);
        if (*cex) {
            goto out;
        }

        /* Keep track of the new maximum range. */
        cgul_big_integer__shift_left(
            cex, prng_range_max, prng_range_max, CGUL_RANDOM_LCG256__BITS >> 1);
        if (*cex) {
            goto out;
        }

        /* While user_range_max is greater than prng_range_max, keep
         * generating more bits.  It is ok if we generate more bits
         * than requested by user_range_max because they will be
         * mapped to that range (below). */
        while (cgul_big_integer__compare(cex,
                                         user_range_max,
                                         prng_range_max) > 0)
        {
            /* Get the next set of bits, but put them in r->tmp. */
            cgul_random_lcg256__next(cex, r, r->tmp);

            /* Discard the non-random low-order bits. (Again, see the
             * doxygen comments for cgul_random_lcg256__next().) */
            cgul_big_integer__shift_right(
                cex, r->tmp, r->tmp, CGUL_RANDOM_LCG256__BITS >> 1);

            /* Shift "next" to make room for the next set of bits. */
            cgul_big_integer__shift_left(
                cex, next, next, CGUL_RANDOM_LCG256__BITS >> 1);

            /* Add the bits in r->tmp to "next". */
            cgul_big_integer__xor(cex, next, next, r->tmp);

            /* Keep track of the new maximum range. */
            cgul_big_integer__shift_left(cex,
                                         prng_range_max,
                                         prng_range_max,
                                         CGUL_RANDOM_LCG256__BITS >> 1);
            if (*cex) {
                goto out;
            }
        }
    }

    /*
     * Map "next" to the range [0 - user_range_max).  The if() selects
     * between using r->m and prng_range_max when performing the
     * mapping.  The difference is that r->m is accurate for the
     * common case where user_range_max is less than r->m;
     * otherwise, prng_range_max is accurate.
     *
     * The reason for this is that LCGs that are based on a power of 2
     * do not have random low-order bits so you cannot just do (next %
     * user_range_max) because this only uses the low-order bits.  The
     * solution is to convert "next" to a percentage of r->m and then
     * scale user_range_max based on the percentage.
     *
     * Typically, this would be something like the following:
     *
     *                       next
     *     user_range_max * ------
     *                       r->m
     *
     * The problem with the equation above is that (next / r->m) is
     * less than one which is in a range that cgul_big_integer cannot
     * handle; instead, we just change the order of operations so that
     * we are always dealing with integer values:
     *
     *     user_range_max * next
     *     ---------------------
     *              r->m
     *
     * Note that using a prime number for r->m would randomize the
     * lower order bits which would save a multiplication here but
     * would add a division to cgul_big_integer__next() because it
     * would no longer be possible to just mask off the high order
     * bits.
     */
    if (prng_range_max) {
        cgul_big_integer__multiply(cex, next, next, user_range_max);
        cgul_big_integer__divide(cex, next, r->tmp, next, prng_range_max);
    } else {
        cgul_big_integer__multiply(cex, next, next, user_range_max);
        cgul_big_integer__divide_newton(cex,
                                        next,
                                        r->tmp,
                                        next,
                                        r->m,
                                        r->m_reciprocal,
                                        r->m_scale);
    }

 out:

    /* Clean up. */
    if (prng_range_max) {
        cgul_big_integer__delete(prng_range_max);
    }

    return;
}
