#include "cgul_random_lfg32_p.h"
#include "cgul_libc.h"
#include "cgul_random_lcg32.h"

/*
 * This file implements a Lagged Fibonacci Generator (LFG) using the
 * primitive polynomial described by the pair (521, 353).
 */

/**
 * The size of the vector that will hold the PRNs until it is their
 * turn to be combined.  The size of the vector cannot be changed
 * without also changing <code>cgul_random_lfg32__lag</code>.
 */
static size_t cgul_random_lfg32__v_size = 521;

/**
 * The second lag parameter.  The first lag parameter is the size of
 * the array, <code>cgul_random_lfg32__v_size</code>.  Both values
 * must be changed in concert.
 */
static size_t cgul_random_lfg32__lag = 353;

/**
 * This is the odd constant in the Weyl Generator often denoted as omega.
 */
static cgul_uint32_t cgul_random_lcg32__omega = 2654435769U;

cgul_random_lfg32_t
cgul_random_lfg32__new(cgul_exception_t* cex)
{
    cgul_uint32_t seed = 0;
    cgul_random_lfg32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a new object seeded with time(). */
    seed = (cgul_uint32_t)time(NULL);
    rv = cgul_random_lfg32__new_from_seed(cex, seed);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

cgul_random_lfg32_t
cgul_random_lfg32__new_from_seed(cgul_exception_t* cex,
                                 cgul_uint32_t seed)
{
    size_t i = 0;
    int has_odd_seed = 0;
    unsigned int seed_count = 0;
    cgul_uint32_t* seeds = NULL;
    cgul_random_lcg32_t lcg32 = NULL;
    cgul_random_lfg32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the temporary cgul_random_lcg32 instance that will be
     * used to seed the circular array. */
    lcg32 = cgul_random_lcg32__new_from_seed(cex, seed);
    if (*cex) {
        goto out;
    }

    /* Allocate the seed array and get its size. */
    cgul_random_lfg32__allocate_seed_array(cex, &seeds, &seed_count);
    if (*cex) {
        goto out;
    }

    /* Initialize the seed array.  At least one of the seed values
     * must be odd. */
    while (!has_odd_seed) {
        for (i = 0 ; i < seed_count ; ++i) {
            seeds[i] = cgul_random_lcg32__next(cex, lcg32);
            if (seeds[i] & 0x1) {
                has_odd_seed = 1;
            }
        }
    }

    /* Allocate the new object from the seed array. */
    rv = cgul_random_lfg32__new_from_seed_array(cex, seeds);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (seeds) {
        free(seeds);
    }
    if (lcg32) {
        cgul_random_lcg32__delete(lcg32);
    }

    return rv;
}

cgul_random_lfg32_t
cgul_random_lfg32__new_from_seed_array(cgul_exception_t* cex,
                                       cgul_uint32_t* seeds)
{
    size_t i = 0;
    int has_odd_seed = 0;
    cgul_random_lfg32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(seeds);

    /* Allocate the object. */
    rv = (cgul_random_lfg32_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Allocate the vector that will hold the PRNs until they are combined. */
    rv->v = (cgul_uint32_t*)cgul_libc__calloc(cex,
                                              cgul_random_lfg32__v_size,
                                              sizeof(*rv->v));
    if (*cex) {
        goto out;
    }

    /* Copy the caller's seed values into our circular array.  At
     * least one of the seed values must be odd.  Note that the
     * comments for this method recommend that the caller use
     * allocate_seed_array() to allocate "seeds" so it should be the
     * correct size. */
    for (i = 0 ; i < cgul_random_lfg32__v_size ; ++i) {
        rv->v[i] = seeds[i];
        if (rv->v[i] & 0x1) {
            has_odd_seed = 1;
        }
    }

    /* Throw an exception if at least one seed was not odd. */
    if (!has_odd_seed) {
        *cex = cgul_exception__new("cgul_random_lfg32__new_from_seed_array: "
                                   "At least one seed must be odd.");
        goto out;
    }

    /* Initialize the index into v to indicate the buffer is exhausted. */
    rv->v_index = cgul_random_lfg32__v_size;

    /* Seed the Weyl Generator. */
    rv->weyl = 166947538U;

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        cgul_random_lfg32__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_random_lfg32__delete(cgul_random_lfg32_t r)
{
    if (r) {

        /* v */
        if (r->v) {
            free(r->v);
        }

        /* Free the object itself. */
        free(r);
    }
}

void
cgul_random_lfg32__allocate_seed_array(cgul_exception_t* cex,
                                       cgul_uint32_t** v,
                                       unsigned int* v_count)
{
    size_t count = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v_count);

    /* Number of elements in the array. */
    count = cgul_random_lfg32__v_size;

    /* Allocate the seed array. */
    if (v) {
        *v = (cgul_uint32_t*)cgul_libc__calloc(cex, count, sizeof(**v));
        if (*cex) {
            goto out;
        }
    }

    /* Return the count. */
    if (v_count) {
        *v_count = count;
    }

 out:
    return;
}

cgul_uint32_t
cgul_random_lfg32__next(cgul_exception_t* cex,
                        cgul_random_lfg32_t r)
{
    cgul_uint32_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);

    /* Refill the buffer if it is exhausted. */
    if (r->v_index >= cgul_random_lfg32__v_size) {
        unsigned int i = 0;
        unsigned int j = cgul_random_lfg32__v_size
                         - cgul_random_lfg32__lag;

        /* Iterate until (i + cgul_random_lfg32__lag) wraps. */
        for ( ; j < cgul_random_lfg32__v_size ; ++i, ++j) {
            r->v[i] += r->v[j];
        }

        /* Wrap j and iterate until i reaches the end of the buffer. */
        for (j = 0 ; i < cgul_random_lfg32__v_size ; ++i, ++j) {
            r->v[i] += r->v[j];
        }

        /* Reset the index. */
        r->v_index = 0;
    }

    /* Iterate the Weyl generator skipping (r->weyl == 0) so that the
     * Weyl Generator's period will be (2^32 - 1) which will be
     * relatively prime with the LFG's period of (2^552 - 2^31). */
    while ((r->weyl += cgul_random_lcg32__omega) == 0) {
        /* empty */
    }

    /* Combine the next PRN with the Weyl Generator. */
    rv = r->v[r->v_index++] + r->weyl;

 out:
    return rv;
}

cgul_uint32_t
cgul_random_lfg32__next_in_range(cgul_exception_t* cex,
                                 cgul_random_lfg32_t r,
                                 cgul_uint32_t range_max)
{
    cgul_uint32_t rv = 0;
    cgul_uint32_t upper_limit = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (range_max == 0) {
        *cex = cgul_exception__new("cgul_random_lfg32__next_in_range: "
                                   "exclusive range_max is zero");
        goto out;
    }

    /* Use integer arithmetic to find the largest multiple of
     * range_max that is less than the maximum random value that could
     * be produced by the PRNG. */
    upper_limit = 0xffffffff - (0xffffffff % range_max);

    /* Iterate the PRNG until it returns a raw value that is less than
     * the largest multiple of range_max.  In effect, we are just
     * discarding the upper few PRNs so that the mod operation is
     * fair.
     *
     * Note that upper_limit is an _exclusive_ upper bound because if
     * it were included (upper_limit % range) would necessarily be
     * zero (because "upper_limit" is defined to be a multiple of
     * "range") which would cause there to be one more potential zero
     * than any other value that could be returned.  This is why
     * "greater than or equal to" is used rather than just "greater
     * than." */
    do {
        rv = cgul_random_lfg32__next(cex, r);
    } while (rv >= upper_limit);

    /* Return a PRN in the user's range that is bias-free. */
    rv %= range_max;

 out:
    return rv;
}

double
cgul_random_lfg32__next_in_unit(cgul_exception_t* cex,
                                cgul_random_lfg32_t r)
{
    double rv = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the next PRN and divide by 2^32 by multiplying by the
     * precomputed reciprocal of 2^32. */
    rv = cgul_random_lfg32__next(cex, r) * CGUL_INVERSE_2_32;

 out:
    return rv;
}
