#include "cgul_random_well32_p.h"
#include "cgul_libc.h"
#include "cgul_random_lcg32.h"

/*
 * This file implements WELL44497b which is documented in the
 * following paper and errata:
 *
 *     http://www.iro.umontreal.ca/~lecuyer/myftp/papers/wellrng.pdf
 *     http://www.iro.umontreal.ca/~lecuyer/myftp/papers/wellrng-errata.txt
 */

/**
 * The size of the vector that will hold the state of the WELL generator.
 */
static size_t cgul_random_well32__v_size = 1391;

cgul_random_well32_t
cgul_random_well32__new(cgul_exception_t* cex)
{
    cgul_uint32_t seed = 0;
    cgul_random_well32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Create a new object seeded with time(). */
    seed = (cgul_uint32_t)time(NULL);
    rv = cgul_random_well32__new_from_seed(cex, seed);
    if (*cex) {
        goto out;
    }

 out:
    return rv;
}

cgul_random_well32_t
cgul_random_well32__new_from_seed(cgul_exception_t* cex,
                                  cgul_uint32_t seed)
{
    size_t i = 0;
    int has_nonzero_seed = 0;
    unsigned int seed_count = 0;
    cgul_uint32_t* seeds = NULL;
    cgul_random_lcg32_t lcg32 = NULL;
    cgul_random_well32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate the temporary cgul_random_lcg32 instance that will be
     * used to seed the circular array. */
    lcg32 = cgul_random_lcg32__new_from_seed(cex, seed);
    if (*cex) {
        goto out;
    }

    /* Allocate the seed array and get its size. */
    cgul_random_well32__allocate_seed_array(cex, &seeds, &seed_count);
    if (*cex) {
        goto out;
    }

    /* Initialize the seed array.  At least one of the seed values
     * must be nonzero. */
    while (!has_nonzero_seed) {
        for (i = 0 ; i < seed_count ; ++i) {
            seeds[i] = cgul_random_lcg32__next(cex, lcg32);
            if (seeds[i]) {
                has_nonzero_seed = 1;
            }
        }
    }

    /* Allocate the new object from the seed array. */
    rv = cgul_random_well32__new_from_seed_array(cex, seeds);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (seeds) {
        free(seeds);
    }
    if (lcg32) {
        cgul_random_lcg32__delete(lcg32);
    }

    return rv;
}

cgul_random_well32_t
cgul_random_well32__new_from_seed_array(cgul_exception_t* cex,
                                        cgul_uint32_t* seeds)
{
    size_t i = 0;
    int has_nonzero_seed = 0;
    cgul_random_well32_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(seeds);

    /* Allocate the object. */
    rv = (cgul_random_well32_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /* Allocate the vector that will hold the state. */
    rv->state = (cgul_uint32_t*)cgul_libc__calloc(cex,
                                                  cgul_random_well32__v_size,
                                                  sizeof(*rv->state));
    if (*cex) {
        goto out;
    }

    /* Allocate the vector that will hold the PRNs until they are returned. */
    rv->v = (cgul_uint32_t*)cgul_libc__calloc(cex,
                                              cgul_random_well32__v_size,
                                              sizeof(*rv->v));
    if (*cex) {
        goto out;
    }

    /* Copy the caller's seed values into our circular array.  At
     * least one of the seed values must be nonzero.  Note that the
     * comments for this method recommend that the caller use
     * allocate_seed_array() to allocate "seeds" so it should be the
     * correct size. */
    for (i = 0 ; i < cgul_random_well32__v_size ; ++i) {
        rv->state[i] = seeds[i];
        if (rv->state[i]) {
            has_nonzero_seed = 1;
        }
    }

    /* Throw an exception if all seeds are zero. */
    if (!has_nonzero_seed) {
        *cex = cgul_exception__new("cgul_random_well32__new_from_seed_array: "
                                   "At least one seed must be nonzero.");
        goto out;
    }

    /* Initialize the index into v to indicate the buffer is exhausted. */
    rv->v_index = -1;

    /* seeds[0] is the hypothetical PRN that would have been returned
     * to the user given the current "state" array.  It is needed to
     * prime the pump so that the the first real PRN can be
     * generated. */
    rv->v[0] = seeds[0];

 out:

    /* Clean up partially initialized objects. */
    if (*cex && rv) {
        cgul_random_well32__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_random_well32__delete(cgul_random_well32_t r)
{
    if (r) {

        /* v */
        if (r->v) {
            free(r->v);
        }

        /* state */
        if (r->state) {
            free(r->state);
        }

        /* Free the object itself. */
        free(r);
    }
}

void
cgul_random_well32__allocate_seed_array(cgul_exception_t* cex,
                                        cgul_uint32_t** v,
                                        unsigned int* v_count)
{
    size_t count = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);
    assert(v_count);

    /* Number of elements in the array. */
    count = cgul_random_well32__v_size;

    /* Allocate the seed array. */
    if (v) {
        *v = (cgul_uint32_t*)cgul_libc__calloc(cex, count, sizeof(**v));
        if (*cex) {
            goto out;
        }
    }

    /* Return the count. */
    if (v_count) {
        *v_count = count;
    }

 out:
    return;
}

cgul_uint32_t
cgul_random_well32__next(cgul_exception_t* cex,
                         cgul_random_well32_t r)
{
    cgul_uint32_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(r);

    /* Refill the buffer if it is exhausted. */
    if (r->v_index < 0) {
        int i = 0;
        int m1 = 23;
        int m2 = 481;
        int m3 = 229;
        cgul_uint32_t z0 = 0;
        cgul_uint32_t z1 = 0;
        cgul_uint32_t z2 = 0;
        cgul_uint32_t mask0 = 0xffff8000;  /* ~mask1 */
        cgul_uint32_t mask1 = 0x00007fff;  /* (p = 15) ==> lower 15 bits set */
        cgul_uint32_t last_xor[] = {0, 0xb729fcec};

        /*
         * NOTE #1: Unlike most other PRNGs, WELL generates two values
         * per iteration both of which have to be logically shifted in
         * into the state array.  One of the values propagates through
         * the state.  The other value is returned to the user and
         * used only in the next iteration.  Logically, the value that
         * is returned to the user is STATE[0] which does not
         * propagate.  The value in STATE[0] is simply overwritten on
         * each iteration.
         *
         * This works out ok when generating one PRN per call to
         * next(), but as luck would have it, when generating the PRNs
         * in batches using a circular array (the way this
         * implementation does it), the logical STATE[0] PRN (which
         * needs to be returned by the user) gets overwritten by the
         * second PRN that propagates.
         *
         * So, this implementation has two arrays.  One array is
         * called "state", and it holds the working state where the
         * second PRN propagates.  The other array is called "v" which
         * stores each value of STATE[0] in turn so that they can be
         * returned to the user.  Note that it is ok to completely
         * remove what is logically STATE[0] from the state because it
         * never propagates through the state.  So long as we can get
         * the values for STATE[0] out of the v array when the are
         * needed, everything works ok.
         */

        /*
         * NOTE #2: In the paper describing the WELL generator, the
         * output is in terms of a base-2 decimal expansion where the
         * newest PRN is logically shifted in at STATE[0] and oldest
         * PRN is shifted out at STATE[R - 1] (where R is the size of
         * the state).  This means the index into our circular buffer
         * is _decremented_ after each PRN is generated.
         */

        /*
         * NOTE #3: In the paper describing the WELL generator, the
         * parameters for many differently sized generators are
         * provided.  In particular, each generator has three
         * different tap points called m1, m2, and m3 (same as in this
         * code).  The order of the tap points is not consistent for
         * all the generators; sometimes, m1 is larger than m2, and
         * sometimes it's not.  Etc.  This comment is just a warning
         * that efficiency requires the code below to take the
         * particular ordering of m1, m2, and m3 for the WELL44497b
         * generator into account.  If you are porting to a different
         * WELL generator, you'll need to make some (minor) changes.
         * You'll also need to change mask0 and mask1 if p is not 15
         * where p tells you how many bits to set in mask1.
         */

        /* Start with the oldest value in the state buffer and work
         * your way down.  Note that "i" always represents what the
         * paper calls "R - 1". */
        i = cgul_random_well32__v_size - 1;

        /* Handle the first iteration where at the start of the
         * iteration the new PRNs are at v[0] and v[1] respectively
         * and where at the end of the iteration the PRNs are at
         * v[cgul_random_well32__v_size - 1] and state[0]
         * respectively.  Note that (i = cgul_random_well32__v_size -
         * 1) here. */
        {
            z0 = (r->state[i] & mask0) ^ (r->state[i - 1] & mask1);

            z1 = r->v[0] ^ (r->v[0] << 24)
                 ^ r->state[m1] ^ (r->state[m1] >> 30);

            z2 = r->state[m2] ^ (r->state[m2] << 10) ^ (r->state[m3] << 26);

            r->state[0] = z1 ^ z2;

            r->v[i] = z0 ^ z1 ^ (z1 >> 20)
                      ^ (((z2 << 9) ^ (z2 >> 23)) & 0xfbffffff)
                      ^ last_xor[(z2 >> 17) & 0x1]
                      ^ r->state[0];
        }

        /* Iterate until m1 wraps (because m1 < m3 < m2 < i). */
        for (--i, --m1, --m2, --m3;
             m1 >= 0;
             --i, --m1, --m2, --m3)
        {
            z0 = (r->state[i] & mask0) ^ (r->state[i - 1] & mask1);

            z1 = r->v[i + 1] ^ (r->v[i + 1] << 24)
                 ^ r->state[m1] ^ (r->state[m1] >> 30);

            z2 = r->state[m2] ^ (r->state[m2] << 10) ^ (r->state[m3] << 26);

            r->state[i + 1] = z1 ^ z2;

            r->v[i] = z0 ^ z1 ^ (z1 >> 20)
                      ^ (((z2 << 9) ^ (z2 >> 23)) & 0xfbffffff)
                      ^ last_xor[(z2 >> 17) & 0x1]
                      ^ r->state[i + 1];
        }

        /* Wrap m1 and iterate until m3 wraps (because m3 < m2 < i). */
        for (m1 = cgul_random_well32__v_size - 1;
             m3 >= 0;
             --i, --m1, --m2, --m3)
        {
            z0 = (r->state[i] & mask0) ^ (r->state[i - 1] & mask1);

            z1 = r->v[i + 1] ^ (r->v[i + 1] << 24)
                 ^ r->state[m1] ^ (r->state[m1] >> 30);

            z2 = r->state[m2] ^ (r->state[m2] << 10) ^ (r->state[m3] << 26);

            r->state[i + 1] = z1 ^ z2;

            r->v[i] = z0 ^ z1 ^ (z1 >> 20)
                      ^ (((z2 << 9) ^ (z2 >> 23)) & 0xfbffffff)
                      ^ last_xor[(z2 >> 17) & 0x1]
                      ^ r->state[i + 1];
        }

        /* Wrap m3 and iterate until m2 wraps (because m2 < i). */
        for (m3 = cgul_random_well32__v_size - 1;
             m2 >= 0;
             --i, --m1, --m2, --m3)
        {
            z0 = (r->state[i] & mask0) ^ (r->state[i - 1] & mask1);

            z1 = r->v[i + 1] ^ (r->v[i + 1] << 24)
                 ^ r->state[m1] ^ (r->state[m1] >> 30);

            z2 = r->state[m2] ^ (r->state[m2] << 10) ^ (r->state[m3] << 26);

            r->state[i + 1] = z1 ^ z2;

            r->v[i] = z0 ^ z1 ^ (z1 >> 20)
                      ^ (((z2 << 9) ^ (z2 >> 23)) & 0xfbffffff)
                      ^ last_xor[(z2 >> 17) & 0x1]
                      ^ r->state[i + 1];
        }

        /* Wrap m2 and iterate until (i - 1) wraps. */
        for (m2 = cgul_random_well32__v_size - 1;
             i - 1 >= 0;
             --i, --m1, --m2, --m3)
        {
            z0 = (r->state[i] & mask0) ^ (r->state[i - 1] & mask1);

            z1 = r->v[i + 1] ^ (r->v[i + 1] << 24)
                 ^ r->state[m1] ^ (r->state[m1] >> 30);

            z2 = r->state[m2] ^ (r->state[m2] << 10) ^ (r->state[m3] << 26);

            r->state[i + 1] = z1 ^ z2;

            r->v[i] = z0 ^ z1 ^ (z1 >> 20)
                      ^ (((z2 << 9) ^ (z2 >> 23)) & 0xfbffffff)
                      ^ last_xor[(z2 >> 17) & 0x1]
                      ^ r->state[i + 1];
        }

        /* Wrap (i - 1) and perform the last iteration. */
        {
            z0 = (r->state[i] & mask0)
                 ^ (r->state[cgul_random_well32__v_size - 1] & mask1);

            z1 = r->v[i + 1] ^ (r->v[i + 1] << 24)
                 ^ r->state[m1] ^ (r->state[m1] >> 30);

            z2 = r->state[m2] ^ (r->state[m2] << 10) ^ (r->state[m3] << 26);

            r->state[i + 1] = z1 ^ z2;

            r->v[i] = z0 ^ z1 ^ (z1 >> 20)
                      ^ (((z2 << 9) ^ (z2 >> 23)) & 0xfbffffff)
                      ^ last_xor[(z2 >> 17) & 0x1]
                      ^ r->state[i + 1];
        }

        /* Reset the index. */
        r->v_index = cgul_random_well32__v_size - 1;
    }

    /* Return the next tempered PRN. */
    rv = r->v[r->v_index];
    rv ^= (rv << 7) & 0x93dd1400;
    rv ^= (rv << 15) & 0xfa118000;

    /* Decrement the index. */
    --r->v_index;

 out:
    return rv;
}

cgul_uint32_t
cgul_random_well32__next_in_range(cgul_exception_t* cex,
                                  cgul_random_well32_t r,
                                  cgul_uint32_t range_max)
{
    cgul_uint32_t rv = 0;
    cgul_uint32_t upper_limit = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Sanity check. */
    if (range_max == 0) {
        *cex = cgul_exception__new("cgul_random_well32__next_in_range: "
                                   "exclusive range_max is zero");
        goto out;
    }

    /* Use integer arithmetic to find the largest multiple of
     * range_max that is less than the maximum random value that could
     * be produced by the PRNG. */
    upper_limit = 0xffffffff - (0xffffffff % range_max);

    /* Iterate the PRNG until it returns a raw value that is less than
     * the largest multiple of range_max.  In effect, we are just
     * discarding the upper few PRNs so that the mod operation is
     * fair.
     *
     * Note that upper_limit is an _exclusive_ upper bound because if
     * it were included (upper_limit % range) would necessarily be
     * zero (because "upper_limit" is defined to be a multiple of
     * "range") which would cause there to be one more potential zero
     * than any other value that could be returned.  This is why
     * "greater than or equal to" is used rather than just "greater
     * than." */
    do {
        rv = cgul_random_well32__next(cex, r);
    } while (rv >= upper_limit);

    /* Return a PRN in the user's range that is bias-free. */
    rv %= range_max;

 out:
    return rv;
}

double
cgul_random_well32__next_in_unit(cgul_exception_t* cex,
                                 cgul_random_well32_t r)
{
    double rv = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get the next PRN and divide by 2^32 by multiplying by the
     * precomputed reciprocal of 2^32. */
    rv = cgul_random_well32__next(cex, r) * CGUL_INVERSE_2_32;

 out:
    return rv;
}
