#ifndef CGUL_RBTREE_NODE_P_H
#define CGUL_RBTREE_NODE_P_H

#include "cgul_common.h"
#include "cgul_rbtree_node.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_rbtree_node</code> class.
 *
 * @see cgul_rbtree_node_with_indexing
 */
struct cgul_rbtree_node {

    /** The heart of the red-black tree. */
    int red;

    /** Pointer to parent node. */
    cgul_rbtree_node_t parent;

    /** Pointer to left node. */
    cgul_rbtree_node_t left;

    /** Pointer to right node. */
    cgul_rbtree_node_t right;

    /**
     * Pointer to older node.  Together with <code>younger</code>,
     * <code>older</code> forms a doubly linked list that has nothing
     * to do with sort order; rather, the <code>older</code> node
     * points to the next older node.  If there is no older node,
     * <code>older</code> is <code>NULL</code>.  By calling
     * <code>cgul_rbtree_node__get_older()</code>, the user can
     * iterate over the nodes in reverse chronological order.
     */
    cgul_rbtree_node_t older;

    /**
     * Pointer to younger node.  Together with <code>older</code>,
     * <code>younger</code> forms a doubly linked list that has
     * nothing to do with sort order; rather, the <code>younger</code>
     * node points to the next younger node.  If there is no younger
     * node, <code>younger</code> is <code>NULL</code>.  By calling
     * <code>cgul_rbtree_node__get_younger()</code>, the user can
     * iterate over the nodes in chronological order.
     */
    cgul_rbtree_node_t younger;

    /** Key used to lookup the value. */
    const void* key;

    /** Value returned by a lookup based on a key. */
    const void* value;
};

/*
 * Private Methods
 */

/**
 * Profiling shows that determining if a 2-3-4 node is full should be
 * an optimized function.  We place it here so that cgul_rbtree can
 * get it inlined.  Unfortunately, MSVC++ does not seem to understand
 * the "inline" reserved word. <p>
 *
 * This #define just requires that n is not red, that n has two
 * children, and that n's two children are also red.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return whether this node is full
 */
#define cgul_rbtree_node__get_234_key_count_is_exactly_three(cex, n) \
    (((*cex) == NULL) \
    && (n) && !((n)->red) \
    && ((n)->left) && ((n)->left->red) \
    && ((n)->right) && ((n)->right->red))

/**
 * Create a new <code>cgul_rbtree_node</code> object.  All new nodes
 * are black by default and have a rank of 1.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_rbtree_node__delete()</code>.  If memory cannot be
 * allocated, <code>NULL</code> is returned, and an exception is
 * thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_rbtree_node</code> instance
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree_node__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  This does not
 * include the <code>key</code> or <code>value</code> pairs.  The
 * value for <code>n</code> should have been instantiated by
 * <code>cgul_rbtree_node__new()</code>.  Do not try to access
 * <code>n</code> after calling this method.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 */
extern CGUL_EXPORT void
cgul_rbtree_node__delete(cgul_rbtree_node_t n);

/**
 * This method is shared between <code>cgul_rbtree_node__new()</code>
 * and the <code>restructor</code> for the <code>cgul_cache()</code>.
 * It initializes new nodes and re-initializes existing nodes.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance to initialize
 */
extern CGUL_EXPORT void
cgul_rbtree_node__initialize(cgul_rbtree_node_t n);

/**
 * Set the key used to lookup this node.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @param[in] key key
 */
extern CGUL_EXPORT void
cgul_rbtree_node__set_key(cgul_exception_t* cex,
                          cgul_rbtree_node_t n,
                          const void* key);

/**
 * Return 1 if this node is red; otherwise, return 0.  If this node is
 * red, it means this node, its parent, and its sibling (if the
 * sibling exists) join together to form a larger 3-child or 4-child
 * node in the corresponding 2-3-4 tree.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return whether this node is red
 */
extern CGUL_EXPORT int
cgul_rbtree_node__is_red(cgul_exception_t* cex,
                         cgul_rbtree_node_t n);

/**
 * Return 1 if this node is black; otherwise, return 0.  If this node
 * is black and has no children, it is a 2-child node in the
 * corresponding 2-3-4 tree.  If this node is black and has one red
 * child, it is a 3-child node in the corresponding 2-3-4 tree.  If
 * this node is black and both children are red, it is a 4-child node
 * in the corresponding 2-3-4 tree.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return whether this node is black
 */
extern CGUL_EXPORT int
cgul_rbtree_node__is_black(cgul_exception_t* cex,
                           cgul_rbtree_node_t n);

/**
 * Return the smallest node in the subtree rooted at <code>n</code>.
 * This is in essence the node that is farthest to the left.  This
 * function should never return <code>NULL</code>.  In the event that
 * <code>n</code> has no left children, <code>n</code> itself is the
 * smallest node of the subtree.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return the smallest node in the subtree rooted at <code>n</code>
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree_node__get_smallest_node_in_subtree(cgul_exception_t* cex,
                                               cgul_rbtree_node_t n);

/**
 * Return the largest node in the subtree rooted at <code>n</code>.
 * This is in essence the node that is farthest to the right.  This
 * function should never return <code>NULL</code>.  In the event that
 * <code>n</code> has no right children, <code>n</code> itself is the
 * largest node of the subtree.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return the largest node in the subtree rooted at <code>n</code>
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree_node__get_largest_node_in_subtree(cgul_exception_t* cex,
                                              cgul_rbtree_node_t n);

/**
 * Return the left child node.  This method does NOT necessarily
 * return the same node as <code>cgul_rbtree_node__prev()</code>.  The
 * difference is that the left child is only the previous node if the
 * left child does not have a right child.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return the left child
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree_node__get_left_child(cgul_exception_t* cex,
                                 cgul_rbtree_node_t n);

/**
 * Return the right child node.  This method does NOT necessarily
 * return the same node as <code>cgul_rbtree_node__next()</code>.  The
 * difference is that the right child is only the next node if the
 * right child does not have a left child.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return the right child
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree_node__get_right_child(cgul_exception_t* cex,
                                  cgul_rbtree_node_t n);

/* Return the sibling for this node.  The value returned has no
 * special relationship to the corresponding 2-3-4 tree.  Instead it
 * is just the sibling in the red-black tree.  If this is the root
 * node, <code>NULL</code> is returned because it obviously does not
 * have a parent so it cannot have a sibling.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return the sibling for this node
 */
extern CGUL_EXPORT cgul_rbtree_node_t
cgul_rbtree_node__get_sibling(cgul_exception_t* cex,
                              cgul_rbtree_node_t n);

/**
 * Return the number key/value pairs held by the corresponding node in
 * the 2-3-4 tree.  Profiling shows that this function is a bottleneck
 * so it is now a macro.  We use macros because MSVC++ does not handle
 * the "inline" keyword.
 *
 * @param[in] cex c-style exception
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @param[out] count number of key/value pairs held by the corresponding node in
 * the 2-3-4 tree.
 */
#define cgul_rbtree_node__get_234_key_count(cex, n, count) \
do { \
    cgul_rbtree_node_t ntest = NULL; \
\
    (count) = 0; \
\
    assert((cex)); \
    if (*(cex)) { \
        break; \
    } \
\
    /* Empty nodes hold no children. */ \
    if ((n) == NULL) { \
        break; \
    } \
\
    /* If n is red, use its black parent as our point of reference for \
     * the 2-3-4 node. */ \
    ntest = (n); \
    if ((n)->red) { \
        ntest = (n)->parent; \
        assert(ntest); \
    } \
\
    assert(!ntest->red);\
\
    /* If n is a black node with no red children, it only holds one \
     * key/value pair. */ \
    if (    ((ntest->left == NULL) || !ntest->left->red) \
         && ((ntest->right == NULL) || !ntest->right->red)) \
    { \
        (count) = 1; \
\
    /* If n is a black node and has both a left red child and a right \
     * red child. */ \
    } else if (ntest->left  && ntest->left->red \
            && ntest->right && ntest->right->red) \
    { \
        (count) = 3; \
\
    /* Otherwise, the node must have exactly two key/value pairs. */ \
    } else { \
        (count) = 2; \
    } \
\
} while(0)

/**
 * Returns 1 if the corresponding 2-3-4 tree's node is leaf node.  You
 * can pass in any node from the red-black tree.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return whether this is a leaf node in the corresponding 2-3-4 tree
 */
extern CGUL_EXPORT int
cgul_rbtree_node__is_234_leaf_node(cgul_exception_t* cex,
                                   cgul_rbtree_node_t n);

/**
 * Return the black depth of the node n.  This only counts the number
 * of black nodes between n and the root.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return black depth
 */
extern CGUL_EXPORT int
cgul_rbtree_node__get_black_depth(cgul_exception_t* cex,
                                  cgul_rbtree_node_t n);

/**
 * Return the raw depth of the node n.  This counts every node
 * (including red nodes) between n and the root.
 *
 * @param[in] n <code>cgul_rbtree_node</code> instance
 * @return raw depth
 */
extern CGUL_EXPORT int
cgul_rbtree_node__get_raw_depth(cgul_exception_t* cex,
                                cgul_rbtree_node_t n);

CGUL_END_C

#endif  /*  CGUL_RBTREE_NODE_P_H  */
