#ifndef CGUL_RBTREE_NODE_WITH_INDEXING_H
#define CGUL_RBTREE_NODE_WITH_INDEXING_H

/**
 * This class implements a red-black tree node which holds one
 * key/value pair.  This class inherits from
 * <code>cguil_rbtree_node</code> extending it to include bookkeeping
 * values that allow the user to index into the
 * <code>cgul_rbtree</code> as if it were an array. <p>
 *
 * This class has no public interface per se, but because of the way
 * this class is structured in memory, you can pass instances of this
 * class to all the public methods for <code>cgul_rbtree_node</code>.
 * This is accomplished using the common C technique of splicing. <p>
 *
 * Your <code>cgul_rbtree</code> should never mix node types in the
 * same tree.  Each <code>cgul_rbtree</code> should have only
 * <code>cgul_rbtree_node</code> or only
 * <code>cgul_rbtree_node_with_indexing</code> nodes but not both. <p>
 *
 * @file
 * @author Paul Serice
 * @brief red-black tree node with indexing
 */

#include "cgul_common.h"
#include "cgul_rbtree_node_p.h"

CGUL_BEGIN_C

/**
 * Opaque pointer to a <code>cgul_rbtree_node_with_indexing</code>
 * instance.
 */
typedef struct cgul_rbtree_node_with_indexing*
               cgul_rbtree_node_with_indexing_t;

/**
 * Data encapsulated by the
 * <code>cgul_rbtree_node_with_indexing</code> class.  This class is
 * used by the <code>cgul_rbtree</code> class whenever the user
 * instantiates a tree with indexing enabled. <p>
 *
 * As you can see below from the <code>base</code> member, this
 * structure inherits from <code>cgul_rbtree_node</code>.  The reason
 * for the separate class defined by this file is to avoid wasting
 * space creating the <code>lcount</code> and <code>rcount</code>
 * members if the user is never going to use them.  Because these
 * extra members would otherwise be including in each node in the
 * tree, the space savings can be substantial. <p>
 */
struct cgul_rbtree_node_with_indexing {

    /**
     * The member data of the base class.  This is a mild form of
     * inheritance which is chosen just because it seems like the best
     * way to prevent the extra members below from consuming memory
     * when they aren't ever going to be used. <p>
     *
     * The usual splicing trick is what will be used to upcast. This
     * must be the first member of this class. <p>
     */
    struct cgul_rbtree_node base;

    /**
     * In order to search by position rather than key, we need to
     * track how many left and right children this node has.  This
     * variable holds the number of left children.
     */
    unsigned long int lcount;

    /**
     * In order to search by position rather than key, we need to
     * track how many left and right children this node has.  This
     * variable holds the number of right children.
     */
    unsigned long int rcount;
};

/*
 * Private Methods
 */

/**
 * Create a new <code>cgul_rbtree_node_with_indexing</code> object.
 * The new object will be initialized with <code>key</code> and
 * <code>value</code> both set to <code>NULL</code>.  All new nodes
 * are black by default and have a rank of 1.  The caller is
 * responsible for freeing the object by calling
 * <code>cgul_rbtree_node_with_indexing__delete()</code>.  If memory
 * cannot be allocated, <code>NULL</code> is returned and an exception
 * is thrown. <p>
 *
 * This class does not take ownership of <code>key</code> or
 * <code>value</code>.  The client must make sure the values pointed
 * to by <code>key</code> and <code>value</code> are not invalidated
 * during the life of the newly created instance.  If the client needs
 * to delete the <code>key</code> or the <code>value</code> or both,
 * the client should first call <code>cgul_rbtree__remove()</code> on
 * the <code>key</code>. <p>
 *
 * The nodes constructed by this method are suitable for inclusion in
 * <code>cgul_rbtree</code> objects constructed with the
 * <code>cgul_rbtree__new_with_indexing()</code> method. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>cgul_rbtree_node_with_indexing</code> instance
 */
extern CGUL_EXPORT cgul_rbtree_node_with_indexing_t
cgul_rbtree_node_with_indexing__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory.  This does not
 * include the <code>key</code> or <code>value</code> pairs.  The
 * value for <code>n</code> should have been instantiated by
 * <code>cgul_rbtree_node_with_indexing__new()</code>.  Do not try to
 * access <code>n</code> after calling this method.
 *
 * @param[in] n <code>cgul_rbtree_node_with_indexing</code> instance
 */
extern CGUL_EXPORT void
cgul_rbtree_node_with_indexing__delete(cgul_rbtree_node_with_indexing_t n);

/**
 * This method is shared between
 * <code>cgul_rbtree_node_with_indexing__new()</code> and the
 * <code>restructor</code> for the <code>cgul_cache()</code>.  It
 * initializes new nodes and re-initializes existing nodes.
 *
 * @param[in] n <code>cgul_rbtree_node_with_indexing</code> instance
 * to initialize
 */
extern CGUL_EXPORT void
cgul_rbtree_node_with_indexing__initialize(cgul_rbtree_node_with_indexing_t n);

CGUL_END_C

#endif  /*  CGUL_RBTREE_NODE_WITH_INDEXING_H  */
