#ifndef CGUL_STREAM_READER_P_H
#define CGUL_STREAM_READER_P_H

#include "cgul_common.h"
#include "cgul_cache.h"
#include "cgul_list.h"
#include "cgul_stream_reader.h"
#include "cgul_vector.h"
#include "cgul_wstring.h"

CGUL_BEGIN_C

/** Data encapsulated by the <code>cgul_stream_reader</code> class. */
struct cgul_stream_reader {

    /**
     * Decoder callback that is set by the client to configure this
     * class for decoding the characters from the block device.  For
     * example, this allows this one class to handle both ASCII and
     * UTF-8.
     */
    cgul_stream_reader__decode_t decode;

    /**
     * Current block reader.  The other block readers are on the
     * <code>block_readers</code> queue.
     */
    cgul_block_reader_t br;

    /**
     * Queue of pending block readers.  The block reader on the front
     * of the queue will be used after the current block reader
     * <code>br</code> is exhausted.
     */
    cgul_list_t block_readers;

    /**
     * Buffer used to hold (at most) one encoded block read from the
     * current block reader.
     */
    unsigned char* block_encoded;

    /**
     * UTF-32 character buffer used to hold one block of decoded characters.
     */
    cgul_wchar_t* block_decoded;

    /**
     * Size of <code>block_encoded</code> in bytes and
     * <code>block_decoded</code> in UTF-32 characters.
     */
    size_t block_size;

    /**
     * Index into <code>block_encoded</code> for the start of the next
     * character to decode.  Before using <code>index_encoded</code>
     * make sure it is less than <code>rcount_encoded</code>.
     */
    size_t index_encoded;

    /**
     * Actual number of bytes that were read from the current block
     * reader and are now actually valid in
     * <code>block_encoded</code>.
     */
    size_t rcount_encoded;

    /**
     * Index into <code>block_decoded</code> for the next character to
     * return to the client.  Before using <code>index_decoded</code>
     * make sure it is less than <code>rcount_decoded</code>.
     */
    size_t index_decoded;

    /**
     * Actual number of bytes that were read from the current block
     * reader and are now actually valid in
     * <code>block_decoded</code>.
     */
    size_t rcount_decoded;

    /**
     * Whether EOF has been reached.
     */
    int eof;

    /**
     * Whether line and column number tracking is enabled.  Disabling
     * should result in about a 10% performance increase.
     */
    int track_line_and_column;

    /**
     * Current line number.
     */
    size_t line_current;

    /**
     * Current column number.
     */
    size_t column_current;

    /**
     * Previous line number.
     */
    size_t line_previous;

    /**
     * Previous column number.
     */
    size_t column_previous;

    /**
     * Whether to normalize EOLs.
     */
    int normalize_eol;

    /**
     * The actual previous character that was read.  This is not
     * necessarily the same as the character returned to the user.  It
     * is used to track when '\r' characters are read.  They are
     * normalized to '\n' and then on the next pass, we need to skip
     * the following '\n' character (if any).
     */
    cgul_wchar_t actual_previous_char;

    /**
     * Presentation name.  If <code>NULL</code>, use the presentation
     * name for the first block reader (if any) instead.
     */
    char* presentation_name;

};

CGUL_END_C

#endif  /*  CGUL_STREAM_READER_P_H  */
