#include "cgul_string_p.h"
#include "cgul_convspec_p.h"
#include "cgul_char.h"
#include "cgul_defensive.h"
#include "cgul_libc.h"
#include "cgul_unicode.h"
#include "cgul_wchar.h"

/* The default capacity for each newly allocated string.  This is the
 * number of characters _not_ including the NUL terminator.  When
 * realloc() is called one is added to the capacity to make room for
 * the NUL terminator. */
static const size_t CGUL_STRING__DEFAULT_MINIMUM_CAPACITY = 128;

/* Array used to convert 4-bits to a hex character. */
static const char cgul_string__hex_map_lower[] = "0123456789abcdef";
static const char cgul_string__hex_map_upper[] = "0123456789ABCDEF";

/* Comments are in cgul_string_p.h. */
void
cgul_string__set_capacity(cgul_exception_t* cex,
                          cgul_string_t s,
                          size_t capacity)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Sanity check. */
    if (capacity == 0) {
        capacity = s->minimum_capacity;
    }

    /* Make sure the capacity is at least large enough to hold the
     * current number of elements plus 1 for the NUL terminator. */
    capacity = CGUL_MAX(capacity, s->length + 1);

#if 0
    fprintf(stderr, "cgul_string__set_capacity = %lu\n", capacity);
    fflush(stderr);
#endif

    /* realloc */
    s->value = (char*)cgul_libc__realloc(cex,
                                         s->value,
                                         sizeof(*s->value) * capacity);
    if (*cex) {
        goto out;
    }

    /* Rememeber the new capacity. */
    s->capacity = capacity;

    /* Clear the extra elements. */
    if (capacity > s->length) {
        memset(s->value + s->length,
               0,
               sizeof(*s->value) * (s->capacity - s->length));
    }

 out:
    return;
}

/* Comments are in cgul_string_p.h. */
void
cgul_string__decrease_capacity_if_necessary(cgul_exception_t* cex,
                                            cgul_string_t s)
{
    size_t actual_new_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* We only have excess capacity when 1/4 capacity is greater than
     * or equal to the size.  So, do nothing when 1/4 capacity is less
     * than the size. */
    if ((s->capacity >> 2) < (s->length + 1)) {
        goto out;
    }

    /* Keep halving the capacity to get the best fit.  It should be at
     * least double the size to avoid thrashing the allocator.  We
     * know from the if() above that we can start iterating with 1/2
     * of s->capacity. */
    actual_new_capacity = s->capacity >> 1;
    while (   ((actual_new_capacity >> 1) > 0)
           && ((actual_new_capacity >> 1) >= ((s->length + 1) << 1)))
    {
        actual_new_capacity >>= 1;
    }

    /* The actual new capacity should never drop below the minimum
     * capacity. */
    actual_new_capacity = CGUL_MAX(actual_new_capacity, s->minimum_capacity);
    assert(actual_new_capacity > 0);

    /* If the capacity is already less than or equal to the actual new
     * capacity, do nothing. */
    if (s->capacity <= actual_new_capacity) {
        goto out;
    }

    /* Sanity check to make sure we are not going to accidentally
     * reduce the capacity below the necessary size. */
    if (actual_new_capacity <= (s->length + 1)) {
        goto out;
    }

    /* Set the reduced capacity. */
    cgul_string__set_capacity(cex, s, actual_new_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Wrapper around <code>cgul_string__increase_capacity_if_necessary()</code>
 * that performs a quick check to determine if there is already enough
 * excess capacity to accommodate the extra size.  If not, the real
 * method is called.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] extra_size extra size in characters excluding NUL-terminator
 */
#define CGUL_STRING__INCREASE_CAPACITY_IF_NECESSARY(cex, s, extra_size) \
do {                                                                    \
    size_t excess_capacity = (s)->capacity - (s)->length - 1;           \
    size_t local_extra_size = (extra_size);                             \
    if (excess_capacity < local_extra_size) {                           \
        cgul_string__increase_capacity_if_necessary(                    \
            (cex), (s), local_extra_size);                              \
    }                                                                   \
} while (0)

/* Comments are in cgul_string_p.h. */
void
cgul_string__increase_capacity_if_necessary(cgul_exception_t* cex,
                                            cgul_string_t s,
                                            size_t extra_size)
{
    size_t needed_capacity = 0;
    size_t target_capacity = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Sanity check. */
    if (extra_size == 0) {
        goto out;
    }

    /*
     * NOTE:
     * ====
     *
     * It is more efficient (on average) to call this method
     * through the CGUL_STRING__INCREASE_CAPACITY_IF_NECESSARY() macro
     * defined above.
     */

    /* Add 1 to extra_size to account for the NUL terminator. */
    extra_size = cgul_defensive__add_size_type(cex, extra_size, 1);
    if (*cex) {
        goto out;
    }

    /* Calculate the needed capacity. */
    needed_capacity =
        cgul_defensive__add_size_type(cex, s->length, extra_size);
    if (*cex) {
        goto out;
    }

    /* If we need more capacity, realloc() it.*/
    if (needed_capacity > s->capacity) {

        /* Try to double the capacity to allow room to grow.  The test
         * against zero is used to detect overflow. */
        target_capacity = s->capacity;
        while (   ((target_capacity << 1) > 0)
               && ((target_capacity << 1) <= (needed_capacity << 1)))
        {
            target_capacity <<= 1;
        }

        /* If there is overflow, just use needed_capacity as-is which
         * doesn't allow for room to grow without reallocating but is
         * large enough to hold the new elements.  Also, we know that
         * needed_capacity did not overflow when calculated above
         * because an exception would have been thrown. */
        if (target_capacity < needed_capacity) {
            target_capacity = needed_capacity;
        }

        /* Set the new capacity. */
        cgul_string__set_capacity(cex, s, target_capacity);
    }

 out:
    return;
}

/**
 * Create the value used in case-insensitive comparisons.  If memory
 * cannot be allocated as needed, an exception is thrown.  You will
 * only need to call this method again if you change the value of the
 * underlying string and want to do another case-insensitive
 * comparison.
 *
 * @param[in] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 */
static void
cgul_string__create_ivalue(cgul_exception_t* cex, cgul_string_t s)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Allocate or reallocate space for s->ivalue. */
    s->ivalue = (char*)cgul_libc__realloc(cex, s->ivalue, s->length + 1);
    if (*cex) {
        goto out;
    }

    /* Copy s->value to s->ivalue converting each character to its
     * lower-case equivalent. */
    for ( i = 0 ; i < s->length ; ++i ) {
        s->ivalue[i] = tolower((unsigned char)s->value[i]);
    }

    /* NUL terminate the string. */
    s->ivalue[i] = '\0';

    /* Remember that s->ivalue is valid. */
    s->is_ivalue_valid = 1;

 out:
    return;
}

/**
 * This method applies the field width specified in
 * <code>cs->field_width</code> to <code>s</code> provided
 * <code>cs->is_field_width</code> present evaluates to true.  If the
 * field is right justified, it will be padded with the
 * <code>pad</code> character.  If the field is left justified, it
 * will be padded with spaces. <p>
 *
 * This method assumes that field width for the conversion
 * specification has already been read off of the stack if necessary
 * and stored in <code>cs->field_width</code>. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] pad pad character for right justification only
 */
static void
cgul_string__apply_field_width(cgul_exception_t* cex,
                               cgul_string_t s,
                               cgul_convspec_t cs,
                               cgul_wchar_t pad)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /*
     * NOTE: The cgul_string__pad_left() and cgul_string__pad_right()
     * methods count characters in s assuming it is a UTF-8 string and
     * inserts the correct amount of padding based on that character
     * count.
     */

    /* If the field width is present, add padding. */
    if (cs->is_field_width_present) {
        /* The '-' flag means we left justify. */
        if (cs->flag_dash) {
            cgul_string__pad_left(cex, s, cs->field_width, (cgul_wchar_t)' ');
        } else {
            /* Otherwise, we right justify. */
            cgul_string__pad_right(cex, s, cs->field_width, pad);
        }
    }

 out:
    return;
}

/**
 * This method appends <code>value</code> to <code>s</code> according
 * to the field width, precision, etc. contained in the conversion
 * specification <code>cs</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_string_with_convspec(cgul_exception_t* cex,
                                         cgul_string_t s,
                                         cgul_convspec_t cs,
                                         const char* value)
{
    int i = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /* Allocate a temporary string. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* If a precision is present, a NUL terminator is not required.
     * So we will handle this as a special case.  Note that
     * cs->precision limits how many characters we copy making it
     * possible to append a string to itself. */
    if (cs->is_precision_present && value) {
        for (i = 0 ; value[i] != '\0' ; ++i) {
            /* We are finished if a precision was specified and we
             * just exceeded it.*/
            if (cs->is_precision_present && (i >= cs->precision)) {
                break;
            }
            /* Append this character to "tmp". */
            cgul_string__append_char(cex, tmp, value[i]);
        }
        /* Apply the field width to "tmp". */
        cgul_string__apply_field_width(cex, tmp, cs, (cgul_wchar_t)' ');
        /* Success. */
        cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));
        goto out;
    }

    /* If a precision is not present, "value" must be NUL terminated
     * so it is safe to call cgul_string__append() which knows how to
     * append a string to itself. */
    cgul_string__append(cex, tmp, value);

    /* Apply the field width to "tmp". */
    cgul_string__apply_field_width(cex, tmp, cs, (cgul_wchar_t)' ');

    /* Success. */
    cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

/**
 * This method appends converts <code>value</code> from a UTF-32
 * string to a UTF-8 string and appends the result to <code>s</code>
 * according to the field width, precision, etc. contained in the
 * conversion specification <code>cs</code>.  If an error occurs, an
 * exception is thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_wstring_with_convspec(cgul_exception_t* cex,
                                          cgul_string_t s,
                                          cgul_convspec_t cs,
                                          const cgul_wchar_t* value)
{
    int i = 0;
    cgul_string_t tmp = NULL;
    char buf[CGUL_WCHAR__MB_LEN_MAX + 1];

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);
    assert(value);

    /* Allocate a temporary string. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* For a string, the precision specifies the maximum number of
     * characters to copy from "value".  Also, if a precision is
     * present, no trailing '\0' character is required.  So, we need
     * to be very careful when copying characters from "value" into
     * "tmp". */
    for (i = 0 ; value[i] != CGUL_WCHAR__NUL ; ++i) {

        /* We are finished if a precision was specified and we just
         * exceeded it.*/
        if (cs->is_precision_present && (i >= cs->precision)) {
            break;
        }

        /* Convert this UTF-32 character into a UTF-8 sequence. */
        cgul_unicode__wctomb(cex, value[i], buf);
        if (*cex) {
            goto out;
        }

        /* Append this UTF-8 character to "tmp". */
        cgul_string__append(cex, tmp, buf);
        if (*cex) {
            goto out;
        }
    }

    /* Apply the field width to "tmp". */
    cgul_string__apply_field_width(cex, tmp, cs, (cgul_wchar_t)' ');

    /* Append "tmp" to "s". */
    cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

/**
 * This method appends <code>value</code> to <code>s</code> according
 * to the field width, precision, etc. contained in the conversion
 * specification <code>cs</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_char_with_convspec(cgul_exception_t* cex,
                                       cgul_string_t s,
                                       cgul_convspec_t cs,
                                       unsigned char value)
{
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /* Allocate a temporary string. */
    tmp = cgul_string__new(cex);

    /* Append the character. */
    cgul_string__append_char(cex, tmp, value);

    /* NOTE: Precision values are ignored for "%c" conversions. */

    /* Apply the field width to "tmp". */
    cgul_string__apply_field_width(cex, tmp, cs, (cgul_wchar_t)' ');

    /* Append "tmp" to "s". */
    cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

/**
 * This method converts the Unicode UTF-32 wide-character in
 * <code>value</code> to a UTF-8 multi-byte charcter sequence and
 * appends the sequence to <code>s</code> according to the field
 * width, precision, etc. contained in the conversion specification
 * <code>cs</code>.  If an error occurs, an exception is thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_wchar_with_convspec(cgul_exception_t* cex,
                                        cgul_string_t s,
                                        cgul_convspec_t cs,
                                        cgul_wchar_t value)
{
    cgul_string_t tmp = NULL;
    char buf[CGUL_WCHAR__MB_LEN_MAX + 1];

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /* Allocate a temporary string. */
    tmp = cgul_string__new(cex);

    /* Convert the UTF-32 character to a UTF-8 multi-byte sequence. */
    cgul_unicode__wctomb(cex, value, buf);

    /* Append the multi-byte sequence to tmp. */
    cgul_string__append(cex, tmp, buf);

    /* NOTE: Precision values are ignored for "%c" conversions. */

    /* Apply the field width to "tmp". */
    cgul_string__apply_field_width(cex, tmp, cs, (cgul_wchar_t)' ');

    /* Append "tmp" to "s". */
    cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

/**
 * This method appends <code>value</code> to <code>s</code> according
 * to the field width, precision, etc. contained in the conversion
 * specification <code>cs</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_long_int_with_convspec(cgul_exception_t* cex,
                                           cgul_string_t s,
                                           cgul_convspec_t cs,
                                           long int value)
{
    int tmp_length = 0;
    cgul_string_t tmp = NULL;
    char* tmp_value = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /* As a special case if the precision is explicitly given as zero
     * and "value" is zero, nothing is appended. */
    if (cs->is_precision_present && (cs->precision == 0) && (value == 0)) {
        goto out;
    }

    /* Allocate a temporary string. */
    tmp = cgul_string__new(cex);

    /* Append value as decimal. */
    cgul_string__append_decimal(cex, tmp, value);

    /* If necessary, strip off the leading sign character so we can
     * append leading zeros.  It may seem quicker to just call
     * cgul_string__append_decimal() above with "-value", but this can
     * overflow if "value" is LONG_MIN. */
    if ((tmp->length > 0) && (tmp->value[0] == '-')) {
        cgul_string__set_value(cex, tmp, &tmp->value[1]);
    }

    /*
     * Check if a precision is present.
     */

    /* If a precision is present, we do not honor the '0' flag. */
    if (cs->is_precision_present) {
        /* Extend the decimal string with zeros to match the precision. */
        tmp_length = (int)cgul_string__get_length(cex, tmp);
        if (tmp_length < cs->precision) {
            cgul_string__pad_right(cex,
                                   tmp,
                                   cs->precision,
                                   '0');
        }
        /* Insert leading '-', '+', or space character. */
        if (value < 0) {
            cgul_string__insert_char(cex, tmp, 0, '-');
        } else {
            if (cs->flag_plus) {
                cgul_string__insert_char(cex, tmp, 0, '+');
            } else if (cs->flag_space) {
                cgul_string__insert_char(cex, tmp, 0, ' ');
            }
        }
        /* Extend the string to match the field width. */
        cgul_string__apply_field_width(cex, tmp, cs, ' ');
        /* Success. */
        cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));
        goto out;
    }

    /*
     * A precision is not present.
     */

    /* So, try to honor the '0' flag. */
    if (cs->flag_zero) {
        /* Extend the string to match the field width. */
        cgul_string__apply_field_width(cex, tmp, cs, '0');
        /* Insert leading '-', '+', or space character.  The trick
         * here is that you have to do so in a why that fits within
         * the '0' padded field width if possible. */
        tmp_value = cgul_string__get_value(cex, tmp);
        if (tmp_value[0] == '0') {
            if (value < 0) {
                tmp_value[0] = '-';
            } else {
                if (cs->flag_plus) {
                    tmp_value[0] = '+';
                } else if (cs->flag_space) {
                    tmp_value[0] = ' ';
                }
            }
        } else {
            /* There wasn't enough room in the field so we have to
             * insert at the beginning of the string. */
            if (value < 0) {
                cgul_string__insert_char(cex, tmp, 0, '-');
            } else {
                if (cs->flag_plus) {
                    cgul_string__insert_char(cex, tmp, 0, '+');
                } else if (cs->flag_space) {
                    cgul_string__insert_char(cex, tmp, 0, ' ');
                }
            }
        }
        /* Success. */
        cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));
        goto out;
    }

    /*
     * A precision is not present and the '0' flag is not present so
     * we do not have to worry about inserting any extra characters at
     * the front of the string.
     */

    /* Insert leading '-', '+', or space character. */
    if (value < 0) {
        cgul_string__insert_char(cex, tmp, 0, '-');
    } else {
        if (cs->flag_plus) {
            cgul_string__insert_char(cex, tmp, 0, '+');
        } else if (cs->flag_space) {
            cgul_string__insert_char(cex, tmp, 0, ' ');
        }
    }

    /* Extend the string to match the field width. */
    cgul_string__apply_field_width(cex, tmp, cs, ' ');

    /* Success. */
    cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

/**
 * This method appends <code>value</code> to <code>s</code> according
 * to the field width, precision, etc. contained in the conversion
 * specification <code>cs</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_size_type_with_convspec(cgul_exception_t* cex,
                                            cgul_string_t s,
                                            cgul_convspec_t cs,
                                            size_t value)
{
    int tmp_length = 0;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /* As a special case if the precision is explicitly given as zero
     * and "value" is zero, nothing is appended. */
    if (cs->is_precision_present && (cs->precision == 0) && (value == 0)) {
        goto out;
    }

    /* Allocate a temporary string. */
    tmp = cgul_string__new(cex);

    /* Append value as decimal. */
    cgul_string__append_unsigned_decimal(cex, tmp, value);

    /* Extend the decimal string with zeros to match the precision. */
    if (cs->is_precision_present) {
        tmp_length = (int)cgul_string__get_length(cex, tmp);
        if (tmp_length < cs->precision) {
            cgul_string__pad_right(cex,
                                   tmp,
                                   cs->precision,
                                   '0');
        }
    }

    /*
     * Expand tmp to match the field width.  Note that the '+' and ' '
     * flags are ignored for unsigned types.
     */
    if (cs->flag_zero && !cs->is_precision_present) {
        cgul_string__apply_field_width(cex, tmp, cs, '0');
    } else {
        cgul_string__apply_field_width(cex, tmp, cs, ' ');
    }

    /* Append "tmp" to "s". */
    cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

/**
 * This method appends <code>value</code> to <code>s</code> according
 * to the field width, precision, etc. contained in the conversion
 * specification <code>cs</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_octal_with_convspec(cgul_exception_t* cex,
                                        cgul_string_t s,
                                        cgul_convspec_t cs,
                                        size_t value)
{
    int tmp_length = 0;
    char* tmp_value = NULL;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /* As a special case if the precision is explicitly given as zero
     * and "value" is zero, nothing is appended. */
    if (cs->is_precision_present && (cs->precision == 0) && (value == 0)) {
        goto out;
    }

    /* Allocate a temporary string. */
    tmp = cgul_string__new(cex);

    /* Append value as octal. */
    cgul_string__append_octal(cex, tmp, value);

    /* Extend the octal string with zeros to match the precision. */
    if (cs->is_precision_present) {
        tmp_length = (int)cgul_string__get_length(cex, tmp);
        if (tmp_length < cs->precision) {
            cgul_string__pad_right(cex,
                                   tmp,
                                   cs->precision,
                                   '0');
        }
    }

    /* If the '#' flag is present, we need to prepend a leading '0'
     * but only if there isn't already a leading '0'. */
    if (cs->flag_hash) {
        tmp_value = cgul_string__get_value(cex, tmp);
        if (tmp_value[0] != '0') {
            cgul_string__insert_char(cex, tmp, 0, '0');
        }
    }

    /*
     * Expand tmp to match the field width.  Note that the '+' and ' '
     * flags are ignored for unsigned types.
     */
    if (cs->flag_zero && !cs->is_precision_present) {
        cgul_string__apply_field_width(cex, tmp, cs, '0');
    } else {
        cgul_string__apply_field_width(cex, tmp, cs, ' ');
    }

    /* Append "tmp" to "s". */
    cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

/**
 * This method appends <code>value</code> to <code>s</code> according
 * to the field width, precision, etc. contained in the conversion
 * specification <code>cs</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_hexadecimal_with_convspec(cgul_exception_t* cex,
                                              cgul_string_t s,
                                              cgul_convspec_t cs,
                                              size_t value)
{
    int tmp_length = 0;
    char* tmp_value = NULL;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /* Sanity check to make sure cs->cspec is either 'x' or 'X'.  This
     * is relied upon in a few places below. */
    if ((cs->cspec != 'x') && (cs->cspec != 'X')) {
        *cex = cgul_exception__new(
                   "Internal Error: "
                   "cgul_string__append_hexadecimal_with_convspec: "
                   "Unexpected cs->cspec: %c",
                   cs->cspec);
        goto out;
    }

    /* As a special case if the precision is explicitly given as zero
     * and "value" is zero, nothing is appended. */
    if (cs->is_precision_present && (cs->precision == 0) && (value == 0)) {
        goto out;
    }

    /* Allocate a temporary string. */
    tmp = cgul_string__new(cex);

    /* Append value as hexadecimal. */
    if (cs->cspec == 'x') {
        cgul_string__append_hexadecimal(cex, tmp, value, 0 /* caps */);
    } else {
        cgul_string__append_hexadecimal(cex, tmp, value, 1 /* caps */);
    }

    /*
     * Check if a precision is present.
     */

    /* If a precision is present, we do not honor the '0' flag. */
    if (cs->is_precision_present) {
        /* Extend the hexadecimal string with zeros to match the precision. */
        tmp_length = (int)cgul_string__get_length(cex, tmp);
        if (tmp_length < cs->precision) {
            cgul_string__pad_right(cex,
                                   tmp,
                                   cs->precision,
                                   '0');
        }
        /* Add leading '0x' if the '#' flag is present. */
        if (cs->flag_hash) {
            if (cs->cspec == 'x') {
                cgul_string__insert(cex, tmp, 0, "0x", 2);
            } else{
                cgul_string__insert(cex, tmp, 0, "0X", 2);
            }
        }
        /* Extend the string to match the field width. */
        cgul_string__apply_field_width(cex, tmp, cs, ' ');
        /* Success. */
        cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));
        goto out;
    }

    /*
     * A precision is not present.
     */

    /* So, try to honor the '0' flag. */
    if (cs->flag_zero) {
        /* Extend the string to match the field width. */
        cgul_string__apply_field_width(cex, tmp, cs, '0');
        /* Now try to honor the '#' flag.  The trick here is that you
         * have to do so in a way that fits within the '0' padded
         * field width if possible.  We do this by checking if the
         * first two characters are both '0'.  This means there is
         * room to overwrite the second '0' character with either 'x'
         * or 'X'. */
        if (cs->flag_hash) {
            tmp_value = cgul_string__get_value(cex, tmp);
            /* Detect the edge case where there is only one leading
             * '0' character.  In this case, we insert a second
             * leading '0' character and fall through. */
            if ((tmp_value[0] == '0') && (tmp_value[1] != '0')) {
                cgul_string__insert_char(cex, tmp, 0, '0');
            }
            /* Check for two leading '0' characters. */
            if ((tmp_value[0] == '0') && (tmp_value[1] == '0')) {
                if (cs->cspec == 'x') {
                    tmp_value[1] = 'x';
                } else {
                    tmp_value[1] = 'X';
                }
            } else {
                /* There wasn't enough room in the field so we have to
                 * insert "0x" or "0X" at the beginning of the
                 * string. */
                if (cs->cspec == 'x') {
                    cgul_string__insert(cex, tmp, 0, "0x", 2);
                } else{
                    cgul_string__insert(cex, tmp, 0, "0X", 2);
                }
            }
        }
        /* Success. */
        cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));
        goto out;
    }

    /*
     * A precision is not present and the '0' flag is not present so
     * we do not have to worry about inserting any extra characters at
     * the front of the string.
     */

    /* Add a leading '0x' if the '#' flag is present. */
    if (cs->flag_hash) {
        if (cs->cspec == 'x') {
            cgul_string__insert(cex, tmp, 0, "0x", 2);
        } else{
            cgul_string__insert(cex, tmp, 0, "0X", 2);
        }
    }

    /* Extend the string to match the field width. */
    cgul_string__apply_field_width(cex, tmp, cs, ' ');

    /* Success. */
    cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

/**
 * This method appends <code>value</code> to <code>s</code> according
 * to the field width, precision, etc. contained in the conversion
 * specification <code>cs</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_pointer_with_convspec(cgul_exception_t* cex,
                                          cgul_string_t s,
                                          cgul_convspec_t cs,
                                          void* value)
{
    int tmp_length = 0;
    char* tmp_value = NULL;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /* Allocate a temporary string. */
    tmp = cgul_string__new(cex);

    /* Append value as a pointer. */
    cgul_string__append_pointer(cex, tmp, value);

    /* The call above always adds a leading "0x" which gets in our way
     * so we remove it. */
    tmp_value = cgul_string__get_value(cex, tmp);
    if ((tmp_value[0] == '0') && (tmp_value[1] == 'x')) {
        cgul_string__set_value(cex, tmp, &tmp_value[2]);
    }

    /*
     * Check if a precision is present.
     */

    /* If a precision is present, we do not honor the '0' flag. */
    if (cs->is_precision_present) {
        /* Extend the hexadecimal string with zeros to match the precision. */
        tmp_length = (int)cgul_string__get_length(cex, tmp);
        if (tmp_length < cs->precision) {
            cgul_string__pad_right(cex,
                                   tmp,
                                   cs->precision,
                                   '0');
        }
        /* Add leading '0x'. */
        cgul_string__insert(cex, tmp, 0, "0x", 2);
        /* Extend the string to match the field width. */
        cgul_string__apply_field_width(cex, tmp, cs, ' ');
        /* Success. */
        cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));
        goto out;
    }

    /*
     * A precision is not present.
     */

    /* So, try to honor the '0' flag. */
    if (cs->flag_zero) {
        /* Extend the string to match the field width. */
        cgul_string__apply_field_width(cex, tmp, cs, '0');
        /* Now try to honor the '#' flag.  The trick here is that you
         * have to do so in a way that fits within the '0' padded
         * field width if possible.  We do this by checking if the
         * first two characters are both '0'.  This means there is
         * room to overwrite the second '0' character with either 'x'
         * or 'X'. */
        tmp_value = cgul_string__get_value(cex, tmp);
        /* Detect the edge case where there is only one leading '0'
         * character.  In this case, we insert a second leading '0'
         * character and fall through. */
        if ((tmp_value[0] == '0') && (tmp_value[1] != '0')) {
            cgul_string__insert_char(cex, tmp, 0, '0');
        }
        /* Check for two leading '0' characters. */
        if ((tmp_value[0] == '0') && (tmp_value[1] == '0')) {
            tmp_value[1] = 'x';
        } else {
            /* There wasn't enough room in the field so we have to
             * insert "0x" or "0X" at the beginning of the string. */
            cgul_string__insert(cex, tmp, 0, "0x", 2);
        }
        /* Success. */
        cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));
        goto out;
    }

    /*
     * A precision is not present and the '0' flag is not present so
     * we do not have to worry about inserting any extra characters at
     * the front of the string.
     */

    /* Add a leading '0x'. */
    cgul_string__insert(cex, tmp, 0, "0x", 2);

    /* Extend the string to match the field width. */
    cgul_string__apply_field_width(cex, tmp, cs, ' ');

    /* Success. */
    cgul_string__append(cex, s, cgul_string__get_value(cex, tmp));

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

/**
 * Return the upper bound on the size of the resulting string when the
 * long double <code>value</code> is converted from a floating point
 * to a string using the <code>field_width</code> and
 * <code>precision</code> that would ordinarily be feed into
 * <code>cgul_string__append_sprintf()</code> with a %e, %f, or %g
 * conversion specifier. <p>
 *
 * The <code>field_width</code> and <code>precision</code> can be
 * negative.  In this case, this function just uses the absolute value
 * of those parameters. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] value value
 * @param[in] field_width field width
 * @param[in] precision precision
 * @return buffer length
 */
static int
cgul_string__calculate_long_double_buffer_length(cgul_exception_t* cex,
                                                 long double value,
                                                 int field_width,
                                                 int precision)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allow for a NUL terminator. */
    rv += 1;

    /* Allow for a decimal point. */
    rv += 1;

    /* Allow for a leading '+' or '-' sign. */
    rv += 1;

    /* Allow for the exponent part which should have a worst case of
     * "e-4966". */
    rv += 6;

    /* Allow for "nan", "inf", "infinity", etc. */
    rv += 24;

    /* Add enough to the buffer length for the worst case when no
     * field width or precision is present.  Under these conditions,
     * the worst case is "%f" where the "long double" has a very large
     * exponent.  For example, if the exponent is "e+4931", it adds
     * 4932 characters to the total length which is one more than the
     * value of the exponent.  For a long double, the largest exponent
     * is a little less than 5000. */
    rv += 5001;

    /* Add the absolute value of the field width. */
    if (field_width < 0) {
        field_width = -field_width;
    }
    rv += field_width;

    /* Add the absolute value of the precision.  We add 6 to the
     * precision because 6 is the default if no precision is given.
     * This also covers the case where a 0 precision is used with %g
     * which causes it to be interpreted as a precision of 1. */
    if (precision < 0) {
        precision = -precision;
    }
    rv += precision + 6;

    /* In case I forgot something. */
    rv += 64;

    /* Just in case the thousands separator somehow slips in on
     * systems that support it, add in 1/3 of the size we currently
     * have. */
    rv += (rv/3) + 1;

 out:
    return rv;
}

/**
 * This method appends <code>value</code> to <code>s</code> according
 * to the field width, precision, etc. contained in the conversion
 * specification <code>cs</code>.  If an error occurs, an exception is
 * thrown. <p>
 *
 * This method assumes that field width and precision for the
 * conversion specification have already been read off of the stack if
 * necessary and stored in <code>cs->field_width</code> and
 * <code>cs->precision</code> respectively. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] s <code>cgul_string</code> instance
 * @param[in] cs conversion specification
 * @param[in] value value to append
 */
static void
cgul_string__append_long_double_with_convspec(cgul_exception_t* cex,
                                              cgul_string_t s,
                                              cgul_convspec_t cs,
                                              long double value)
{
    int sprintf_rv = 0;
    char* buf = NULL;
    int buf_length = 0;
    cgul_string_t format = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(cs);

    /* Calculate the length we need for our buffer. */
    buf_length =
        cgul_string__calculate_long_double_buffer_length(
            cex,
            value,
            (cs->is_field_width_present ? cs->field_width : 0),
            (cs->is_precision_present   ? cs->precision   : 0));

    /* Allocate the buffer. */
    buf = (char*)cgul_libc__malloc(cex, buf_length);
    if (*cex) {
        goto out;
    }

    /* Allocate the format string. */
    format = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Because we have promoted all "double" values to "long double"
     * values, we have to insure that "cs" tracks this change. */
    cgul_convspec__set_is_length_modifier_present(cex, cs, 1);
    cgul_convspec__set_length_modifier(cex, cs, "L");

    /* Generate a format string from the conversion specification. */
    cgul_convspec__serialize(cex, cs, format);
    if (*cex) {
        goto out;
    }

    /* Use sprintf() to convert from floating point to string. */
    sprintf_rv = sprintf(buf, cgul_string__get_value(cex, format), value);
    if (sprintf_rv < 0) {
        *cex = cgul_exception__new(
                   "cgul_string__append_long_double_with_convspec: "
                   "sprintf(): %s",
                   strerror(errno));
        goto out;
    }

    /* Sanity check the number of characters sprintf() reports writing
     * with the size of the buffer we allocated. */
    if (sprintf_rv >= buf_length) {
        *cex = cgul_exception__new(
                   "cgul_string__append_long_double_with_convspec: "
                   "Internal Error: sprintf() wrote beyond buffer!!!");
        goto out;
    }

    /* Success. */
    cgul_string__append(cex, s, buf);

 out:

    /* Clean up. */
    if (format) {
        cgul_string__delete(format);
    }
    if (buf) {
        free(buf);
    }

    return;
}

/**
 * Initialize <code>s</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] s cgul string
 * @param[in] value initial value
 */
static void
cgul_string__initialize(cgul_exception_t* cex,
                        cgul_string_t s,
                        const char* value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Free the old underlying arrays if any. */
    if (s->value) {
        free(s->value);
    }
    if (s->ivalue) {
        free(s->ivalue);
    }

    /* Initialize all member variables. */
    s->value = NULL;
    s->ivalue = NULL;
    s->is_ivalue_valid = 1;
    s->length = 0;
    s->capacity = 0;
    s->minimum_capacity = CGUL_STRING__DEFAULT_MINIMUM_CAPACITY;

    /* Initialize the capacity. */
    cgul_string__set_capacity(cex, s, s->minimum_capacity);
    if (*cex) {
        goto out;
    }

    /* Now that it is safe to call methods, use set_value() to set the value. */
    cgul_string__set_value(cex, s, value);

 out:
    return;
}

cgul_string_t
cgul_string__new(cgul_exception_t* cex)
{
    cgul_string_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (cgul_string_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    cgul_string__initialize(cex, rv, "");
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_string__delete(rv);
        rv = NULL;
    }

    return rv;
}

cgul_string_t
cgul_string__new_from_cstring(cgul_exception_t* cex,
                              const char* value)
{
    cgul_string_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(value);

    /* Create an empty string object. */
    rv = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Set the value. */
    cgul_string__set_value(cex, rv, value);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up partially constructed objects. */
    if (*cex && rv) {
        cgul_string__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
cgul_string__delete(cgul_string_t s)
{
    if (s) {
        if (s->value) {
            free(s->value);
        }
        if (s->ivalue) {
            free(s->ivalue);
        }
        free(s);
    }
}

char*
cgul_string__get_value(cgul_exception_t* cex, cgul_string_t s)
{
    char* rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    rv = s->value;

 out:
    return rv;
}

char*
cgul_string__take_value(cgul_exception_t* cex, cgul_string_t s)
{
    char* rv= NULL;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /*
     * If you reach here, the rest of this function will complete
     * without throwing an exception.
     */

    /* Swap the newly allocated string with s thereby resetting s. */
    cgul_string__swap(cex, s, tmp);

    /* The client is taking ownership of what is now tmp->value. */
    rv = tmp->value;
    tmp->value = NULL;

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return rv;
}

void
cgul_string__set_value(cgul_exception_t* cex,
                       cgul_string_t s,
                       const char* value)
{
    int is_interior_ptr = 0;
    size_t i = 0;
    size_t value_length = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);
    assert(value);

    /* Do not copy the string to itself. */
    if (s->value == value) {
        goto out;
    }

    /* Determine the string length of "value". */
    value_length = strlen(value);

    /* Determine if "value" is a pointer to a substring inside of s->value. */
    is_interior_ptr = (   (value >= s->value)
                       && (value <= s->value + s->length));

    /* Make sure the underlying buffer is large enough but only if
     * "value" is not an interior pointer.  Without this extra check,
     * the realloc() call could invalidate "value", and in this case
     * there is no reason to call realloc() because we know s->value
     * is large enough to hold any of its own substrings. */
    if (!is_interior_ptr) {
        CGUL_STRING__INCREASE_CAPACITY_IF_NECESSARY(cex, s, value_length);
        if (*cex) {
            goto out;
        }
    }

    /* Copy "value" to "s->value" including the NUL terminator. */
    for (i = 0 ; i <= value_length ; ++i) {
        s->value[i] = value[i];
    }
    s->length = value_length;
    s->is_ivalue_valid = 0;

    /* See if the capacity can be decreased.  (Note that we *increase*
     * the capacity if necessary before attempting to copy, but
     * sometimes we are setting the string to a shorter value.) */
    cgul_string__decrease_capacity_if_necessary(cex, s);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

char
cgul_string__get_value_at(cgul_exception_t* cex,
                          cgul_string_t s,
                          size_t index)
{
    char rv = '\0';

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    if (index >= s->length) {
        *cex = cgul_exception__new("cgul_string__get_value_at: "
                                   "index out of bounds: %lu",
                                   (unsigned long)index);
        goto out;
    }

    rv = s->value[index];

 out:
    return rv;
}

extern void
cgul_string__set_value_at(cgul_exception_t* cex,
                          cgul_string_t s,
                          size_t index,
                          char c)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Verify that "index" is within bounds. */
    if (index >= s->length) {
        *cex = cgul_exception__new("cgul_string__set_value_at: "
                                   "index out of bounds: %lu",
                                   (unsigned long)index);
        goto out;
    }

    s->value[index] = c;

 out:
    return;
}

/* This method is part of the "cacheable" interface defined by
 * "cgul_cache".  It is required to not throw any exception. */
void
cgul_string__clear(cgul_string_t s)
{
    cgul_exception_t local = NULL;
    cgul_string__set_length(&local, s, 0);
}

signed char
cgul_string__to_schar(cgul_exception_t* cex,
                      cgul_string_t s,
                      int base)
{
    long int tmp = 0;
    long int schar_min = SCHAR_MIN;
    long int schar_max = SCHAR_MAX;
    signed char rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Convert to a long. */
    tmp = cgul_string__to_long(cex, s, base);
    if (*cex) {
        goto out;
    }

    /* Verify that tmp is within the bounds of a signed char. */
    if ((tmp < schar_min) || (tmp > schar_max)) {
        *cex = cgul_exception__new("cgul_string__to_schar: out of bounds: %ld",
                                   tmp);
        goto out;
    }

    rv = (signed char)tmp;

out:
    return rv;
}

unsigned char
cgul_string__to_uchar(cgul_exception_t* cex,
                      cgul_string_t s,
                      int base)
{
    unsigned long int tmp = 0;
    unsigned long int uchar_max = UCHAR_MAX;
    unsigned char rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Convert to unsigned long. */
    tmp = cgul_string__to_unsigned_long(cex, s, base);
    if (*cex) {
        goto out;
    }

    /* Verify that tmp is within the bounds of an unsigned char. */
    if (tmp > uchar_max) {
        *cex = cgul_exception__new("cgul_string__to_uchar: "
                                   "out of bounds: %lu", tmp);
        goto out;
    }

    rv = (unsigned char)tmp;

 out:
    return rv;
}

short int
cgul_string__to_short(cgul_exception_t* cex,
                      cgul_string_t s,
                      int base)
{
    long int tmp = 0;
    long int shrt_min = SHRT_MIN;
    long int shrt_max = SHRT_MAX;
    short int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Convert to a long. */
    tmp = cgul_string__to_long(cex, s, base);
    if (*cex) {
        goto out;
    }

    /* Verify that tmp is within the bounds of a short. */
    if ((tmp < shrt_min) || (tmp > shrt_max)) {
        *cex = cgul_exception__new("cgul_string__to_short: out of bounds: %ld",
                                   tmp);
        goto out;
    }

    rv = (short int)tmp;

out:
    return rv;
}

unsigned short int
cgul_string__to_ushort(cgul_exception_t* cex,
                       cgul_string_t s,
                       int base)
{
    unsigned long int tmp = 0;
    unsigned long int ushrt_max = USHRT_MAX;
    unsigned short int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Convert to unsigned long. */
    tmp = cgul_string__to_unsigned_long(cex, s, base);
    if (*cex) {
        goto out;
    }

    /* Verify that tmp is within the bounds of an unsigned short. */
    if (tmp > ushrt_max) {
        *cex = cgul_exception__new("cgul_string__to_ushort: "
                                   "out of bounds: %lu", tmp);
        goto out;
    }

    rv = (unsigned short int)tmp;

 out:
    return rv;
}

int
cgul_string__to_int(cgul_exception_t* cex, cgul_string_t s, int base)
{
    int rv = 0;
    long int tmp = 0;
    long int int_min = INT_MIN;
    long int int_max = INT_MAX;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Convert to a long. */
    tmp = cgul_string__to_long(cex, s, base);
    if (*cex) {
        goto out;
    }

    /* Verify that tmp is within the bounds of an int. */
    if ((tmp < int_min) || (tmp > int_max)) {
        *cex = cgul_exception__new("cgul_string__to_int: out of bounds: %ld",
                                   tmp);
        goto out;
    }

    rv = (int)tmp;

out:
    return rv;
}

unsigned int
cgul_string__to_unsigned(cgul_exception_t* cex,
                         cgul_string_t s,
                         int base)
{
    int rv = 0;
    unsigned long int tmp = 0;
    unsigned long int uint_max = UINT_MAX;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Convert to unsigned long. */
    tmp = cgul_string__to_unsigned_long(cex, s, base);
    if (*cex) {
        goto out;
    }

    /* Verify that tmp is within the bounds of an unsigned int. */
    if (tmp > uint_max) {
        *cex = cgul_exception__new("cgul_string__to_unsigned: "
                                   "out of bounds: %lu", tmp);
        goto out;
    }

    rv = (unsigned int)tmp;

 out:
    return rv;
}

long int
cgul_string__to_long(cgul_exception_t* cex,
                     cgul_string_t s,
                     int base)
{
    long int rv = 0;
    long int tmp = 0;
    char* first_bad_char = NULL;
    const char* curr = NULL;  /* Pointer to current position. */
    const char* r = NULL;     /* Remainder of the string.     */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    /* You have to manually check for an empty string. */
    for ( curr = s->value ; *curr != '\0' ; ++curr) {
        if (!isspace((int)*curr)) {
            break;
        }
    }
    if (*curr == '\0') {
        *cex = cgul_exception__new("cgul_string__to_long: blank string");
        goto out;
    }

    /* Do the conversion. */
    errno = 0;
    tmp = strtol(curr, &first_bad_char, base);

    /* Check for errors. */
    if (errno) {
        *cex = cgul_exception__new("cgul_string__to_long: "
                                   "strtol: %s", strerror(errno));
        goto out;
    }

    /* Allow trailing space. */
    r = first_bad_char;
    for ( ; *r != '\0' ; ++r) {
        if (!isspace((int)*r)) {
            break;
        }
    }
    if (*r != '\0') {
        *cex = cgul_exception__new("cgul_string__to_long: extra chars");
        goto out;
    }

    /* Success. */
    rv = tmp;

 out:
    return rv;
}

unsigned long int
cgul_string__to_unsigned_long(cgul_exception_t* cex,
                              cgul_string_t s,
                              int base)
{
    unsigned long int rv = 0;
    int is_negative = 0;
    unsigned long int tmp = 0;
    char* first_bad_char = NULL;
    const char* curr = NULL;  /* Pointer to current position. */
    const char* r = NULL;     /* Remainder of the string.     */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    /* You have to manually check for a leading minus sign or an empty
     * string. */
    for ( curr = s->value ; *curr != '\0' ; ++curr) {
        if (!isspace((int)*curr)) {
            break;
        }
    }
    if (*curr == '\0') {
        *cex = cgul_exception__new("cgul_string__to_unsigned_long: "
                                   "blank string");
        goto out;
    }

    /* You want to allow "-0" to succeed.  So we have to wait to see
     * how the conversion goes. */
    if (*curr == '-') {
        is_negative = 1;
    }

    /* Do the conversion. */
    errno = 0;
    tmp = strtoul(curr, &first_bad_char, base);

    /* Check for errors. */
    if (errno) {
        *cex = cgul_exception__new("cgul_string__to_unsigned_long: "
                                   "strtoul: %s", strerror(errno));
        goto out;
    }

    /* Allow trailing space. */
    r = first_bad_char;
    for ( ; *r != '\0' ; ++r) {
        if (!isspace((int)*r)) {
            break;
        }
    }
    if (*r != '\0') {
        *cex = cgul_exception__new("cgul_string__to_unsigned_long: "
                                   "extra chars");
        goto out;
    }

    /* Only "-0" is allowed. */
    if (is_negative && (tmp != 0)) {
        *cex = cgul_exception__new("cgul_string__to_unsigned_long: "
                                   "not unsigned: %s", s->value);
        goto out;
    }

    /* Success. */
    rv = tmp;

 out:
    return rv;
}

double
cgul_string__to_double(cgul_exception_t* cex,
                       cgul_string_t s)
{
    double rv = 0.0;
    double tmp = 0.0;
    char* first_bad_char = NULL;
    const char* curr = NULL;  /* Pointer to current position. */
    const char* r = NULL;     /* Remainder of the string.     */

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    /* You have to manually check for an empty string. */
    for ( curr = s->value ; *curr != '\0' ; ++curr) {
        if (!isspace((int)*curr)) {
            break;
        }
    }
    if (*curr == '\0') {
        *cex = cgul_exception__new("cgul_string__to_double: blank string");
        goto out;
    }

    /* Do the conversion. */
    errno = 0;
    tmp = strtod(curr, &first_bad_char);

    /* Check for errors. */
    if (errno) {
        *cex = cgul_exception__new("cgul_string__to_double: "
                                   "strtod: %s", strerror(errno));
        goto out;
    }

    /* Allow trailing space. */
    r = first_bad_char;
    for ( ; *r != '\0' ; ++r) {
        if (!isspace((int)*r)) {
            break;
        }
    }
    if (*r != '\0') {
        *cex = cgul_exception__new("cgul_string__to_double: "
                                   "extra chars");
        goto out;
    }

    /* Success. */
    rv = tmp;

 out:
    return rv;
}

void
cgul_string__append(cgul_exception_t* cex,
                    cgul_string_t s,
                    const char* value)
{
    size_t value_length = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* If "value" is NULL, insert "(nil)". */
    if (value == NULL) {
        value = "(nil)";
    }

    /* Calculate the string length of "value". */
    value_length = strlen(value);

    /* Add capacity if necessary. */
    CGUL_STRING__INCREASE_CAPACITY_IF_NECESSARY(cex, s, value_length);
    if (*cex) {
        goto out;
    }

    /* We cannot use strcpy() because "value" could be the same as
     * "s->value".  Because we have already calculated the length of
     * "value", we can use memcpy() instead. */
    memcpy(&s->value[s->length], value, value_length);

    /* Calculate the new length. */
    s->length += value_length;

    /* NUL terminate the string. */
    s->value[s->length] = '\0';

    /* Invalidate the case-insensitive string. */
    s->is_ivalue_valid = 0;

 out:
    return;
}

void
cgul_string__append_block(cgul_exception_t* cex,
                          cgul_string_t s,
                          const void* block,
                          size_t block_size)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);
    assert(block);

    CGUL_STRING__INCREASE_CAPACITY_IF_NECESSARY(cex, s, block_size);
    if (*cex) {
        goto out;
    }

    memcpy(&s->value[s->length], block, block_size);
    s->length += block_size;
    s->value[s->length] = '\0';
    s->is_ivalue_valid = 0;

 out:
    return;
}

int
cgul_string__append_sprintf(cgul_exception_t* cex,
                            cgul_string_t s,
                            const char* format,
                            ...)
{
    va_list args;
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(format);

    va_start(args, format);
    rv = cgul_string__append_vsprintf(cex, s, format, args);
    va_end(args);

 out:
    return rv;
}

int
cgul_string__append_vsprintf(cgul_exception_t* cex,
                             cgul_string_t s,
                             const char* format,
                             va_list args)
{
    int rv = -1;
    char c = '\0';
    unsigned char arg_as_char = '\0';
    cgul_wchar_t arg_as_wchar = CGUL_WCHAR__NUL;
    char* arg_as_string = NULL;
    cgul_wchar_t* arg_as_wstring = NULL;
    void* arg_as_pointer = NULL;
    long int arg_as_long_int = 0;
    long unsigned int arg_as_size_type = 0;
    long double arg_as_long_double = 0.0;
    unsigned long int index = 0;
    size_t starting_length = 0;
    size_t ending_length = 0;
    const char* length_modifier = NULL;
    cgul_convspec_t cs = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(format);

    /* Allocate a cgul_convspec object that knows how to reduce
     * individual converstion specifications (like "%-3.4d") to their
     * component parts. */
    cs = cgul_convspec__new(cex);
    if (*cex) {
        goto out;
    }

    /* Get the starting length. */
    starting_length = cgul_string__get_length(cex, s);

    /* Iterate over each character in the format string. */
    index = 0;
    while (format[index] != '\0') {

        c = format[index];

        /*
         * Handle literal characters.
         */
        if (c != '%') {
            cgul_string__append_char(cex, s, c);
            ++index;
            continue;
        }

        /*
         * Reduce the conversion specificiation to its component parts.
         */
        cgul_convspec__deserialize(cex, cs, format, &index);
        if (*cex) {
            goto out;
        }

        /*
         * Field Width
         */

        /* Get the field width off of the stack if necessary. */
        if (cs->is_field_width_present && cs->is_field_width_on_stack) {
            cs->field_width = va_arg(args, int);
            cs->is_field_width_on_stack = 0;
        }

        /* Normalize the field width.  A negative field width should
         * be treated as a positive field width with the '-' flag
         * set. */
        if (cs->field_width < 0) {
            cs->field_width = -(cs->field_width);
            cs->flag_dash = 1;
        }

        /*
         * Precision
         */

        /* Get the precision off of the stack if necessary. */
        if (cs->is_precision_present && cs->is_precision_on_stack) {
            cs->precision = va_arg(args, int);
            cs->is_precision_on_stack = 0;
        }

        /* Normalize the precision.  The man page says that a negative
         * precision is the same as a zero precision, but testing
         * shows that when a negative precision is specified
         * (typically by passing in a negative integer on the stack),
         * the default precision is used. */
        if (cs->precision < 0) {
            cs->is_precision_present = 0;
            cs->precision = 0;
        }

        /*
         * Conversion Specifier
         */

        /* Determine the type we are converting. */
        switch (cs->cspec) {

        /* %% */
        case '%':
            cgul_string__append_char(cex, s, '%');
            break;

        /* %c, %lc */
        case 'c':
            /* Check for a length modifier. */
            if (cs->is_length_modifier_present) {
                length_modifier =
                    cgul_convspec__get_length_modifier(cex, cs);
                if (strcmp(length_modifier, "l") == 0) {
                    /* Get the value off of the stack. */
                    arg_as_wchar = va_arg(args, cgul_wchar_t);
                    /* Append the value. */
                    cgul_string__append_wchar_with_convspec(
                        cex, s, cs, arg_as_wchar);
                    break;
                }
            }
            /* Get the value off of the stack. */
            arg_as_char = va_arg(args, int);
            /* Append the value. */
            cgul_string__append_char_with_convspec(
                cex, s, cs, arg_as_char);
            break;

        /* %d, %i, %ld, %li, %hd, %hi */
        case 'd':
        case 'i':
            /* Check for a length modifier. */
            if (cs->is_length_modifier_present) {
                length_modifier =
                    cgul_convspec__get_length_modifier(cex, cs);
                if (strcmp(length_modifier, "l") == 0) {
                    /* Get the long value off of the stack. */
                    arg_as_long_int = va_arg(args, long int);
                } else if (strcmp(length_modifier, "h") == 0) {
                    /* Get the short value off of the stack. */
                    arg_as_long_int = (long int)va_arg(args, int);
                } else {
                    *cex = cgul_exception__new(
                               "cgul_string__append_vsprintf: "
                               "unsupported length modifier for %%d: %s",
                               length_modifier);
                    goto out;
                }
            } else {
                /* Get the int value off of the stack. */
                arg_as_long_int = (long int)va_arg(args, int);
            }
            /* Append the value. */
            cgul_string__append_long_int_with_convspec(
                cex, s, cs, arg_as_long_int);
            break;

        /* %e, %E, %f, %F, %g, %G */
        case 'e':
        case 'E':
        case 'f':
        case 'F':
        case 'g':
        case 'G':
            /* Check for a length modifier. */
            if (cs->is_length_modifier_present) {
                length_modifier =
                    cgul_convspec__get_length_modifier(cex, cs);
                if (strcmp(length_modifier, "L") == 0) {
                    /* Get the long double value off of the stack. */
                    arg_as_long_double = va_arg(args, long double);
                } else {
                    *cex = cgul_exception__new(
                               "cgul_string__append_vsprintf: "
                               "unsupported length modifier for float: %s",
                               length_modifier);
                    goto out;
                }
            } else {
                /* Get the double value off of the stack. */
                arg_as_long_double = (long double)va_arg(args, double);
            }
            /* Append the value. */
            cgul_string__append_long_double_with_convspec(
                cex, s, cs, arg_as_long_double);
            break;

        /* %m */
        case 'm':
            cgul_string__append_string_with_convspec(
                cex, s, cs, strerror(errno));
            break;

        /* %o, %lo, %ho, %zo */
        case 'o':
            /* Check for a length modifier. */
            if (cs->is_length_modifier_present) {
                length_modifier =
                    cgul_convspec__get_length_modifier(cex, cs);
                if (strcmp(length_modifier, "z") == 0) {
                    /* Get the size_t value off of the stack. */
                    arg_as_size_type = va_arg(args, size_t);
                } else if (strcmp(length_modifier, "l") == 0) {
                    /* Get the long value off of the stack. */
                    arg_as_size_type =
                        (size_t)va_arg(args, long unsigned int);
                } else if (strcmp(length_modifier, "h") == 0) {
                    /* Get the short value off of the stack. */
                    arg_as_size_type =
                        (size_t)va_arg(args, unsigned int);
                } else {
                    *cex = cgul_exception__new(
                               "cgul_string__append_vsprintf: "
                               "unsupported length modifier for %%o: %s",
                               length_modifier);
                    goto out;
                }
            } else {
                /* Get the int value off of the stack. */
                arg_as_size_type =
                    (size_t)va_arg(args, unsigned int);
            }
            /* Append the value. */
            cgul_string__append_octal_with_convspec(
                cex, s, cs, arg_as_size_type);
            break;

        /* %p */
        case 'p':
            arg_as_pointer = va_arg(args, void*);
            cgul_string__append_pointer_with_convspec(
                cex, s, cs, arg_as_pointer);
            break;

        /* %s, %ls */
        case 's':
            /* Check for a length modifier. */
            if (cs->is_length_modifier_present) {
                length_modifier =
                    cgul_convspec__get_length_modifier(cex, cs);
                if (strcmp(length_modifier, "l") == 0) {
                    /* Get the value off of the stack. */
                    arg_as_wstring = (cgul_wchar_t*)va_arg(args, char*);
                    /* Append the value. */
                    cgul_string__append_wstring_with_convspec(
                        cex, s, cs, arg_as_wstring);
                    break;
                }
            }
            /* Get the value off of the stack. */
            arg_as_string = va_arg(args, char*);
            /* Append the value. */
            cgul_string__append_string_with_convspec(
                cex, s, cs, arg_as_string);
            break;

        /* %u, %lu, %hu, %zu */
        case 'u':
            /* Check for a length modifier. */
            if (cs->is_length_modifier_present) {
                length_modifier =
                    cgul_convspec__get_length_modifier(cex, cs);
                if (strcmp(length_modifier, "z") == 0) {
                    /* Get the size_t value off of the stack. */
                    arg_as_size_type = va_arg(args, size_t);
                } else if (strcmp(length_modifier, "l") == 0) {
                    /* Get the long value off of the stack. */
                    arg_as_size_type =
                        (size_t)va_arg(args, long unsigned int);
                } else if (strcmp(length_modifier, "h") == 0) {
                    /* Get the short value off of the stack. */
                    arg_as_size_type =
                        (size_t)va_arg(args, unsigned int);
                } else {
                    *cex = cgul_exception__new(
                               "cgul_string__append_vsprintf: "
                               "unsupported length modifier for %%u: %s",
                               length_modifier);
                    goto out;
                }
            } else {
                /* Get the int value off of the stack. */
                arg_as_size_type =
                    (size_t)va_arg(args, unsigned int);
            }
            /* Append the value. */
            cgul_string__append_size_type_with_convspec(
                cex, s, cs, arg_as_size_type);
            break;

        /* %x, %X %lx, %lX, %hx, %hX, %zx, %zX */
        case 'x':
        case 'X':
            /* Check for a length modifier. */
            if (cs->is_length_modifier_present) {
                length_modifier =
                    cgul_convspec__get_length_modifier(cex, cs);
                if (strcmp(length_modifier, "z") == 0) {
                    /* Get the size_t value off of the stack. */
                    arg_as_size_type = va_arg(args, size_t);
                } else if (strcmp(length_modifier, "l") == 0) {
                    /* Get the long value off of the stack. */
                    arg_as_size_type =
                        (size_t)va_arg(args, long unsigned int);
                } else if (strcmp(length_modifier, "h") == 0) {
                    /* Get the short value off of the stack. */
                    arg_as_size_type =
                        (size_t)va_arg(args, unsigned int);
                } else {
                    *cex = cgul_exception__new(
                               "cgul_string__append_vsprintf: "
                               "unsupported length modifier for %%x: %s",
                               length_modifier);
                    goto out;
                }
            } else {
                /* Get the int value off of the stack. */
                arg_as_size_type =
                    (size_t)va_arg(args, unsigned int);
            }
            /* Append the value. */
            cgul_string__append_hexadecimal_with_convspec(
                cex, s, cs, arg_as_size_type);
            break;

        /* Unsupported */
        default:
            /* We do not support this conversion specifier. */
            *cex = cgul_exception__new(
                       "cgul_string__append_vsprintf: "
                       "unsupported conversion specifier: %c",
                       cs->cspec);
            goto out;
        }

        /* Break on error. */
        if (*cex) {
            goto out;
        }
    }

    /* Success. */
    ending_length = cgul_string__get_length(cex, s);
    rv = (int)(ending_length - starting_length);

 out:

    /* Clean up. */
    if (cs) {
        cgul_convspec__delete(cs);
    }

    return rv;
}

void
cgul_string__append_char(cgul_exception_t* cex,
                         cgul_string_t s,
                         char c)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    CGUL_STRING__INCREASE_CAPACITY_IF_NECESSARY(cex, s, 1);
    if (*cex) {
        goto out;
    }

    s->value[s->length] = c;
    ++s->length;
    s->value[s->length] = '\0';
    s->is_ivalue_valid = 0;

 out:
    return;
}

void
cgul_string__append_hchar(cgul_exception_t* cex,
                          cgul_string_t s,
                          cgul_hchar_t* utf16,
                          size_t utf16_length)
{
    size_t index = 0;
    size_t byte_count = 0;
    char buf[CGUL_HCHAR__MB_LEN_MAX + 1] = {0};

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Sanity check. */
    if (!utf16 || (utf16_length == 0)) {
        goto out;
    }

    /* Convert. */
    byte_count = cgul_unicode__hctomb(cex, utf16, utf16_length, &index, buf);
    if (*cex) {
        goto out;
    }

    /* Append. */
    cgul_string__append_block(cex, s, buf, byte_count);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

void
cgul_string__append_wchar(cgul_exception_t* cex,
                          cgul_string_t s,
                          cgul_wchar_t utf32)
{
    size_t i = 0;
    size_t rcount = 0;
    char tmp[CGUL_WCHAR__MB_LEN_MAX + 1] = {0};

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    /* ASCII (7-bit) characters encode to themselves (allowing us to
     * avoid the call to cgul_unicode__wctomb()). */
    if (utf32 < 0x80) {
        cgul_string__append_char(cex, s, (char)utf32);
        goto out;
    }

    /* Convert from UTF-32 to UTF-8 writing the result to "tmp" and
     * the number of bytes written to "rcount". */
    rcount = cgul_unicode__wctomb(cex, utf32, tmp);
    if (*cex) {
        goto out;
    }

    /* Increase capacity (if necessary) to accommodate appending "tmp". */
    CGUL_STRING__INCREASE_CAPACITY_IF_NECESSARY(cex, s, rcount);
    if (*cex) {
        goto out;
    }

    /* Copy the bytes in "tmp" to "s". */
    for (i = 0 ; i < rcount ; ++i) {
        s->value[s->length + i] = tmp[i];
    }

    /* Update the other string members to match. */
    s->length += rcount;
    s->value[s->length] = '\0';
    s->is_ivalue_valid = 0;

 out:
    return;
}

void
cgul_string__append_decimal(cgul_exception_t* cex,
                            cgul_string_t s,
                            long int value)
{
    char tmp[256];

    assert(cex);
    if (*cex) {
        goto out;
    }

    sprintf(tmp, "%ld", value);
    cgul_string__append(cex, s, tmp);

 out:
    return;
}

void
cgul_string__append_unsigned_decimal(cgul_exception_t* cex,
                                     cgul_string_t s,
                                     size_t value)
{
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Handle zero specially. */
    if (value == 0) {
        cgul_string__append_char(cex, s, '0');
        goto out;
    }

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Insert each decimal digit into the temporary string in reverse order. */
    for ( ; value ; value /= 10) {
        /* Borrow "hex_map_lower" because the mod operation limits us
         * to [0-9]. */
        cgul_string__append_char(
            cex, tmp, cgul_string__hex_map_lower[value % 10]);
        if (*cex) {
            goto out;
        }
    }

    /* Reverse the octal digits to put them in the correct order. */
    cgul_string__reverse(cex, tmp);

    /* Append the octal digits in the correct order to s. */
    cgul_string__append_block(cex, s, tmp->value, tmp->length);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

void
cgul_string__append_octal(cgul_exception_t* cex,
                          cgul_string_t s,
                          size_t value)
{
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Handle zero specially. */
    if (value == 0) {
        cgul_string__append_char(cex, s, '0');
        goto out;
    }

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Insert each octal digit into the temporary string in reverse order. */
    for ( ; value ; value >>= 3) {
        /* Borrow "hex_map_lower" because the 0x7 mask limits us to [0-7]. */
        cgul_string__append_char(
            cex, tmp, cgul_string__hex_map_lower[value & 0x7]);
        if (*cex) {
            goto out;
        }
    }

    /* Reverse the octal digits to put them in the correct order. */
    cgul_string__reverse(cex, tmp);

    /* Append the octal digits in the correct order to s. */
    cgul_string__append_block(cex, s, tmp->value, tmp->length);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

void
cgul_string__append_hexadecimal(cgul_exception_t* cex,
                                cgul_string_t s,
                                size_t value,
                                int caps)
{
    const char* hex_map = NULL;
    cgul_string_t tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Handle zero specially. */
    if (value == 0) {
        cgul_string__append_char(cex, s, '0');
        goto out;
    }

    /* Allocate. */
    tmp = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Determine whether to use the upper-case or lower-case mapping
     * to hexadecimal digits. */
    hex_map = (caps ? cgul_string__hex_map_upper : cgul_string__hex_map_lower);

    /* Insert each hex digit into the temporary string in reverse order. */
    for ( ; value ; value >>= 4) {
        cgul_string__append_char(cex, tmp, hex_map[value & 0xf]);
        if (*cex) {
            goto out;
        }
    }

    /* Reverse the hex digits to put them in the correct order. */
    cgul_string__reverse(cex, tmp);

    /* Append the hex digits in the correct order to s. */
    cgul_string__append_block(cex, s, tmp->value, tmp->length);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (tmp) {
        cgul_string__delete(tmp);
    }

    return;
}

void
cgul_string__append_double(cgul_exception_t* cex,
                           cgul_string_t s,
                           double value,
                           int precision)
{
    char tmp[256];

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Limit the precision to make sure the result can fit in our buffer. */
    if (precision > 176) {
        precision = 176;
    }

    sprintf(tmp, "%.*g", precision, value);
    cgul_string__append(cex, s, tmp);

 out:
    return;
}

void
cgul_string__append_pointer(cgul_exception_t* cex,
                            cgul_string_t s,
                            const void* ptr)
{
    size_t i = 0;
    size_t i_max = 0;
    size_t tmp = 0;
    char buf[256];

    assert(cex);
    if (*cex) {
        goto out;
    }

    /*
     * Different compilers convert pointers differently especially
     * whether they pad with leading zeros.  GCC even uses "(nil)"
     * instead of "0x0" for a NULL pointer.  So, the loop below does
     * the conversion by hand in order to get consistent results.
     */

    /* Pointer arithmetic is very limited.  In fact, here it gets in
     * the way because "ptr + 2" depends on the size of the thing
     * pointed to.  If you think of shifting bits as repeated
     * multiplication (or division) by two and if you think of
     * multiplication (or division) as repeated addition (or
     * subtraction), it is not too surprising that the compiler
     * refuses to allow simply shifting of bits because it is not sure
     * of the mathematical consequences (I guess).  (For some reason,
     * this is true even if "ptr" is a "char*" with size of 1 and
     * presumably normal math.)  So the work around is to cast the
     * pointer to an unsigned integral type.  Given that size_t is
     * logically related to pointers, I think it makes the most sense
     * to cast "ptr" to a size_t. */
    assert(sizeof(size_t) >= sizeof(void*));
    tmp = (size_t)ptr;

    /* Calculate the maximum number of hex digits present in a pointer. */
    i_max = sizeof(void*) * 2;
    assert(i_max < sizeof(buf) - 1);

    /* Iterate over "tmp" inserting each hex digit into "buf". */
    for (i = 0 ; i < i_max ; ++i, tmp >>= 4) {
        /* Get the hex digit for the four least-significant bits and
         * add it to "buf" in reverse order because we need to present
         * the most-significant hex digit first. */
        buf[i_max - i - 1] =
            cgul_string__hex_map_lower[tmp & 0xf];
    }

    /* NUL terminate the string. */
    buf[i_max] = '\0';

    /* Append the result to s. */
    cgul_string__append(cex, s, "0x");
    cgul_string__append(cex, s, buf);

 out:
    return;
}

void
cgul_string__insert(cgul_exception_t* cex,
                    cgul_string_t s,
                    size_t index,
                    const char* value,
                    size_t value_length)
{
    size_t i = 0;
    size_t i_max = 0;
    size_t src = 0;
    size_t dst = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(value);

    /* Verify that "index" is within bounds. */
    if (index > s->length) {
        *cex = cgul_exception__new("cgul_string__insert: "
                                   "index out of bounds: %lu > %lu",
                                   (unsigned long)index,
                                   (unsigned long)s->length);
        goto out;
    }

    /* Add extra capacity if necessary to accommodate the extra characters. */
    CGUL_STRING__INCREASE_CAPACITY_IF_NECESSARY(cex, s, value_length);
    if (*cex) {
        goto out;
    }

    /* Shift the characters after and including the insertion point
     * down to the new end of the string.  Notice that this includes
     * shifting the NUL terminator. */
    src = s->length;
    dst = s->length + value_length;
    i_max = s->length - index;
    for (i = 0 ; i <= i_max ; ++i, --src, --dst) {
        s->value[dst] = s->value[src];
    }

    /* Insert the characters from "value". */
    for (i = 0 ; i < value_length ; ++i) {
        s->value[index + i] = value[i];
    }

    /* Calculate the new length. */
    s->length += value_length;

    /* NUL terminate the string. */
    s->value[s->length] = '\0';

    /* Invalidate the case-insensitive string. */
    s->is_ivalue_valid = 0;

 out:
    return;
}

void
cgul_string__insert_char(cgul_exception_t* cex,
                         cgul_string_t s,
                         size_t index,
                         char value)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(value);

    /* Have cgul_string__insert() do the work. */
    cgul_string__insert(cex, s, index, &value, 1);

 out:
    return;
}

void
cgul_string__pad_right(cgul_exception_t* cex,
                       cgul_string_t s,
                       size_t width,
                       cgul_wchar_t pad)
{
    const char* utf8 = NULL;
    size_t utf8_length = 0;
    const cgul_wchar_t* utf32 = NULL;
    size_t utf32_length = 0;
    cgul_wstring_t ws = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    /* Allocate a UTF-32 string to make this process easier. */
    ws = cgul_wstring__new(cex);

    /* Unpack s. */
    utf8 = cgul_string__get_value(cex, s);
    utf8_length = cgul_string__get_length(cex, s);

    /* Convert s to UTF-32. */
    cgul_unicode__mbstowcs(cex, utf8, utf8_length, 0 /* skip_bom */, ws);

    /* Pad right. */
    cgul_wstring__pad_right(cex, ws, width, pad);

    /* Unpack ws. */
    utf32 = cgul_wstring__get_value(cex, ws);
    utf32_length = cgul_wstring__get_length(cex, ws);

    /* Convert back to UTF-8. */
    cgul_string__set_value(cex, s, "");
    cgul_unicode__wcstombs(cex, utf32, utf32_length, 0 /* skip_bom */, s);

 out:

    /* Clean up. */
    if (ws) {
        cgul_wstring__delete(ws);
    }

    return;
}

void
cgul_string__pad_left(cgul_exception_t* cex,
                      cgul_string_t s,
                      size_t width,
                      cgul_wchar_t pad)
{
    const char* utf8 = NULL;
    size_t utf8_length = 0;
    const cgul_wchar_t* utf32 = NULL;
    size_t utf32_length = 0;
    cgul_wstring_t ws = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    /* Allocate a UTF-32 string to make this process easier. */
    ws = cgul_wstring__new(cex);

    /* Unpack s. */
    utf8 = cgul_string__get_value(cex, s);
    utf8_length = cgul_string__get_length(cex, s);

    /* Convert s to UTF-32. */
    cgul_unicode__mbstowcs(cex, utf8, utf8_length, 0 /* skip_bom */, ws);

    /* Pad left. */
    cgul_wstring__pad_left(cex, ws, width, pad);

    /* Unpack ws. */
    utf32 = cgul_wstring__get_value(cex, ws);
    utf32_length = cgul_wstring__get_length(cex, ws);

    /* Convert back to UTF-8. */
    cgul_string__set_value(cex, s, "");
    cgul_unicode__wcstombs(cex, utf32, utf32_length, 0 /* skip_bom */, s);

 out:

    /* Clean up. */
    if (ws) {
        cgul_wstring__delete(ws);
    }

    return;
}

void
cgul_string__trim(cgul_exception_t* cex,
                  cgul_string_t s)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Trim the string.  Because padding right does not require a
     * memmove(), it is done first in the hope that it will reduce the
     * amount of work memmove() has to do for trim_left(). */
    cgul_string__trim_right(cex, s);
    cgul_string__trim_left(cex, s);

 out:
    return;
}

void
cgul_string__trim_left(cgul_exception_t* cex,
                       cgul_string_t s)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Find the first character in s that is not white space. */
    for (i = 0 ; i < s->length ; ++i) {
        if (!cgul_char__isspace(cex, s->value[i])) {
            break;
        }
    }

    /* Nothing to do if there is no leading white space. */
    if (i == 0) {
        goto out;
    }

    /* Trim the string. */
    memmove(s->value, &s->value[i], s->length - i);
    s->length -= i;
    s->value[s->length] = '\0';
    s->is_ivalue_valid = 0;

    /* Decrease capacity if necessary. */
    cgul_string__decrease_capacity_if_necessary(cex, s);

 out:
    return;
}

void
cgul_string__trim_right(cgul_exception_t* cex,
                        cgul_string_t s)
{
    size_t i = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Iterate in reverse order over the characters in s.  Note that
     * "i" is always one beyond the character we are checking making
     * it easier to test to make sure we do not read off the front of
     * the string. */
    for (i = s->length ; i > 0 ; --i) {
        if (!cgul_char__isspace(cex, s->value[i - 1])) {
            break;
        }
    }

    /* Nothing to do if there is no trailing white space. */
    if (i == s->length) {
        goto out;
    }

    /* Trim the string. */
    s->length = i;
    s->value[s->length] = '\0';
    s->is_ivalue_valid = 0;

    /* Decrease capacity if necessary. */
    cgul_string__decrease_capacity_if_necessary(cex, s);

 out:
    return;
}

void
cgul_string__reverse(cgul_exception_t* cex,
                     cgul_string_t s)
{
    size_t i = 0;
    size_t j = 0;
    char tmp = '\0';

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    /* If the string is empty or has only one charcter, there is
     * nothing to reverse. */
    if (s->length <= 1) {
        goto out;
    }

    for ( j = s->length - 1 ; i < j ; ++i, --j ) {
        tmp = s->value[i];
        s->value[i] = s->value[j];
        s->value[j] = tmp;
    }
    s->is_ivalue_valid = 0;

 out:
    return;
}

size_t
cgul_string__get_length(cgul_exception_t* cex,
                        cgul_string_t s)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    rv = s->length;

 out:
    return rv;
}

size_t
cgul_string__set_length(cgul_exception_t* cex,
                        cgul_string_t s,
                        size_t length)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Throw an exception if the user tries to increase the length. */
    if (length > s->length) {
        *cex = cgul_exception__new(
                   "cgul_string__set_length: "
                   "attempt to increase length beyond capacity");
        goto out;
    }

    /* Return the previous length. */
    rv = s->length;

    /* Truncate value. */
    s->value[length] = '\0';

    /* If ivalue is valid, truncate it. */
    if (s->is_ivalue_valid) {
        s->ivalue[length] = '\0';
    }

    /* Set the new length. */
    s->length = length;

    /* Decrease capacity if necessary. */
    cgul_string__decrease_capacity_if_necessary(cex, s);

 out:
    return rv;
}

void
cgul_string__externally_adjusted(cgul_exception_t* cex,
                                 cgul_string_t s)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);
    assert(s->value);

    s->length = strlen(s->value);
    s->is_ivalue_valid = 0;
    cgul_string__decrease_capacity_if_necessary(cex, s);

 out:
    return;
}

void
cgul_string__trim_capacity(cgul_exception_t* cex,
                           cgul_string_t s)
{
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Use realloc() to trim the buffer. */
    s->value = (char*)cgul_libc__realloc(cex,
                                         s->value,
                                         sizeof(char) * (s->length + 1));
    if (*cex) {
        goto out;
    }

    /* Remember the new capacity. */
    s->capacity = s->length + 1;

 out:
    return;
}

size_t
cgul_string__get_minimum_capacity(cgul_exception_t* cex,
                                  cgul_string_t s)
{
    size_t rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Return the minimum capacity to the user. */
    rv = s->minimum_capacity;

 out:
    return rv;
}

void
cgul_string__set_minimum_capacity(cgul_exception_t* cex,
                                  cgul_string_t s,
                                  size_t minimum_capacity)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s);

    /* Set the new minimum_capacity. */
    if (minimum_capacity == 0) {
        s->minimum_capacity = CGUL_STRING__DEFAULT_MINIMUM_CAPACITY;
    } else {
        s->minimum_capacity = minimum_capacity;
    }

    /* Realloc the underlying buffer (if necessary). */
    cgul_string__set_capacity(cex, s, s->minimum_capacity);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

int
cgul_string__compare(cgul_exception_t* cex,
                     cgul_string_t s1,
                     cgul_string_t s2)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s1);
    assert(s2);

    /* Do locale-sensitive comparison on case-sensitive strings. */
    errno = 0;
    rv = strcoll(s1->value, s2->value);
    if (errno) {
        *cex = cgul_exception__new("cgul_string__compare: "
                                   "strcoll: %s", strerror(errno));
    }

 out:
    return rv;
}

int
cgul_string__icompare(cgul_exception_t* cex,
                      cgul_string_t s1,
                      cgul_string_t s2)
{
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s1);
    assert(s2);

    /* Make sure the "ivalue" strings are up to date. */
    if (!s1->is_ivalue_valid) {
        cgul_string__create_ivalue(cex, s1);
    }
    if (!s2->is_ivalue_valid) {
        cgul_string__create_ivalue(cex, s2);
    }

    /* Do locale-sensitive comparison on case-insensitive strings. */
    errno = 0;
    rv = strcoll(s1->ivalue, s2->ivalue);
    if (errno) {
        *cex = cgul_exception__new("cgul_string__icompare: "
                                   "strcoll: %s", strerror(errno));
    }

 out:
    return rv;
}

void
cgul_string__swap(cgul_exception_t* cex,
                  cgul_string_t s1,
                  cgul_string_t s2)
{
    struct cgul_string tmp;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(s1);
    assert(s2);

    /* Nothing to swap if they point to the same objects. */
    if (s1 == s2) {
        goto out;
    }

    /* A shallow copy, by design, is exactly what is needed. */
    tmp = *s1;
    *s1 = *s2;
    *s2 = tmp;

 out:
    return;
}
