#ifndef CGUL_TRIE_P_H
#define CGUL_TRIE_P_H

#include "cgul_common.h"
#include "cgul_cache.h"
#include "cgul_trie.h"
#include "cgul_trie_node.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_trie</code> class.
 */
struct cgul_trie {

    /**
     * Whether this <code>cgul_trie</code> object is indexed.  If it
     * is indexed, it means <code>cgul_trie__find_at()</code> and
     * <code>cgul_trie__remove_at()</code> are available and will not
     * throw an exception if called.  We go to all the trouble of
     * having two different classes that can be used as nodes in our
     * <code>cgul_trie</code> because the indexing feature costs three
     * extra <code>unsigned long</code> values per node.  It also
     * slows down inserts and deletes by about 15%.
     */
    int is_indexed;

    /**
     * Root node.
     */
    cgul_trie_node_t root;

    /**
     * Number of key/value pairs stored in the trie.
     */
    unsigned long int size;

    /**
     *  The oldest node in the trie.  This is the first node that was
     *  inserted.
     */
    cgul_trie_node_t oldest;

    /**
     *  The youngest node in the trie.  This is the last node that was
     *  inserted.
     */
    cgul_trie_node_t youngest;

    /**
     * For efficiency, the <code>cgul_trie</code> can keep a cache of
     * nodes that can be reused.  In many situations, this can greatly
     * reduce the number of calls to <code>malloc()</code> which can
     * greatly improve performance and reduce memory fragmentation.
     */
    cgul_cache_t cache;

};

/**
 * Data needed for <code>cgul_trie__clear_cb()</code>.
 */
typedef struct cgul_trie__clear_cb_data {

    /** The <code>cgul_cache</code> the client is using for caching keys. */
    cgul_cache_t keys_cache;

    /** The <code>cgul_cache</code> the client is using for caching values. */
    cgul_cache_t values_cache;

    /** Return value indicating whether any nodes have been removed. */
    int rv;

} *cgul_trie__clear_cb_data_t;

CGUL_END_C

#endif  /*  CGUL_TRIE_P_H  */
