#ifndef CGUL_WSTRING_P_H
#define CGUL_WSTRING_P_H

#include "cgul_common.h"
#include "cgul_wstring.h"

CGUL_BEGIN_C

/**
 * Data encapsulated by the <code>cgul_wstring</code> class.
 */
struct cgul_wstring {

    /** The user's wide-character string. */
    cgul_wchar_t* value;

    /**
     * Length of the user's wide-character string.  This is the number
     * of wide characters in the string, not the number of bytes.
     */
    size_t length;

    /**
     * Current capacity of this cgul_wstring object.  The capacity is
     * the number of wide-characters (not bytes) the string can hold
     * including the trailing NUL terminator.
     */
    size_t capacity;

    /**
     * Minimum capacity below which <code>capacity</code> is not
     * allowed to fall.
     */
    size_t minimum_capacity;

};

/**
 * Set the capacity to <code>capacity</code> for the string
 * <code>ws</code>.  If <code>capacity</code> is less than the current
 * size, the current size is used instead.  If <code>capacity</code>
 * is zero, the default capacity is used instead.  If memory cannot be
 * allocated, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 * @param[in] capacity new capacity
 */
extern CGUL_EXPORT void
cgul_wstring__set_capacity(cgul_exception_t* cex,
                           cgul_wstring_t ws,
                           size_t capacity);

/**
 * If necessary, decrease capacity for the string <code>ws</code> such
 * that the capacity is reduced by 1/2 if the size is 1/4th the
 * capacity.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws <code>cgul_wstring</code> instance
 */
extern CGUL_EXPORT void
cgul_wstring__decrease_capacity_if_necessary(cgul_exception_t* cex,
                                             cgul_wstring_t ws);

/**
 * If necessary, increase capacity to accommodate appending a string
 * that is <code>extra_size</code> wide-characters (not bytes) long.
 * When calculating <code>extra_size</code>, the caller does not need
 * to include the character associated with the NUL terminator because
 * an extra character will be added automatically.  If more memory
 * cannot be allocated, an exception is thrown, but the underlying
 * string is not altered.
 *
 * @param[in,out] cex c-style exception
 * @param[in] ws this string
 * @param[in] extra_size minimum number of extra wide-characters to
 * add to the capacity
 */
extern CGUL_EXPORT void
cgul_wstring__increase_capacity_if_necessary(cgul_exception_t* cex,
                                             cgul_wstring_t ws,
                                             size_t extra_size);

CGUL_END_C

#endif  /*  CGUL_WSTRING_P_H  */
