/**
 * The program merges multiple bdf font files into a single bdf font
 * file that is written to standard output.  The fonts should appear
 * on the command in order of priority.  So, if an encoding appears in
 * multiple input BDFs, the encoding that appears in the output BDF is
 * the one that comes from the input BDF that is listed first on the
 * command line. <p>
 *
 * @note The FONT line that appears in the output BDF is taken
 * verbatim from the FONT line from the BDF that is listed first on
 * the command line.  It is recommended that user review this line to
 * make sure it is still appropriate for the merged font.
 *
 * @file merge multiple bdf font files
 */

#include <assert.h>
#include "cgul.h"

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    int i = 0;
    cgul_bdf_t bdf = NULL;
    cgul_exception_t local = NULL;

    /* Verify the command line. */
    if (argc < 2) {
        fprintf(stderr,
                "Usage: main_bdf_merge <font1.bdf> [<font2.bdf> ...]\n");
        rv = 1;
        goto out;
    }

    /* Allocate the bdf object. */
    bdf = cgul_bdf__new(&local);
    if (local) {
        goto out;
    }

    /* Load each BDF file into the "bdf" object.  The process of
     * loading the next BDF file merges it with previous ones.  Note
     * that we explicitly disable use of the BDF index file which only
     * slows things down in this scenario. */
    for (i = 1 ; i < argc ; ++i) {
        cgul_bdf__load_explicit(&local, bdf, argv[i], NULL);
        if (local) {
            goto out;
        }
    }

    /* Print the result to stdout. */
    cgul_bdf__write_file(&local, bdf, stdout);
    if (local) {
        goto out;
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (bdf) {
        cgul_bdf__delete(bdf);
    }

    return rv;
}
