/**
 * NOTE: The best way to test hash-related code is to use one of the
 * following programs:
 *
 *     test/main_hash_list_new
 *     test/main_hash_list_new_custom
 *
 * which stores all the lines from an "old" file in a hash and
 * searches the hash for all the lines in a "new" file listing which
 * lines are new.
 */

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include "cgul.h"

static int
get_next_word(cgul_exception_t* cex, FILE* f, cgul_string_t word)
{
    int c = 0;
    int rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Clear word. */
    cgul_string__set_value(cex, word, "");

    /* Find the first character that is not a white space character. */
    for (;;) {
        c = cgul_libc__fgetc(cex, f);
        if (*cex || (c == EOF)) {
            goto out;
        }
        if (!isspace(c)) {
            break;
        }
    }

    cgul_string__append_char(cex, word, (char)c);

    /* Find the rest of the characters that comprise the word. */
    for (;;) {
        c = cgul_libc__fgetc(cex, f);
        if (*cex) {
            goto out;
        }
        if ((c == EOF) || isspace(c)) {
            break;
        }
        cgul_string__append_char(cex, word, (char)c);
    }

    rv = 1;

 out:
    return rv;
}

static void
free_keys(cgul_hash_t h)
{
    cgul_exception_t local = NULL;
    cgul_hash_node_t n = cgul_hash__get_front(&local, h);
    for ( ; n != NULL ; n = cgul_hash_node__get_next(&local, n)) {
        free((void*)cgul_hash_node__get_key(&local, n));
    }
}

static void
print_stats(cgul_exception_t* cex, cgul_hash_t h)
{
    char* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    tmp = cgul_hash__get_stats(cex, h);
    if (*cex) {
        goto out;
    }

    fflush(NULL);
    fprintf(stderr, "%s\n", tmp);

 out:
    if (tmp) {
        free(tmp);
    }
}

static void
print_histogram(cgul_exception_t*cex, cgul_hash_t h)
{
    unsigned long int i = 0;
    unsigned long int value = 0;
    cgul_hash_node_t n = NULL;
    const char** v = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Get a sorted array of keys. */
    v =(const char**) cgul_hash__get_sorted_keys(cex, h, NULL);

    /* Iterate over the keys. */
    for (i = 0 ; v[i] ; ++i) {
        if ((n = cgul_hash__find(cex, h, v[i])) != NULL) {
            value = (unsigned long int)cgul_hash_node__get_value(cex, n);
            printf("%s  %lu\n", v[i], value);
        }
    }

 out:

    if (v) {
        free(v);
    }

    return;
}

static void
add_to_histogram(cgul_exception_t* cex, cgul_hash_t h, cgul_string_t s) {
    int insert_rv = 0;
    unsigned long int count = 0;
    char* key = NULL;
    cgul_hash_node_t n = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(h);
    assert(s);

    key = cgul_string__get_value(cex, s);
    insert_rv = cgul_hash__insert(cex, h, key, &n);
    if (*cex) {
        goto out;
    }
    if( insert_rv ) {
        /* Let cgul_string know we've taken ownership of key. */
        cgul_string__take_value(cex, s);
        cgul_hash_node__set_value(cex, n, (const void*)1);
    } else {
        count = (unsigned long int)cgul_hash_node__get_value(cex, n);
        cgul_hash_node__set_value(cex, n, (const void*)++count);
    }

 out:
    return;
}

int main(int argc, char* argv[])
{
    int rv = 0;
    cgul_hash_t h = NULL;
    cgul_string_t s = NULL;
    cgul_exception_t local = NULL;

    s = cgul_string__new(&local);
    h = cgul_hash__new(&local);

    if (local) {
        goto out;
    }

    while( get_next_word(&local, stdin, s) ) {
        add_to_histogram(&local, h, s);
        if (local) {
            goto out;
        }
    }

    print_histogram(&local, h);
    print_stats(&local, h);

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** main_hash: Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    if (h) {
        free_keys(h);
        cgul_hash__delete(h);
    }

    if (s) {
        cgul_string__delete(s);
    }

    return rv;
}
