/**
 * Generate a list of prime numbers using the Sieve of Eratosthenes
 * without any preconceived upper bound on the sieve.  The basic idea
 * is to use cgul_heap as a min-heap to control multiple streams of
 * numbers where each stream comprises all the multiples of a
 * particular prime number.  Whenever the next prime number is
 * detected, a new stream is created and filtered through the
 * min-heap.
 *
 * @file
 * @brief Sieve of Eratosthenes
 */

#include "cgul.h"

/**
 * Data that will be stored in the min-heap to lazily generate all the
 * multiples of a particular prime number.
 */
typedef struct prime_data {

    /**
     * Prime.
     */
    unsigned long int prime;

    /**
     * The next multiple of <code>prime</code>.
     */
    unsigned long int multiple;

} *prime_data_t;

/**
 * Compare the two <code>prime_data_t</code> elements <code>e1</code>
 * and <code>e2</code> returning less than zero, zero, or greater than
 * zero if <code>e1</code> is greater than [sic], equal to, or less
 * than [sic] <code>e2</code>. <p>
 *
 * @note This function returns the negated value of the natural
 * comparison in order for <code>cgul_heap</code> to function as a
 * min-heap (instead of functioning as a default max-heap). <p>
 */
int
main__compare_prime_data(const void* e1,
                         const void* e2)
{
    int rv = 0;
    unsigned long int int1 = ((prime_data_t)e1)->multiple;
    unsigned long int int2 = ((prime_data_t)e2)->multiple;

    if (int1 < int2) {
        rv = 1;
    } else if (int1 > int2) {
        rv = -1;
    } else {
        rv = 0;
    }

    return rv;
}

/**
 * Push a new <code>prime_data_t</code> instance onto the heap
 * <code>heap</code> for the prime number <code>prime</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] heap heap
 * @param[in] prime prime number
 */
void
main__push_prime_data(cgul_exception_t* cex,
                      cgul_heap_t heap,
                      unsigned long int prime)
{
    prime_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(heap);

    /* Allocate the prime data. */
    data = cgul_libc__calloc(cex, 1, sizeof(*data));
    if (*cex) {
        goto out;
    }

    /* Initialize the prime data. */
    data->prime = prime;
#if 0
    data->multiple = data->prime + data->prime;
#else
    /* Optimized initialization of the first multiple of "prime" that
     * needs to be tested.  This works because all the multiples of
     * "prime" that are less than "prime * prime" are already in the
     * heap.  For example, if (prime == 7), then because 5 is a prime
     * and less than 7 it is already in the heap such that (7 * 5) is
     * already scheduled to be tested as (5 * 7).  This also works for
     * composites like 6 because (2 * 3 * 7) is already scheduled to
     * be tested twice: once for 2 and once for 3.  Testing shows that
     * this can reduce run time by about a third. */
    data->multiple = data->prime * data->prime;
#endif

    /* Push the data onto the heap. */
    cgul_heap__push(cex, heap, data);
    if (*cex) {
        goto out;
    }

    /* The heap now owns the data. */
    data = NULL;

 out:

    /* Clean up. */
    if (data) {
        free(data);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    unsigned long int next = 0;
    prime_data_t top = NULL;
    cgul_heap_t heap = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    heap = cgul_heap__new(&local, main__compare_prime_data);
    if (local) {
        goto out;
    }

    /* Loop until "next" underflows. */
    for (next = 2 ; next > 0 ; ++next) {

        /* Get the top value. */
        top = cgul_heap__top(&local, heap);
        if (!top || (top->multiple > next)) {

            /* "next" is a prime (because it is not the multiple of
             * any other prime). */
            printf("%lu\n", next);

            /* Add a new prime entry for "next". */
            main__push_prime_data(&local, heap, next);
            if (local) {
                goto out;
            }

            continue;
        }

        /* If you reach here, "next" is composite.  So, calculate the
         * next multiple for an prime that is a multiple of next.
         * This is how the min-heap is used to filter the multiple
         * streams of prime number multiples. */
        while ((top = cgul_heap__top(&local, heap)) != NULL) {

            /* Stop looping when the minimum value on the heap is
             * greater than "next" because that prime number multiple
             * hasn't been tested yet. */
            if (top->multiple > next) {
                break;
            }

            /* Pop the prime number data. */
            top = cgul_heap__pop(&local, heap);

            /* Update the prime number data to reflect the next
             * multiple in the stream associated with this prime
             * number. */
            top->multiple += top->prime;

            /* Push the prime number data. */
            cgul_heap__push(&local, heap, top);
            if (local) {
                goto out;
            }
        }

    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (heap) {
        cgul_heap__free_values(heap);
        cgul_heap__delete(heap);
    }

    return rv;
}
