#include "cgul.h"

/**
 * Compare the integer elements <code>e1</code> and <code>e2</code>
 * returning less than zero, zero, or greater than zero if
 * <code>*e1</code> is less than, equal to, or greater than
 * <code>*e2</code> respectively.
 *
 * @param[in] e1 first element
 * @param[in] e2 second element
 * @return comparison of e1 with e2
 */
int
main__compare_integers(const void* e1,
                       const void* e2)
{
    int rv = 0;
    int int1 = *(int*)e1;
    int int2 = *(int*)e2;
    
    if (int1 < int2) {
        rv = -1;
    } else if (int1 > int2) {
        rv = 1;
    } else {
        rv = 0;
    }
    
    return rv;
}

/**
 * Append the integer <code>n</code> to the vector <code>v</code>.  If
 * an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v vector
 * @param[in] n integer to append
 */
void
main__append_integer(cgul_exception_t* cex,
                     cgul_vector_t v,
                     int n)
{
    int* buf = NULL;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Allocate space for a new integer. */
    buf = cgul_libc__calloc(cex, 1, sizeof(*buf));
    if (*cex) {
        goto out;
    }

    /* Append the integer to the back of the vector. */
    *buf = n;
    cgul_vector__push_back(cex, v, buf);
    if (*cex) {
        goto out;
    }

    /* The vector now owns the integer. */
    buf = NULL;

 out:

    /* Clean up. */
    if (buf) {
        free(buf);
        buf = NULL;
    }
    
    return;
}

/**
 * Print the vector of integers <code>v</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] v vector
 */
static void
main__print_vector(cgul_exception_t* cex,
                   cgul_vector_t v)
{
    int* n = NULL;
    size_t i = 0;
    size_t i_max = 0;
    
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(v);

    /* Iterate over the vector. */
    i_max = cgul_vector__get_size(cex, v);
    for (i = 0 ; i < i_max ; ++i) {
        /* Get the next integer. */
        n = cgul_vector__at(cex, v, i);
        if (*cex) {
            goto out;
        }
        /* Print the integer. */
        printf("v[%d] = %d\n", (int)i, *n);
    }

 out:
    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    int i = 0;
    int* n = NULL;
    cgul_vector_t v = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    v = cgul_vector__new(&local);
    if (local) {
        goto out;
    }

    /* Initialize the vector from 1 to 8 (inclusive). */
    for (i = 1 ; i < 9 ; ++i) {
        main__append_integer(&local, v, i);
        if (local) {
            goto out;
        }
    }
    printf("===== Vector initialized. \n");
    main__print_vector(&local, v);

    /* Heapify the vector. */
    cgul_heap_primitives__heapify(&local,
                                  cgul_vector__get_array(&local, v),
                                  cgul_vector__get_size(&local, v),
                                  main__compare_integers);
    if (local) {
        goto out;
    }
    printf("===== Vector heapified. \n");
    main__print_vector(&local, v);

    /* Append a 9 to the vector and use shift_up() to maintain heap order. */
    main__append_integer(&local, v, 9);
    cgul_heap_primitives__shift_up(&local,
                                   cgul_vector__get_array(&local, v),
                                   cgul_vector__get_size(&local, v),
                                   cgul_vector__get_size(&local, v) - 1,
                                   main__compare_integers);
    printf("===== Value of 9 added to heap. \n");
    main__print_vector(&local, v);
    
    /* Replace the value of 9 with 0 and use shift_down() to maintain
     * heap order.  (Because this is a max-heap, and 9 is currently
     * the largest value on the heap, it is the first value in the
     * vector. */
    n = cgul_vector__at(&local, v, 0);
    if (local) {
        goto out;
    }
    *n = 0;
    cgul_heap_primitives__shift_down(&local,
                                     cgul_vector__get_array(&local, v),
                                     cgul_vector__get_size(&local, v),
                                     0,  /* index */
                                     0,  /* use_floyds_method */
                                     main__compare_integers);
    printf("===== Value of 0 replaced value of 9. \n");
    main__print_vector(&local, v);

 out:

    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (v) {
        cgul_vector__free_values(v);
        cgul_vector__delete(v);
    }

    return rv;
}
