#include "cgul.h"

/**
 * Load the binary file <code>fin</code> into the string
 * <code>msg</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] msg message
 * @param[in] fin input file
 */
static void
main__load_file(cgul_exception_t* cex,
                cgul_string_t msg,
                FILE* fin)
{
    size_t rcount = 0;
    size_t buf_size = 16384;
    char* buf = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(msg);
    assert(fin);

    /* Allocate. */
    buf = cgul_libc__calloc(cex, 1, buf_size);
    if (*cex) {
        goto out;
    }

    /* Iterate over each block in fin. */
    for (;;) {

        /* Read the block and append it to msg. */
        rcount = cgul_libc__fread(cex, buf, 1, buf_size, fin);
        if (*cex) {
            goto out;
        }

        /* EOF */
        if (rcount == 0) {
            break;
        }

        /* Append the block to msg. */
        cgul_string__append_block(cex, msg, buf, rcount);
    }

 out:

    /* Clean up. */
    if (buf) {
        free(buf);
    }

    return;
}

/**
 * Load the binary file with name <code>fname</code> into the string
 * <code>msg</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] msg message
 * @param[in] fname file name
 */
static void
main__load_fname(cgul_exception_t* cex,
                 cgul_string_t msg,
                 const char* fname)
{
    FILE* fin = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(msg);
    assert(fname);

    /* Open the file. */
    fin = cgul_libc__fopen(cex, fname, "rb");
    if (*cex) {
        goto out;
    }

    /* Load the file. */
    main__load_file(cex, msg, fin);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up. */
    if (fin) {
        fclose(fin);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    const char* fname = NULL;
    const char* boundary = NULL;
    cgul_string_t msg = NULL;
    cgul_exception_t local = NULL;

    /* Verify command line. */
    if (argc != 3) {
        fprintf(stderr, "Usage: main_mime_split_memory <file> <boundary>\n");
        rv = 1;
        goto out;
    }

    /* Set parameters from the command line. */
    fname = argv[1];
    boundary = argv[2];

    /* Allocate. */
    msg = cgul_string__new(&local);
    if (local) {
        goto out;
    }

    /* Load the file into memory.  You could also use mmap() here. */
    main__load_fname(&local, msg, fname);
    if (local) {
        goto out;
    }

    /* Split the MIME message. */
    cgul_mime__split_memory(&local,
                            cgul_string__get_value(&local, msg),
                            cgul_string__get_length(&local, msg),
                            ".",
                            "mime-preamble",
                            "mime-header",
                            "mime-body",
                            "mime-epilogue",
                            boundary,
                            NULL);  /* auto eol */
    if (local) {
        goto out;
    }

 out:

    /* Catch exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (msg) {
        cgul_string__delete(msg);
    }

    return rv;
}
