#include <iostream>
#include <stdexcept>
#include "cgul_cxx.h"

using namespace std;

/**
 * Load the binary file <code>fin</code> into the string
 * <code>msg</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in] msg message
 * @param[in] fin input file
 */
static void
main__load_file(cgul_string_cxx& msg,
                FILE* fin)
{
    size_t rcount = 0;
    size_t buf_size = 16384;
    char* buf = NULL;

    assert(fin);

    try {

        /* Allocate. */
        buf = (char*)cgul_libc_cxx::calloc(1, buf_size);

        /* Iterate over each block in fin. */
        for (;;) {

            /* Read the block and append it to msg. */
            rcount = cgul_libc_cxx::fread(buf, 1, buf_size, fin);

            /* EOF */
            if (rcount == 0) {
                break;
            }

            /* Append the block to msg. */
            msg.append_block(buf, rcount);
        }

    } catch (...) {

        /* Clean up. */
        if (buf) {
            free(buf);
        }

        /* Rethrow. */
        throw;
    }

    /* Clean up. */
    if (buf) {
        free(buf);
    }
}

/**
 * Load the binary file with name <code>fname</code> into the string
 * <code>msg</code>.  If an error occurs, an exception is thrown.
 *
 * @param[in] msg message
 * @param[in] fname file name
 */
static void
main__load_fname(cgul_string_cxx& msg,
                 const char* fname)
{
    FILE* fin = NULL;

    assert(fname);

    try {

        /* Open the file. */
        fin = cgul_libc_cxx::fopen(fname, "rb");

        /* Load the file. */
        main__load_file(msg, fin);

    } catch (...) {

        /* Clean up. */
        if (fin) {
            fclose(fin);
        }

        /* Rethrow. */
        throw;
    }

    /* Clean up. */
    if (fin) {
        fclose(fin);
    }
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    const char* fname = NULL;
    const char* boundary = NULL;
    cgul_string_cxx msg;

    try {

        /* Verify command line. */
        if (argc != 3) {
            cerr << "Usage: main_mime_split_memory <file> <boundary>" << endl;
            rv = 1;
            goto out;
        }

        /* Set parameters from the command line. */
        fname = argv[1];
        boundary = argv[2];

        /* Load the file into memory.  You could also use mmap() here. */
        main__load_fname(msg, fname);

        /* Split the MIME message. */
        cgul_mime_cxx::split_memory(msg.get_value(),
                                    msg.get_length(),
                                    ".",
                                    "mime-preamble",
                                    "mime-header",
                                    "mime-body",
                                    "mime-epilogue",
                                    boundary,
                                    NULL);  /* auto eol */

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;

    } catch (const cgul_exception_cxx& e) {
        cerr << "*** Error: " << e.get_message() << endl;
        rv = 1;
    }

 out:
    return rv;
}
