/*
 * For cgul_stream_reader, test automatic detection of UTF-8, UTF-16,
 * and UTF-32 streams.  This program reads all the streams and
 * converts them to UTF-8 for display so they should all be the same.
 */

#include "cgul.h"

/**
 * Convert the Unicode stream <code>sr</code> to UTF-8 and print the
 * result.
 *
 * @param[in,out] cex c-style exception
 * @param[in] sr <code>cgul_stream_reader</code> instance
 */
static void
main__print_stream(cgul_exception_t* cex,
                   cgul_stream_reader_t sr)
{
    cgul_wchar_t utf32 = CGUL_WCHAR__NUL;
    cgul_string_t s = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(sr);

    /* Allocate. */
    s = cgul_string__new(cex);
    if (*cex) {
        goto out;
    }

    /* Iterate over each character in the stream. */
    for (;;) {

        /* Read the next character. */
        if (!cgul_stream_reader__read(cex, sr, &utf32)) {
            if (*cex) {
                goto out;
            }
            /* EOF */
            break;
        }

        /* Append the UTF-32 character to the UTF-8 accumulator. */
        cgul_string__append_wchar(cex, s, utf32);
        if (*cex) {
            goto out;
        }
    }

    /* Print the result. */
    puts(cgul_string__get_value(cex, s));

 out:

    /* Clean up. */
    if (s) {
        cgul_string__delete(s);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_string_t utf8 = NULL;
    cgul_hstring_t utf16 = NULL;
    cgul_wstring_t utf32 = NULL;
    cgul_block_reader_t br = NULL;
    cgul_block_reader_memory_t brmemory = NULL;
    cgul_stream_reader_t sr = NULL;
    cgul_exception_t local = NULL;

    /* Allocate. */
    utf8 = cgul_string__new(&local);
    utf16 = cgul_hstring__new(&local);
    utf32 = cgul_wstring__new(&local);
    brmemory = cgul_block_reader_memory__new(&local);
    sr = cgul_stream_reader__new(&local);
    if (local) {
        goto out;
    }

    /* Upcast the block reader. */
    br = cgul_block_reader_memory__as_block_reader(&local, brmemory);

    /*************************************************************************
     * UTF-8
     *************************************************************************/

    /* Create a UTF-8 stream with a Unicode Byte-Order Mark (BOM). */
    cgul_string__append_wchar(&local, utf8, (cgul_wchar_t)0xFEFF);
    cgul_string__append(&local, utf8, "Hello, World!");
    if (local) {
        goto out;
    }

    /* Open the stream. */
    cgul_block_reader_memory__open(&local,
                                   brmemory,
                                   cgul_string__get_value(&local, utf8),
                                   cgul_string__get_length(&local, utf8));
    cgul_stream_reader__reset(&local, sr);
    cgul_stream_reader__add_block_reader(&local, sr, br);
    if (local) {
        goto out;
    }

    /* Print the stream. */
    main__print_stream(&local, sr);
    if (local) {
        goto out;
    }

    /*************************************************************************
     * UTF-16
     *************************************************************************/

    /* Create a UTF-16 stream with a Unicode Byte-Order Mark (BOM). */
    cgul_unicode__mbstohcs(&local,
                           cgul_string__get_value(&local, utf8),
                           cgul_string__get_length(&local, utf8),
                           /* skip_bom */ 0,
                           utf16);
    if (local) {
        goto out;
    }

    /* Open the stream. */
    cgul_block_reader_memory__open(&local,
                                   brmemory,
                                   cgul_hstring__get_value(&local, utf16),
                                   2 * cgul_hstring__get_length(&local, utf16));
    cgul_stream_reader__reset(&local, sr);
    cgul_stream_reader__add_block_reader(&local, sr, br);
    if (local) {
        goto out;
    }

    /* Print the stream. */
    main__print_stream(&local, sr);
    if (local) {
        goto out;
    }

    /*************************************************************************
     * UTF-32
     *************************************************************************/

    /* Create a UTF-32 stream with a Unicode Byte-Order Mark (BOM). */
    cgul_unicode__mbstowcs(&local,
                           cgul_string__get_value(&local, utf8),
                           cgul_string__get_length(&local, utf8),
                           /* skip_bom */ 0,
                           utf32);
    if (local) {
        goto out;
    }

    /* Open the stream. */
    cgul_block_reader_memory__open(&local,
                                   brmemory,
                                   cgul_wstring__get_value(&local, utf32),
                                   4 * cgul_wstring__get_length(&local, utf32));
    cgul_stream_reader__reset(&local, sr);
    cgul_stream_reader__add_block_reader(&local, sr, br);
    if (local) {
        goto out;
    }

    /* Print the stream. */
    main__print_stream(&local, sr);
    if (local) {
        goto out;
    }

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (sr) {
        cgul_stream_reader__delete(sr);
    }
    if (brmemory) {
        cgul_block_reader_memory__delete(brmemory);
    }
    if (utf32) {
        cgul_wstring__delete(utf32);
    }
    if (utf16) {
        cgul_hstring__delete(utf16);
    }
    if (utf8) {
        cgul_string__delete(utf8);
    }

    return rv;
}
