#include "cgul.h"

/**
 * Insert a copy of the key <code>key</code> into the trie
 * <code>trie</code>.  The caller is responsible for calling
 * <code>free()</code> on the key stored in the trie.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] trie <code>cgul_trie</code> instance
 * @param[in] key
 */
static void
main__insert_key(cgul_exception_t* cex,
                 cgul_trie_t trie,
                 const char* key)
{
    int insert_rv = 0;
    char* tmp = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Duplicate the key. */
    tmp = cgul_libc__strdup(cex, key);
    if (*cex) {
        goto out;
    }

    /* Insert the duplicated key. */
    insert_rv = cgul_trie__insert(cex, trie, tmp, NULL);
    if (*cex) {
        goto out;
    }
    if (insert_rv) {
        /* Transfer ownership of the duplicated key to the trie. */
        tmp = NULL;
    }

 out:

    /* Clean up. */
    if (tmp) {
        free(tmp);
    }

    return;
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    char* key_out = NULL;
    const char* ceiling_key = NULL;
    cgul_trie_t trie = NULL;
    cgul_trie_node_t node = NULL;
    cgul_exception_t local = NULL;

    /* Allocate*/
    trie = cgul_trie__new(&local);
    if (local) {
        goto out;
    }

    /* Make sure an empty trie returns NULL as the ceiling for the
     * empty string. */
    node = cgul_trie__find_ceiling(&local, trie, "");
    cgul_libc__printf(&local, "empty trie && ceiling(\"\") => node=%p\n", node);
    assert(!node);

    /* Make sure an empty trie returns NULL as the ceiling for a
     * non-empty string. */
    node = cgul_trie__find_ceiling(&local, trie, "foo");
    cgul_libc__printf(&local, "ceiling(\"foo\") => node=%p\n", node);
    assert(!node);

    /* Insert a key smaller than "foo". */
    main__insert_key(&local, trie, "baz");
    if (local) {
        goto out;
    }

    /* Make sure the ceiling returns the single key given the empty string. */
    node = cgul_trie__find_ceiling(&local, trie, "");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "\"baz\" key too && ceiling(\"\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "baz") == 0);

    /* Make sure the ceiling still returns NULL for an non-empty string. */
    node = cgul_trie__find_ceiling(&local, trie, "foo");
    cgul_libc__printf(&local, "ceiling(\"foo\") => node=%p\n", node);
    assert(!node);

    /* Insert the empty string. */
    main__insert_key(&local, trie, "");
    if (local) {
        goto out;
    }

    /* Make sure the ceiling returns the the empty string. */
    node = cgul_trie__find_ceiling(&local, trie, "");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "empty key too && ceiling(\"\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "") == 0);

    /* Make sure the ceiling is still NULL. */
    node = cgul_trie__find_ceiling(&local, trie, "foo");
    cgul_libc__printf(&local, "ceiling(\"foo\") node=%p\n", node);
    assert(!node);

    /* Insert a key larger than "foo". */
    main__insert_key(&local, trie, "quux");
    if (local) {
        goto out;
    }

    /* Make sure the ceiling is the larger key. */
    node = cgul_trie__find_ceiling(&local, trie, "foo");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "\"quux\" key too && ceiling(\"foo\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "quux") == 0);

    /* Insert key with a "foo" prefix.  Note that "foo" is still not
     * inserted at this point. */
    main__insert_key(&local, trie, "foobar");
    if (local) {
        goto out;
    }

    /* Make sure the ceiling is found along the prefix path. */
    node = cgul_trie__find_ceiling(&local, trie, "foo");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "\"foobar\" key too && ceiling(\"foo\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "foobar") == 0);

    /* Insert "foo". */
    main__insert_key(&local, trie, "foo");
    if (local) {
        goto out;
    }

    /* Make sure the ceiling is now "foo". */
    node = cgul_trie__find_ceiling(&local, trie, "foo");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "\"foo\" key too && ceiling(\"foo\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "foo") == 0);

    /* Make sure search keys that pass "foo" but that do not reach
     * "foobar" find "foobar" as the ceiling. */
    node = cgul_trie__find_ceiling(&local, trie, "foob");  /* sic */
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "ceiling(\"foob\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "foobar") == 0);

    /* Ditto, but one character more. */
    node = cgul_trie__find_ceiling(&local, trie, "fooba");  /* sic */
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "ceiling(\"fooba\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "foobar") == 0);

    /* Make sure we can now find "foobar" exactly. */
    node = cgul_trie__find_ceiling(&local, trie, "foobar");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "ceiling(\"foobar\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "foobar") == 0);

    /* Make sure we can now find "quux" when using a search key that
     * is along the same path as "foobar" but is a longer string. */
    node = cgul_trie__find_ceiling(&local, trie, "foobarbaz");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "ceiling(\"quux\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "quux") == 0);

    /* Make sure we can now find "foobar" when it is greater than a
     * string that varies only in the last character. */
    node = cgul_trie__find_ceiling(&local, trie, "fooban");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "ceiling(\"fooban\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "foobar") == 0);

    /* Make sure we can get NULL when our search key has the largest
     * key as a prefix. */
    node = cgul_trie__find_ceiling(&local, trie, "quuxar");
    assert(!node);
    cgul_libc__printf(&local, "ceiling(\"quuxar\") => node=%p\n", node);
    assert(!node);

    /* Make sure we can find the ceiling when it exists on the right subtree. */
    node = cgul_trie__find_ceiling(&local, trie, "bba");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "ceiling(\"bba\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "foo") == 0);

    /* Make sure we can find the ceiling when it exists on the right subtree. */
    node = cgul_trie__find_ceiling(&local, trie, "grip");
    assert(node);
    ceiling_key = cgul_trie_node__get_key(&local, node);
    cgul_libc__printf(&local, "ceiling(\"grip\") => ceiling=-->%s<--\n", ceiling_key);
    assert(strcmp(ceiling_key, "quux") == 0);

    /* Remove the "foo" key. */
    if (!cgul_trie__remove(&local, trie, "foo", &key_out, NULL)) {
        local = cgul_exception__new("Unable to remove key: foo");
        goto out;
    }
    free(key_out);
    key_out = NULL;

 out:

    /* Handle exceptions. */
    if (local) {
        const char* msg = cgul_exception__get_message(local);
        fprintf(stderr, "*** Error: %s\n", msg);
        cgul_exception__delete(local);
        local = NULL;
        rv = 1;
    }

    /* Clean up. */
    if (trie) {
        cgul_trie__free_keys(trie);
        cgul_trie__delete(trie);
    }

    return rv;
}
