/**
 * This program prints a range of Unicode characters in UTF-8 format.
 */

#include "cgul_cxx.h"

/**
 * The program name.
 */
static const char* progname = "main_unicode_range_cxx";

/**
 * Inclusive starting code point for the range of Unicode characters.
 */
static cgul_wchar_t main_start;

/**
 * Inclusive stopping code point for the range of Unicode characters.
 */
static cgul_wchar_t main_stop;

/**
 * Whether the user wants help.
 */
static int user_wants_help = 0;

/**
 * Whether the user wants a leading byte-order mark (BOM) printed.
 * This is convenient under Windows where a leading BOM marks the file
 * as UTF-8.
 */
static int user_wants_bom = 0;

/**
 * Usage message.
 */
static char usage_message[] = "\
Usage: %s [options] <start> <stop>\n\
\n\
    Options:\n\
\n\
            -b : Print a leading byte-order mark (BOM).\n\
\n\
            -h : Print help message.\n\
\n\
    Example:\n\
\n\
        %s -b 0 0x10ffff > unicode.txt\n\
";

/**
 * Print the usage message to the file <code>f</code>.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in] f file
 */
static void
print_usage(FILE* f)
{
    cgul_libc_cxx::fprintf(f, "\n");
    cgul_libc_cxx::fprintf(f, usage_message, progname, progname);
    cgul_libc_cxx::fprintf(f, "\n\n");
}

/**
 * Convert the string <code>s</code> to a <code>cgul_wchar_t</code>
 * and return the result.  If an error occurs, an exception is thrown.
 *
 * @param[in] s string
 * @return converstion of <code>s</code> to a <code>cgul_wchar_t</code>
 */
static cgul_wchar_t
string_to_wchar(const char* s)
{
    unsigned long int n = 0;
    cgul_string_cxx tmp;

    /* Load s into tmp. */
    tmp.set_value(s);

    /* Convert to an unsigned long. */
    try {
        n = tmp.to_unsigned_long(0 /* base */);
    } catch (const cgul_exception_cxx& e) {
        /* Throw a more descriptive exception. */
        throw cgul_exception_cxx("unable to convert to unicode: %s", s);
    }

    /* The valid Unicode range is 0 - 0x10ffff. */
    if (n > 0x10ffff) {
        throw cgul_exception_cxx("out of range for a unicode: %s", s);
    }

    /* Success. */
    return (cgul_wchar_t)n;
}

/**
 * Parse the command-line options and set global values.
 *
 * @param[in] argc argument count
 * @param[in] argv argument vector
 */
static void
parse_options(int argc,
              char** argv)
{
    int i = 1;
    int is_start_set = 0;
    int is_stop_set = 0;
    const char* arg = NULL;

    /* Iterate over the command-line arguments. */
    for (i = 1 ; i < argc ; ++i) {
        arg = argv[i];

        /* -bom */
        if (   (strcmp(arg, "-b") == 0)
            || (strcmp(arg, "-bom") == 0)
            || (strcmp(arg, "--bom") == 0))
        {
            user_wants_bom = 1;
            continue;
        }

        /* -help */
        if (   (strcmp(arg, "-h") == 0)
            || (strcmp(arg, "-help") == 0)
            || (strcmp(arg, "--help") == 0))
        {
            user_wants_help = 1;
            goto out;
        }

        /* Check for invalid flags. */
        if (arg[0] == '-') {
            throw cgul_exception_cxx("invalid argument: %s", arg);
        }

        /* Check for the starting code point. */
        if (!is_start_set) {
            main_start = string_to_wchar(arg);
            is_start_set = 1;
            continue;
        }

        /* Check for the stopping code point. */
        if (!is_stop_set) {
            main_stop = string_to_wchar(arg);
            is_stop_set = 1;
            continue;
        }

        /* Invalid command line. */
        throw cgul_exception_cxx("invalid argument: %s", arg);
    }

    /* Make sure both main_start and main_stop have been set. */
    if (!is_start_set || !is_stop_set) {
        throw cgul_exception_cxx("invalid command line");
    }

    /* make sure main_start is less than main_stop. */
    if (main_start >= main_stop) {
        throw cgul_exception_cxx(
                  "starting point is greater than stopping point");
    }

 out:
    return;
}

/**
 * Print the inclusive range <code>[start - stop]</code> of Unicode
 * characters to the file <code>f</code> encoded as UTF-8.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in] f file
 * @param[in] start inclusive starting Unicode code point
 * @param[in] stop inclusive stopping Unicode code point
 */
static void
print_range(FILE* f,
            cgul_wchar_t start,
            cgul_wchar_t stop)
{
    cgul_wchar_t i = CGUL_WCHAR__NUL;
    char buf[CGUL_WCHAR__MB_LEN_MAX + 1] = {0};

    /* Bounds check start. */
    if (start > 0x10ffff) {
        throw cgul_exception_cxx("print_range: start is out of bounds: %lu",
                                 (unsigned long)start);
    }

    /* Bounds check stop. */
    if (stop > 0x10ffff) {
        throw cgul_exception_cxx("print_range: stop is out of bounds: %lu",
                                 (unsigned long)stop);
    }

    /* Write the byte-order mark (BOM) to mark the resulting file as
     * UTF-8 under Windows. */
    if (user_wants_bom) {
        cgul_unicode_cxx::wctomb(0xfeff, buf);
        cgul_libc_cxx::fprintf(f, "%s", buf);
    }

    /* Iterate over each Unicode character in the range. */
    for (i = start ; i <= stop ; ++i) {

        /* Skip the UTF-16 surrogate pair range. */
        if ((i >= 0xd800) && (i <= 0xdfff)) {
            cgul_libc_cxx::fprintf(
                stderr,
                "Warning: skipping UTF-16 surrogate pair range: "
                "[0xd800 - 0xdfff]\n");
            i = 0xdfff;
            continue;
        }

        /* Convert UTF-32 to UTF-8. */
        cgul_unicode_cxx::wctomb(i, buf);

        /* Print UTF-8 to file. */
        cgul_libc_cxx::fprintf(f, "%#08lx: %s\n", (unsigned long)i, buf);
    }
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;

    try {

        /* Parse the command line options. */
        parse_options(argc, argv);

        /* Check if the user wants help. */
        if (user_wants_help) {
            print_usage(stdout);
            goto out;
        }

        /* Print the range of Unicode characters. */
        print_range(stdout, main_start, main_stop);

    } catch (const cgul_exception_cxx& e) {
        fprintf(stderr, "*** Error: %s\n", e.get_message());
        rv = 1;
    }

 out:
    return rv;
}
