#include "cgul_cxx.h"

/**
 * Load the tree such that each line of text in the file
 * <code>fname</code> is a key in the tree.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in] t tree
 * @param[in] fname file name
 */
static void
load_tree(cgul_rbtree_cxx& t,
          const char* fname)
{
    int insert_rv = 0;
    const char* line = NULL;
    cgul_wstring_cxx ws;
    cgul_crlf_file_cxx crlf;

    assert(fname);

    /* Open the file. */
    crlf.open_fname("tower_of_babel.txt");

    /* Iterate over each line in the text file. */
    for (;;) {

        /* Get the next line. */
        line = crlf.get_line();
        if (!line) {
            /* EOF */
            break;
        }

        /* Clear the temporary string. */
        ws.set_value(CGUL_WCHAR__EMPTY_STRING);

        /* Convert the line from UTF-8 to UTF-32. */
        cgul_unicode_cxx::mbstowcs(line, strlen(line), 0 /* skip bom */, ws);

        /* Insert the string into the red-black tree. */
        insert_rv = t.insert(ws.get_value(), NULL);
        if (!insert_rv) {
            fprintf(stderr, "*** Warning: duplicate key: \"%s\"\n", line);
        } else {
            /* The red-black tree now owns the key. */
            ws.take_value();
        }
    }
}

/**
 * Print a wide-character string exposing the Unicode code points so
 * that the keys can be visually compared for correct ordering.
 *
 * @param[in] ws wide-character string
 */
static void
print_key(cgul_wchar_t* ws)
{
    size_t i = 0;

    assert(ws);

    for (i = 0 ; ws[i] ; ++i) {
        if (i) {
            printf(" ");
        }
        printf("%lu", (unsigned long)ws[i]);
    }
    puts("");
}

/**
 * Print the keys in the red-black tree <code>t</code> exposing the
 * Unicode code points so that the keys can be visually compared for
 * correct ordering.
 *
 * @param[in] t tree
 */
static void
print_keys(cgul_rbtree_cxx& t)
{
    cgul_wchar_t* ws = NULL;
    cgul_rbtree_node_cxx* n = NULL;

    /* Iterate over each key. */
    for (n = t.get_front() ; n ; n = n->get_next()) {
        ws = (cgul_wchar_t*)n->get_key();
        print_key(ws);
    }
}

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    cgul_rbtree_cxx* t = NULL;

    try {

        /* Allocate. */
        t = new cgul_rbtree_cxx(
                    (cgul_rbtree_cxx::compare_t)cgul_wchar_cxx::wcscmp);
        if (!t) {
            throw cgul_exception_cxx("out of memory");
        }

        /* Load the tree.  Each key in the tree will be a line from
         * the text file. */
        load_tree(*t, "tower_of_babel.txt");

        /* Print the lines to visually verify they are sorted. */
        print_keys(*t);

    } catch (const cgul_exception_cxx& e) {
        fprintf(stderr, "*** Error: %s\n", e.get_message());
        rv = 1;
    }

    /* Clean up. */
    if (t) {
        t->free_keys();
        delete t;
    }

    return rv;
}
